#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.core.urlresolvers import reverse
from django.test import TestCase


from rest_framework import status
from rest_framework.test import APITestCase

from django.contrib.auth.models import User
from django.conf import settings

from .models import SupervisionTrack, Profile

from datetime import datetime, timedelta
import datetime
import simplejson as json

#----------------------------------------------------------


class AccountTestCase(APITestCase):

    def setUp(self):
        self.tearDown()


        now = datetime.datetime.now()
        expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_BLOCKAGE_AFTER_FIRST_REJECTION_DAYS)

        # Create the users
        self.password = '1234'

        self.firstgodfather = User.objects.create_user('firstgodfather', 'firstgodfather@test.org', self.password)
        self.secondgodfather = User.objects.create_user('secondgodfather', 'secondgodfather@test.org', self.password)
        self.invalidgodfather = User.objects.create_user('invalidgodfather', 'invalidgodfather@test.org', self.password)

        self.accepteduser = User.objects.create_user('accepteduser', 'accepteduser@test.org', self.password)
        self.rejecteduser = User.objects.create_user('rejecteduser', 'rejecteduser@test.org', self.password)
        self.yearrevalidationuser = User.objects.create_user('yearrevalidationuser', 'yearrevalidationuser@test.org', self.password)
        self.blockeduser = User.objects.create_user('blockeduser', 'blockeduser@test.org', self.password)

        # Update the profiles
        #godfathers
        self.firstgodfather.profile.status = Profile.ACCEPTED
        self.secondgodfather.profile.status = Profile.ACCEPTED
        self.invalidgodfather.profile.status = Profile.BLOCKED
        self.firstgodfather.profile.is_godfather = True
        self.secondgodfather.profile.is_godfather = True
        self.invalidgodfather.profile.is_godfather = True
        self.invalidgodfather.is_active = False

        self.firstgodfather.save()
        self.secondgodfather.save()
        self.invalidgodfather.save()
        self.firstgodfather.profile.save()
        self.secondgodfather.profile.save()
        self.invalidgodfather.profile.save()

        #users
        self.accepteduser.profile.status = Profile.ACCEPTED
        self.rejecteduser.profile.status = Profile.REJECTED
        self.yearrevalidationuser.profile.status = Profile.YEARREVALIDATION
        self.blockeduser.profile.status = Profile.BLOCKED

        # Create the supervision tracks
        #Accepted user
        #Create and assign key
        self.accepteduser.profile.supervision_key = self.accepteduser.profile._generate_current_supervision_key()
        supervisiontrack = SupervisionTrack.objects.create(
            supervisee = self.accepteduser,
            godfather = self.firstgodfather,
            is_valid = True,
            start_date = now,
            expiration_date = now + expiration_date_delta,
            last_validation_date = now
            )

        #Assign key to supervision track
        supervisiontrack.supervision_key = self.accepteduser.profile.supervision_key
        supervisiontrack.save()
        self.accepteduser.profile.supervision.add(supervisiontrack)
        self.accepteduser.profile.save()
        self.accepteduser.save()

        #Rejected user
        #No supervision key but a rejection date
        self.rejecteduser.profile.supervision_key = None
        self.rejecteduser.profile.rejection_date = now + expiration_date_delta
        self.rejecteduser.profile.save()
        self.rejecteduser.save()

        #YearRevalidation user
        #Create and assign key
        self.yearrevalidationuser.profile.supervision_key = self.yearrevalidationuser.profile._generate_current_supervision_key()
        supervisiontrack = SupervisionTrack.objects.create(
            supervisee = self.yearrevalidationuser,
            godfather = self.firstgodfather,
            is_valid = True,
            start_date = now,
            expiration_date = now + expiration_date_delta,
            last_validation_date = now
            )

        #Assign key to supervision track
        supervisiontrack.supervision_key = self.yearrevalidationuser.profile.supervision_key
        supervisiontrack.save()
        self.yearrevalidationuser.profile.supervision.add(supervisiontrack)
        self.yearrevalidationuser.profile.save()
        self.yearrevalidationuser.save()

        #Blocked user
        self.blockeduser.profile.supervision_key = None
        self.blockeduser.profile.rejection_date = None
        self.blockeduser.is_active = False
        self.blockeduser.profile.save()
        self.blockeduser.save()


    def tearDown(self):
        user = None
        pass

    def test_check_access_and_setup(self):

        #Godfathers
        user = self.firstgodfather
        logged_in = self.client.login(username=user.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(user.username, 'firstgodfather')
        logged_out = self.client.logout()

        user = self.secondgodfather
        logged_in = self.client.login(username=user.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(user.username, 'secondgodfather')
        logged_out = self.client.logout()

        user = self.invalidgodfather
        logged_in = self.client.login(username=user.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(user.username, 'invalidgodfather')
        logged_out = self.client.logout()

        #Users
        user = self.accepteduser
        logged_in = self.client.login(username=user.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(user.username, 'accepteduser')
        logged_out = self.client.logout()

        user = self.rejecteduser
        logged_in = self.client.login(username=user.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(user.username, 'rejecteduser')
        logged_out = self.client.logout()

        user = self.yearrevalidationuser
        logged_in = self.client.login(username=user.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(user.username, 'yearrevalidationuser')
        logged_out = self.client.logout()

        user = self.blockeduser
        logged_in = self.client.login(username=user.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(user.username, 'blockeduser')
        logged_out = self.client.logout()


#----------------------------------------------------------


class AccountListTestCase(AccountTestCase):

    def setUp(self):
        super(AccountListTestCase, self).setUp()

        self.url = reverse('api_accounts:list_supervisee')

    def test_anonymous_user(self):
        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

    def test_logged_in_user_not_godfather(self):

        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisor account')

        logged_out = self.client.logout()

    def test_logged_in_user_godfather(self):

        client = self.firstgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'firstgodfather')

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 2)
        self.assertEqual(content[0]['supervisee']['username'], 'accepteduser')
        self.assertEqual(content[1]['supervisee']['username'], 'yearrevalidationuser')

        logged_out = self.client.logout()


#----------------------------------------------------------


class AccountRevalidationTestCase(AccountTestCase):

    def setUp(self):
        super(AccountRevalidationTestCase, self).setUp()

        self.url = reverse('api_accounts:revalidate_account')

    def test_anonymous_user(self):
        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

    def test_logged_in_accepted_user(self):
        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['error']
        self.assertEqual(errormsg, "You don't need to revalidate at the moment, your supervision is still valid")

        logged_out = self.client.logout()

    def test_logged_in_rejected_user(self):
        client = self.rejecteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'rejecteduser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['error']
        self.assertEqual(errormsg, 'You are not allowed to perform this action, you first need to get a valid supervision')

        logged_out = self.client.logout()

    def test_logged_in_yearrevalidation_user(self):

        client = self.yearrevalidationuser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        logged_out = self.client.logout()

    def test_logged_in_yearrevalidation_user_twice_validation(self):

        client = self.yearrevalidationuser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['error']
        self.assertEqual(errormsg, "You don't need to revalidate at the moment, your supervision is still valid")

        logged_out = self.client.logout()

    def test_logged_in_blocked_user(self):

        client = self.blockeduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'blockeduser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        logged_out = self.client.logout()

    def test_logged_in_user_godfather(self):

        client = self.firstgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'firstgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisee account')

        logged_out = self.client.logout()

    def test_logged_in_invalid_godfather(self):

        client = self.invalidgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'invalidgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        logged_out = self.client.logout()


#----------------------------------------------------------


class AccountAddSupervisorTestCase(AccountTestCase):

    def setUp(self):
        super(AccountAddSupervisorTestCase, self).setUp()

        self.url = reverse('api_accounts:add_supervisor', kwargs={'supervisor_name': self.secondgodfather.username})

    def test_anonymous_user(self):
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

    def test_logged_in_accepted_user_supervision_request_to_invalid_godfather(self):
        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)

        self.url = reverse('api_accounts:add_supervisor', kwargs={'supervisor_name': self.invalidgodfather.username})
        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['error']
        self.assertEqual(errormsg, 'Not a valid supervisor request')

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)
        self.assertEqual(previous_supervision_track, new_supervision_track)
        self.assertEqual(previous_supervision_track.is_valid, True)
        self.assertEqual(previous_supervision_key, new_supervision_key)
        self.assertEqual(client.profile.status, Profile.ACCEPTED)
        logged_out = self.client.logout()

    def test_logged_in_accepted_user(self):

        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)

        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertNotEqual(previous_supervision_track.godfather, new_supervision_track.godfather)
        self.assertEqual(client.profile.status, Profile.REJECTED)
        logged_out = self.client.logout()

    def test_logged_in_rejected_user(self):

        client = self.rejecteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'rejecteduser')

        previous_supervision_key = client.profile.supervision_key
        self.assertEqual(previous_supervision_key, None)

        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertEqual(new_supervision_track.godfather.username, "secondgodfather")
        self.assertEqual(client.profile.status, Profile.REJECTED)
        logged_out = self.client.logout()

    def test_logged_in_rejected_user_twice(self):

        client = self.rejecteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'rejecteduser')

        previous_supervision_key = client.profile.supervision_key
        self.assertEqual(previous_supervision_key, None)

        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertEqual(new_supervision_track.godfather.username, "secondgodfather")
        self.assertEqual(client.profile.status, Profile.REJECTED)
        #change supervisor second request
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['error']
        self.assertEqual(errormsg, "You are not able to perform this action as you already have a pending request")

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertEqual(new_supervision_track.godfather.username, "secondgodfather")
        self.assertEqual(client.profile.status, Profile.REJECTED)
        logged_out = self.client.logout()

    def test_logged_in_yearrevalidation_user(self):

        client = self.yearrevalidationuser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)

        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertNotEqual(previous_supervision_track.godfather, new_supervision_track.godfather)
        self.assertEqual(client.profile.status, Profile.REJECTED)

        logged_out = self.client.logout()

    def test_logged_in_blocked_user(self):

        client = self.blockeduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'blockeduser')

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        #This reactivation can be done through online request ONLY

        logged_out = self.client.logout()

    def test_logged_in_user_godfather(self):

        client = self.firstgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'firstgodfather')

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisee account')

        logged_out = self.client.logout()

    def test_logged_in_invalid_godfather(self):

        client = self.invalidgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'invalidgodfather')

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        logged_out = self.client.logout()


#----------------------------------------------------------


class AccountValidateSuperviseeTestCase(AccountTestCase):

    def setUp(self):
        super(AccountValidateSuperviseeTestCase, self).setUp()

        self.url = reverse('api_accounts:validate_supervisee', kwargs={'supervisee_name': self.accepteduser.username})

    def test_anonymous_user(self):
        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

    def test_logged_in_accepted_user(self):
        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisor account')

        logged_out = self.client.logout()

    def test_logged_in_rejected_user(self):
        client = self.rejecteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'rejecteduser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisor account')

        logged_out = self.client.logout()

    def test_logged_in_yearrevalidation_user(self):

        client = self.yearrevalidationuser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisor account')

        logged_out = self.client.logout()

    def test_logged_in_blocked_user(self):

        client = self.blockeduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'blockeduser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        logged_out = self.client.logout()

    def test_logged_in_user_godfather_validate_already_accepted_user(self):

        client = self.firstgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'firstgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['error']
        self.assertEqual(errormsg, "You can't validate this supervisee as his supervision is still valid")

        logged_out = self.client.logout()

    def test_logged_in_user_secondgodfather_validates_previously_accepted_user(self):

        self.url = reverse('api_accounts:add_supervisor', kwargs={'supervisor_name': self.secondgodfather.username})

        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)

        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertEqual(new_supervision_track.is_valid, False)
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertNotEqual(previous_supervision_track.godfather, new_supervision_track.godfather)
        self.assertEqual(client.profile.status, Profile.REJECTED)
        logged_out = self.client.logout()

        self.url = reverse('api_accounts:validate_supervisee', kwargs={'supervisee_name': self.accepteduser.username})
        client = self.secondgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'secondgodfather')

        response = self.client.put(self.url, format='json')

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        logged_out = self.client.logout()

        client = User.objects.get(username=self.accepteduser.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertEqual(new_supervision_track.is_valid, True)
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertNotEqual(previous_supervision_track.godfather, new_supervision_track.godfather)
        self.assertEqual(client.profile.status, Profile.ACCEPTED)


    def test_logged_in_user_secondgodfather_validates_rejected_user(self):

        self.url = reverse('api_accounts:add_supervisor', kwargs={'supervisor_name': self.secondgodfather.username})

        client = self.rejecteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'rejecteduser')

        previous_supervision_key = client.profile.supervision_key
        self.assertEqual(previous_supervision_key, None)

        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertEqual(new_supervision_track.godfather.username, "secondgodfather")
        self.assertEqual(client.profile.status, Profile.REJECTED)
        logged_out = self.client.logout()

        self.url = reverse('api_accounts:validate_supervisee', kwargs={'supervisee_name': self.rejecteduser.username})
        client = self.secondgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'secondgodfather')

        response = self.client.put(self.url, format='json')

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        logged_out = self.client.logout()

        client = User.objects.get(username=self.rejecteduser.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        self.assertEqual(new_supervision_track.is_valid, True)
        self.assertNotEqual(new_supervision_track.godfather, "secondgodfather")
        self.assertEqual(client.profile.status, Profile.ACCEPTED)

    def test_logged_in_user_secondgodfather_validates_yearrevalidation_user(self):

        self.url = reverse('api_accounts:add_supervisor', kwargs={'supervisor_name': self.secondgodfather.username})

        client = self.yearrevalidationuser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)

        #change supervisor
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        client = User.objects.get(username=client.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertNotEqual(previous_supervision_track.godfather, new_supervision_track.godfather)
        self.assertEqual(client.profile.status, Profile.REJECTED)

        logged_out = self.client.logout()

        self.url = reverse('api_accounts:validate_supervisee', kwargs={'supervisee_name': self.yearrevalidationuser.username})
        client = self.secondgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'secondgodfather')

        response = self.client.put(self.url, format='json')

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        logged_out = self.client.logout()

        client = User.objects.get(username=self.yearrevalidationuser.username)
        new_supervision_key = client.profile.supervision_key
        new_supervision_track = SupervisionTrack.objects.get(supervision_key = new_supervision_key)

        #check track change and supervisor
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertEqual(new_supervision_track.is_valid, True)
        self.assertNotEqual(previous_supervision_key, new_supervision_key)
        self.assertNotEqual(previous_supervision_track.godfather, new_supervision_track.godfather)
        self.assertEqual(client.profile.status, Profile.ACCEPTED)

    def test_logged_in_user_godfather_validation_try(self):

        client = self.firstgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'firstgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['error']
        self.assertEqual(errormsg, "You can't validate this supervisee as his supervision is still valid")

        logged_out = self.client.logout()

    def test_logged_in_invalid_godfather(self):

        client = self.invalidgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'invalidgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        logged_out = self.client.logout()


#----------------------------------------------------------


class AccountRemoveSuperviseeTestCase(AccountTestCase):

    def setUp(self):
        super(AccountRemoveSuperviseeTestCase, self).setUp()

        self.url = reverse('api_accounts:remove_supervisee', kwargs={'supervisee_name': self.accepteduser.username})

    def test_anonymous_user(self):
        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

    def test_logged_in_accepted_user(self):
        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisor account')

        logged_out = self.client.logout()

    def test_logged_in_rejected_user(self):
        client = self.rejecteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'rejecteduser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisor account')

        logged_out = self.client.logout()

    def test_logged_in_yearrevalidation_user(self):

        client = self.yearrevalidationuser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Not a supervisor account')

        logged_out = self.client.logout()

    def test_logged_in_blocked_user(self):

        client = self.blockeduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'blockeduser')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        logged_out = self.client.logout()

    def test_logged_in_godfather_remove_accepted_user(self):
        #accepted user before
        client = self.accepteduser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)
        self.assertEqual(client.profile.status, Profile.ACCEPTED)
        logged_out = self.client.logout()

        #accepted user removed
        client = self.firstgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'firstgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        logged_out = self.client.logout()

        #accepted user after
        client = User.objects.get(username=self.accepteduser.username)

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'accepteduser')

        new_supervision_key = client.profile.supervision_key
        self.assertEqual(new_supervision_key, None)
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertEqual(client.profile.status, Profile.REJECTED)
        logged_out = self.client.logout()

    def test_logged_in_godfather_remove_yearrevalidation_user(self):
        self.url = reverse('api_accounts:remove_supervisee', kwargs={'supervisee_name': self.yearrevalidationuser.username})

        #yearrevalidationuser user before
        client = self.yearrevalidationuser

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        previous_supervision_key = client.profile.supervision_key
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)

        self.assertEqual(previous_supervision_track.is_valid, True)
        self.assertEqual(client.profile.status, Profile.YEARREVALIDATION)
        logged_out = self.client.logout()

        #yearrevalidationuser user removed
        client = self.firstgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'firstgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        logged_out = self.client.logout()

        #yearrevalidationuser user after
        client = User.objects.get(username=self.yearrevalidationuser.username)

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertTrue(logged_in)
        self.assertEqual(client.username, 'yearrevalidationuser')

        new_supervision_key = client.profile.supervision_key
        self.assertEqual(new_supervision_key, None)
        previous_supervision_track = SupervisionTrack.objects.get(supervision_key = previous_supervision_key)
        self.assertEqual(previous_supervision_track.is_valid, False)
        self.assertEqual(client.profile.status, Profile.REJECTED)
        logged_out = self.client.logout()

    def test_logged_in_invalid_godfather(self):

        client = self.invalidgodfather

        logged_in = self.client.login(username=client.username, password=self.password)
        self.assertFalse(logged_in)
        self.assertEqual(client.username, 'invalidgodfather')

        response = self.client.put(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        content = json.loads(response.content)
        self.assertEqual(len(json.loads(response.content)), 1)
        errormsg = content['detail']
        self.assertEqual(errormsg, 'Authentication credentials were not provided.')

        logged_out = self.client.logout()
