#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from rest_framework import serializers

from ..code.serializers import CodeSerializer, CodeCreationSerializer
from ..libraries.serializers import LibraryReferenceSerializer
from ..dataformats.serializers import ReferencedDataFormatSerializer
from ..attestations.serializers import AttestationSerializer
from ..experiments.serializers import ExperimentSerializer
from ..experiments.models import Experiment

from .models import Algorithm

import simplejson as json

import beat.core.algorithm


#----------------------------------------------------------


class AlgorithmCreationSerializer(CodeCreationSerializer):
    class Meta(CodeCreationSerializer.Meta):
        model = Algorithm
        beat_core_class = beat.core.algorithm


#----------------------------------------------------------


class AlgorithmSerializer(CodeSerializer):
    attestations = AttestationSerializer(many=True)
    result_dataformat = serializers.SerializerMethodField()
    parameters = serializers.SerializerMethodField()
    inputs = serializers.SerializerMethodField()
    outputs = serializers.SerializerMethodField()
    referencing_experiments = serializers.SerializerMethodField()
    referenced_libraries = LibraryReferenceSerializer(many=True)
    referenced_dataformats = ReferencedDataFormatSerializer(many=True, source='all_referenced_dataformats')
    needed_dataformats = ReferencedDataFormatSerializer(many=True, source='all_needed_dataformats')

    class Meta(CodeSerializer.Meta):
        model = Algorithm


    def get_result_dataformat(self, obj):
        return json.loads(obj.result_dataformat) if obj.result_dataformat else None

    def get_parameters(self, obj):
        return json.loads(obj.parameters) if obj.parameters else None

    def get_inputs(self, obj):
        return map(lambda x: {
                  'name': x.name,
                  'dataformat': x.dataformat.fullname(),
                  'channel': x.channel
            }, obj.endpoints.filter(input=True).order_by('name'))

    def get_outputs(self, obj):
        return map(lambda x: {
                        'name': x.name,
                        'dataformat': x.dataformat.fullname(),
                        'channel': x.channel,
                    }, obj.endpoints.filter(input=False).order_by('name'))

    def get_referencing_experiments(self, obj):
        user = self.context.get('user')
        experiments = Experiment.objects.for_user(user, True).filter(blocks=obj.blocks.all()).distinct()
        serializer = ExperimentSerializer(experiments, many=True)
        referencing_experiments = serializer.data

        # Put the pending experiments first
        ordered_result = filter(lambda x: x['creation_date'] is None, \
                referencing_experiments)
        ordered_result += filter(lambda x: x['creation_date'] is not None, \
                referencing_experiments)

        return ordered_result


#----------------------------------------------------------


class FullAlgorithmSerializer(AlgorithmSerializer):

    class Meta(AlgorithmSerializer.Meta):
        default_fields = AlgorithmSerializer.Meta.default_fields + AlgorithmSerializer.Meta.extra_fields
