#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.conf import settings
from django.core.urlresolvers import reverse

from ..experiments.models import Experiment
from ..toolchains.models import Toolchain
from ..algorithms.models import Algorithm
from ..dataformats.models import DataFormat

from .signals import unlocked
from .exceptions import AlreadyUnlockedError

import random
from datetime import datetime, timedelta

class AttestationManager(models.Manager):

    def get_by_natural_key(self, number):
        return self.get(number=number)

    def create_attestation(self, experiment):

        # Generate a unique attestation number
        used_numbers = [attestation.number for attestation in self.model.objects.all()]
        number = 0
        while (number == 0) or number in used_numbers:
            number = random.randint(100000, 2**31)

        creation_date = datetime.now()
        expiration_date = creation_date + timedelta(days=settings.EXPIRATION_DELTA)

        attestation  = self.model(number=number,
                                  experiment=experiment,
                                  toolchain=experiment.toolchain,
                                  locked=True,
                                  creation_date=creation_date,
                                  expiration_date=expiration_date,
                                  publication_date=None)

        attestation.save()
        return attestation

    def locked(self):
        return self.filter(locked=True)

    def published(self):
        return self.filter(locked=False)


class Attestation(models.Model):

    number           = models.IntegerField()
    experiment       = models.OneToOneField(Experiment, related_name='attestation')
    locked           = models.BooleanField(default=True)
    creation_date    = models.DateTimeField()
    publication_date = models.DateTimeField(null=True, blank=True)
    expiration_date = models.DateTimeField(null=True, blank=True)
    toolchain        = models.ForeignKey(Toolchain, related_name='attestations',
                                         null=True, blank=True)
    dataformats      = models.ManyToManyField(DataFormat, related_name='attestations',
                                              blank=True)
    algorithms       = models.ManyToManyField(Algorithm, related_name='attestations',
                                              blank=True)


    objects = AttestationManager()

    def __str__(self):
        return "Attestation #%d, for experiment %s" % (self.number, self.experiment.fullname())

    def natural_key(self):
        return (self.number,)

    def get_absolute_url(self):
        return reverse(
                'attestations:view',
                args=(
                    self.number,
                    ),
                )

    def modifiable(self):
        return not(self.locked) and super(Attestation, self).modifiable()

    def deletable(self):
        return not(self.locked) and super(Attestation, self).deletable()

    def list_of_referenced_formats(self):
        s = '['
        for dataformat in self.dataformats.all():
            s += "'%s'," % dataformat.fullname()

        if s[-1] == ',':
            s = s[:-1]

        return s + ']'

    def share(self, visible_algorithms_names):
        # Build algorithms sharing information
        own_needed_algorithms = self.experiment.referenced_algorithms.filter(author=self.experiment.author)

        algorithms_infos = {}

        visible_algorithms = filter(lambda x: x.fullname() in visible_algorithms_names,
                                    own_needed_algorithms)
        for visible_algorithm in visible_algorithms:
            algorithms_infos[visible_algorithm.fullname()] = {'opensource': False}

        public_algorithms = filter(lambda x: x.fullname() not in visible_algorithms_names,
                                   own_needed_algorithms)
        for public_algorithm in public_algorithms:
            algorithms_infos[public_algorithm.fullname()] = {'opensource': True}

        # Share the experiment
        self.experiment.share(algorithms_infos=algorithms_infos)

    def unlock(self, visible_algorithms_names):
        if self.locked:
            # Share the experiment
            self.share(visible_algorithms_names=visible_algorithms_names)
            # Unlock the attestation
            self.locked           = False
            self.publication_date = datetime.now()
            self.save()
            unlocked.send(self)
        else:
            raise AlreadyUnlockedError
