/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/**
 * beat.attestations.dialogs.js
 * Implementation of attestation-related dialogs
 */
if (beat === undefined) var beat = {};
if (beat.attestations === undefined) beat.attestations = {};
if (beat.attestations.dialogs === undefined) beat.attestations.dialogs = {};

/**
 * Implements a modal dialog based on Bootstrap3 that allows the user to unlock
 * the current attestation. The user is presented with a clone of the above
 * form in modal window and decides to accept, select/unselect private
 * algorithms for open-sourcing and clicking the "Ok" button.
 *
 * Parameters:
 *
 *   number (int): Number of the attestation to be unlocked
 *   panel_id (str): The DOM identity of the panel containing my forms
 *   api_url (str): The URL towards the unlocking API
 *   redirect_url (str): The URL where to redirect the user if the operation is
 *       successful
 */
beat.attestations.dialogs.modal_unlock = function(number, panel_id, api_url, redirect_url) {

  var panel = $('#' + panel_id).clone();
  panel.show();

  var dialog = new BootstrapDialog({
    title: '<i class="fa fa-warning"></i> Confirm unlocking of attestation #' + number,
    message: panel,
    type: BootstrapDialog.TYPE_PRIMARY,
    buttons: [
    {
      label: 'Cancel',
      cssClass: 'btn-default',
      action: function(the_dialog) {
        the_dialog.close();
      }
    },
    {
      label: 'Unlock',
      cssClass: 'btn-primary',
      action: function(the_dialog) {
        $.ajaxSetup({
          beforeSend: function(xhr, settings) {
            var csrftoken = $.cookie('csrftoken');
            xhr.setRequestHeader('X-CSRFToken', csrftoken);
          }
        });

        var promise = $.ajax({
            type: 'POST',
            url: api_url,
            data: JSON.stringify({
              visible_algorithms: panel.find('input:not(:checked)').map(function () { return this.value; }).get(),
            }),
            dataType: 'json',
            contentType: 'application/json; charset=utf-8',
        });

        promise.done(function(data) {
          the_dialog.close();
          window.location.href = redirect_url;
        });

        promise.fail(function(jqXHR, textStatus, errorThrown) {
          if ((jqXHR.status == 403) && (jqXHR.responseText.length > 0))
            alert(jqXHR.responseText);
          else
            alert('Error: ' + errorThrown);
          the_dialog.close();
        });

        the_dialog.close();
      }
    },
    ],
  });
  dialog.realize();
  dialog.open();
}


/**
 * Uses bootstrap3 to display a modal window to delete an attestation.
 * Attestation URLs are differnt from other contributions, unfortunately.
 *
 * This method will display an alert box if a problem occurs with the deletion,
 * or will just redirect the client to the `redirect` URL in case of success.
 *
 * Parameters:
 *
 * name (int): The number of the attestation to delete
 * url (str): The URL where to post the DELETE query (e.g.
 *   "/api/v1/attestations/12345678999/")
 * redirect (str): The URL where to redirect the client after the operation
 *   finished (e.g. "/attestations/user/")
 **/
beat.attestations.dialogs.modal_delete = function(name, url, redirect) {

  var dialog = new BootstrapDialog({
    title: '<i class="fa fa-warning"></i> Confirm deletion of attestation #' +
      name,
    message: 'Once you delete this attestation there is no going back. Be certain.',
    type: BootstrapDialog.TYPE_PRIMARY,
    buttons: [
    {
      label: 'Cancel',
      cssClass: 'btn-default',
      action: function(the_dialog) {
        the_dialog.close();
      }
    },
    {
      label: 'OK',
      cssClass: 'btn-primary',
      action: function(the_dialog) {
        $.ajaxSetup({
          beforeSend: function(xhr, settings) {
            var csrftoken = $.cookie('csrftoken');
            xhr.setRequestHeader('X-CSRFToken', csrftoken);
          }
        });
        var promise = $.ajax({type: "DELETE", url: url});

        promise.done(function(data) {
            the_dialog.close();
            window.location.href = redirect;
        });

        promise.fail(function(jqXHR, textStatus, errorThrown) {
            if ((jqXHR.status == 403) && (jqXHR.responseText.length > 0))
              alert(jqXHR.responseText);
            else
              alert('Error: ' + errorThrown);
            the_dialog.close();
        });

        the_dialog.close();
      }
    },
    ],
    });
  dialog.realize();
  dialog.open();

}
