#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import time
import signal
import datetime
import operator
import traceback
import subprocess

import logging
logger = logging.getLogger(__name__)

import psutil
import simplejson

from django.db import utils
from django.db import models
from django.db import transaction
from django.conf import settings
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.models import Group

from guardian.shortcuts import get_perms

import beat.core.stats
import beat.core.data
import beat.core.execution
from beat.core.dock import Host

from ..code.models import Code
from ..common.models import Shareable, ShareableManager
from ..common.texts import Messages
from ..statistics.utils import updateStatistics


#----------------------------------------------------------


class EnvironmentManager(ShareableManager):

    def get_by_natural_key(self, name, version):
        return self.get(name=name, version=version)

    def get_by_fullname(self, fullname):
        name, version = fullname.rsplit(' ', 1)
        return self.get_by_natural_key(name, version[1:-1])


class Environment(Shareable):
    """Defines a software environment to run algorithms"""

    name = models.CharField(
        max_length=200,
        help_text=Messages['name'],
        )

    version = models.CharField(
        max_length=20,
        help_text='Free-style version for this environment (normally read from the Worker/Scheduler available environments)',
        )

    short_description = models.CharField(
        max_length=100,
        default='',
        blank=True,
        help_text=Messages['short_description'],
        )

    description = models.TextField(
        default='',
        blank=True,
        help_text=Messages['description'],
        )

    creation_date = models.DateTimeField(
        'Creation date',
        auto_now_add = True,
        )

    active = models.BooleanField(
        default=True,
        help_text='If this environment can be used in experiments',
        )

    previous_version = models.ForeignKey('self',
        related_name='next_versions',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        )

    objects = EnvironmentManager()

    #_____ Meta parameters __________

    class Meta:
        unique_together = ('name', 'version')

    #_____ Overrides __________

    def __str__(self):
        return self.fullname()


    def natural_key(self):
        return (self.name, self.version)


    #_____ Utilities __________

    def fullname(self):
        return '%s (%s)' % (self.name, self.version)

    def get_absolute_url(self):

        return reverse(
            'backend:view-environment',
            args=(self.name, self.version,),
            )

    def get_admin_change_url(self):
        return reverse('admin:backend_environment_change', args=(self.id,))

    def queues_for(self, user):
        """Returns all queues associated to this environment for which the user
        has the 'can_access' permission"""

        return [q for q in self.queues.all() if user.has_perm('backend.can_access', q)]


    def as_dict(self):
        '''Returns a representation as a dictionary'''

        return dict(
            name=self.name,
            version=self.version,
            short_description=self.short_description,
            description=self.description,
            )


class EnvironmentLanguage(models.Model):

    environment = models.ForeignKey(Environment,
        related_name='languages'
        )

    language = models.CharField(max_length=1, choices=Code.CODE_LANGUAGE,
                                default=Code.PYTHON)


#----------------------------------------------------------


def _cleanup_zombies():
    '''Cleans-up eventual zombie subprocesses launched by the worker'''

    for child in psutil.Process().children(recursive=True):
        try:
            if child.status() == psutil.STATUS_ZOMBIE:
                child.wait()
        except psutil.NoSuchProcess:
            # process is gone meanwhile, which is ok
            pass


class WorkerManager(models.Manager):

    def get_by_natural_key(self, name):
        return self.get(name=name)


class Worker(models.Model):

    name = models.CharField(
        max_length=100,
        help_text=Messages['name'],
        unique=True,
        )

    active = models.BooleanField(
        help_text=u'If this worker is usable presently',
        default=False,
        )

    update = models.BooleanField(
        help_text=u'If this worker state must be updated at the next cycle',
        default=False,
        )

    updated = models.DateTimeField(null=True, auto_now=True)

    cores = models.PositiveIntegerField()

    memory = models.PositiveIntegerField(default=0, help_text='In megabytes')

    used_cores = models.PositiveIntegerField(default=0, help_text='In %')

    used_memory = models.PositiveIntegerField(default=0, help_text='In %')

    info = models.TextField(null=True, blank=True,
        help_text='Informative message from the worker')

    objects = WorkerManager()


    #_____ Overrides __________


    def __str__(self):
        retval = '%s (%d cores, %d Mb)' % (self.name, self.cores, self.memory)
        if not self.active: retval += ' [INACTIVE]'
        return retval


    def natural_key(self):
        return (self.name,)


    def get_admin_change_url(self):
        return reverse('admin:backend_worker_change', args=(self.id,))


    def load(self):
        '''Calculates the number of cores in use or to be used in the future'''
        return sum([j.job.block.queue.cores_per_slot for j in self.splits.all()])


    def current_load(self):
        '''Calculates the number of cores being used currently'''
        return sum([j.job.block.queue.cores_per_slot for j in self.splits.filter(status=Job.PROCESSING)])


    def available_cores(self):
        '''Calculates the number of available cores considering current load'''

        return max(self.cores - self.load(), 0)


    def deactivate(self, reason):
        '''Deactivates the current worker for a reason, that is registered'''

        self.info = reason
        self.active = False


    def activate(self, reason=None):
        '''Reactivates the worker, deletes any associated information'''

        self.info = reason
        self.active = True


    def as_dict(self):
        '''Returns a dictionary-like representation'''

        return dict(cores=self.cores, memory=self.memory)


    def check_environments(self, environments):
        '''Checks that this worker has access to all environments it needs

        This method will check if the found set of environments (in the
        dictionary ``environments``) contains, at least, one environment for
        each environment object this worker is supposed to be able to execute
        user algorithms for.


        Parameters:

          environments (dict): A dictionary of environments found by using
          :py:func:`utils.find_environments` in which, keys represent the
          natural keys of Django database environments.


        Returns:

          list: A list of missing environments this worker can be assigned to
              work with, but where not found

          list: A list of unused environments this worker cannot be assigned to
              work with, but where nevertheless found

        '''

        slots = Slot.objects.filter(worker=self)
        queues = Queue.objects.filter(slots__in=slots)
        wishlist = Environment.objects.filter(queues__in=queues, active=True)
        wishlist = wishlist.order_by('id').distinct()

        required = [k.fullname() for k in wishlist]
        missing = [k for k in required if k not in environments]
        unused = [k for k in environments if k not in required]

        return missing, unused


    def update_state(self):
        '''Updates state on the database based on current machine readings'''

        # check I have at least all cores and memory I'm supposed to have
        cores = psutil.cpu_count()
        ram = psutil.virtual_memory().total/(1024*1024)
        self.info = ''

        if cores < self.cores:
            logger.warn("Worker `%s' only has %d cores which is less then " \
                "the value declared on the database - it's not a problem, " \
                "but note this self may get overloaded", self, cores)
            self.info += 'only %d cores;' % cores

        if ram < self.memory:
            logger.warn("Worker `%s' only has %d Mb of RAM which is less " \
                "then the value declared on the database - it's not a " \
                "problem, but note this self may get overloaded", self,
                ram)
            self.info += 'only %d Mb of RAM;' % ram

        with transaction.atomic():
            self_ = Worker.objects.select_for_update().get(pk=self.pk) #lock

            # update process and memory usage
            self.used_cores = int(psutil.cpu_percent())
            self.used_memory = int(psutil.virtual_memory().percent)

            # save current self state
            self.active = True
            self.update = False
            self.save()


    def terminate(self):
        '''Cleanly terminates a particular worker at the database

        .. note::

           This method does not destroy running or assigned processes that may
           be running or assigned to this worker. This is implemented in this
           way to allow for a clean replacement of the worker program w/o an
           interruption of the backend service.

        '''

        # disables worker, so no more splits can be assigned to it
        with transaction.atomic():
            self_ = Worker.objects.select_for_update().get(pk=self.pk)
            self_.active = False
            self_.used_cores = 0
            self_.used_memory = 0
            self_.info = 'Worker deactivated by system administrator'
            self_.save()

        # cancel job splits which should be cancelled anyways
        for j in JobSplit.objects.filter(worker=self, status=Job.CANCEL,
            end_date__isnull=True, process_id__isnull=False):
            if psutil.pid_exists(j.process_id):
                os.kill(j.process_id, signal.SIGTERM)

        # cleans-up zombie processes that may linger
        _cleanup_zombies()


    def shutdown(self):
        '''Removes all running/assigned jobs from the queue, shuts down

        This method should be used with care as it may potentially cancel all
        assigned splits for the current worker.

        '''

        self.terminate()

        message = 'Cancelled on forced worker shutdown (maintenance)' \
            ' - you may retry submitting your experiment shortly'

        # cancel job splits which were not yet started
        for j in JobSplit.objects.filter(worker=self, status=Job.QUEUED,
            start_date__isnull=True, process_id__isnull=True):
            j.end(Result(status=1, usrerr=message))

        # cancel job splits which are running
        for j in JobSplit.objects.filter(worker=self, status=Job.PROCESSING,
            end_date__isnull=True, process_id__isnull=False):
            j._cancel()



    def work(self, environments, process):
        '''Launches user code on isolated processes

        This function is supposed to be called asynchronously, by a
        scheduled agent, every few seconds. It examines job splits assigned
        to the current host and launches an individual process to handle
        these splits. The process is started locally and the process ID
        stored with the split.

        Job split cancelling is executed by setting the split state as
        ``CANCEL`` and waiting for this function to handle it.


        Parameters:

          environments (dict): A dictionary containing installed
            environments, their description and execute-file paths.

          process (str): The path to the ``process.py`` program to use for
            running the user code on isolated processes.

        '''

        # refresh state from database and update state if required
        self.refresh_from_db()
        if self.update: self.update_state()

        # cancel job splits by killing associated processes
        for j in JobSplit.objects.filter(worker=self, status=Job.CANCEL,
            end_date__isnull=True):
            if j.process_id is not None and psutil.pid_exists(j.process_id):
                os.kill(j.process_id, signal.SIGTERM)
            else: # process went away without any apparent reason
                with transaction.atomic():
                    message = "Split %d/%d running at worker `%s' for " \
                        "block `%s' of experiment `%s' finished before " \
                        "even starting. Force-cancelling job split at " \
                        "database..." % (j.split_index+1,
                            j.job.block.required_slots,
                            self,
                            j.job.block.name,
                            j.job.block.experiment.fullname(),
                            )
                    logger.error(message)
                    j.end(Result(status=1, usrerr=settings.DEFAULT_USER_ERROR))

        # cmdline base argument
        cmdline = [process]
        if settings.DEBUG:
            cmdline += ['-vv']
        else:
            cmdline += ['-v']

        # start newly assigned job splits
        with transaction.atomic():
            splits = JobSplit.objects.select_for_update().filter(worker=self,
                status=Job.QUEUED, start_date__isnull=True,
                process_id__isnull=True)
            for split in splits:
                # if we get to this point, then we launch the user process
                # -> see settings.WORKER_DETACH_CHILDREN for more info
                kwargs = dict()
                if settings.WORKER_DETACH_CHILDREN:
                    kwargs['preexec_fn'] = os.setpgrp
                subprocess.Popen(cmdline + [str(split.pk)], **kwargs)
                split.status = Job.PROCESSING #avoids re-running
                split.save()

        # cleans-up zombie processes that may linger
        _cleanup_zombies()


    def __enter__(self):
        self.update_state()
        return self


    def __exit__(self, *exc):
        self.terminate()
        return False #propagate exceptions


#----------------------------------------------------------


class QueueManager(models.Manager):

    def get_by_natural_key(self, name):
        return self.get(name=name)


class Queue(models.Model):

    name = models.CharField(max_length=100, help_text=Messages['name'],
        unique=True)

    memory_limit = models.PositiveIntegerField(help_text='In megabytes')

    time_limit = models.PositiveIntegerField(help_text='In minutes')

    cores_per_slot = models.PositiveIntegerField()

    max_slots_per_user = models.PositiveIntegerField()

    environments = models.ManyToManyField(
        Environment,
        related_name='queues',
        )

    objects = QueueManager()

    #_____ Meta parameters __________

    class Meta:
        permissions = [
            ['can_access', _('Can access queue')]
        ]


    #_____ Overrides __________

    def __str__(self):
        return '%s (%d minutes, %d megabytes, %d cores per slot, %d slots max per user)' % (
            self.name,
            self.time_limit,
            self.memory_limit,
            self.cores_per_slot,
            self.max_slots_per_user
            )


    def natural_key(self):
        return (self.name,)


    def get_admin_change_url(self):
        return reverse('admin:backend_queue_change', args=(self.id,))

    #_____ Utilities __________

    def number_of_slots(self):
        '''Total number of slots considering all assigned worker/slots'''

        r = self.slots.filter(worker__active=True)
        return r.aggregate(nslots=models.Sum('quantity'))['nslots'] or 0


    def availability(self):
        '''Returns the availability for this queue in terms of number of slots

        This method does not take into consideration the occupation of this
        queue slots caused by jobs on other queues. It only looks to its inner
        occupancy and reports on that.

        Returns an integer between 0 and :py:meth:`Queue.slots`.
        '''

        running = JobSplit.objects.filter(job__block__in=self.blocks.all(),
            status=Job.PROCESSING).count()
        return max(self.number_of_slots() - running, 0)


    def worker_availability(self):
        '''Returns an ordered dictionary indicating the availability of workers
           according to their queue priority.

        The dictionary contains, as value, the number of slots available per
        worker

        The order of workers is sorted by:

          * slot priority (the higher, the better)
          * load (the lower, the better)
          * name (alphabetically)

        '''

        workers = [(k.worker, -k.priority, k.worker.load(), k.worker.name) \
            for k in self.slots.filter(worker__active=True)]

        workers = sorted(workers, key=operator.itemgetter(1,2,3))

        return [w[0] for w in workers]


    def splits(self):
        '''Lists all job splits currently associated to this queue'''

        return JobSplit.objects.filter(job__block__queue=self)


    def as_dict(self):
        '''Returns a representation as a dictionary'''

        return {
            'memory-limit': self.memory_limit,
            'time-limit': self.time_limit,
            'cores-per-slot': self.cores_per_slot,
            'max-slots-per-user': self.max_slots_per_user,
            'environments': [k.fullname() for k in self.environments.all()],
            'slots': dict([(s.worker.name, dict(quantity=s.quantity,
              priority=s.priority)) for s in self.slots.all()]),
            'groups': [k.name for k in Group.objects.all() if 'can_access' in get_perms(k, self)]
            }


class SlotManager(models.Manager):

    def get_by_natural_key(self, queue_name, worker_name):
        return self.get(queue__name=queue_name, worker__name=worker_name)


class Slot(models.Model):

    queue = models.ForeignKey(Queue, related_name='slots',
        on_delete=models.CASCADE)

    worker = models.ForeignKey(Worker, related_name='slots',
        on_delete=models.CASCADE)

    quantity = models.PositiveIntegerField(
        'Number of slots',
        help_text='Number of processing slots to dedicate in this worker for a given queue'
        )

    priority = models.PositiveIntegerField(
        default=0,
        help_text='Priority of these slots on the defined queue'
        )

    objects = SlotManager()

    #_____ Meta parameters __________

    class Meta:
        unique_together = ('queue', 'worker')

    #_____ Overrides __________

    def __str__(self):
        return '%s - %s (slots: %d, priority: %d)' % (self.queue, self.worker, self.quantity, self.priority)


    def natural_key(self):
        return (self.queue.name, self.worker.name)


#----------------------------------------------------------


def _merge_strings(s):
    if len(s) == 1: return s[0]
    s = [k.strip() for k in s]
    if any(s):
        return '\n'.join(['Process %d: %s' % (i,k) for i,k in enumerate(s)])
    else:
        return ''


#----------------------------------------------------------


class Result(models.Model):
    '''Logging and status information concerning block or job execution.
    '''

    # exit status code
    status = models.IntegerField()
    stdout = models.TextField(null=True, blank=True)
    stderr = models.TextField(null=True, blank=True)
    usrerr = models.TextField(null=True, blank=True)
    _stats = models.TextField(null=True, blank=True)
    timed_out = models.BooleanField(default=False)
    cancelled = models.BooleanField(default=False)


    def __str__(self):
        status = 'success' if self.status == 0 else 'failed'
        retval = 'Result(%s' % status
        if self.stdout: retval += ', stdout=' + self.stdout
        if self.stderr: retval += ', stderr=' + self.stderr
        if self.usrerr: retval += ', usrerr=' + self.usrerr
        retval += ')'
        return retval


    def _get_stats(self):
        if self._stats is not None:
            return beat.core.stats.Statistics(simplejson.loads(self._stats))
        else:
            return beat.core.stats.Statistics()


    def _set_stats(self, v):
        self._stats = simplejson.dumps(v.as_dict())


    stats = property(_get_stats, _set_stats)


#----------------------------------------------------------


class Job(models.Model):
    '''Class describing the execution of a Job on the backend'''

    QUEUED     = 'N' #Block.PENDING
    PROCESSING = 'P' #Block.PROCESSING
    COMPLETED  = 'C' #Block.COMPLETED
    FAILED     = 'F' #Block.FAILED
    SKIPPED    = 'S' #Block.SKIPPED
    CANCELLED  = 'L' #Block.CANCELLED
    CANCEL     = 'K' #Job was asked to be killed

    STATUS = (
        (QUEUED,     'Queued'),
        (PROCESSING, 'Processing'),
        (COMPLETED,  'Completed'),
        (FAILED,     'Failed'),
        (SKIPPED,    'Skipped'),
        (CANCELLED,  'Cancelled'),
        (CANCEL,     'Cancel'),
    )

    block = models.OneToOneField('experiments.Block', null=True,
        on_delete=models.CASCADE, related_name='job')

    status = models.CharField(max_length=1, choices=STATUS, default=QUEUED)

    result = models.OneToOneField(Result, null=True, on_delete=models.CASCADE,
        related_name='job')

    runnable_date = models.DateTimeField(null=True, blank=True)

    start_date = models.DateTimeField(null=True, blank=True)

    end_date = models.DateTimeField(null=True, blank=True)

    parent = models.OneToOneField('self', related_name='child', null=True,
        on_delete=models.SET_NULL)

    split_errors = models.PositiveIntegerField(default=0)


    def _get_child(self):
        return self.child if hasattr(self, 'child') else None

    def _set_child(self, val):
        val.parent = self
        val.save()

    child_ = property(_get_child, _set_child)


    def __str__(self):

        return "Job(%s, %s, splits=%d, status=%s, cores=%d)" % \
            (self.block.name, self.block.experiment.name,
                self.block.required_slots, self.status,
                self.block.queue.cores_per_slot)


    def done(self):
        '''Says whether the job has finished or not'''

        return self.status in (Job.COMPLETED, Job.SKIPPED, Job.FAILED,
            Job.CANCELLED)



    def _copy(self, other):
        '''Copy state from another block'''

        self.refresh_from_db()

        if self.done(): return

        self.start_date = other.start_date
        self.end_date = other.end_date
        self.status = other.status

        # update status of parent jobs
        self.save()
        self._cascade_updates()
        self.block._update_state(None)


    def _make_runnable(self):
        '''Tries to make this job runnable - if it is cached, we skip it'''

        # lock self - avoids concurrent update from scheduler/worker subsystem
        self_ = Job.objects.select_for_update().get(pk=self.pk)

        # checks for the presence of output caches - if they exist and
        # checksum, skip and update related blocks
        if all([k.exists() for k in self.block.outputs.all()]):
            if all([k.check_checksums() for k in self.block.outputs.all()]):
                self.status = Job.SKIPPED
                self.split_errors = 0
                self.end_date = datetime.datetime.now()
                self.save()
                self._cascade_updates() #to similar blocks
                self.block._update_state()
                return
            else:
                logger.warning("Trying to make block `%s' runnable, but " \
                    "indexes do not checksum - waiting...", self.block)
                self.split_errors += 1
                self.save()
                return

        # else, flag it as runnable
        self.runnable_date = datetime.datetime.now()

        # runs index splitting once, for all created splits
        self._split_indices()


    def _split_indices(self):
        '''Runs the index splitting machinery once for all associated splits'''

        # no index spliting is required
        if self.block.required_slots == 1:
            self.save()
            s = JobSplit(job=self, split_index=0)
            s.save()
            return

        indices = []

        conf = simplejson.loads(self.block.command)

        try:

            # For all synchronized inputs with the current block, append the
            # list of generated object indices. This is necessary for an
            # informed decision on where to split the processing
            sync = [conf['inputs'][i] for i in conf['inputs'] if conf['inputs'][i]['channel']==conf['channel']]
            for i in sync:
                indices.append(beat.core.data.load_data_index(
                  settings.CACHE_ROOT, str(i['path'])))

            # Determine N splits using the possible indices for split:
            indices = beat.core.data.foundSplitRanges(indices,
                self.block.required_slots)

            self.split_errors = 0
            self.save()

            if len(indices) == 0:
                message = "Index splitting for block `%s' of experiment " \
                    "`%s' could not be completed: not splittable!" % \
                    (self.block.name, self.block.experiment.fullname())
                logger.error(message)
                self._cancel(usrerr=settings.DEFAULT_USER_ERROR)

            # if you get to this point, the splitting has succeeded,
            # create the necessary splits and assign the ranges
            for i, s in enumerate(indices):
                split_indices = indices[i]
                s = JobSplit(
                    job=self,
                    split_index=i,
                    start_index=split_indices[0],
                    end_index=split_indices[1],
                    )
                s.save()

        except Exception as e:

            self.split_errors += 1
            self.save()

            if self.split_errors > settings.MAXIMUM_SPLIT_ERRORS: #stop
                message = "Index splitting for block `%s' of experiment " \
                    "`%s' could not be completed due to an index split " \
                    "error: %s" % (self.block.name,
                        self.block.experiment.fullname(),
                        traceback.format_exc())
                logger.error(message)
                self._cancel(usrerr=settings.DEFAULT_USER_ERROR)


    def _cascade_updates(self):
        '''Cascade updates to children before I'm deleted.
        '''

        if hasattr(self, 'child'):
            if self.status == Job.CANCELLED:
                if self.parent: #I have a parent, so must give to child
                    parent = self.parent
                    self.parent = None
                    self.child.parent = parent
                else: #child is the new parent
                    child = self.child
                    self.child.parent = None
                    # does this unblock the child to run?
                    if child.block.is_runnable(): child._make_runnable()

            else:
                self.child._copy(self)

        if self.parent and self.status == Job.CANCELLED:
            self.parent = None


    def _update_state(self):
        '''Update self state based on associated job states

        This method is not part of the Job's public API. It is supposed to be
        called by children splits or itself.
        '''

        # lock
        self_ = Job.objects.select_for_update().get(pk=self.pk)

        if self_.done(): return

        # If this process has a parent, then don't try to get split
        # statuses
        if not self.parent:

            split_statuses = self.splits.values_list('status', flat=True)

            if self.start_date is None:
                qs = self.splits.filter(start_date__isnull=False).\
                    order_by('start_date')
                if qs:
                    self.start_date = qs.first().start_date
                else:
                    self.start_date = datetime.datetime.now()

            # Process main status and status from job results
            if Job.FAILED in split_statuses:
                self.status = Job.FAILED

            elif Job.CANCELLED in split_statuses:
                self.status = Job.CANCELLED

            elif (Job.PROCESSING in split_statuses) or \
                (Job.QUEUED in split_statuses and \
                Job.COMPLETED in split_statuses) or \
                (Job.CANCEL in split_statuses):
                self.status = Job.PROCESSING

            elif all([s == Job.SKIPPED for s in split_statuses]):
                self.status = Job.SKIPPED

            elif Job.QUEUED not in split_statuses:
                self.status = Job.COMPLETED

            else:
                self.status = Job.QUEUED

        # if required, erase dangling files, update own results
        timings = None
        if self.done() and self.status != Job.CANCELLED:
            # compute final timings and update parent block
            if self.status != Job.SKIPPED:
                diff_timings = self._merge_results()
                # delete all splits w/o results (still queued)
                self.splits.filter(result__isnull=True).delete()
                self.end_date = self.splits.order_by('-end_date').\
                    first().end_date
                updateStatistics(self.result.stats, self.end_date)
                Result.objects.filter(split__in=self.splits.all()).delete()
                seqtime = sum(diff_timings)
                if self.start_date is None:
                    queuing = 0
                else:
                    queuing = (self.start_date - \
                        self.runnable_date).total_seconds()
                if not seqtime:
                    speed_up_real = 1.0
                    speed_up_maximal = 1.0
                else:
                    speed_up_real = float(seqtime) / \
                        (self.end_date - self.start_date).total_seconds()
                    speed_up_maximal = float(seqtime) / max(diff_timings)
                timings = dict(
                    queuing = queuing,
                    linear_execution = seqtime,
                    speed_up_real = speed_up_real,
                    speed_up_maximal = speed_up_maximal,
                    )
                self.runnable_date = None
                self.erase_dangling_files()

        # updates the dependents and child state
        self.save()

        self._cascade_updates()
        self.block._update_state(timings)


    def erase_dangling_files(self):
        '''Erase dangling files produced by this job in case of errors'''

        l = []
        failed = self.status in (Job.FAILED, Job.CANCELLED)

        if failed:
            for o in self.block.outputs.all(): l += o.files()

        for f in l:
          logger.info("Erasing output file `%s' because Job `%s' failed", f,
              self)
          os.remove(f)


    def _cancel(self, usrerr=None):
        '''Cancel the execution of this job

        As a consequence: delete all associated jobs, mark end_date and set
        cancelled state. This method should only be called by the owning Block.
        '''

        # lock
        self_ = Job.objects.select_for_update().get(pk=self.pk)

        if self_.done(): return

        logger.info("Marking job `%s' as 'cancelled'", self)
        self.runnable_date = None
        self.start_date = None
        if self.splits.count():
            for s in self.splits.all(): s._cancel()
        else:
            self.status = Job.CANCELLED
            if usrerr:
                r = Result(status=1, usrerr=usrerr)
                r.save()
                self.result = r
            self.save()
            self.block._update_state()
            self._cascade_updates()


    def _merge_results(self):
        '''Merge results from jobs, if any exist'''

        # update results
        job_results = Result.objects.filter(pk__in=self.splits.filter(result__isnull=False).values_list('result', flat=True))

        diff_timings = [(k[0]-k[1]).total_seconds() for k in \
            self.splits.filter(end_date__isnull=False,
              start_date__isnull=False).values_list('end_date', 'start_date')]

        status = sum([k.status for k in job_results])
        stdout = _merge_strings([k.stdout for k in job_results])
        stderr = _merge_strings([k.stderr for k in job_results])
        usrerr = _merge_strings([k.usrerr for k in job_results])

        # merge beat.core statistics
        if job_results:
          stats = job_results[0].stats
          for k in job_results[1:]: stats += k.stats
          stats = stats
        else:
          stats = beat.core.stats.Statistics()

        cancelled = any([k.cancelled for k in job_results])
        timed_out = any([k.timed_out for k in job_results])

        r = Result(status=status, stdout=stdout, stderr=stderr, usrerr=usrerr,
            timed_out=timed_out, cancelled=cancelled)
        r.stats = stats
        r.save()
        self.result = r

        return diff_timings


#----------------------------------------------------------


class JobSplit(models.Model):
    '''Class describing a part of job of an experiment'''

    worker = models.ForeignKey(Worker, null=True, on_delete=models.SET_NULL,
        related_name='splits')

    job = models.ForeignKey(Job, null=True, on_delete=models.CASCADE,
        related_name='splits')

    split_index = models.PositiveIntegerField()

    start_index = models.PositiveIntegerField(null=True)

    end_index = models.PositiveIntegerField(null=True)

    cache_errors = models.PositiveIntegerField(default=0)

    status = models.CharField(max_length=1, choices=Job.STATUS,
        default=Job.QUEUED)

    result = models.OneToOneField(Result, null=True, on_delete=models.CASCADE,
        related_name='split')

    start_date = models.DateTimeField(null=True)

    end_date = models.DateTimeField(null=True)

    process_id = models.PositiveIntegerField(null=True)


    host = None


    class Meta:
        unique_together = ('job', 'split_index')


    def __str__(self):

        return "JobSplit(%s, index=%d, state=%s)%s" % \
            (self.job, self.split_index, self.status,
                ('@%s' % self.worker) if self.worker else '')


    def done(self):
        '''Says whether the job has finished or not'''

        return self.status in (
            Job.COMPLETED,
            Job.SKIPPED,
            Job.FAILED,
            Job.CANCELLED,
            )


    @transaction.atomic
    def schedule(self, worker):
        '''Schedules this split to be executed on a given worker

        Parameters:

            worker (:py:class:Worker): The worker this job split was actually
              submitted to, if there is one.

        '''

        # lock self - avoids concurrent update from scheduler/worker
        # subsystem
        self_ = JobSplit.objects.select_for_update().get(pk=self.pk)
        worker_ = Worker.objects.select_for_update().get(pk=worker.pk)

        self.worker = worker
        self.save()

        logger.info("Job split %s scheduled at `%s' was assigned to `%s'",
            self, self.job.block.queue, self.worker)


    def signal_io_error(self):
        '''Marks the split as having an IOError (cache sync issues, likely)
        '''

        tries = 0

        while True:

            try:

                with transaction.atomic():

                    # lock self - avoids concurrent update from
                    # scheduler/worker subsystem
                    self_ = JobSplit.objects.select_for_update().get(pk=self.pk)

                    if self_.start_date is not None: return

                    self.cache_errors += 1
                    self.status = Job.QUEUED
                    self.save()

                    break

            except utils.OperationalError:
                tries += 1
                if tries > settings.MAXIMUM_SPLIT_SAVE_RETRIES:
                    raise
                else:
                  logger.info("Database error caught starting `%s': retrying " \
                      "in 1 second (%d/%d)...", self, tries,
                      settings.MAXIMUM_SPLIT_SAVE_RETRIES)
                  # wait a second and retry
                  time.sleep(1)


    def start(self):
        '''Marks the job as started, acknowledging scheduler assignment

        Once this function is called, a second call no longer alters anything.
        '''

        tries = 0

        while True:

            try:

                with transaction.atomic():

                    # lock self - avoids concurrent update from
                    # scheduler/worker subsystem
                    self_ = JobSplit.objects.select_for_update().get(pk=self.pk)

                    if self_.start_date is not None: return

                    self.start_date = datetime.datetime.now()
                    self.process_id = os.getpid()

                    self.status = Job.PROCESSING

                    self.save()

                    logger.info("Job split `%s' was just started.", self)

                    self.job._update_state()

                    break

            except utils.OperationalError:
                tries += 1
                if tries > settings.MAXIMUM_SPLIT_SAVE_RETRIES:
                    raise
                else:
                  logger.info("Database error caught starting `%s': retrying " \
                      "in 1 second (%d/%d)...", self, tries,
                      settings.MAXIMUM_SPLIT_SAVE_RETRIES)
                  # wait a second and retry
                  time.sleep(1)


    def _cancel(self):
        '''Marks this job as cancelled.'''

        # If this split is running, then wait
        if self.status == Job.PROCESSING:
            with transaction.atomic():
                # lock self - avoids concurrent update from scheduler/worker
                # subsystem
                self_ = JobSplit.objects.select_for_update().get(pk=self.pk)
                self.status = Job.CANCEL
                self.save()

            logger.info("Job split `%s' is currently processing. Waiting " \
                "for worker to cancel split remotely.", self)

        else: #just end it
            self.end(None, Job.CANCELLED)


    def end(self, result, status=None):
        '''Marks the job as finished on the state database

        Disassociates the worker from this job. Once this function is called, a
        second call no longer alters anything.


        Parameters:

        result (:py:class:`Result`): The result of the task

        status (str): One of the possible (single character) Job statuses, in
          case ``result`` is not provided. Notice that, if ``result`` is
          provided, this variable is **ignored** and the state
          (``Job.COMPLETED`` or ``Job.FAILED``) is filled in from
          ``result.status``. A ``result.status`` of 0 (zero) indicates a
          successful task (set job status to ``Job.COMPLETED``), whereas if the
          status is different than zero, the job status is set to
          ``Job.FAILED``.

        '''

        tries = 0

        while True:

            try:

                with transaction.atomic():

                    # lock self - avoids concurrent update from
                    # scheduler/worker subsystem
                    self_ = JobSplit.objects.select_for_update().get(pk=self.pk)

                    if self_.done(): return

                    if status:
                        logger.info("Marking job split `%s' as '%s'", self,
                            status)

                    if self.start_date is None:
                        self.start_date = datetime.datetime.now()
                    self.end_date = datetime.datetime.now()
                    self.worker = None #frees worker slot

                    if result:
                        # special condition to handle cancelled jobs
                        # which should be marked as cancelled, unless
                        # they have finished before I could act on them
                        # in those, cases, preserve the result as the
                        # caches are already setup
                        if result.status != 0 and self_.status == Job.CANCEL:
                            self.status = Job.CANCELLED
                            if result.id is not None: result.delete()
                        else:
                            self.status = Job.COMPLETED if \
                                result.status == 0 else Job.FAILED
                            if result.status in (-15, 15, -9, 9):
                                # job received a term/kill signal
                                if not result.usrerr:
                                    result.usrerr = 'User process was ' \
                                        'terminated by an external agent'
                            if result.id is None: result.save()
                            self.result = result

                    else:
                      self.status = status

                    self.save()

                    logger.info("Job split `%s' finished executing", self)

                    self.job._update_state()

                    break

            except utils.OperationalError:
                tries += 1
                if tries > settings.MAXIMUM_SPLIT_SAVE_RETRIES:
                    raise
                else:
                  logger.info("Database error caught ending `%s': retrying " \
                      "in 1 second (%d/%d)...", self, tries,
                      settings.MAXIMUM_SPLIT_SAVE_RETRIES)
                  # wait a second and retry
                  time.sleep(1)


    def try_end(self, result):
        '''Tries to end the split - ignores if the split was deleted'''

        try:
            self.refresh_from_db()
        except JobSplit.DoesNotExist:
            logger.warn("Job split(pk=%d) does not exist. Likely cancelled, " \
                "so ignoring result `%s'", self.pk, result)
        self.end(result)


    def process(self, cache=settings.CACHE_ROOT, docker_images_cache=None):
        '''Process assigned job splits using beat.core

        This task executes the user algorithm on a subprocess. It also serves
        the data to the user process so it works like an I/O daemon.

        If ``required_slots == 1``, then this job takes care of the whole data
        set.  Otherwise, it takes care of a subset of the input data that is
        synchronised with this block, determined by ``split_index``.


        Parameters:

          cache (str, Optional): The path leading to the root of the cache to
            use for this run. If not set, use the global default at
            ``settings.CACHE_ROOT``.

        '''

        logger.info("Starting to process split `%s' (pid=%d)...", self,
            os.getpid())

        self.executor = None

        config = simplejson.loads(str(self.job.block.command))

        if settings.DATASETS_UID is not None:
          config['datasets_uid'] = settings.DATASETS_UID

        if settings.DATASETS_ROOT_PATH is not None:
          config['datasets_root_path'] = settings.DATASETS_ROOT_PATH

        # setup range if necessary
        if self.job.block.required_slots > 1:

            if (self.start_index) is None or (self.end_index is None):
                message = "The split %d/%d (pid=%d) running on worker `%s' " \
                    "for block `%s' of experiment `%s' could not " \
                    "be completed: indexes are missing!" % \
                    (self.split_index+1, self.job.block.required_slots,
                      os.getpid(), self.worker, self.job.block.name,
                      self.job.block.experiment.fullname())
                logger.error(message)
                self.try_end(Result(status=1,
                    usrerr=settings.DEFAULT_USER_ERROR))

            config['range'] = [self.start_index, self.end_index]

        # For reference, this bit of code should match (or be very similar) to
        # the one at beat.cmdline.experiments:run_experiment()

        try:

            if JobSplit.host is None:
                JobSplit.host = Host(images_cache=docker_images_cache,
                                     raise_on_errors=not(getattr(settings, 'TEST_CONFIGURATION', False)))

            self.executor = beat.core.execution.DockerExecutor(
                JobSplit.host, settings.PREFIX, config, cache
            )

            if not self.executor.valid:
                err = ''
                for e in self.executor.errors: err += '  * %s\n' % e
                message = "Failed to load execution information for split " \
                    "%d/%d running at worker `%s', for block `%s' of " \
                    "experiment `%s': %s" % (self.split_index+1,
                            self.job.block.required_slots,
                            self.worker, self.job.block.name,
                            self.job.block.experiment.fullname(), err)
                raise RuntimeError(message)

            queue = self.job.block.queue

            logger.info("Running `%s' on worker request",
                self.executor.algorithm.name)

            # n.b.: with executor may crash on the database view setup
            with self.executor:
                self.start()
                result = self.executor.process(
                    virtual_memory_in_megabytes=queue.memory_limit,
                    max_cpu_percent=int(100*float(queue.cores_per_slot)), #allows for 150%
                    timeout_in_minutes=queue.time_limit
                )

            self.try_end(Result(
                status=result['status'],
                stdout=result['stdout'],
                stderr=result['stderr'],
                usrerr=result['user_error'],
                _stats=simplejson.dumps(result['statistics'], indent=2),
                ))
            logger.info("Split `%s' (pid=%d) ended gracefully", self,
                os.getpid())

        except IOError:
            logger.warn("Split `%s' (pid=%d) execution raised an IOError: %s",
                self, os.getpid(), traceback.format_exc())
            self.signal_io_error()
            if self.cache_errors > settings.MAXIMUM_IO_ERRORS:
                logger.info("Split `%s' reached the maximum number of IO " \
                    "errors (%d > %d). Force failing this split." % \
                    (self, self.cache_errors, settings.MAXIMUM_IO_ERRORS))
                logger.error(traceback.format_exc())
                self.try_end(Result(status=1,
                  usrerr=settings.DEFAULT_USER_ERROR))
            else:
                logger.info("Split `%s' will be retried (%d/%d)",
                    self, self.cache_errors, settings.MAXIMUM_IO_ERRORS)

        except Exception:
            logger.error("Split `%s' (pid=%d) ended with an error: %s",
                self, os.getpid(), traceback.format_exc())
            self.try_end(Result(status=1, usrerr=settings.DEFAULT_USER_ERROR))

        self.executor = None
