/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.backend.models.js

   Declaration of the backend-related classes
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.backend === undefined)
    beat.backend = {}

if (beat.backend.models === undefined)
    beat.backend.models = {}


/******************************* CLASS: EnvironmentsList ********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Allows to easily manipulate a list of environments
//
// declaration: JSON declaration of the list, as returned by the Web API
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsList = function(declaration)
{
    // Attributes
    this.data = new Array();

    // Initialisation
    for (var i = 0; i < declaration.length; ++i)
    {
        var environment = {
            name: declaration[i].name,
            version: declaration[i].version,
            languages: declaration[i].languages,
            queues: [],
        };

        var queue_names = Object.keys(declaration[i].queues);
        for (var j = 0; j < queue_names.length; ++j)
        {
            var queue = {
                name:     queue_names[j],
                nb_slots: declaration[i].queues[queue_names[j]].nb_slots,
                memory_limit: declaration[i].queues[queue_names[j]].memory_limit,
                time_limit: declaration[i].queues[queue_names[j]].time_limit,
                nb_cores_per_slot: declaration[i].queues[queue_names[j]].nb_cores_per_slot,
                max_slots_per_user: declaration[i].queues[queue_names[j]].max_slots_per_user,
            };

            environment.queues.push(queue);
        }

        this.data.push(environment);
    }

    this.length = this.data.length;
}


//----------------------------------------------------------------------------------------
// Retrieve an environment by name
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsList.prototype.get = function(name)
{
    for (var i = 0; i < this.data.length; i++)
    {
        var environment = this.data[i];

        if (environment.name == name)
            return environment;
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieve an environment by index
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsList.prototype.getByIndex = function(index)
{
    if (index < this.data.length)
        return this.data[index];

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieve an environment by name
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsList.prototype.contains = function(name, version)
{
    for (var i = 0; i < this.data.length; i++)
    {
        var environment = this.data[i];

        if ((environment.name == name) && (environment.version == version))
            return true;
    }

    return false;
}


//----------------------------------------------------------------------------------------
// Retrieve a list of environments supporting a specific language
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsList.prototype.filter = function(language)
{
    var result = new beat.backend.models.EnvironmentsList([]);

    for (var i = 0; i < this.data.length; i++)
    {
        var environment = this.data[i];

        if (environment.languages.indexOf(language) >= 0)
            result.data.push(JSON.parse(JSON.stringify(environment)));
    }

    result.length = result.data.length;

    return result;
}


//----------------------------------------------------------------------------------------
// Retrieve the list of supported languages across all the environments
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsList.prototype.languages = function()
{
    var languages = [];

    for (var i = 0; i < this.data.length; i++)
    {
        var environment = this.data[i];

        for (var j = 0; j < environment.languages.length; j++)
        {
            var language = environment.languages[j];

            if (languages.indexOf(language) < 0)
                languages.push(language);
        }
    }

    return languages;
}


//----------------------------------------------------------------------------------------
// Returns an iterator over the environments
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsList.prototype.iterator = function()
{
    return new beat.backend.models.EnvironmentsIterator(this);
}



/***************************** CLASS: EnvironmentsIterator ******************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Allows to easily iterate through a list of datasets
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsIterator = function(environments)
{
    // Attributes
    this.environments = environments;
    this.next_index   = 0;
}


//----------------------------------------------------------------------------------------
// Indicates if the iteration can continue
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsIterator.prototype.hasMoreData = function()
{
    return (this.next_index < this.environments.length);
}


//----------------------------------------------------------------------------------------
// Returns the next algorithm
//----------------------------------------------------------------------------------------
beat.backend.models.EnvironmentsIterator.prototype.next = function()
{
    var environment = this.environments.getByIndex(this.next_index);

    this.next_index += 1;

    return environment;
}
