#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from rest_framework import permissions
from .permissions import IsAuthor

class CommonContextMixin(object):
    """
    Apply this mixin to any view or viewset to get the
    request user in the serializer context
    optionnaly the request object format
    """
    def get_serializer_context(self):
        context = super(CommonContextMixin, self).get_serializer_context()
        context['user'] = self.request.user

        object_format = self.request.GET.get('object_format', None)
        if object_format is not None:
            if object_format not in ['json', 'string']:
                object_format = 'json'

            context['object_format'] = object_format

        return context


class SerializerFieldsMixin(object):
    """
    Apply this mixin to any view or viewset to get the
    list of fields to return
    """
    def get_serializer_fields(self, request, allow_sharing=False, exclude_fields=[]):
        # Process the query string
        fields = None
        query_params = request.query_params

        if query_params.has_key('fields'):
            fields = query_params['fields'].split(',')
        else:
            fields = self.get_serializer_class().Meta.default_fields

        if query_params.has_key('include_fields'):
            include_fields = query_params['include_fields'].split(',')
            fields.extend(include_fields)

        if not(allow_sharing):
            exclude_fields = ['sharing'] + exclude_fields

        if request.user.is_anonymous():
            exclude_fields = ['is_owner'] + exclude_fields

        fields = filter(lambda x: x not in exclude_fields, fields)

        return fields


class IsAuthorOrReadOnlyMixin(object):
    """
    Apply this mixin to any view or viewset. Allows read for
    all and modification only by author
    """
    def get_permissions(self):
        if self.request.method == 'GET':
            self.permission_classes = [permissions.AllowAny]
        else:
            self.permission_classes = [permissions.IsAuthenticated, IsAuthor]
        return super(IsAuthorOrReadOnlyMixin, self).get_permissions()


class IsAdminOrReadOnlyMixin(object):
    """
    Apply this mixin to any view or viewset. Allows read for
    all and modification only by admin
    """
    def get_permissions(self):
        if self.request.method == 'GET':
            self.permission_classes = [permissions.AllowAny]
        else:
            self.permission_classes = [permissions.IsAuthenticated, permissions.IsAdminUser]
        return super(IsAdminOrReadOnlyMixin, self).get_permissions()
