/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.databases.models.js

   Declaration of the databases-related classes
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.databases === undefined)
    beat.databases = {}

if (beat.databases.models === undefined)
    beat.databases.models = {}


/*********************************** CLASS: Dataset *************************************/


//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.databases.models.Dataset = function(database, protocol, set, outputs, description,
                                         short_description)
{
    // Attributes
    this.name              = database + '/' + protocol + '/' + set;
    this.database          = database;
    this.protocol          = protocol;
    this.set               = set;
    this.outputs           = outputs;
    this.description       = description;
    this.short_description = short_description;
}


//----------------------------------------------------------------------------------------
// Retrieves the details about an output of the dataset
//----------------------------------------------------------------------------------------
beat.databases.models.Dataset.prototype.output = function(name)
{
    for (var i = 0; i < this.outputs.length; ++i)
    {
        if (this.outputs[i].name == name)
            return this.outputs[i];
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Indicates if the dataset is compatible with the provided dataset block
//
// The result object fields means:
//  - compatible: The dataset can be used in the block (when not taking data formats
//                restrictions into account)
//  - usable:     The dataset can be used in the block with the current data formats
//                restrictions
//  - perfect:    The dataset is a perfect match for the block (identical set of outputs)
//  - distance:   The distance between the dataset and the block
//----------------------------------------------------------------------------------------
beat.databases.models.Dataset.prototype.isCompatible = function(dataset_block, dataformats)
{
    var result = {
        compatible: true,
        usable:     true,
        perfect:    false,
        distance:   0,
    }

    for (var i = 0; i < dataset_block.outputs.length; ++i)
    {
        var block_output = dataset_block.outputs[i];

        var dataset_output = this.output(block_output.name);
        if (!dataset_output)
        {
            result.compatible = false;
            break;
        }

        if (result.usable && dataformats)
        {
            var needed_dataformat = block_output.neededDataformat(dataformats);
            result.usable = dataformats.areCompatible(dataset_output.dataformat, needed_dataformat);

            if (dataset_output.dataformat != needed_dataformat)
                result.distance += 1;
        }
    }

    if (result.compatible)
    {
        result.perfect = (dataset_block.outputs.length == this.outputs.length);
        result.distance += (this.outputs.length - dataset_block.outputs.length) * dataset_block.outputs.length +
                           (1.0 - dataset_block.name.score(this.set, 0.5));
    }

    return result;
}



/********************************* CLASS: DatasetsList **********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Allows to easily manipulate a list of datasets
//
// declaration: JSON declaration of the list, as returned by the Web API
//----------------------------------------------------------------------------------------
beat.databases.models.DatasetsList = function(declaration)
{
    // Attributes
    this.data = new Array();

    // Initialisation
    for (var j = 0; j < declaration.length; ++j)
    {
        var database = declaration[j];

        for (var k = 0; k < database.protocols.length; ++k)
        {
            var protocol = database.protocols[k];

            for (var l = 0; l < protocol.datasets.length; ++l)
            {
                var set = protocol.datasets[l];

                this.data.push(new beat.databases.models.Dataset(
                                                    database.name, protocol.name, set.name,
                                                    set.outputs, database.description,
                                                    database.short_description
                                                ));
            }
        }
    }

    this.length = this.data.length;
}


//----------------------------------------------------------------------------------------
// Retrieve a dataset by name
//----------------------------------------------------------------------------------------
beat.databases.models.DatasetsList.prototype.get = function(name)
{
    for (var i = 0; i < this.data.length; i++)
    {
        var dataset = this.data[i];

        if (dataset.name == name)
            return dataset;
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieve a dataset by index
//----------------------------------------------------------------------------------------
beat.databases.models.DatasetsList.prototype.getByIndex = function(index)
{
    if (index < this.data.length)
        return this.data[index];

    return null;
}


//----------------------------------------------------------------------------------------
// Returns an iterator over the datasets
//----------------------------------------------------------------------------------------
beat.databases.models.DatasetsList.prototype.iterator = function()
{
    return new beat.databases.models.DatasetsIterator(this);
}



/******************************* CLASS: DatasetsIterator ********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Allows to easily iterate through a list of datasets
//----------------------------------------------------------------------------------------
beat.databases.models.DatasetsIterator = function(datasets)
{
    // Attributes
    this.datasets   = datasets;
    this.next_index = 0;
}


//----------------------------------------------------------------------------------------
// Indicates if the iteration can continue
//----------------------------------------------------------------------------------------
beat.databases.models.DatasetsIterator.prototype.hasMoreData = function()
{
    return (this.next_index < this.datasets.length);
}


//----------------------------------------------------------------------------------------
// Returns the next algorithm
//----------------------------------------------------------------------------------------
beat.databases.models.DatasetsIterator.prototype.next = function()
{
    var dataset = this.datasets.getByIndex(this.next_index);

    this.next_index += 1;

    return dataset;
}
