#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db.models.signals import pre_delete, post_delete, pre_save
from django.dispatch import receiver
from ..team.models import Team

from .models import DataFormat


#----------------------------------------------------------


# These two auto-delete files from filesystem when they are unneeded:
@receiver(post_delete, sender=DataFormat)
def auto_delete_file_on_delete(sender, instance, **kwargs):
    """Deletes file from filesystem when ``DataFormat`` object is deleted.
    """
    if instance.declaration_file:
        instance.declaration_file.delete(save=False)

    if instance.description_file:
        instance.description_file.delete(save=False)


@receiver(pre_save, sender=DataFormat)
def auto_delete_file_on_change(sender, instance, **kwargs):
    """Deletes file from filesystem when ``DataFormat`` object is changed."""

    if not instance.pk:
        return False

    try:
        old_file = DataFormat.objects.get(pk=instance.pk).declaration_file
    except DataFormat.DoesNotExist:
        return False

    if old_file != instance.declaration_file:
        old_file.delete(save=False)

    try:
        old_descr = DataFormat.objects.get(pk=instance.pk).description_file
    except DataFormat.DoesNotExist:
        return False

    if old_descr != instance.description_file:
        old_descr.delete(save=False)


#----------------------------------------------------------


def build_team_dataformat_set(team):
    all_dataformats = []

    for dataformat in team.shared_dataformats.all():
        all_dataformats.append(dataformat)
        all_dataformats.extend(dataformat.all_needed_dataformats())

    return set(all_dataformats)


def build_user_dataformat_set(user):
    all_dataformats = []

    for dataformat in user.dataformats.all():
        all_dataformats.append(dataformat)
        all_dataformats.extend(dataformat.all_needed_dataformats())

    return set(all_dataformats)


def process_dataformats(team):
    team_dataformats = build_team_dataformat_set(team)

    for member in team.members.all():
        user_dataformats = build_user_dataformat_set(member)
        common_dataformats = team_dataformats & user_dataformats
        for dataformat in common_dataformats:
            dataformat.share(users=[member])


@receiver(pre_delete, sender=Team)
def on_team_delete(sender, **kwargs):
    team = kwargs.get('instance')
    process_dataformats(team)
