#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import simplejson as json

from django.contrib.auth.models import User
from django.conf import settings
from django.core.urlresolvers import reverse

from ...common.testutils import tearDownModule
from ..models import DataFormat

from core import DataFormatsAPIBase, DataFormatSharingAPIBase

from beat.core.dataformat import Storage, DataFormat as CoreDataFormat


class DataFormatsListRetrieval(DataFormatsAPIBase):
    def setUp(self):
        super(DataFormatsListRetrieval, self).setUp()

        self.url = reverse('api_dataformats:all')
        self.owner_url = reverse('api_dataformats:list_create', args=['jackdoe'])

    def test_only_public_dataformats_for_anonymous_user(self):
        response = self.client.get(self.url)
        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 2)

        self.assertEqual(content[0]['name'], 'jackdoe/format3/1')
        self.assertEqual(content[0]['short_description'], 'short description 3')
        self.assertEqual(content[0]['extend'], 'johndoe/format1/1')
        self.assertEqual(content[0]['accessibility'], 'public')
        self.assertFalse(content[0].has_key('is_owner'))

        self.assertEqual(content[1]['name'], 'johndoe/format1/1')
        self.assertEqual(content[1]['short_description'], 'short description 1')
        self.assertEqual(content[1]['extend'], None)
        self.assertEqual(content[1]['accessibility'], 'public')
        self.assertFalse(content[1].has_key('is_owner'))


    def test_all_accessible_dataformats(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.get(self.url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 3)

        self.assertEqual(content[0]['name'], 'jackdoe/format3/1')
        self.assertEqual(content[0]['short_description'], 'short description 3')
        self.assertEqual(content[0]['extend'], 'johndoe/format1/1')
        self.assertEqual(content[0]['accessibility'], 'public')
        self.assertEqual(content[0]['is_owner'], False)

        self.assertEqual(content[1]['name'], 'johndoe/format1/1')
        self.assertEqual(content[1]['short_description'], 'short description 1')
        self.assertEqual(content[1]['extend'], None)
        self.assertEqual(content[1]['accessibility'], 'public')
        self.assertEqual(content[1]['is_owner'], True)

        self.assertEqual(content[2]['name'], 'johndoe/format_private/1')
        self.assertEqual(content[2]['short_description'], 'short private description')
        self.assertEqual(content[2]['extend'], None)
        self.assertEqual(content[2]['accessibility'], 'private')
        self.assertEqual(content[2]['is_owner'], True)


    def test_own_dataformats(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.get(self.owner_url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 4)

        self.assertEqual(content[0]['name'], 'jackdoe/format2/1')
        self.assertEqual(content[0]['short_description'], 'short description 2')
        self.assertEqual(content[0]['extend'], None)
        self.assertEqual(content[0]['accessibility'], 'private')

        self.assertEqual(content[1]['name'], 'jackdoe/format3/1')
        self.assertEqual(content[1]['short_description'], 'short description 3')
        self.assertEqual(content[1]['extend'], 'johndoe/format1/1')
        self.assertEqual(content[1]['accessibility'], 'public')

        self.assertEqual(content[2]['name'], 'jackdoe/team_format1/1')
        self.assertEqual(content[2]['short_description'], 'team short description 1')
        self.assertEqual(content[2]['extend'], None)
        self.assertEqual(content[2]['accessibility'], 'confidential')

        self.assertEqual(content[3]['name'], 'johndoe/format1/1')
        self.assertEqual(content[3]['short_description'], 'short description 1')
        self.assertEqual(content[3]['extend'], None)
        self.assertEqual(content[3]['accessibility'], 'public')


    def test_other_user_dataformats(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.get(self.owner_url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 1)

        self.assertEqual(content[0]['name'], 'jackdoe/format3/1')
        self.assertEqual(content[0]['short_description'], 'short description 3')
        self.assertEqual(content[0]['extend'], 'johndoe/format1/1')
        self.assertEqual(content[0]['accessibility'], 'public')


    def test_user_dataformats_for_anonymous_user(self):
        response = self.client.get(self.owner_url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 1)

        self.assertEqual(content[0]['name'], 'jackdoe/format3/1')
        self.assertEqual(content[0]['short_description'], 'short description 3')
        self.assertEqual(content[0]['extend'], 'johndoe/format1/1')
        self.assertEqual(content[0]['accessibility'], 'public')


    def test_team_dataformats(self):
        self.client.login(username='jacquelinedoe', password='1234')

        response = self.client.get(self.owner_url)

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, list))
        self.assertEqual(len(content), 2)

        self.assertEqual(content[0]['name'], 'jackdoe/format3/1')
        self.assertEqual(content[0]['short_description'], 'short description 3')
        self.assertEqual(content[0]['extend'], 'johndoe/format1/1')
        self.assertEqual(content[0]['accessibility'], 'public')

        self.assertEqual(content[1]['name'], 'jackdoe/team_format1/1')
        self.assertEqual(content[1]['short_description'], 'team short description 1')
        self.assertEqual(content[1]['extend'], None)
        self.assertEqual(content[1]['accessibility'], 'confidential')


class DataFormatsNameCheck(DataFormatsAPIBase):

    def setUp(self):
        super(DataFormatsNameCheck, self).setUp()

        self.url = reverse('api_dataformats:check_name')

        user = User.objects.get(username='jackdoe')
        (format4, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format-4',
            short_description='short description 4',
            )
        assert format4, errors


    def test_no_check_for_anonymous_user(self):
        response = self.client.post(self.url)
        self.checkResponse(response, 403)


    def test_bad_request_for_post_method_without_needed_data(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.post(self.url,
                                    json.dumps({}),
                                    content_type='application/json')
        self.checkResponse(response, 400)


    def test_not_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'valid-name1',
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'valid-name1')
        self.assertFalse(content['used'])


    def test_not_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'invalid name_1 #due-to{}%&é',
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'invalid-name_1--due-to-----')
        self.assertFalse(content['used'])


    def test_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'format2',
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'format2')
        self.assertTrue(content['used'])


    def test_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'format 4',
            }), content_type='application/json')

        content = self.checkResponse(response, 200, content_type='application/json')

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'format-4')
        self.assertTrue(content['used'])



class DataFormatCreation(DataFormatsAPIBase):

    def setUp(self):
        super(DataFormatCreation, self).setUp()

        self.url = reverse('api_dataformats:list_create', args=['jackdoe'])

        user = User.objects.get(username='jackdoe')
        (format4, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format-4',
            short_description='short description 4',
            )
        assert format4, errors


    def checkDataFormatFile(self, name):
        serializer = CoreDataFormat(settings.PREFIX, name)
        assert serializer.valid, '\n  * %s' % '\n  * '.join(serializer.errors)
        return serializer.type()


    def test_no_creation_for_anonymous_user(self):
        response = self.client.post(self.url)
        self.checkResponse(response, 403)


    def test_no_creation_for_other_user(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_dataformats:list_create', args=['johndoe'])
        response = self.client.post(url)
        self.checkResponse(response, 403)


    def test_bad_request_for_post_method_without_needed_data(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.post(self.url,
                                    json.dumps({}),
                                    content_type='application/json')
        self.checkResponse(response, 400)


    def test_not_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'valid-name1',
                'short_description': 'blah'
            }), content_type='application/json')

        url = reverse('api_dataformats:object', args=['jackdoe', 'valid-name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'valid-name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            dataformat = DataFormat.objects.get(author__username='jackdoe', name='valid-name1')
        except:
            self.assertTrue(False)

        self.assertEqual(dataformat.short_description, 'blah')

        self.checkDataFormatFile(dataformat.fullname())


    def test_not_used_valid_name__no_description(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'valid-name1'
            }), content_type='application/json')

        url = reverse('api_dataformats:object', args=['jackdoe', 'valid-name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'valid-name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            dataformat = DataFormat.objects.get(author__username='jackdoe', name='valid-name1')
        except:
            self.assertTrue(False)

        self.assertEqual(dataformat.short_description, '')

        self.checkDataFormatFile(dataformat.fullname())


    def test_not_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'invalid name1',
                'short_description': 'blah'
            }), content_type='application/json')

        url = reverse('api_dataformats:object', args=['jackdoe', 'invalid-name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'invalid-name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            dataformat = DataFormat.objects.get(author__username='jackdoe', name='invalid-name1')
        except:
            self.assertTrue(False)

        self.assertEqual(dataformat.short_description, 'blah')

        self.checkDataFormatFile(dataformat.fullname())


    def test_not_used_invalid_name__no_description(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'invalid name1'
            }), content_type='application/json')

        url = reverse('api_dataformats:object', args=['jackdoe', 'invalid-name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'invalid-name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            dataformat = DataFormat.objects.get(author__username='jackdoe', name='invalid-name1')
        except:
            self.assertTrue(False)

        self.assertEqual(dataformat.short_description, '')

        self.checkDataFormatFile(dataformat.fullname())


    def test_used_valid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'format2',
                'description': 'blah'
            }), content_type='application/json')

        self.checkResponse(response, 400)


    def test_used_valid_name__no_description(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'format2',
            }), content_type='application/json')

        self.checkResponse(response, 400)


    def test_used_invalid_name(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'format 4',
                'description': 'blah'
            }), content_type='application/json')

        self.checkResponse(response, 400)


    def test_used_invalid_name__no_description(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'format 4',
            }), content_type='application/json')

        self.checkResponse(response, 400)

    def test_creation_with_declaration(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'name': 'valid-name1',
                'short_description': 'blah',
                'declaration': { 'value': 'int32' }
            }), content_type='application/json')

        url = reverse('api_dataformats:object', args=['jackdoe', 'valid-name1', 1])
        content = self.checkResponse(response, 201,
                                     content_type='application/json',
                                     location=url)

        self.assertTrue(isinstance(content, dict))
        self.assertEqual(content['name'], 'valid-name1')
        self.assertTrue(content['url'].endswith(url))

        try:
            dataformat = DataFormat.objects.get(author__username='jackdoe', name='valid-name1')
        except:
            self.assertTrue(False)

        self.assertEqual(dataformat.short_description, 'blah')
        data = self.checkDataFormatFile(dataformat.fullname())
        self.assertEqual(data.value, 0)



class DataFormatUpdate(DataFormatsAPIBase):
    def setUp(self):
        super(DataFormatUpdate, self).setUp()

        self.url = reverse('api_dataformats:object', args=['johndoe', 'format_private', 1])
        self.format1_url = reverse('api_dataformats:object', args=['johndoe', 'format1', 1])
        self.format4_url = reverse('api_dataformats:object', args=['johndoe', 'format4', 1])

    def test_no_update_for_anonymous_user(self):
        response = self.client.put(self.url,
            json.dumps({
                'description': 'blah',
                'declaration': { 'value': 'int32' }
            }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_to_update_with_invalid_username(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['unknown', 'format_private', 1])
        response = self.client.put(url,
            json.dumps({
                'description': 'blah',
                'declaration': { 'value': 'int32' }
            }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_to_update_with_invalid_dataformat_name(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'unknown', 1])
        response = self.client.put(url,
            json.dumps({
                'description': 'blah',
                'declaration': { 'value': 'int32' }
            }), content_type='application/json')

        self.checkResponse(response, 404)


    def test_no_update_without_content(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_dataformats:object', args=['jackdoe', 'format2'])
        response = self.client.put(url)
        self.checkResponse(response, 400)


    def test_fail_to_update_a_shared_dataformat(self):
        self.client.login(username='johndoe', password='1234')

        DataFormat.objects.get(name='format1').share()

        response = self.client.put(self.format1_url,
            json.dumps({
                'description': 'blah',
                'declaration': { 'value': 'int32' }
            }), content_type='application/json')

        self.checkResponse(response, 403)


    def test_successfull_update(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.put(self.url,
            json.dumps({
                'description': 'blah',
                'declaration': { 'value': 'float64' }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format_private')
        self.assertEqual(dataformat.description, 'blah')

        storage = Storage(settings.PREFIX,
                dataformat.fullname())
        assert storage.exists
        with open(storage.json.path, 'rt') as f: declaration = json.load(f)
        self.assertEqual(declaration['value'], 'float64')


    def test_successfull_update_of_description_only(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.put(self.format1_url,
            json.dumps({
                'description': 'blah',
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format1')
        self.assertEqual(dataformat.description, 'blah')


    def test_successfull_update_of_declaration_only(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.put(self.url,
            json.dumps({
                'declaration': { 'value': 'float64' }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format_private')

        storage = Storage(settings.PREFIX,
                dataformat.fullname())
        assert storage.exists
        with open(storage.json.path, 'rt') as f: declaration = json.load(f)
        self.assertEqual(declaration['value'], 'float64')


    def test_successfull_update_of_declaration__extension_addition(self):
        self.client.login(username='johndoe', password='1234')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=User.objects.get(username='johndoe'),
            name='format4',
            short_description='short description 4',
            declaration={
                'value2': 'float64'
            },
            )
        assert dataformat, errors

        response = self.client.put(self.url,
            json.dumps({
                'declaration': {
                    '#extends': 'johndoe/format4/1',
                    'value': 'float64'
                }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format_private', version=1)

        self.assertTrue(dataformat.extend is not None)
        self.assertEqual(dataformat.extend.name, 'format4')
        self.assertEqual(dataformat.referenced_formats.count(), 0)

        storage = Storage(settings.PREFIX,
                dataformat.fullname())
        assert storage.exists
        with open(storage.json.path, 'rt') as f: declaration = json.load(f)
        self.assertEqual(declaration['value'], 'float64')
        self.assertEqual(declaration['#extends'], 'johndoe/format4/1')


    def test_successfull_update_of_declaration__extension_removal(self):
        self.client.login(username='johndoe', password='1234')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=User.objects.get(username='johndoe'),
            name='format4',
            short_description='short description 4',
            declaration={
                '#extends': 'johndoe/format1/1',
                'value2': 'float64'
            },
            )
        assert dataformat, errors

        response = self.client.put(self.format4_url,
            json.dumps({
                'declaration': {
                    'value2': 'float64'
                }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format4', version=1)

        self.assertTrue(dataformat.extend is None)
        self.assertEqual(dataformat.referenced_formats.count(), 0)


    def test_successfull_update_of_declaration__extension_keeping(self):
        self.client.login(username='johndoe', password='1234')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=User.objects.get(username='johndoe'),
            name='format4',
            short_description='short description 4',
            declaration={
                '#extends': 'johndoe/format1/1',
                'value2': 'float64'
            },
            )
        assert dataformat, errors

        response = self.client.put(self.format4_url,
            json.dumps({
                'declaration': {
                    '#extends': 'johndoe/format1/1',
                    'value2': 'int32'
                }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format4', version=1)

        self.assertTrue(dataformat.extend is not None)
        self.assertEqual(dataformat.extend.name, 'format1')
        self.assertEqual(dataformat.referenced_formats.count(), 0)


    def test_successfull_update_of_declaration__composition_addition(self):
        self.client.login(username='johndoe', password='1234')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=User.objects.get(username='johndoe'),
            name='format4',
            short_description='short description 4',
            declaration={
                'value2': 'float64'
            },
            )
        assert dataformat, errors

        response = self.client.put(self.url,
            json.dumps({
                'declaration': {
                    'field1': 'johndoe/format4/1',
                    'value': 'float64'
                }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format_private', version=1)

        self.assertTrue(dataformat.extend is None)
        self.assertEqual(dataformat.referenced_formats.count(), 1)
        self.assertEqual(dataformat.referenced_formats.filter(name='format4').count(), 1)

        storage = Storage(settings.PREFIX,
                dataformat.fullname())
        assert storage.exists
        with open(storage.json.path, 'rt') as f: declaration = json.load(f)

        self.assertEqual(declaration['value'], 'float64')
        self.assertEqual(declaration['field1'], 'johndoe/format4/1')


    def test_successfull_update_of_declaration__composition_removal(self):
        self.client.login(username='johndoe', password='1234')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=User.objects.get(username='johndoe'),
            name='format4',
            short_description='short description 4',
            declaration={
                'value2': 'float64',
                'field1': 'johndoe/format1/1',
            },
            )
        assert dataformat, errors

        response = self.client.put(self.format4_url,
            json.dumps({
                'declaration': {
                    'value2': 'float64'
                }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format4', version=1)

        self.assertTrue(dataformat.extend is None)
        self.assertEqual(dataformat.referenced_formats.count(), 0)


    def test_successfull_update_of_declaration__composition_keeping(self):
        self.client.login(username='johndoe', password='1234')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=User.objects.get(username='johndoe'),
            name='format4',
            short_description='short description 4',
            declaration={
                'value2': 'float64',
                'field1': 'johndoe/format1/1',
            },
            )
        assert dataformat, errors

        response = self.client.put(self.format4_url,
            json.dumps({
                'declaration': {
                    'value2': 'int32',
                    'field1': 'johndoe/format1/1',
                }
            }), content_type='application/json')

        self.checkResponse(response, 204)

        dataformat = DataFormat.objects.get(author__username='johndoe', name='format4', version=1)

        self.assertTrue(dataformat.extend is None)
        self.assertEqual(dataformat.referenced_formats.count(), 1)
        self.assertEqual(dataformat.referenced_formats.filter(name='format1').count(), 1)



class DataFormatRetrieval(DataFormatsAPIBase):

    def setUp(self):
        super(DataFormatRetrieval, self).setUp()

        user = User.objects.get(username='jackdoe')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format4',
            short_description='short description 4',
            )
        assert dataformat, errors

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format5',
            short_description='short description 5',
            )
        assert dataformat, errors
        dataformat.share(users=[User.objects.get(username='johndoe')])


    def test_no_retrieval_of_confidential_format_for_anonymous_user(self):
        url = reverse('api_dataformats:object', args=['johndoe', 'format_private', 1])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_fail_to_retrieve_with_invalid_username(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['unknown', 'format', 1])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_fail_to_retrieve_with_invalid_dataformat_name(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'unknown', 1])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_no_retrieval_of_confidential_format(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['johndoe', 'format_private', 1])
        response = self.client.get(url)
        self.checkResponse(response, 404)


    def test_successful_retrieval_of_public_format_for_anonymous_user(self):
        url = reverse('api_dataformats:object', args=['jackdoe', 'format3', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['accessibility'], 'public')
        self.assertEqual(data['description'], 'long description 3')
        self.assertTrue(isinstance(data['declaration'], dict))
        self.assertFalse(data.has_key('is_owner'))

        self.assertFalse(data.has_key('sharing'))


    def test_successful_retrieval_of_public_format(self):
        self.client.login(username='johndoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'format3', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], False)
        self.assertEqual(data['accessibility'], 'public')
        self.assertEqual(data['description'], 'long description 3')
        self.assertTrue(isinstance(data['declaration'], dict))

        self.assertFalse(data.has_key('sharing'))


    def test_successful_retrieval_of_confidential_format(self):
        self.client.login(username='johndoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'format5', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], False)
        self.assertEqual(data['accessibility'], 'confidential')
        self.assertEqual(data['short_description'], 'short description 5')
        self.assertTrue(isinstance(data['declaration'], dict))

        self.assertFalse(data.has_key('sharing'))


    def test_successful_retrieval_of_own_public_format(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'format3', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], True)
        self.assertEqual(data['accessibility'], 'public')
        self.assertEqual(data['description'], 'long description 3')
        self.assertTrue(isinstance(data['declaration'], dict))

        self.assertTrue(isinstance(data['sharing'], dict))
        self.assertEqual(data['sharing']['status'], 'public')
        self.assertFalse(data['sharing'].has_key('shared_with'))


    def test_successful_retrieval_of_own_private_format(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'format2', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], True)
        self.assertEqual(data['accessibility'], 'private')
        self.assertEqual(data['description'], 'long description 2')
        self.assertTrue(isinstance(data['declaration'], dict))

        self.assertTrue(isinstance(data['sharing'], dict))
        self.assertEqual(data['sharing']['status'], 'private')


    def test_successful_retrieval_of_own_publicly_shared_format(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'format5', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], True)
        self.assertEqual(data['accessibility'], 'confidential')
        self.assertEqual(data['short_description'], 'short description 5')
        self.assertTrue(isinstance(data['declaration'], dict))

        self.assertTrue(isinstance(data['sharing'], dict))
        self.assertEqual(data['sharing']['status'], 'shared')
        self.assertEqual(data['sharing']['shared_with'], ['johndoe'])


    def test_successful_retrieval_of_own_team_shared_format(self):
        self.client.login(username='jackdoe', password='1234')

        url = reverse('api_dataformats:object', args=['jackdoe', 'team_format1', 1])
        response = self.client.get(url)
        data = self.checkResponse(response, 200, content_type='application/json')

        self.assertEqual(data['is_owner'], True)
        self.assertEqual(data['accessibility'], 'confidential')
        self.assertEqual(data['short_description'], 'team short description 1')
        self.assertTrue(isinstance(data['declaration'], dict))

        self.assertTrue(isinstance(data['sharing'], dict))
        self.assertEqual(data['sharing']['status'], 'shared')
        self.assertEqual(data['sharing']['shared_with_team'], ['jackdoe/teamdoe'])


class DataFormatDeletion(DataFormatsAPIBase):

    def setUp(self):
        super(DataFormatDeletion, self).setUp()

        user = User.objects.get(username='jackdoe')

        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user,
            name='format4',
            short_description='short description 4',
            )
        assert dataformat, errors


    def test_no_deletion_for_anonymous_user(self):
        url = reverse('api_dataformats:object', args=['johndoe', 'format3', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 403)


    def test_fail_to_delete_with_invalid_username(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_dataformats:object', args=['unknown', 'format1', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 403)


    def test_fail_to_delete_with_invalid_dataformat_name(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_dataformats:object', args=['jackdoe', 'unknown', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 404)


    def test_no_deletion_of_used_format(self):
        self.client.login(username='johndoe', password='1234')
        url = reverse('api_dataformats:object', args=['johndoe', 'format1', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 403)


    def test_no_deletion_of_not_owned_format(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_dataformats:object', args=['johndoe', 'format3', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 403)


    def test_successful_deletion_of_private_format(self):
        self.client.login(username='jackdoe', password='1234')
        url = reverse('api_dataformats:object', args=['jackdoe', 'format4', 1])
        response = self.client.delete(url)
        self.checkResponse(response, 204)



class DataFormatSharingAPI_Failures(DataFormatSharingAPIBase):
    def setUp(self):
        super(DataFormatSharingAPI_Failures, self).setUp()

        self.url = reverse('api_dataformats:share', args=['jackdoe', 'personal', 1])

    def test_fail_sharing_for_anonymous_user(self):
        response = self.client.post(self.url,
                                    json.dumps({}),
                                    content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_sharing_for_unauthorized_user(self):
        self.client.login(username='johndoe', password='1234')
        response = self.client.post(self.url,
                                    json.dumps({}),
                                    content_type='application/json')

        self.checkResponse(response, 403)


    def test_fail_sharing_with_unknown_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users': ['doejack'],
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')
        self.assertTrue(data is not None)
        self.assertTrue(isinstance(data['users'], list))


    def test_fail_sharing_with_unknown_team(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams': ['doejack'],
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')
        self.assertTrue(data is not None)
        self.assertTrue(isinstance(data['teams'], list))


class NotSharedDataFormat_SharingAPI(DataFormatSharingAPIBase):
    def setUp(self):
        super(NotSharedDataFormat_SharingAPI, self).setUp()

        self.url = reverse('api_dataformats:share', args=['jackdoe', 'personal', 1])

    def test_sharing_as_public_for_one_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_one_team(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe', 'janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_multiple_teams(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


class PublicForOneUserDataFormat_SharingAPI(DataFormatSharingAPIBase):
    def setUp(self):
        super(PublicForOneUserDataFormat_SharingAPI, self).setUp()

        self.url = reverse('api_dataformats:share', args=['jackdoe', 'public_for_one_user', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_the_same_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_another_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_a_team(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'shared_with_team': ['jackdoe/teamdoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe', 'jacquelinedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe', 'jacquelinedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_multiple_teams(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)



class PublicForOneTeamDataFormat_SharingAPI(DataFormatSharingAPIBase):
    def setUp(self):
        super(PublicForOneTeamDataFormat_SharingAPI, self).setUp()

        self.url = reverse('api_dataformats:share', args=['jackdoe', 'public_for_one_team', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
            'shared_with_team': ['jackdoe/teamdoe']
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_team', reference)


    def test_sharing_as_public_for_the_same_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
            'shared_with_team': ['jackdoe/teamdoe']
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_team', reference)


    def test_sharing_as_public_for_a_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['janedoe'],
            'shared_with_team': ['jackdoe/teamdoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_team', reference)


    def test_sharing_as_public_for_another_team(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe2'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_team', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe', 'jacquelinedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['janedoe', 'jacquelinedoe'],
            'shared_with_team': ['jackdoe/teamdoe']
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_team', reference)


    def test_sharing_as_public_for_multiple_teams(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe2', 'jackdoe/teamdoe3'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2', 'jackdoe/teamdoe3'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_one_team', reference)



class PublicForAllDataFormat_SharingAPI(DataFormatSharingAPIBase):
    def setUp(self):
        super(PublicForAllDataFormat_SharingAPI, self).setUp()

        self.url = reverse('api_dataformats:share', args=['jackdoe', 'public_for_all', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_one_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users': ['johndoe']
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_one_team(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams': ['jackdoe/teamdoe']
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users': ['johndoe', 'janedoe']
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_multiple_teams(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams': ['jackdoe/teamdoe', 'jackdoe/teamdoe2']
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)



class NotSharedDataFormat_ExtensionOfNotSharedDataFormat_SharingAPI(DataFormatSharingAPIBase):
    def setUp(self):
        super(NotSharedDataFormat_ExtensionOfNotSharedDataFormat_SharingAPI, self).setUp()

        self.url = reverse('api_dataformats:share', args=['jackdoe', 'extension_of_personal', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_personal', reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_one_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_personal', reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_one_team(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_personal', reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe', 'janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_personal', reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)


    def test_sharing_as_public_for_multiple_teams(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_personal', reference)
        self.checkDBSharingPreferences('jackdoe', 'personal', reference)



class NotSharedDataFormat_ExtensionOfPublicFormat_SharingAPI(DataFormatSharingAPIBase):
    def setUp(self):
        super(NotSharedDataFormat_ExtensionOfPublicFormat_SharingAPI, self).setUp()

        self.url = reverse('api_dataformats:share', args=['jackdoe', 'extension_of_public', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'public',
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_public', reference)

        reference = {
            'status': 'public',
        }

        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_one_user(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_public', reference)

        reference = {
            'status': 'public',
        }

        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_one_team(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_public', reference)

        reference = {
            'status': 'public',
        }

        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['johndoe', 'janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_public', reference)

        reference = {
            'status': 'public',
        }

        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)


    def test_sharing_as_public_for_multiple_teams(self):
        self.client.login(username='jackdoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('jackdoe', 'extension_of_public', reference)

        reference = {
            'status': 'public',
        }

        self.checkDBSharingPreferences('jackdoe', 'public_for_all', reference)



class NotSharedDataFormat_ExtensionOfPublicForOneUserFormat_SharingAPI(DataFormatSharingAPIBase):
    def setUp(self):
        super(NotSharedDataFormat_ExtensionOfPublicForOneUserFormat_SharingAPI, self).setUp()

        self.url = reverse('api_dataformats:share', args=['johndoe', 'extension_of_external', 1])

    def test_sharing_as_public_for_all(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')

        self.assertTrue(isinstance(data, list))
        self.assertTrue(len(data) > 0)


    def test_sharing_as_public_for_base_format_author(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['jackdoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['jackdoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('johndoe', 'extension_of_external', reference)

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe'],
        }

        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)


    def test_sharing_as_public_for_one_user(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')

        self.assertTrue(isinstance(data, list))
        self.assertTrue(len(data) > 0)


    def test_sharing_as_public_for_one_team(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')

        self.assertTrue(isinstance(data, list))
        self.assertTrue(len(data) > 0)


    def test_sharing_as_public_for_multiple_users(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users' : ['janedoe', 'jacquelinedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')

        self.assertTrue(isinstance(data, list))
        self.assertTrue(len(data) > 0)


    def test_sharing_as_public_for_multiple_teams(self):
        self.client.login(username='johndoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'teams' : ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
            }), content_type='application/json')

        data = self.checkResponse(response, 400, content_type='application/json')

        self.assertTrue(isinstance(data, list))
        self.assertTrue(len(data) > 0)


    def test_sharing_as_public_for_another_authorized_user(self):

        DataFormat.objects.get(name='public_for_one_user').share(users=[User.objects.get(username='janedoe')])


        self.client.login(username='johndoe', password='1234')

        response = self.client.post(self.url,
            json.dumps({
                'users': ['janedoe'],
            }), content_type='application/json')

        data = self.checkResponse(response, 200, content_type='application/json')

        reference = {
            'status': 'shared',
            'shared_with': ['janedoe'],
        }

        self.checkJSONSharingPreferences(data, reference)
        self.checkDBSharingPreferences('johndoe', 'extension_of_external', reference)

        reference = {
            'status': 'shared',
            'shared_with': ['johndoe', 'janedoe'],
        }

        self.checkDBSharingPreferences('jackdoe', 'public_for_one_user', reference)
