#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.conf import settings

import beat.core.hash

import os
import glob

import logging
logger = logging.getLogger(__name__)


#----------------------------------------------------------


class CachedFileManager(models.Manager):

    def get_by_natural_key(self, hash):
        return self.get(hash=hash)


#----------------------------------------------------------


class CachedFile(models.Model):

    NOT_CACHED = 'N'
    PROCESSING = 'P'
    CACHED     = 'C'

    STATUS = (
        (NOT_CACHED, 'Not cached'),
        (PROCESSING, 'Processing'),
        (CACHED,     'Cached'),
    )


    blocks = models.ManyToManyField('Block', related_name='outputs', blank=True)
    hash = models.CharField(max_length=64, unique=True)
    status = models.CharField(max_length=1, choices=STATUS, default=NOT_CACHED)

    # the total amount of time this block took to run considering the
    # wall-clock time.
    linear_execution_time = models.FloatField(default=0.)

    # the real speed-up obtained by running this block using X slots
    speed_up_real = models.FloatField(default=0.)

    # the maximum obtainable speed-up that could be achieved if all slots
    # were running in parallel. Essentially linear_execution_time /
    # maximum_slot_time
    speed_up_maximal = models.FloatField(default=0.)

    # the time this block waited to be executed
    queuing_time = models.FloatField(default=0.)

    stdout = models.TextField(null=True, blank=True)
    stderr = models.TextField(null=True, blank=True)
    error_report = models.TextField(null=True, blank=True)

    # other statistics of interest
    cpu_time = models.FloatField(default=0.)
    max_memory = models.BigIntegerField(default=0)
    data_read_size = models.BigIntegerField(default=0)
    data_read_nb_blocks = models.IntegerField(default=0)
    data_read_time = models.FloatField(default=0.)
    data_written_size = models.BigIntegerField(default=0)
    data_written_nb_blocks = models.IntegerField(default=0)
    data_written_time = models.FloatField(default=0.)

    objects = CachedFileManager()


    def __str__(self):
        return 'CachedFile(%s, %s, %d blocks)' % (
                    self.hash, self.get_status_display(), self.blocks.count()
                )


    def natural_key(self):
        return (self.hash,)


    def path(self):
        '''Returns the full path prefix to the cached file on disk'''

        return beat.core.hash.toPath(self.hash, suffix='')


    def absolute_path(self, cache=settings.CACHE_ROOT):
        '''Returns the full path prefix to the cached file on disk'''

        return os.path.join(cache, self.path())


    def files(self, cache=settings.CACHE_ROOT):
        '''Checks if any file belonging to this cache exist on disk'''

        return glob.glob(self.absolute_path(cache) + '*')


    def update(self, block_status):
        from . import Block

        if (block_status == Block.DONE) and (self.status != CachedFile.CACHED):
            self.status = CachedFile.CACHED
            self.save()

        elif (block_status == Block.PROCESSING) and (self.status != CachedFile.PROCESSING):
            self.status = CachedFile.PROCESSING
            self.save()

        elif (block_status != Block.DONE) and (self.status == CachedFile.PROCESSING):
            self.status = CachedFile.NOT_CACHED
            self.save()


    def exists(self, cache=settings.CACHE_ROOT):
        '''Checks if any file belonging to this cache exist on disk'''

        return bool(self.files(cache))


    def check_checksums(self, cache=settings.CACHE_ROOT):
        '''Checks if the cached files checksum properly'''

        abs_path = self.absolute_path(cache)

        data_files = sorted(glob.glob(abs_path + '*.index')) + \
                     sorted(glob.glob(abs_path + '*.data'))

        checksum_files = sorted(glob.glob(abs_path + '*.index.checksum')) + \
                         sorted(glob.glob(abs_path + '*.data.checksum'))

        if len(data_files) != len(checksum_files):
            logger.warn("Number of files (%d) is different from checksums (%d) " \
                        "for cache `%s'", len(data_files), len(checksum_files),
                        abs_path)
            return False

        for data_file, checksum_file in zip(data_files, checksum_files):
            with open(checksum_file, 'rt') as f:
                recorded = f.read().strip()

            actual = beat.core.hash.hashFileContents(data_file)

            if actual != recorded:
                logger.warn("Checksum for file `%s' does not match (%s != %s)",
                            data_file, actual, recorded)
                return False

        return True
