#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.dispatch import receiver

from ..team.models import Team
from .models import Experiment, Block

from datetime import datetime


# These two auto-delete files from filesystem when they are unneeded:
@receiver(models.signals.post_delete, sender=Experiment)
def auto_delete_file_on_delete(sender, instance, **kwargs):
    """Deletes file from filesystem when ``Experiment`` object is deleted.
    """
    if instance.declaration_file:
        instance.declaration_file.delete(save=False)

    if instance.description_file:
        instance.description_file.delete(save=False)


@receiver(models.signals.pre_save, sender=Experiment)
def auto_delete_file_on_change(sender, instance, **kwargs):
    """Deletes file from filesystem when ``Experiment`` object is changed."""

    if not instance.pk:
        return False

    try:
        old_file = Experiment.objects.get(pk=instance.pk).declaration_file
    except Experiment.DoesNotExist:
        return False

    if old_file != instance.declaration_file:
        old_file.delete(save=False)

    try:
        old_descr = Experiment.objects.get(pk=instance.pk).description_file
    except Experiment.DoesNotExist:
        return False

    if old_descr != instance.description_file:
        old_descr.delete(save=False)


@receiver(models.signals.pre_save, sender=Block)
def log_dates_on_state_change(sender, instance, **kwargs):
    """Logs the relevant dates upon the setting of a block's state"""

    if not instance.pk: #object creation, we may ignore
        return False

    old_block = Block.objects.filter(pk=instance.pk)

    if not old_block: #restoring from archive, ignore
        return False

    old_status = old_block[0].status
    new_status = instance.status

    if old_status != new_status: #status has changed

        if new_status == Block.PENDING: #admin reset
            instance.results.all().delete()

        elif new_status == Block.PROCESSING: #started to process
            instance.start_date = datetime.now()

        else: #finishing as FAILED or CACHED
            instance.end_date = datetime.now()


#_________ Algorithms _________

def build_user_algorithm_set(user):
    all_algorithms = []

    for experiment in user.experiments.all():
        all_algorithms.extend(experiment.referenced_algorithms.all())

    return set(all_algorithms)

def process_algorithms(team):
    team_algorithms = set(team.shared_algorithms.all() | team.usable_algorithms.all())

    for member in team.members.all():
        user_algorithms = build_user_algorithm_set(member)
        common_algorithms = team_algorithms & user_algorithms
        for algorithm in common_algorithms:
            algorithm.share(public=False, users=[member])


#_________ Toolchains _________

def build_user_toolchain_set(user):
    all_toolchains = []

    for experiment in user.experiments.all():
        all_toolchains.append(experiment.toolchain)

    return set(all_toolchains)


def process_toolchains(team):
    team_toolchains = set(team.shared_toolchains.all())

    for member in team.members.all():
        user_toolchains = build_user_toolchain_set(member)
        common_toolchains = team_toolchains & user_toolchains
        for toolchain in common_toolchains:
            toolchain.share(users=[member])


@receiver(models.signals.pre_delete, sender=Team)
def on_team_delete(sender, **kwargs):
    team = kwargs.get('instance')
    process_algorithms(team)
    process_toolchains(team)
