/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/**
 * beat.experiments.dialogs.js
 * Implementation of experiment-related dialogs
 */
if (beat === undefined) var beat = {};
if (beat.experiments === undefined) beat.experiments = {};
if (beat.experiments.dialogs === undefined) beat.experiments.dialogs = {};


/************** CLASS: AlgorithmMapping ********************************/

/**
 * Constructor
 *
 * Parameters:
 *   dialog_id (str): ID of the DOM element to use
 *   url_prefix (str): The URL prefix to use for creating links for dataformat
 *     types
 */
beat.experiments.dialogs.AlgorithmMapping = function(dialog_id, url_prefix) {
  var dialog = this;
  this.id = dialog_id;

  // Initialisations - BEAT stuff
  this.block = null;
  this.configuration = null;
  this.algorithm = null;
  this.dataformats = null;
  this.smart_mapping = null;

  // Initialisations - Dialog status
  this.input_infos = [];
  this.output_infos = [];
  this.onConfigurationModified = null;
  this.onClose = null;

  // This flag tells at any moment if the user form is modifiable at all
  this.modifiable = true;

  this.url_prefix = url_prefix;
  if (this.url_prefix === undefined)
    this.url_prefix = '';

  this.modifications = {
    inputs_mapping: [],
    outputs_mapping: [],
  };

  this.errors = {
    inputs_mapping: [],
    outputs_mapping: [],
  };

  this.message = $(document.createElement('div'));

  // Create the dialog options, so we can realize it later
  this.dialog = new BootstrapDialog({
    title: '',
    message: this.message,
    closable: true,
    size: BootstrapDialog.SIZE_WIDE,
    buttons: this.buttons,
    // When hiding, destroys descriptions
    onhidden: function(dialogWindow) {
      $('#' + dialog.id + ' dl#inputs dd').empty();
      $('#' + dialog.id + ' dl#outputs dd').empty();
      dialog.block = null;
    },

    buttons: [
    {
      label: 'Close', //cancels the mapping modification action
      id: 'button-' + dialog_id + '-cancel',
      action: function(dialogWindow) {
        dialogWindow.close();
      }
    },
    {
      label: 'Modify', //performs the mapping modification
      id: 'button-' + dialog_id + '-modify',
      cssClass: 'btn-primary',
      action: function(dialogWindow) {
        if (dialog.block !== null) {
          // Save the mapping of the inputs
          for (var i = 0; i < dialog.block.nbInputs(); ++i) {
            var input = dialog.block.inputs[i];

            var algorithm_input_name;
            if (dialog.modifications.inputs_mapping[input.index] !== undefined)
              algorithm_input_name = dialog.modifications.inputs_mapping[input.index];
            else
              algorithm_input_name = dialog.input_infos[input.index].reference_algorithm_endpoint;

            dialog.configuration.setComponentInputMapping(dialog.block.name,
                algorithm_input_name,
                input.name
                );

            for (var j = 0; j < dialog.algorithm.inputs.length; ++j) {
              if (dialog.algorithm.inputs[j].name == algorithm_input_name) {
                input.dataformat = dialog.algorithm.inputs[j].dataformat;
                break;
              }
            }
          }

          // Save the mapping of the outputs
          for (var i = 0; i < dialog.block.nbOutputs(); ++i) {
            var output = dialog.block.outputs[i];

            var algorithm_output_name;
            if (dialog.modifications.outputs_mapping[output.index] !== undefined)
              algorithm_output_name = dialog.modifications.outputs_mapping[output.index];
            else
              algorithm_output_name = dialog.output_infos[output.index].reference_algorithm_endpoint;

            dialog.configuration.setComponentOutputMapping(dialog.block.name,
                algorithm_output_name,
                output.name
                );

            for (var j = 0; j < dialog.algorithm.outputs.length; ++j) {
              if (dialog.algorithm.outputs[j].name == algorithm_output_name) {
                output.dataformat = dialog.algorithm.outputs[j].dataformat;
                break;
              }
            }
          }

          // Close the dialog
          dialog.block = null;
          if (dialog.onConfigurationModified) dialog.onConfigurationModified();
        }
        dialogWindow.close();
      }//action
    },
    ],
  });

  this.dialog.options.message.addClass('algorithm_mapping');

}


//-----------------------------------------------


beat.experiments.dialogs.AlgorithmMapping.prototype.hasModifications = function()
{
  if (this.block === null) return false;

  for (var i = 0; i < this.modifications.inputs_mapping.length; ++i) {
    if (this.modifications.inputs_mapping[i] !== undefined)
      return true;
  }

  for (var i = 0; i < this.modifications.outputs_mapping.length; ++i) {
    if (this.modifications.outputs_mapping[i] !== undefined)
      return true;
  }

  return false;
}


//-----------------------------------------------


beat.experiments.dialogs.AlgorithmMapping.prototype.hasErrors = function()
{
  for (var i = 0; i < this.errors.inputs_mapping.length; ++i) {
    if (this.errors.inputs_mapping[i])
      return true;
  }

  for (var i = 0; i < this.errors.outputs_mapping.length; ++i) {
    if (this.errors.outputs_mapping[i])
      return true;
  }

  return false;
}


//-----------------------------------------------

/**
 * This function is called just before showing the modal window and right after
 * errors have been detected. In both cases, we expect the dialog to be fully
 * realized by the calling time.
 */
beat.experiments.dialogs.AlgorithmMapping.prototype.updateButtons = function() {
  if (this.modifiable) {
    var cancel_button = this.dialog.getButton('button-' + this.id + '-cancel');
    if (this.hasModifications()) cancel_button.text('Cancel');
    else cancel_button.text('Close');

    var save_button = this.dialog.getButton('button-' + this.id + '-modify');
    if (this.hasModifications() && !this.hasErrors()) {
      save_button.addClass('btn-success');
      save_button.enable();
    }
    else {
      if (save_button) save_button.disable();
    }
  }
}


//--------------------------------------------------------------------------
// Display the informations about the provided block/algorithm configuration
//--------------------------------------------------------------------------
beat.experiments.dialogs.AlgorithmMapping.prototype.display = function(block, algorithm, configuration, smart_mapping)
{
  // Cleanup
  this.dialog.options.message.empty();

  // Initialisations
  this.block         = block;
  this.configuration = configuration;
  this.algorithm     = algorithm;
  this.smart_mapping = smart_mapping;
  this.input_infos   = [];
  this.output_infos  = [];

  // Add the controls allowing to change the mapping of the inputs and outputs
  // Decides if the mapping to be shown is user-modifiable (in case there are
  // multiple options for the mapping) or not (only one works for the current
  // setup)
  this.modifiable = this._addInputControls();
  this.modifiable = this._addOutputControls() || this.modifiable;

  // Update the interface options
  this.dialog.options.title = "<i class=\"fa fa-random\"></i> Input/Output mapping for block <strong>\"" + block.name + '"</strong>';

  var p = $(document.createElement('p'));
  p.addClass('help');

  if (!this.modifiable) { //final patch, in case it is not modifiable anyways
    this.dialog.options.title += ' <i>(readonly)</i>';
    this.dialog.options.buttons = [this.dialog.options.buttons[0]];
    p.text('This panel allows you to verify the endpoint mappings that connect your algorithm endpoints (on the right) to the indicated toolchain block endpoints (indicated on the left). You cannot modify the mappings as this is the only possible fit given dataformats used by the algorithm in question and those required by input/output endpoints.');
    this.dialog.options.message.prepend(p);
  }
  else {
    p.text('This panel allows you to modify the endpoint mappings that connect your algorithm endpoints (on the right) to the indicated toolchain block endpoints (indicated on the left). You may only modify endpoints which are not locked by input/output requirements, or for which the wiring is not unique (e.g. two endpoints with the same dataformat). Changes to other endpoints are disabled.');
    this.dialog.options.message.prepend(p);
  }

  this.dialog.realize();
  this.dialog.open();

  this.updateButtons();

}


//-----------------------------------------------


beat.experiments.dialogs.AlgorithmMapping.prototype._addInputControls = function()
{
  // Declarations
  var dialog = this;
  var modifiable = false;
  var table = $(document.createElement('table'));
  table.addClass('table table-hover table-striped');
  this.dialog.options.message.append(table);
  var tbody = $(document.createElement('tbody'));
  table.append(tbody);
  var dom_tbody = tbody[0];

  // Iterate through all the inputs of the block, group by group
  var channels = this.block.inputChannels();
  for (var i = 0; i < channels.length; ++i) {
    var channel_inputs = this.block.inputsOfChannel(channels[i]);

    for (var j = 0; j < channel_inputs.length; ++j)
    {
      var block_input = channel_inputs[j];

      this.modifications.inputs_mapping[block_input.index] = undefined;

      // Create the controls allowing to modify the input
      var tr = document.createElement('tr');
      tr.id = 'input__' + block_input.name;

      var first_cell = document.createElement('th');
      first_cell.className = 'first';
      tr.appendChild(first_cell);

      if (j == 0)
      {
        if (i == 0)
        {
          first_cell.className += ' category';
          first_cell.textContent = 'Inputs:';
        }
        else
        {
          tr.className = 'separator';
        }
      }

      var name_cell = document.createElement('td');
      name_cell.className = 'name';
      name_cell.textContent = block_input.name;
      tr.appendChild(name_cell);

      var dataformat_cell = document.createElement('td');
      dataformat_cell.className = 'dataformat';
      tr.appendChild(dataformat_cell);

      var dataformat = block_input.connections[0].output.dataformat;
      if (dataformat)
      {
        var a = document.createElement('a');
        a.href = this.url_prefix + '/dataformats/' + dataformat + '/';
        a.textContent = dataformat;
        a.target = '_blank';

        dataformat_cell.appendChild(a);
      }
      else
      {
        dataformat_cell.textContent = '<undefined data format>';
        dataformat_cell.className += ' undefined';
      }

      if (!this.readonly)
      {
        var inputs_selector_cell = document.createElement('td');

        var inputs_selector = document.createElement('select');
        inputs_selector.className = 'form-control input-sm';
        inputs_selector.input_index = block_input.index;

        inputs_selector_cell.appendChild(inputs_selector);
        tr.appendChild(inputs_selector_cell);

        var error_cell = document.createElement('td');

        var span_error = document.createElement('span');
        span_error.className = 'error';
        span_error.style.display = 'none';

        error_cell.appendChild(span_error);
        tr.appendChild(error_cell);

        // Keep some informations in memory about the input (to be used later)
        this.input_infos[block_input.index] = {
          selector:                     inputs_selector,
          error_field:                  span_error,
          block_endpoint:               block_input,
          reference_algorithm_endpoint: undefined,
        };

        this._updateInputChoices(inputs_selector, block_input);

        modifiable = (inputs_selector.length > 1) || modifiable;

        // Handler for input changes
        $(inputs_selector).change(function() {
          dialog._onAlgorithmEndpointChanged(this.value, this.input_index, dialog.input_infos,
              dialog.modifications.inputs_mapping,
              dialog.errors.inputs_mapping);
        });
      }
      else
      {
        var algorithm_input_cell = document.createElement('td');

        var span = document.createElement('span');

        var input_name = this.configuration.algorithmInputName(this.block.name,
            block_input.name);

        if (input_name != null)
          span.textContent = input_name;
        else
          span.textContent = '(none)';

        algorithm_input_cell.appendChild(span);
        tr.appendChild(algorithm_input_cell);
      }

      dom_tbody.appendChild(tr);
    }
  }


  return modifiable;
}


//-----------------------------------------------


beat.experiments.dialogs.AlgorithmMapping.prototype._addOutputControls = function()
{
  // Declarations
  var dialog = this;
  var modifiable = false;
  var table = this.dialog.options.message.find('tbody')[0];


  // Iterate through all the outputs of the block, group by group
  var channels = this.block.outputChannels();
  for (var i = 0; i < channels.length; ++i)
  {
    var channel_outputs = this.block.outputsOfChannel(channels[i]);

    for (var j = 0; j < channel_outputs.length; ++j)
    {
      var block_output = channel_outputs[j];

      this.modifications.outputs_mapping[block_output.index] = undefined;

      // Create the controls allowing to modify the output
      var tr = document.createElement('tr');
      tr.id = 'output__' + block_output.name;

      var first_cell = document.createElement('th');
      first_cell.className = 'first';
      tr.appendChild(first_cell);

      if (j == 0)
      {
        if (i == 0)
        {
          first_cell.className += ' category';
          first_cell.textContent = 'Outputs:';
          tr.className = 'outputs';
        }
        else
        {
          tr.className = 'separator';
        }
      }

      var name_cell = document.createElement('td');
      name_cell.className = 'name';
      name_cell.textContent = block_output.name;
      tr.appendChild(name_cell);

      var dataformat_cell = document.createElement('td');
      dataformat_cell.className = 'dataformat';
      tr.appendChild(dataformat_cell);

      var dataformat = block_output.connections[0].input.dataformat;
      if (dataformat)
      {
        var a = document.createElement('a');
        a.href = this.url_prefix + '/dataformats/' + dataformat + '/';
        a.textContent = dataformat;
        a.target = '_blank';

        dataformat_cell.appendChild(a);
      }
      else
      {
        dataformat_cell.textContent = '<undefined data format>';
        dataformat_cell.className += ' undefined';
      }

      if (!this.readonly)
      {
        var outputs_selector_cell = document.createElement('td');

        var outputs_selector = document.createElement('select');
        outputs_selector.className = 'form-control input-sm';
        outputs_selector.output_index = block_output.index;

        outputs_selector_cell.appendChild(outputs_selector);
        tr.appendChild(outputs_selector_cell);

        var error_cell = document.createElement('td');

        var span_error = document.createElement('span');
        span_error.className = 'error';
        span_error.style.display = 'none';

        error_cell.appendChild(span_error);
        tr.appendChild(error_cell);

        // Keep some informations in memory about the output (to be used later)
        this.output_infos[block_output.index] = {
          selector:                     outputs_selector,
          error_field:                  span_error,
          block_endpoint:               block_output,
          reference_algorithm_endpoint: undefined,
        };

        this._updateOutputChoices(outputs_selector, block_output);

        modifiable = (outputs_selector.length > 1) || modifiable;

        // Handler for output changes
        $(outputs_selector).change(function() {
          dialog._onAlgorithmEndpointChanged(this.value, this.output_index, dialog.output_infos,
              dialog.modifications.outputs_mapping,
              dialog.errors.outputs_mapping);
        });
      }
      else
      {
        var algorithm_output_cell = document.createElement('td');

        var span = document.createElement('span');

        var output_name = this.configuration.algorithmOutputName(this.block.name,
            block_output.name);

        if (output_name != null)
          span.textContent = output_name;
        else
          span.textContent = '(none)';

        algorithm_output_cell.appendChild(span);
        tr.appendChild(algorithm_output_cell);
      }

      table.appendChild(tr);
    }
  }


  return modifiable;
}


//-----------------------------------------------


beat.experiments.dialogs.AlgorithmMapping.prototype._updateInputChoices = function(
    inputs_selector, block_input)
{
  // Not in read-only mode (should not even reach this point)
  if (this.readonly)
    return;

  var block_configuration = this.configuration.component(this.block.name);

  $(inputs_selector).empty();

  this.input_infos[block_input.index].reference_algorithm_endpoint = undefined;

  // Add the compatible algorithm inputs to the selector element
  var groups = this.smart_mapping.possibilities[block_input.channel];
  for (var i = 0; i < groups.length; ++i)
  {
    var algorithm_inputs = this.algorithm.inputsOfChannel(groups[i].group);

    for (var j = 0; j < algorithm_inputs.length; ++j)
    {
      var algorithm_input = algorithm_inputs[j];

      if (this.dataformats.areCompatible(block_input.connections[0].output.dataformat, algorithm_input.dataformat))
      {
        var option   = document.createElement('option');
        option.value = algorithm_input.name;
        option.text  = algorithm_input.name;

        inputs_selector.appendChild(option);

        if ((block_configuration != null) && (block_configuration.inputs[algorithm_input.name] == block_input.name))
        {
          inputs_selector.selectedIndex = inputs_selector.length - 1;
          this.input_infos[block_input.index].reference_algorithm_endpoint = algorithm_input.name;
        }
      }
    }
  }

  if (inputs_selector.length > 1)
    $(inputs_selector).enable();
  else
    $(inputs_selector).disable();
}


//-----------------------------------------------


beat.experiments.dialogs.AlgorithmMapping.prototype._updateOutputChoices = function(
    outputs_selector, block_output)
{
  // Not in read-only mode (should not even reach this point)
  if (this.readonly)
    return;

  var block_configuration = this.configuration.component(this.block.name);

  $(outputs_selector).empty();

  this.output_infos[block_output.index].reference_algorithm_endpoint = undefined;

  // Add the compatible algorithm outputs to the selector element
  var groups = this.smart_mapping.possibilities[block_output.channel];
  for (var i = 0; i < groups.length; ++i)
  {
    var algorithm_outputs = this.algorithm.outputsOfChannel(groups[i].group);

    for (var j = 0; j < algorithm_outputs.length; ++j)
    {
      var algorithm_output = algorithm_outputs[j];

      if (this.dataformats.areCompatible(algorithm_output.dataformat, block_output.connections[0].input.dataformat))
      {
        var option   = document.createElement('option');
        option.value = algorithm_output.name;
        option.text  = algorithm_output.name;

        outputs_selector.appendChild(option);

        if ((block_configuration != null) && (block_configuration.outputs[algorithm_output.name] == block_output.name))
        {
          outputs_selector.selectedIndex = outputs_selector.length - 1;
          this.output_infos[block_output.index].reference_algorithm_endpoint = algorithm_output.name;
        }
      }
    }
  }

  if (outputs_selector.length > 1)
    $(outputs_selector).enable();
  else
    $(outputs_selector).disable();
}


//-----------------------------------------------


beat.experiments.dialogs.AlgorithmMapping.prototype._onAlgorithmEndpointChanged = function(
    value, endpoint_index, infos_list,
    modifications_list, errors_list)
{
  // Not in read-only mode (should not even reach this point)
  if (this.readonly)
    return;

  $('#' + this.id + ' select').removeClass('automatic');

  var infos = infos_list[endpoint_index];

  // Retrieve the previous value
  var previous_value = modifications_list[endpoint_index];
  if (previous_value === undefined)
    previous_value = infos.reference_algorithm_endpoint;

  // Determine if the groups of the new and previous endpoints are the same
  var is_output = infos.block_endpoint.is_output;

  var previous_group = (is_output ? this.algorithm.output(previous_value).channel
      : this.algorithm.input(previous_value).channel);

  var new_group = (is_output ? this.algorithm.output(value).channel
      : this.algorithm.input(value).channel);

  // No group change
  if (previous_group == new_group)
  {
    // Memorize the modification
    if (value == infos.reference_algorithm_endpoint)
      modifications_list[endpoint_index] = undefined;
    else
      modifications_list[endpoint_index] = value;

    // Change the value of the other selector that currently has our new value
    for (var i = 0; i < infos_list.length; ++i)
    {
      if (i == endpoint_index)
        continue;

      var infos2 = infos_list[i];
      if (infos2.selector.value == value)
      {
        var changed = false;
        for (var j = 0; j < infos2.selector.options.length; ++j)
        {
          if (infos2.selector.options[j].value == previous_value)
          {
            infos2.selector.selectedIndex = j;
            changed = true;

            $(infos2.selector).addClass('automatic');

            if (previous_value == infos2.reference_algorithm_endpoint)
              modifications_list[i] = undefined;
            else
              modifications_list[i] = previous_value;
            break;
          }
        }

        if (changed)
          break;
      }
    }
  }

  // Different group
  else
  {
    // Automatically reassign all the endpoints of the previous and new groups
    var block_endpoints1 = (is_output ? this.block.outputsOfChannel(infos.block_endpoint.channel)
        : this.block.inputsOfChannel(infos.block_endpoint.channel));

    var algorithm_endpoints1 = (is_output ? this.algorithm.outputsOfChannel(new_group)
        : this.algorithm.inputsOfChannel(new_group));

    var algorithm_endpoints2 = (is_output ? this.algorithm.outputsOfChannel(previous_group)
        : this.algorithm.inputsOfChannel(previous_group));

    var channel2 = null;
    for (var i = 0; i < modifications_list.length; ++i)
    {
      if (modifications_list[i] == value)
      {
        channel2 = infos_list[i].block_endpoint.channel;
        break;
      }
    }

    if (channel2 === null)
    {
      var block_configuration = this.configuration.component(this.block.name);

      channel2 = (is_output ? this.block.output(block_configuration.outputs[value]).channel
          : this.block.input(block_configuration.inputs[value]).channel);
    }

    var block_endpoints2 = (is_output ? this.block.outputsOfChannel(channel2)
        : this.block.inputsOfChannel(channel2));

    var mapping = {};
    mapping[value] = infos.block_endpoint.name;

    this.configuration._autoAssignAlgorithmEndpoints(block_endpoints1, algorithm_endpoints1,
        mapping, is_output, this.dataformats)

      this.configuration._autoAssignAlgorithmEndpoints(block_endpoints2, algorithm_endpoints2,
          mapping, is_output, this.dataformats)

      // Memorize the modifications
      var algorithm_endpoint_names = Object.keys(mapping);
    for (var i = 0; i < algorithm_endpoint_names.length; ++i)
    {
      var algorithm_endpoint_name = algorithm_endpoint_names[i];
      var block_endpoint_name = mapping[algorithm_endpoint_name];

      for (var j = 0; j < infos_list.length; ++j)
      {
        var infos2 = infos_list[j];

        if (infos2.block_endpoint.name == block_endpoint_name)
        {
          infos2.selector.value = algorithm_endpoint_name;
          $(infos2.selector).addClass('automatic');

          if (algorithm_endpoint_name == infos2.reference_algorithm_endpoint)
            modifications_list[infos2.block_endpoint.index] = undefined;
          else
            modifications_list[infos2.block_endpoint.index] = algorithm_endpoint_name;

          break;
        }
      }
    }
  }

  // Conflicts handling
  var counters = {};
  for (var i = 0; i < infos_list.length; ++i)
  {
    var infos2 = infos_list[i];
    if (counters[infos2.selector.value] !== undefined)
      counters[infos2.selector.value] += 1;
    else
      counters[infos2.selector.value] = 1;
  }

  for (var i = 0; i < infos_list.length; ++i)
  {
    var infos2 = infos_list[i];
    if (counters[infos2.selector.value] > 1)
    {
      errors_list[i] = true;

      $(infos2.error_field).text("Conflict!");
      $(infos2.error_field).show();
    }
    else
    {
      errors_list[i] = false;

      $(infos2.error_field).text("");
      $(infos2.error_field).hide();
    }
  }

  this.updateButtons();
}


/**
 * Implements a modal dialog based on Bootstrap3 that allows the user to cancel
 * the current experiment. The user is presented with a small modal window that
 * will allow him to abort the cancel action or continue and really cancel the
 * experiment.
 *
 * Parameters:
 *
 *   fullname (str): The full name of the experiment
 *   api_url (str): The URL towards the cancelling API for this particular
 *       experiment.
 *   redirect_url (str): The URL where to redirect the user if the operation is
 *       successful
 */
beat.experiments.dialogs.modal_cancel = function(fullname, api_url, redirect_url) {
  var dialog = new BootstrapDialog({
    title: '<i class="fa fa-warning"></i> Confirm experiment cancellation',
    message: 'Blocks which were successfuly completed will be cached as ' +
      'usual. Running this experiment again will pick-up from that point. ' +
      'You may abort this action by clicking "Abort" or closing this window.' +
      ' You may proceed cancelling the experiment by clicking "Cancel".',
    type: BootstrapDialog.TYPE_PRIMARY,
    buttons: [
    {
      label: 'Abort',
      cssClass: 'btn-default',
      action: function(the_dialog) {
        the_dialog.close();
      }
    },
    {
      label: 'Cancel',
      cssClass: 'btn-primary',
      action: function(the_dialog) {
        $.ajaxSetup({
          beforeSend: function(xhr, settings) {
            var csrftoken = $.cookie('csrftoken');
            xhr.setRequestHeader('X-CSRFToken', csrftoken);
          }
        });

        $.ajax({
          type: "POST",
          url: api_url,
        }).done(function(data) {
          the_dialog.close();
          window.location.href = redirect_url;
        }).fail(function(jqXHR, textStatus, errorThrown) {
          if ((jqXHR.status == 403) && (jqXHR.responseText.length > 0))
            alert(jqXHR.responseText);
          else
            alert('Error: ' + errorThrown);
          the_dialog.close();
        });
        the_dialog.close();
      }
    },
    ],
  });
  dialog.realize();
  dialog.open();
}


/**
 * Implements a modal dialog based on Bootstrap3 that allows the user to modify
 * the current experiment's name (only possible if the experiment is not
 * running). The user is presented with a small modal window that will allow
 * him to set the new experiment name. It is possible to cancel the action at
 * any moment by pressing "Cancel" or hitting the ESC key.
 *
 * Parameters:
 *
 *   fullname (str): The full name of the experiment to change name
 *   api_url (str): The URL towards the API for changing the experiment name
 *   redirect_url (str): The URL where to redirect the user if the operation is
 *       successful
 */
beat.experiments.dialogs.modal_rename = function(fullname, api_url, redirect_url) {

  // Retrieve the original name
  var parts = experiment_fullname.split('/');
  var original_name = parts[parts.length - 1];

  // Create the dialog
  $('#' + dialog_id).dialog({
    autoOpen: false,
    resizable: false,
    width: Math.min($(window).width() * 0.6, 700),
    position: { my: "center", at: "center", of: window },
    modal: true,
    closeOnEscape: true,
    buttons: [
    {
      id: 'button-' + dialog_id + '-renameit',
      text: 'Rename it',
      click: function() {
        var name = $('#' + dialog_id + ' input')[0].value.trim();

        if (name.length == 0)
        {
          alert("You can't enter an empty name!");
          return;
        }

        if (name == original_name)
        {
          alert("The new name is the same as the previous one!");
          return;
        }

        $.ajaxSetup({
          beforeSend: function(xhr, settings) {
            var csrftoken = $.cookie('csrftoken');
            xhr.setRequestHeader('X-CSRFToken', csrftoken);
          }
        });

        $.ajax({
          type: "PUT",
          url: api_url + experiment_fullname + '/',
          data: JSON.stringify({
            name: name,
          }),
          contentType: "application/json; charset=utf-8",
          dataType: "json",

          success: function(data) {
            $('#' + dialog_id).dialog("close");
            window.location.href = base_url + data.name + '/';
          },

          error: function(jqXHR, textStatus, errorThrown) {
            if ((jqXHR.status == 400) && (jqXHR.responseText.length > 0))
              alert(jqXHR.responseText);
            else
              alert('Error: ' + errorThrown);

            $(this).dialog("close");
          }
        });
      }
    },
    {
      id: 'button-' + dialog_id + '-cancel',
      text: 'Cancel',
      click: function() {
        $(this).dialog("close");
      }
    }
    ]
  });

  // Initialise the dialog content
  $('#' + dialog_id + ' input')[0].value = original_name;

  // Display the dialog
  $('#' + dialog_id).dialog("open");
}

/**
 * Setup an input box so that it does automatic experiment name completion, so
 * the user knows what **not** to choose.
 *
 * This function requires 'bootstrap3-typeahead' to be installed
 *
 * Parameters:
 *
 *   selector (str): A jQuery selector for the input box to rig
 *   data (Array): A JS array containing the existing names for the user
 *     experiments
 *   own (str): My own name, if one was set, so we don't mistakenly highlight
 *     my own (previously set) experiment, which continues to be a valid
 *     choice.
 */
beat.experiments.dialogs.name_typeahead = function(selector, data, own) {

  var element = $(selector);

  var check_validity = function() {
    var currval = element.val();
    if (!currval.length) {
      element.parent().removeClass('has-success').addClass('has-error');
      element.siblings('span.glyphicon').removeClass('glyphicon-ok').addClass('glyphicon-remove');
      return; //ok, we don't need to check the rest
    }
    if (own !== undefined && own == currval) {
      element.parent().removeClass('has-error').addClass('has-success');
      element.siblings('span.glyphicon').removeClass('glyphicon-remove').addClass('glyphicon-ok');
      return; //ok, we don't need to check the rest
    }
    if ($.inArray(currval, data) >= 0) {
      element.parent().removeClass('has-success').addClass('has-error');
      element.siblings('span.glyphicon').removeClass('glyphicon-ok').addClass('glyphicon-remove');
    }
    else {
      element.parent().removeClass('has-error').addClass('has-success');
      element.siblings('span.glyphicon').removeClass('glyphicon-remove').addClass('glyphicon-ok');
    }
  }

  check_validity(); //checks if the currently value set is OK
  element.on('keyup', check_validity);

  element.typeahead({
    source: data,
    items: 10, //how many to show at a time
    minLength: 0, //show suggestions even if no text available
    autoSelect: false, //user may want to leave it where it is
    afterSelect: check_validity, //updates after the user selected an item
    showHintOnFocus: true, //dropdown as soon as focus on
  });

}
