#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


from django import template
from django.conf import settings
from ...common.texts import Messages as Texts

register = template.Library()


@register.inclusion_tag('libraries/panels/table.html', takes_context=True)
def library_table(context, objects, owner, id):
    '''Composes a library list table

    This panel primarily exists for user's library list page.

    Parameters:

        objects (iterable): An iterable containing Library objects
        owner (bool): A flag indicating if the list is being created for the
          owner of his personal list page on the user-microsite or not.
        id: The HTML id to set on the generated table. This is handy for the
          filter functionality normally available on list pages.

    '''
    return {
            'request': context['request'],
            'objects': objects,
            'owner': owner,
            'panel_id': id,
            }


@register.inclusion_tag('libraries/panels/actions.html', takes_context=True)
def library_actions(context, obj, display_count):
    '''Composes the action buttons for a particular library

    This panel primarily exists for showing action buttons for a given
    library taking into consideration it is being displayed for a given user.

    Parameters:

        obj (Library): The library object concerned for which the
          buttons will be drawn.
        display_count (bool): If the set of buttons should include one with the
          number of experiments using this library.

    '''
    request = context['request']
    return {
            'request': request,
            'object': obj,
            'display_count': display_count,
            'open_source': obj.open_source(request.user),
            }


@register.inclusion_tag('libraries/panels/sharing.html', takes_context=True)
def library_sharing(context, obj):
    '''Composes the current sharing properties and a form to change them

    Parameters:

        obj (Library): The library object concerned for which the
          buttons will be drawn.

    '''
    return {
            'request': context['request'],
            'object': obj,
            'owner': context['request'].user == obj.author,
            'users': context['users'],
            'teams': context['teams'],
            }


@register.inclusion_tag('libraries/panels/editor.html', takes_context=True)
def library_editor(context, obj):
    request = context['request']
    return {
            'owner': request.user == obj.author,
            'object': obj,
            'texts': Texts,
            'open_source': obj.open_source(request.user),
            }


@register.assignment_tag(takes_context=True)
def visible_referrer_libraries(context, obj):
    '''Calculates the visible usage for a given library and requestor'''

    return obj.referencing.for_user(context['request'].user, True).order_by('-creation_date').distinct()


@register.assignment_tag(takes_context=True)
def visible_referrer_algorithms(context, obj):
    '''Calculates the visible usage for a given library and requestor'''

    return obj.used_by_algorithms.for_user(context['request'].user, True).order_by('-creation_date').distinct()


@register.assignment_tag(takes_context=True)
def visible_referrers_count(context, obj):
    '''Calculates the visible usage for a given library and requestor'''

    return visible_referrer_libraries(context, obj).count() + \
            visible_referrer_algorithms(context, obj).count()
