#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.contrib.auth.models import User

from ...common.testutils import tearDownModule
from ...team.models import Team

from ..models import Library

from .core import LibrariesAccessibilityFunctionsBase



class NotSharedLibrary_CheckTeamAccessibilityFunction(LibrariesAccessibilityFunctionsBase):

    def test_team_has_no_access(self):
        library = Library.objects.get(name='personal')

        errors = library.is_accessible(public=False, teams=['jackdoe/teamdoe'])
        self.assertEqual(len(errors), 1)

        errors = library.is_accessible(public=True, teams=['jackdoe/teamdoe'])
        self.assertEqual(len(errors), 1)


class UsableByOneTeamLibrary_CheckAccessibilityFunction(LibrariesAccessibilityFunctionsBase):

    def test_authorized_team_has_usable_access(self):
        library = Library.objects.get(name='usable_by_one_team')
        name = 'jackdoe/teamdoe'

        errors = library.is_accessible(public=False, teams=[name])
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_one_team/1' isn't accessible to the following teams: {}".format(name))


    def test_unauthorized_team_has_no_access(self):
        library = Library.objects.get(name='usable_by_one_team')
        name = 'jackdoe/teamdoe2'

        errors = library.is_accessible(public=False, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_one_team/1' isn't usable by the following teams: {}".format(name))

        errors = library.is_accessible(public=True, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_one_team/1' isn't accessible to the following teams: {}".format(name))


    def test_mixed_case(self):
        library = Library.objects.get(name='usable_by_one_team')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']
        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_one_team/1' isn't usable by the following teams: {}".format(teams[1]))

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_one_team/1' isn't accessible to the following teams: {}".format(', '.join(teams)))


class UsableByTeamsLibrary_CheckAccessibilityFunction(LibrariesAccessibilityFunctionsBase):

    def test_authorized_team_has_usable_access(self):
        library = Library.objects.get(name='usable_by_teams')
        teams = ['jackdoe/teamdoe']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_teams/1' isn't accessible to the following teams: {}".format(', '.join(teams)))


    def test_authorized_teams_has_usable_access(self):
        library = Library.objects.get(name='usable_by_teams')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_teams/1' isn't accessible to the following teams: {}".format(', '.join(teams)))


    def test_unauthorized_team_has_no_access(self):
        library = Library.objects.get(name='usable_by_teams')
        name = 'jackdoe/teamdoe3'

        errors = library.is_accessible(public=False, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_teams/1' isn't usable by the following teams: {}".format(name))

        errors = library.is_accessible(public=True, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_teams/1' isn't accessible to the following teams: {}".format(name))


    def test_mixed_case(self):
        library = Library.objects.get(name='usable_by_teams')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2', 'jackdoe/teamdoe3']
        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_teams/1' isn't usable by the following teams: {}".format(teams[2]))

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_teams/1' isn't accessible to the following teams: {}".format(', '.join(teams)))


class UsableByAllUsersLibrary_CheckTeamAccessibilityFunction(LibrariesAccessibilityFunctionsBase):

    def test_teams_have_usable_access(self):
        library = Library.objects.get(name='usable_by_all')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/usable_by_all/1' isn't accessible to the following teams: {}".format(', '.join(teams)))


class PublicForOneTeamLibrary_CheckAccessibilityFunction(LibrariesAccessibilityFunctionsBase):

    def test_authorized_team_has_full_access(self):
        library = Library.objects.get(name='public_for_one_team')
        name = 'jackdoe/teamdoe'

        errors = library.is_accessible(public=False, teams=[name])
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=[name])
        self.assertEqual(len(errors), 0)


    def test_unauthorized_team_has_no_access(self):
        library = Library.objects.get(name='public_for_one_team')
        name = 'jackdoe/teamdoe2'

        errors = library.is_accessible(public=False, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_one_team/1' isn't usable by the following teams: {}".format(name))

        errors = library.is_accessible(public=True, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_one_team/1' isn't accessible to the following teams: {}".format(name))


    def test_mixed_case(self):
        library = Library.objects.get(name='public_for_one_team')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_one_team/1' isn't usable by the following teams: {}".format(teams[1]))

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_one_team/1' isn't accessible to the following teams: {}".format(teams[1]))


class PublicForTeamsLibrary_CheckAccessibilityFunction(LibrariesAccessibilityFunctionsBase):

    def test_authorized_team_has_full_access(self):
        library = Library.objects.get(name='public_for_teams')
        teams = ['jackdoe/teamdoe']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 0)


    def test_authorized_teams_has_full_access(self):
        library = Library.objects.get(name='public_for_teams')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 0)


    def test_unauthorized_team_has_no_access(self):
        library = Library.objects.get(name='public_for_teams')
        name = 'jackdoe/teamdoe3'

        errors = library.is_accessible(public=False, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_teams/1' isn't usable by the following teams: {}".format(name))

        errors = library.is_accessible(public=True, teams=[name])
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_teams/1' isn't accessible to the following teams: {}".format(name))


    def test_mixed_case(self):
        library = Library.objects.get(name='public_for_teams')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2', 'jackdoe/teamdoe3']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_teams/1' isn't usable by the following teams: {}".format(teams[2]))

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0], "The library 'jackdoe/public_for_teams/1' isn't accessible to the following teams: {}".format(teams[2]))


class PublicForAllUsersLibrary_CheckTeamAccessibilityFunction(LibrariesAccessibilityFunctionsBase):

    def test_teams_have_full_access(self):
        library = Library.objects.get(name='public_for_all')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        errors = library.is_accessible(public=False, teams=teams)
        self.assertEqual(len(errors), 0)

        errors = library.is_accessible(public=True, teams=teams)
        self.assertEqual(len(errors), 0)


class NotSharedLibrary_TeamShareFunction(LibrariesAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_one_team(self):
        library = Library.objects.get(name='personal')
        team = ['jackdoe/teamdoe']

        library.share(public=False, teams=team)

        reference = {
            'status': 'shared',
            'shared_with_team': [],
            'usable_by_team': team,
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_usable_by_multiple_teams(self):
        library = Library.objects.get(name='personal')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']
        library.share(public=False, teams=teams)

        reference = {
            'status': 'shared',
            'shared_with_team': [],
            'usable_by_team': teams,
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_one_team(self):
        library = Library.objects.get(name='personal')
        team = ['jackdoe/teamdoe']
        library.share(public=True, teams=team)

        reference = {
            'status': 'shared',
            'shared_with_team': team,
            'usable_by_team': [],
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_multiple_teams(self):
        library = Library.objects.get(name='personal')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']
        library.share(public=True, teams=teams)

        reference = {
            'status': 'shared',
            'shared_with_team': teams,
            'usable_by_team': [],
        }

        self.checkObjectDBSharingPreferences(library, reference)


class UsableByOneTeamLibrary_ShareFunction(LibrariesAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_another_team(self):
        library = Library.objects.get(name='usable_by_one_team')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']

        library.share(public=False, teams=teams[1:2])

        reference = {
            'status': 'shared',
            'shared_with_team': [],
            'usable_by_team': teams,
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_usable_by_multiple_teams(self):
        library = Library.objects.get(name='usable_by_one_team')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2', 'jackdoe/teamdoe3']

        library.share(public=False, teams=teams[1:])

        reference = {
            'status': 'shared',
            'shared_with_team': [],
            'usable_by_team': teams,
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_another_team(self):
        library = Library.objects.get(name='usable_by_one_team')

        library.share(public=True, teams=['jackdoe/teamdoe2'])

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe2'],
            'usable_by_team': ['jackdoe/teamdoe'],
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_multiple_teams(self):
        library = Library.objects.get(name='usable_by_one_team')
        teams = ['jackdoe/teamdoe2', 'jackdoe/teamdoe3']
        library.share(public=True, teams=teams)

        reference = {
            'status': 'shared',
            'shared_with_team': teams,
            'usable_by_team': ['jackdoe/teamdoe'],
        }

        self.checkObjectDBSharingPreferences(library, reference)


class UsableByAllLibrary_TeamShareFunction(LibrariesAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_one_team(self):
        library = Library.objects.get(name='usable_by_all')

        library.share(public=False, teams=['jackdoe/teamdoe'])

        reference = {
            'status': 'usable',
            'shared_with_team': [],
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_usable_by_multiple_teams(self):
        library = Library.objects.get(name='usable_by_all')

        library.share(public=False, teams=['jackdoe/teamdoe', 'jackdoe/teamdoe2'])

        reference = {
            'status': 'usable',
            'shared_with_team': [],
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_one_team(self):
        library = Library.objects.get(name='usable_by_all')
        team = ['jackdoe/teamdoe']
        library.share(public=True, teams=team)

        reference = {
            'status': 'usable',
            'shared_with_team': team,
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_multiple_teams(self):
        library = Library.objects.get(name='usable_by_all')
        teams = ['jackdoe/teamdoe', 'jackdoe/teamdoe2']
        library.share(public=True, teams=teams)

        reference = {
            'status': 'usable',
            'shared_with_team': teams,
        }

        self.checkObjectDBSharingPreferences(library, reference)


class PublicForOneTeamLibrary_ShareFunction(LibrariesAccessibilityFunctionsBase):

    def test_sharing_as_usable_by_another_team(self):
        library = Library.objects.get(name='public_for_one_team')
        team = ['jackdoe/teamdoe2']
        library.share(public=False, teams=team)

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe'],
            'usable_by_team': team,
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_usable_by_multiple_teams(self):
        library = Library.objects.get(name='public_for_one_team')
        teams = ['jackdoe/teamdoe2', 'jackdoe/teamdoe3']
        library.share(public=False, teams=teams)

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe'],
            'usable_by_team': teams,
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_another_team(self):
        library = Library.objects.get(name='public_for_one_team')

        library.share(public=True, teams=['jackdoe/teamdoe2'])

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2'],
            'usable_by_team': [],
        }

        self.checkObjectDBSharingPreferences(library, reference)


    def test_sharing_as_public_for_multiple_teams(self):
        library = Library.objects.get(name='public_for_one_team')

        library.share(public=True, teams=['jackdoe/teamdoe2', 'jackdoe/teamdoe3'])

        reference = {
            'status': 'shared',
            'shared_with_team': ['jackdoe/teamdoe', 'jackdoe/teamdoe2', 'jackdoe/teamdoe3'],
            'usable_by_team': [],
        }

        self.checkObjectDBSharingPreferences(library, reference)

class TeamDeletionPropagation(LibrariesAccessibilityFunctionsBase):

    def test_propagation(self):

        jackdoe = User.objects.get(username='jackdoe')
        johndoe = User.objects.get(username='johndoe')
        janedoe = User.objects.get(username='janedoe')

        team_to_delete = Team.objects.create(name='team_to_delete', owner=jackdoe)
        team_to_delete.members.add(janedoe, johndoe)

        DECLARATION = """{
          "language": "python"
        }"""

        (library, errors) = Library.objects.create_library(
            author=jackdoe,
            name='library1',
            short_description='some description',
            declaration=DECLARATION,
            )
        assert library, errors

        library.share(True, teams=[team_to_delete])

        self.assertEqual(Library.objects.for_user(janedoe).count(), 4)
        self.assertEqual(Library.objects.for_user(johndoe).count(), 8)

        jackdoe_db_library1 = Library.objects.get(author=jackdoe, name='library1')
        self.assertEqual(jackdoe_db_library1.shared_with_team.all().count(), 1)

        team_to_delete.delete()

        self.assertEqual(Library.objects.for_user(janedoe).count(), 3)
        self.assertEqual(Library.objects.for_user(johndoe).count(), 7)
