/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * GroupTableFieldSelector
 * Desc:
 *  Handles the choosing of table columns
 */
angular.module('reportApp')
.directive("groupTableFieldSelector", ['GroupsService', 'ExperimentsService', function(GroupsService, ExperimentsService){
    return {
        scope: {
            id: '=',
            group: '=',
            // currently selected columns for the table
            colsSelected: '=',
            // function to execute when the user clicks the submit button
            buttonAction: '&',
            // title for the menu
            title: '@',
            // text  for the submit button
            buttonText: '@'
        },
        link: function(scope){
            // bootstrap's auto dropdown toggling is disabled for the table creation dropdown
            // add a click handler for the table creation dropdown submit button to toggle
            // manually
            scope.clickButton = (e) => {
                $(`#${scope.id}`).dropdown('toggle');
                scope.buttonAction()();
            };

            scope.tableables = () => {
                // start with the tableables generated in experimentsservice
                const tableables = ExperimentsService.tableables;
                const fieldArr = Object.entries(tableables)
                // only look at fields that are from an experiment in the group
                .filter(([e, fields]) => scope.group.experiments.includes(e))
                // get the names of the fields
                .map(([e, fields]) => Object.keys(fields))
                // make one big array of all field names
                .reduce((arr, fArr) => [...arr, ...fArr], [])
                ;

                // converting to and from a Set is a simple way of
                // removing dups
                const arr = Array.from(new Set(fieldArr));

                return arr;
            };

            // has this fieldName already been processed?
            // need to look at the already-processed field names
            scope.isUniqueTableable = (expName, fieldName) => {
                const tableables = ExperimentsService.tableables;
                const concatNames = (eName, fName) => `${eName}.${fName}`;

                // see if this field is a repeat
                const isRepeat = Object.entries(tableables)
                .filter(([e, fields]) => {
                    // only look at tableables of exps that are in group
                    let isInGroup = scope.group.experiments.includes(e);
                    // and have already been looked at
                    let alreadyChecked = Object.keys(tableables).indexOf(e) < Object.keys(tableables).indexOf(expName);
                    return isInGroup && alreadyChecked;
                })
                // get field names
                .map(([e, fields]) => Object.keys(fields))
                // flatten
                .reduce((arr, fArr) => [...arr, ...fArr], [])
                // does this flattened array have this field name in it?
                .includes(fieldName)
                ;

                // if it isnt a repeat, its unique!
                return !isRepeat;
            };

            // many tableable fields are fields from an analyzer, a block, or something else
            // these fields have a field group name, a '.', and an actual field name
            // find these group names and use them to subdivide the list of fields in the menu
            scope.tableablesGroups = () => {
                let groupNames = scope.tableables()
                .filter(f => f.includes('.'))
                .map(f => f.split('.')[0]);

                const sorted = Array.from(new Set(groupNames))
                .sort((a, b) => groupNames.indexOf(a) - groupNames.indexOf(b))

                return sorted;
            };

            // finds the actual field name whether its in a field group or not
            scope.groupName = (field) => field.includes('.') ? field.split('.')[0] : field;

            // finds the actual field name whether its in a field group or not
            scope.subfieldName = (field) => field.includes('.') ? field.split('.').slice(1).join('.') : field;

            scope.shouldShowField = (fName, gName) => {
                const isUnique = scope.isUniqueTableable(fName);
                const isInGroup = scope.groupName(fName) == gName;

                return isUnique && isInGroup;
            }

            // toggle the selection of a field
            scope.toggleField = (fName) => {
                let idx = scope.colsSelected.indexOf(fName);
                if(idx > -1){
                    scope.colsSelected.splice(idx, 1);
                } else {
                    scope.colsSelected.push(fName);
                    scope.colsSelected.sort((a, b) => scope.tableables().indexOf(a) - scope.tableables().indexOf(b));
                }
            };
        },
        template: `
<button id='{{ id }}' type="button" class="btn btn-default dropdown-toggle" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">
    {{ title }}
    <span class="caret"></span>
</button>
<div class='dropdown-menu' ng-click="$event.stopPropagation();" style='width:300px;'>
    <h4>Select columns to show in Table</h4>
    <form style='max-height: 200px; overflow-y: scroll;'>
        <fieldset>
            <h4>General</h4>
            <div
                ng-repeat='fName in tableables()'
                ng-if="isUniqueTableable(fName) && !fName.includes('.')">
                <label>
                    <input
                        type='checkbox'
                        value='{{ fName }}'
                        ng-checked='colsSelected.includes(fName)'
                        ng-click='toggleField(fName)'>
                        {{ fName }}
                </label>
            </div>
        </fieldset>
        <fieldset ng-repeat='gName in tableablesGroups()'>
            <h4>{{ gName }}</h4>
            <div
                ng-repeat='fName in tableables()'
                ng-if='shouldShowField(fName, gName)'>
                <label>
                    <input
                        type='checkbox'
                        value='{{ fName }}'
                        ng-checked='colsSelected.includes(fName)'
                        ng-click='toggleField(fName)'>
                        {{ subfieldName(fName) }}
                </label>
            </div>
        </fieldset>
    </form>
    <!--
    <select multiple ng-model='colsSelected'>
        <optgroup label='General'>
            <option
                ng-repeat='fName in tableables()'
                ng-if="isUniqueTableable(fName) && !fName.includes('.')"
                value='{{ fName }}'>
                {{ fName }}
            </option>
        </optgroup>
        <optgroup ng-repeat='gName in tableablesGroups()' label='{{ gName }}'>
            <option
                ng-repeat='fName in tableables()'
                ng-if='isUniqueTableable(fName) && fName.startsWith(gName)'
                value='{{ fName }}'>
                {{ subfieldName(fName) }}
            </option>
        </optgroup>
    </select>
    !-->
    <button class='btn btn-default' ng-click='clickButton($event)'>{{ buttonText }}</button>
</div>
`
    }
}]);
