/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * groupPanelExperiments
 * Desc:
 *  displays the experiments panel of the group -
 *  a table of experiments in the group, their databases/protocols, and aliases.
 *  Also has a menu for adding (compatible) experiments to the group.
 */
angular.module('reportApp').directive("groupPanelExperiments", ['GroupsService', 'ExperimentsService', 'UrlService', function(GroupsService, ExperimentsService, UrlService){
    return {
        scope: {
            group: '='
        },
        link: function(scope){
            scope.experiments = ExperimentsService.experiments;
            scope.dropdownId = `${scope.group.name}_exp_add_dropdown`;

            scope.getExpName = (expName) => scope.experiments[expName] ? expName : expName.split('/').pop();
            const getExp = (expName) => scope.experiments[expName] || scope.experiments[expName.split('/').pop()];

            // find experiments that are not in the group but are
            // compatible with the existing experiments (if any)
            scope.expsNotInGroup = () => {
                return ExperimentsService.experimentNames
                // exps not in group
                .filter(e => !scope.group.experiments.includes(e))
                // exp has compatible analyzer
                .filter(e => scope.group.analyzer === '' || ExperimentsService.getAnalyzerFromExpName(e) === scope.group.analyzer)
                ;
            };

            // collects an array of formatted databases & protocols of an experiment
            // format is "<database name>@<protocol name>"
            scope.getExpDatabases = (expName) => {
                const expObj = getExp(expName);
                if(!expObj){
                    return;
                }

                let dbs = expObj.declaration.datasets;
                return Array.from(new Set(Object.values(dbs).map(db => `${db.database}@${db.protocol}`)));
            };

            scope.getAnalyzerFromExpName = ExperimentsService.getAnalyzerFromExpName;
            scope.getExpUrl = UrlService.getExperimentUrl;
            scope.getBlockUrl = UrlService.getBlockUrl;
            scope.getDatabaseUrl = UrlService.getDatabaseUrl;
            scope.isViewmode = UrlService.isViewmode;
        },
        template: `
<div id="{{group.name}}-explist-heading" class="panel-heading" role="tab">
    <h4 class="panel-title">
        <a
            class=''
            role="button"
            data-toggle="collapse"
            data-parent="#{{group.name}}-explist-heading"
            href="#collapse-{{group.name}}-explist"
            aria-expanded="true"
            aria-controls="collapse-{{group.name}}-explist">
            Experiments
        </a>
        <div ng-if='!isViewmode()' class='btn-group'>
            <div class="dropdown">
                <button
                    class="btn btn-default dropdown-toggle"
                    ng-class='{disabled: expsNotInGroup().length === 0}'
                    type="button"
                    id="{{ dropdownId }}"
                    data-toggle="dropdown"
                    aria-haspopup="true"
                    aria-expanded="true">
                    Add Experiment
                    <span class="caret"></span>
                </button>
                <ul class="dropdown-menu" aria-labelledby="{{ dropdownId }}">
                    <li
                        ng-repeat='exp in expsNotInGroup()'
                        ng-click='group.addExperiment(exp, getAnalyzerFromExpName(exp))'>
                        <a>{{ exp }}</a>
                    </li>
                </ul>
            </div>
        </div>
        <i style='margin-left: 5px;' ng-if='group.analyzer.length > 0'>
            Analyzer: <a href='{{ getBlockUrl(group.analyzer) }}'>{{ group.analyzer }}</a>
        </i>
    </h4>
</div>
<div id="collapse-{{group.name}}-explist"
    class="panel-collapse collapse in"
    role="tabpanel"
    aria-labelledby="{{group.name}}-explist-heading">
    <div class="panel-body">
        <table ng-if='group.experiments.length > 0' class="table table-striped table-hover">
            <thead>
                <tr>
                    <th ng-if='!isViewmode()'></th>
                    <th>Alias</th>
                    <th>Experiment</th>
                    <th>Databases/Protocols</th>
                </tr>
            </thead>
            <tbody>
                <tr ng-repeat='expName in group.experiments'>
                    <td ng-if='!isViewmode()'>
                        <div class='btn-group action-buttons'>
                            <span
                                style='cursor: pointer;'
                                ng-click='group.removeExperiment(expName)'
                                class="btn-delete"
                                data-toggle="tooltip" data-placement="top" title="Remove Experiment from Group">
                                <i class="fa fa-times fa-lg"></i>
                            </span>
                        </div>
                    </td>
                    <td ng-if='!isViewmode()'><input ng-model='group.aliases[expName]'></input></td>
                    <td ng-if='isViewmode()'><span>{{ group.aliases[expName] }}</span></td>
                    <td><a href='{{ getExpUrl(expName) }}'>{{ getExpName(expName) }}</a></td>
                    <td>
                        <span ng-repeat='db in getExpDatabases(expName)'>
                            <a href='{{ getDatabaseUrl(db.split("@")[0]) }}'>{{ db }}</a>
                            &nbsp;
                        </span>
                    </td>
                </tr>
            </tbody>
        </table>
    </div>
</div>
`
    };
}]);
