/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * groupPlotItem
 * Desc:
 *  displays a plot report item (basically a container for the plots code to insert into)
 */
angular.module('reportApp')
.directive("groupPlotItem", ['ExperimentsService', 'PlotService', '$timeout', 'UrlService', function(ExperimentsService, PlotService, $timeout, UrlService){
    return {
        scope: {
            group: '=',
            itemId: '=',
            content: '='
        },
        link: function(scope){
            const group = scope.group;
            scope.domId = `${scope.group.name}_${scope.itemId}`;

            // container for the plots applet to insert into
            scope.renderDivId = `${scope.domId}-render`;

            // the callback for when the plot renders
            // (called every time the plot re-renders, e.g. user changes config/merged
            const updatePlotConfig = (selectedPlotter, selectedConfig, isMerged) => {
                scope.content.merged = isMerged;
                scope.content.savedPlotter = selectedPlotter;
                scope.content.savedConfig = selectedConfig;
            };

            // wait until the container html element is rendered.
            // angular will run these functions called with $timeout
            // after everything has been rendered
            $timeout(function() {
                PlotService.addPlot(scope.group, scope.itemId, scope.renderDivId, updatePlotConfig);
            });

            let plotTimer;

            const updatePlot = () => {
                clearTimeout(plotTimer);

                const queueUpdate = () => {
                    let el = document.querySelector(`#${scope.renderDivId}`);
                    // if the container is rendered and it already has had a render,
                    // redo the render
                    if(el && el.childNodes.length > 0){
                        el.innerHTML = '';
                        return PlotService.addPlot(scope.group, scope.itemId, scope.renderDivId, updatePlotConfig);
                    }
                };

                plotTimer = setTimeout(queueUpdate, 1000);
            };

            // if the user selected different plotter or config, rerender
            scope.$watch(
                () => `${scope.content.savedPlotter}|${scope.content.savedConfig}`,
                updatePlot
            );

            scope.toggleMerged = () => {
                scope.content.merged = !scope.content.merged;
                updatePlot();
            };

            scope.getPlotter = () => PlotService.getPlotter(scope.content);
            scope.getPossiblePlotters = () => PlotService.plotters.filter(p => p.dataformat === scope.content.type).map(p => p.name);
            scope.getPossibleConfigs = () => PlotService.getPossibleConfigs(scope.getPlotter());
            scope.isViewmode = UrlService.isViewmode;
        },
        template: `
<div panel-container dom-id='domId'>
    <header>
        <editable-label obj='content' field='itemName'></editable-label>
        <div ng-if='!isViewmode()' class='btn-group action-buttons'>
            <span
                ng-click='group.removeReportItem(itemId)'
                class="btn btn-default btn-delete"
                data-toggle="tooltip" data-placement="top" title="Delete Plot">
                <i class="fa fa-times fa-lg"></i>
            </span>
            <span drag-handle handle-helper-class='dragItem'></span>
        </div>
        <div class='btn-group' role='group'>
            <download-link
                class='btn-group'
                dom-id='{{ domId }}'
                group='group'
                item-id='itemId'>
            </download-link>
            <button ng-if='group.experiments.length > 1' class='btn btn-default' ng-click='toggleMerged()'>
                <span ng-if='content.merged'><i class='fa fa-expand'></i> Expand</span>
                <span ng-if='!content.merged'><i class='fa fa-compress'></i> Merge</span>
            </button>
            <div class='btn-group' role='group'>
                <button
                    ng-class='{disabled: getPossiblePlotters().length === 1}'
                    class='btn btn-default dropdown-toggle'
                    type="button"
                    data-toggle="dropdown"
                    aria-haspopup="true"
                    aria-expanded="false">
                    Plotter: {{ content.savedPlotter }}
                    <span class="caret"></span>
                </button>
                <ul class="dropdown-menu">
                    <li ng-click='content.savedPlotter = p' ng-repeat='p in getPossiblePlotters()'><a>{{ p }}</a></li>
                </ul>
            </div>
            <div class='btn-group' role='group'>
                <button
                    ng-class='{disabled: getPossibleConfigs().length === 1}'
                    class='btn btn-default dropdown-toggle'
                    type="button"
                    data-toggle="dropdown"
                    aria-haspopup="true"
                    aria-expanded="false">
                    Config: {{ content.savedConfig }}
                    <span class="caret"></span>
                </button>
                <ul class="dropdown-menu">
                    <li ng-click='content.savedConfig = c' ng-repeat='c in getPossibleConfigs()'><a>{{ c }}</a></li>
                </ul>
            </div>
        </div>
    </header>
    <content>
        <div id='{{ renderDivId }}'></div>
    </content>
</div>
`
    };
}]);
