/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * tableItem
 * Desc:
 *  displays a table report item and lets the user
 *  manage this table's selected cols and float precision
 */
angular.module('reportApp')
.directive("groupTableItem", ['GroupsService', 'ExperimentsService', 'UrlService', function(GroupsService, ExperimentsService, UrlService){
    return {
        scope: {
            group: '=',
            itemId: '=',
            content: '='
        },
        link: function(scope){
            // aliases
            scope.fields = scope.content.fields;
            // ids
            scope.domId = `${scope.group.name}_${scope.itemId}`;
            scope.colSelectorId = `${scope.domId}_columnSelector`;

            // 1 - 10
            // probably the most concise way of generating 1-10 computationally
            // also vastly more complex than writing [1,2,3,4,5,6,7,9,10]
            scope.floatingPointRange = [...(new Array(10)).keys()].map(i => i + 1);

            // init the chosen cols for the table with the saved cols
            scope.chosenCols = Array.from(scope.fields);

            // save new cols choice
            // due to how angular handles functions passed as props to an element,
            // it must return a function that does the actual work.
            scope.saveChosenCols = () => () => {
                // we want to keep the selected columns in order,
                // so try to add new columns in their correct indices,
                // and rm cols by mutating the array

                const newCols = scope.chosenCols
                .filter(c => !scope.fields.includes(c));

                const rmCols = scope.fields
                .filter(f => !scope.chosenCols.includes(f) && f !== 'Experiment');

                rmCols.forEach(rf => scope.fields.splice(scope.fields.indexOf(rf), 1));

                newCols.forEach(nf => scope.fields.push(nf));
            };

            // toggle val for viewing CSV
            scope.isViewingCSV = { val: false };
            scope.toggleViewingCSV = () => {
                scope.isViewingCSV.val = !scope.isViewingCSV.val;
            };

            // aliases
            scope.fields = scope.content.fields;

            // add 'expName' to the beginning of the fields to show in the table
            // if it isnt already there
            if(scope.fields.length === 0 || scope.fields[0] !== 'Experiment'){
                scope.fields.unshift('Experiment');
            }

            // get possible table entries
            scope.tableables = ExperimentsService.tableables || {};

            // gets the field type (int, float, string, nothing)
            scope.getFieldType = (field) => {
                if(field === scope.fields[0]){
                    return 'string';
                }

                let hasFieldObj = Object.values(scope.tableables)
                .find(o => o[field]);
                let fVal = hasFieldObj ? hasFieldObj[field] : {};
                let type;
                if(fVal.type){
                    type = fVal.type;
                } else if(Number.isSafeInteger(fVal)){
                    type = 'integer';
                } else if(Number.isFinite(fVal)){
                    type = 'float';
                } else if(typeof fVal === 'string'){
                    type = 'string';
                } else {
                    type = undefined;
                }

                return type;
            };
            // gets the field val for the given exp
            scope.getFieldVal = (expName, field) => {
                const alias = scope.group.aliases[expName].length > 0 ?
                    scope.group.aliases[expName] : expName;

                const name = scope.tableables[alias] ?
                    alias : expName;

                let fVal = scope.tableables[name] ? scope.tableables[name][field] : undefined;
                let val;

                if(field === scope.fields[0]){
                    val = alias;
                } else if(!fVal){
                    val = '-';
                } else {
                    let tmp;
                    if(fVal.value){
                        tmp = fVal.value;
                    } else {
                        tmp = fVal;
                    }

                    let type = scope.getFieldType(field);
                    if(type && type.startsWith('float')){
                        val = tmp.toFixed(parseInt(scope.content.precision));
                    } else {
                        val = tmp;
                    }
                }

                return val;
            };

            // need to nest actual value in an obj to get angular
            // to watch it correctly
            scope.sortField = { val: 'Experiment', isReversed: false };
            // sort rows (one row per exp)
            scope.sortFunc = (expName) => {
                return scope.getFieldType(scope.sortField.val) ?
                    scope.getFieldVal(expName, scope.sortField.val) : expName;
            };
            // sets the new sort field and direction
            scope.setSortField = (field) => {
                if(scope.sortField.val === field){
                    scope.sortField.isReversed = !scope.sortField.isReversed;
                } else {
                    scope.sortField.val = field;
                    scope.sortField.isReversed = false;
                }
            };

            // a different view of the table
            scope.getCSV = () => {
                let fields = scope.fields;
                let exps = scope.group.experiments
                // clone arr
                .map(e => `${e}`)
                .sort((ea, eb) => (scope.sortField.isReversed ? -1 : 1) * (scope.sortFunc(ea) < scope.sortFunc(eb) ? -1 : 1))
                ;

                let str = '';

                let fieldsStr = fields
                .map(f => `${f}(${scope.getFieldType(f)})`)
                .join(',');

                let expsStrs = exps
                .map(e => fields.map(f => `${scope.getFieldVal(e, f)}`).join(','))
                .join('\n');

                str = `${fieldsStr}\n${expsStrs}`;

                return str;
            };

            // get experiment url for linking to exps
            // returns the url if successful else false
            scope.getExperimentUrl = (expName) => {
                // if theres more than 2 '/' in the expName,
                // its the absolute URL, and the user has access
                // to view the experiment.
                if(expName.split('/').length > 3){
                    return UrlService.getExperimentUrl(expName);
                }

                // else, its the short name
                // (just the exp name, not including author/toolchain)
                // and the user cant see the exp
                return false;
            };

            scope.sortableOptions = {
                items: `th:not(:first-child)`
            };

            scope.isViewmode = UrlService.isViewmode;
        },
        template: `
<div panel-container dom-id='domId'>
    <header>
        <editable-label obj='content' field='itemName'></editable-label>
        <div ng-if='!isViewmode()' class='btn-group action-buttons'>
            <span
                ng-click='group.removeReportItem(itemId)'
                class="btn btn-default btn-delete"
                data-toggle="tooltip" data-placement="top" title="Delete Table">
                <i class="fa fa-times fa-lg"></i>
            </span>
            <span drag-handle handle-helper-class='dragItem'></span>
        </div>
        <div class="btn-group" role="group" role='tab'>
            <div ng-if='!isViewmode()' class="btn-group" role="group"
                group-table-field-selector
                id='colSelectorId'
                group='group'
                cols-selected='chosenCols'
                button-action='saveChosenCols()'
                title="Choose Columns"
                button-text="Save Column Choices"
                >
            </div>
            <div class='btn-group' role='group'>
                <button class='btn btn-default' id="{{domId}}-precision" type="button" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">
                    Float Precision: {{ content.precision }}
                    <span class="caret"></span>
                </button>
                <ul class="dropdown-menu" aria-labelledby="{{domId}}-precision">
                    <li ng-click='content.precision = i' ng-repeat='i in floatingPointRange'><a>{{ i }}</a></li>
                </ul>
            </div>
            <button class='btn btn-default' ng-click='toggleViewingCSV()'>
                Toggle CSV View
            </button>
        </div>
    </header>
    <content>
        <div class='panel-body'>
            <div ng-if='isViewingCSV.val'>
                <pre>{{ getCSV() }}</pre>
            </div>
            <div ng-if='!isViewingCSV.val' style='height: 100%; overflow-x: auto;'>
                <table class="table table-striped table-hover">
                    <thead>
                        <tr ui-sortable='sortableOptions' ng-model='fields'>
                            <th ng-repeat='field in fields'>
                                <span
                                    ng-if="sortField.val == field"
                                    class='glyphicon'
                                    ng-class="{
                                        'glyphicon-chevron-up': sortField.isReversed,
                                        'glyphicon-chevron-down': !sortField.isReversed
                                        }"
                                    >
                                </span>
                                <a role='button' ng-click='setSortField(field)'>
                                    {{ field }} <i>({{ getFieldType(field) }})</i>
                                </a>
                            </th>
                        </tr>
                    </thead>
                    <tbody>
                        <tr ng-repeat="exp in group.experiments | orderBy:sortFunc:sortField.isReversed">
                            <td ng-repeat='field in fields'>
                                <a ng-if='$index == 0 && getExperimentUrl(exp)' href='{{ getExperimentUrl(exp) }}'>
                                    {{ getFieldVal(exp, field) }}
                                </a>
                                <span ng-if='!$index == 0 || !getExperimentUrl(exp)'>
                                    {{ getFieldVal(exp, field) }}
                                </span>
                            </td>
                        </tr>
                    </tbody>
                </table>
            </div>
        </div>
    </content>
</div>
`
    };
}]);
