/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * groupTextItem
 * Desc:
 *  displays a text report item:
 *  - sends the raw RST to the server to be compiled,
 *      and displays the returned HTML
 *  - the user can edit the RST using codemirror
 *  - the compile is async and doesnt require refreshing the page
 */
angular.module('reportApp')
.directive("groupTextItem", ['GroupsService', '$sce', 'UrlService', 'reportFactory', function(GroupsService, $sce, UrlService, reportFactory){
    return {
        scope: {
            group: '=',
            reportItem: '=',
        },
        link: function(scope){
            // aliases
            // angular requires that compiling raw html be sanitized
            scope.trustAsHtml = $sce.trustAsHtml;
            scope.item = scope.reportItem;
            scope.domId = `${scope.group.name}_${scope.item.id}`;

            scope.isViewmode = UrlService.isViewmode;

            // codemirror options
            scope.uicmOptions = {
                mode: 'rst',
                readOnly: scope.isViewmode()
            };

            // handle compiling content
            // holds the last response from the server
            scope.compiledContent = { val: '' };
            scope.compileContent = () => {
                let url = UrlService.getCompileRstUrl();
                let content = !scope.isViewmode() ? scope.item.content.text :
                    `${scope.group.name}|${scope.group.reportItems.indexOf(scope.reportItem)}`;

                return reportFactory.compileRST(url, content)
                .then(data => {
                    // when compiled, save the raw html
                    scope.compiledContent.val = data.data.html_str;
                });
            };

            // handle edit/save/cancel buttons
            scope.isSrcMode = { val: false };
            // when editing, use a buffer to hold the raw text
            scope.unsavedContent = { val: `${scope.item.content.text}` };
            // save the buffer to the actual report item content
            scope.saveAction = () => {
                scope.item.content.text = scope.unsavedContent.val;
                scope.compileContent();
                scope.isSrcMode.val = false;
            };
            // discard buffer and use report item content
            scope.cancelAction = () => {
                scope.unsavedContent.val = `${scope.item.content.text}`;
                scope.isSrcMode.val = false;
            };

            // compile the content when loaded
            scope.compileContent();
        },
        template: `
<div panel-container dom-id='domId'>
    <header>
        <editable-label obj='item.content' field='itemName'></editable-label>
        <div ng-if='!isViewmode()' class='btn-group action-buttons'>
            <span
                ng-click='group.removeReportItem(item.id)'
                class="btn btn-default btn-delete"
                data-toggle="tooltip" data-placement="top" title="Delete Text Block">
                <i class="fa fa-times fa-lg"></i>
            </span>
            <span drag-handle handle-helper-class='dragItem'></span>
        </div>
        <button ng-if='isViewmode()' class='btn btn-primary' ng-click='isSrcMode.val = !isSrcMode.val'>
            Toggle Source View
        </button>
    </header>
    <content>
        <div ng-if='!isViewmode()' class='row'>
            <div class='col-sm-10'>
                <div ng-if='!isSrcMode.val' ng-bind-html='trustAsHtml(compiledContent.val)'></div>
                <div ng-if='isSrcMode.val'>
                    <ui-codemirror ng-model='unsavedContent.val' ui-codemirror-opts='uicmOptions'></ui-codemirror>
                    <p class='help-block'>
                    Describe the object thoroughly using <a href="http://docutils.sourceforge.net/rst.html">reStructuredText mark-up</a><br><i class="fa fa-thumbs-up"></i> The ruler at 80 columns indicate suggested <a href="https://en.wikipedia.org/wiki/POSIX">POSIX line breaks</a> (for readability).<br><i class="fa fa-thumbs-up"></i> The editor will automatically enlarge to accomodate the entirety of your input<br><i class="fa fa-thumbs-up"></i> Use <a href="http://codemirror.net/doc/manual.html#commands">keyboard shortcuts</a> for search/replace and faster editing. For example, use Ctrl-F (PC) or Cmd-F (Mac) to search through this box
                    </p>
                </div>
            </div>
            <div class='col-sm-2'>
                <div class="pull-right action-buttons">
                    <a
                        ng-if='!isSrcMode.val'
                        ng-click='isSrcMode.val = !isSrcMode.val'
                        class="btn btn-primary btn-sm">
                        <i class="fa fa-edit fa-lg"></i> Edit
                    </a>
                    <a
                        ng-if='isSrcMode.val'
                        ng-click='cancelAction()'
                        class="btn btn-danger btn-sm">
                        <i class="fa fa-times fa-lg"></i> Cancel
                    </a>
                    <a
                        ng-if='isSrcMode.val'
                        ng-click='saveAction()'
                        class="btn btn-success btn-sm">
                        <i class="fa fa-save fa-lg"></i> Save
                    </a>
                </div>
            </div>
        </div>
        <div ng-if='isViewmode()' class='row'>
            <div class='col-sm-12'>
                <div ng-if='!isSrcMode.val' ng-bind-html='trustAsHtml(compiledContent.val)'></div>
                <div ng-if='isSrcMode.val'>
                    <i>Readonly view</i>
                    <ui-codemirror ng-model='item.content.text' ui-codemirror-opts='uicmOptions'></ui-codemirror>
                </div>
            </div>
        </div>
    </content>
</div>
`
    };
}]);
