/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * ExperimentsService
 * Desc:
 *  Manages the experiments data, including storing it and generating
 *  different views of the data
 */
angular.module('reportApp').factory('ExperimentsService', ['experimentFactory', 'GroupsService', 'UrlService', function(experimentFactory, GroupsService, UrlService){
    // holds the raw exp data received from the server
    const expData = {};
    // holds the view generated from `getTableablesFromExpName`
    const tableData = {};
    // holds the view generated from `getPlottablesFromExpName`
    const plotData = {};
    const rawExpNames = [];

    // lots of experiments have fields that are differently-named
    // but mean the same thing.
    // These fields are the *only* available field in the obj.
    const getVarProp = (obj, dontLogErrors) => {
        let keys = Object.keys(obj);
        return obj[keys[0]];
    };

    // analyzer
    const getAnalyzerFromExpName = (expName) => getVarProp(expData[expName].declaration.analyzers, true).algorithm;

    // get all possible objects that could be shown in a table
    const getTableablesFromExpName = (expName) => {
        let currExp = expData[expName];

        // get all result objs that have a type that doesnt start with 'plot/'
        // these results can be displayed in a table
        let results = Object.entries(getVarProp(currExp.results))
        // type not starting with plot
        .filter(([res_name, obj]) => !obj.type.startsWith('plot/'))
        // after we've filtered fields, make it into an obj
        .map(([res_name, obj]) => { return {[res_name]: obj}; })
        // concat objs via reduce
        .reduce((o, curr) => Object.assign(o, curr), {})
        ;

        // get block parameters
        let blockVars = Object.entries(currExp.declaration.blocks)
        // only look at blocks with parameters
        .filter(([blockName, block]) => block.parameters)
        // loop through each block
        .map(([blockName, block]) => {
            // flatten the block's execution infos
            return Object.entries(block.parameters)
            .map(([fieldName, field]) => {
                return { [`${blockName}.${fieldName}`]: field };
            })
            .reduce((o, fieldObj) => Object.assign(o, fieldObj), {})
            ;
        })
        // flatten the exp's blocks' params
        .reduce((o, blockObj) => Object.assign(o, blockObj), {})
        ;

        // get global experiment variables
        let globalVars = Object.entries(currExp.declaration.globals)
        // properties that have a '/' in the name
        .filter(([name, obj]) => name.includes('/'))
        .map(([algName, alg]) => {
            // flatten the alg's fields
            return Object.entries(alg)
            .map(([fieldName, field]) => {
                return { [`${algName}.${fieldName}`]: field };
            })
            .reduce((o, fieldObj) => Object.assign(o, fieldObj), {})
            ;
        })
        // flatten the exp's algs' fields
        .reduce((o, algObj) => Object.assign(o, algObj), {})
        ;

        // get block timing infos
        let blockTiming = Object.entries(currExp.execution_info)
        // loop through each block
        .map(([blockName, block]) => {
            // flatten the block's execution infos
            return Object.entries(block)
            .map(([fieldName, field]) => {
                return { [`${blockName}.${fieldName}`]: field };
            })
            .reduce((o, fieldObj) => Object.assign(o, fieldObj), {})
            ;
        })
        // flatten the exp's execution infos
        .reduce((o, blockObj) => Object.assign(o, blockObj), {})
        ;

        // get total timing info
        let expTiming = {
            'total_execution_time': Object.entries(blockTiming)
            .filter(([name, val]) => name.includes('linear_execution_time'))
            .map(([name, val]) => val)
            .reduce((n, v) => n + v, 0)
        };

        // concat all objs
        let allObj = Object.assign({},
            results,
            globalVars,
            blockVars,
            blockTiming,
            expTiming
        );

        return allObj;
    };

    // get all result objs that have a type that starts with 'plot/'
    // these results can be displayed in a plot
    const getPlottablesFromExpName = (expName) => {
        return Object.entries(getVarProp(expData[expName].results))
        .filter(([res_name, obj]) => obj.type.startsWith('plot/'))
        .map(([res_name, obj]) => Object.assign(obj, { label: res_name }));
    };

    // make sure there are no invalid experiments in the report's groups
    // invalid -> in a group but not in the report
    const cleanGroups = (validExpNames) => {
        GroupsService.groups.forEach(g => {
            g.experiments.forEach(e => {
                // when the user can see the exp, the full name of the exp is available,
                // including the user who made it, and the toolchain
                const fullNameValid = validExpNames.includes(e);
                // when the exp isnt able to be seen by the user,
                // only the actual name of the exp is available
                const partNameValid = validExpNames.includes(e.split('/').pop());
                if(!(fullNameValid || partNameValid)){
                    g.removeExperiment(e);
                }
            });
        });
    };

    // fetch the exp data and process it
    const loadExperiments = () => {
        let expFetch;
        const namePath = UrlService.getByNamePath();
        const numberPath = UrlService.getByNumberPath();

        if(namePath && namePath.length > 0){
            // get exps by report num
            const user = namePath.split('/')[1];
            const name = namePath.split('/')[2];
            expFetch = experimentFactory.getAllExperimentResultsForAuthor(user, name, '');
        } else if(numberPath && numberPath.length > 0){
            //by report author
            const num = numberPath.split('/')[1];
            expFetch = experimentFactory.getAllExperimentResults('', num);
        }

        return expFetch
        .then((exps) => {
            // save the exp names as they were given by the server
            Object.keys(exps).forEach(e => rawExpNames.push(e));

            // process the given exp data
            Object.entries(exps)
            .forEach(([expName, exp]) => {
                // depending on the permissions of the user
                // when viewing this report
                // the author/toolchain may be hidden.
                // to be safe, always process
                // both the given name (either the full or short name)
                const shortName = expName.split('/').pop();
                for(let n of [shortName, expName]){
                    expData[n] = exp;
                    tableData[n] = getTableablesFromExpName(n);
                    plotData[n] = getPlottablesFromExpName(n);
                }
            });

            cleanGroups(rawExpNames);

            return expData;
        });
    };

    const cachedDeletedExperiments = [];

    const deleteExperiment = function(expName) {
        // delete from all groups
        GroupsService.groups.forEach(g => g.removeExperiment(expName));

        // delete from expNames
        const expNameIdx = this.experimentNames.indexOf(expName);
        this.experimentNames.splice(expNameIdx, 1);

        // delete from expData
        delete this.experiments[expName];

        cachedDeletedExperiments.push(expName);
    };

    loadExperiments();

    return {
        experimentNames: rawExpNames,
        experiments: expData,
        plottables: plotData,
        tableables: tableData,
        getAnalyzerFromExpName,
        loadExperiments,
        deleteExperiment,
        cachedDeletedExperiments
    };
}]);
