/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * GroupsService
 * Desc:
 *  The main datastore for the reports app, holding the tree of
 *  relationships between groups, experiments, aliases, and report items
 */
angular.module('reportApp').factory('GroupsService', ['reportFactory', function(reportFactory){
    let groupsServiceInstance = {};
    // experiments of reports are in arbitrary groups,
    // in a many-to-many relationship
    let groupData = [];

    // is the report editable?
    // set by the reportController when saving report data
    groupsServiceInstance.isEditable = undefined;
    groupsServiceInstance.setEditable = (val) => {
        if(val !== true && val !== false){
            throw new Error(`invalid isEditable: ${JSON.stringify(val)}`);
        }
        if(groupsServiceInstance.isEditable !== undefined){
            throw new Error(`isEditable already set: ${JSON.stringify(groupsServiceInstance.isEditable)}`);
        }
        groupsServiceInstance.isEditable = val;
    };

    // represents a Group in the report
    // has a name and a list of experiments that belong to it
    class Group {
        constructor (name) {
            this._name = name;
            this._analyzer = '';
            this._experimentNames = new Set();
            this._reportItems = [];
            this._aliases = {};
        }

        // get the experiment names in this group
        get experiments () {
            return Array.from(this._experimentNames);
        }

        // get the report items in this group
        get reportItems () {
            return Array.from(this._reportItems);
        }

        // get the group name
        get name () {
            return this._name;
        }

        // get the analyzer of the experiments in this group
        get analyzer () {
            return this._analyzer;
        }

        // gets the aliases for the experiments in the group
        get aliases () {
            return this._aliases;
        }

        set analyzer (analyzer) {
            this._analyzer = analyzer;
        }

        // add an exp to this group
        // optionally sets the analyzer
        // initializes the new exp's alias to its name
        addExperiment (expName, analyzer) {
            let res = this._experimentNames.add(expName);
            if(this._experimentNames.size === 1 && analyzer && analyzer.length > 0){
                this.analyzer = analyzer;
            }
            if(!this.aliases[expName]){
                let autoAlias = expName.split('/').pop();
                this.setAliasToExperiment(autoAlias, expName);
            }

            return res;
        }

        // rm an exp from this group as well as its alias
        removeExperiment (expName) {
            let res = this._experimentNames.delete(expName);
            if(this._experimentNames.size === 0){
                this.analyzer = '';
                this.reportItems.forEach(i => this.removeReportItem(i.id));
            }
            this.unsetExperimentAlias(expName);

            return res;
        }

        // add an item (table, plot, or text block) to this group
        // if the id already exists, it just replaces the old
        // element with the new one
        addReportItem (id, content) {
            let newEl = {
                id,
                content
            };

            let alreadyAddedEl = this._reportItems.find(i => i.id === id);
            if(alreadyAddedEl){
                let idx = this._reportItems.indexOf(alreadyAddedEl);
                return this._reportItems.splice(idx, 1, newEl);
            } else {
                return this._reportItems.push(newEl);
            }
        }

        // rm a report item from this group
        removeReportItem (id) {
            let idx = this._reportItems
            .indexOf(this._reportItems.find(o => o.id === id));

            return this._reportItems.splice(idx, 1);
        }

        // (re)sets an alias to an experiment in the group
        setAliasToExperiment (alias, expName) {
            if(!this.experiments.includes(expName)){
                return false;
            }

            this._aliases[expName] = alias;
        }

        // unsets an alias for an experiment
        unsetExperimentAlias (expName) {
            return delete this._aliases[expName];
        }
    };

    // gets groups
    groupsServiceInstance.groups = groupData;

    // serializes groups as an object with form:
    // {
    //  <group name 1>: {
    //      experiments: [],
    //      reportItems: [],
    //      analyzer: '',
    //      aliases: {},
    //      idx: 1
    //      },
    //  ...
    // }
    // the 'idx' saves the ordering of the groups
    groupsServiceInstance.serializeGroups = () => {
        return groupData
        .map((g, i) => { return {
            [g.name]: {
                experiments: g.experiments,
                reportItems: g.reportItems,
                analyzer: g.analyzer,
                aliases: g.aliases,
                idx: i
            }
        };
        })
        .reduce((o, g) => Object.assign(o, g), {});
    };

    // create a new group for the report
    // returns false if it already exists
    // returns the newly added group if successful
    groupsServiceInstance.createGroup = (name) => {
        if(typeof name !== 'string'){
            throw new Error(`new group name is not a string: ${JSON.stringify(name)}`);
        }

        if(groupData.find(g => g.name === name)){
            return false;
        }

        let g = new Group(name);

        groupData.push(g);
        return g;
    };

    // delete a group
    // via MUTATING the groupdata
    groupsServiceInstance.deleteGroup = (name) => {
        let idx = groupData.indexOf(groupData.find(g => g.name === name));
        if (idx > -1) {
            groupData.splice(idx, 1);
        }
    };

    // load group info from the serialized format:
    // {
    //  <group name 1>: {
    //      experiments: [],
    //      reportItems: [],
    //      analyzer: '',
    //      aliases: {},
    //      idx: 1
    //      },
    //  ...
    // }
    groupsServiceInstance.loadGroups = (data) => {
        // wipe data
        groupData.splice(0, groupData.length);
        let safeData = data || {};

        Object.entries(safeData)
        // sometimes we get an empty string for name for some reason
        .filter(([groupName, gData]) => groupName.length > 0)
        // sort using the saved index values
        .sort(([aName, a], [bName, b]) => {
            // if these are undefined, the sort func handles it
            let ia = a.idx;
            let ib = b.idx;
            return ia - ib;
        })
        .forEach(([groupName, gData]) => {
            let g = groupsServiceInstance.createGroup(groupName);

            // default group data to empty
            let analyzer = gData.analyzer || '';
            let experiments = gData.experiments || [];
            let reportItems = gData.reportItems || [];
            let aliases = gData.aliases || {};

            // save fields to group
            // by MUTATING the group obj
            g.analyzer = analyzer;
            experiments.forEach(n => g.addExperiment(n));
            reportItems.forEach(i => g.addReportItem(i.id, i.content));
            Object.entries(aliases).forEach(([e, a]) => g.setAliasToExperiment(a, e));
        });
    };

    return groupsServiceInstance;
}]);
