/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * PlotService
 * Desc:
 *  Manages the plots in the report,
 *  including:
 *  - Adding new/saved
 *  - Configuring
 *  - Deleting
 *  - Rendering
 */
angular.module('reportApp').factory('PlotService', ['UrlService', function(UrlService){
    const ps = {
        // these are provided by ReportService
        plotters: [],
        defaultPlotters: [],
        plotterParameters: [],
        reportNumber: undefined
    };

    // this 'queue' idea is the solution to trying to load plots before we receive all the data
    // from the server that we need (report info, exp info, etc.).
    // until the queue is processed (after which, no queue is needed),
    // plots that want to be rendered are added to a queue to wait.
    const queue = [];
    let noQueueNeeded = false;

    const getDefaults = (plotType) => ps.defaultPlotters.find(p => p.dataformat === plotType);

    // data to be used in to interact with the plot
    // users can change the plotter used if theres more than 1 available for that
    // plot type
    const getPossiblePlotters = (itemContent) =>
        ps.plotters.filter(p => p.dataformat === itemContent.type).map(p => p.name);

    // users can choose which params to use if more than 1 available
    // for that plotter
    const getPossibleConfigs = (plotter) =>
        ps.plotterParameters.filter(pp => plotter.id === pp.plotter).map(pp => pp.name);

    const getPlotter = (itemContent) => {
        const savedPlotterName = itemContent.savedPlotter;
        // defaults obj, in case we're using defaults
        const defaults = getDefaults(itemContent.type);

        // a plot obj can have different plotters,
        const plotter = ps.plotters.find(p => p.name === itemContent.savedPlotter) ||
            ps.plotters.find(p => p.name === defaults.plotter);

        return plotter;
    };

    const getPlotterConfig = (itemContent) => {
        // defaults obj, in case we're using defaults
        const defaults = getDefaults(itemContent.type);

        // each plotter obj can each have different configurations (plotter parameter instances)
        const config = ps.plotterParameters.find(pp => pp.name === itemContent.savedConfig) ||
            ps.plotterParameters.find(pp => pp.name === defaults.parameter);

        return config;
    };

    // constructs the payload to send to the server
    const constructPlotInfo = (group, itemId) => {
        const content = group.reportItems.find(i => i.id === itemId).content;
        const plotter = getPlotter(content);
        const config = getPlotterConfig(content);

        // the data to be sent to the server
        const requestData = {
            report_number: ps.reportNumber,
            // exps in the group
            experiment: group.experiments,
            // group's analyzer
            analyzer: [group.analyzer],
            // ?
            output: [content.name],
            // plotter to use
            plotter: plotter.name,
            // config to use
            parameter: config.name,
            // string for making the legend in the plot
            legend: group.experiments.map(e => group.aliases[e]).join('&'),
            // whether one plot with all the exps' data,
            // or one plot for each exp
            merged: content.merged === undefined ? true : content.merged
        };

        const possiblePlotters = getPossiblePlotters(content);
        const possibleConfigs = getPossibleConfigs(plotter);

        const returnStruct = [requestData, possiblePlotters, possibleConfigs];

        return returnStruct;
    };

    const fetchDownload = (requestData, contentType) => {
        const urlPrefix = '';

        // override the 'merged' property to always request the merged version.
        // requested the unmerged version only downloads the first image.
        requestData[0] = Object.assign({}, requestData[0], { merged: true });

        return new Promise((resolve, reject) => {
            beat.experiments.utils.getPlotData(
                // url_prefix
                urlPrefix,
                // spread out the request data to fill the next 3 spots
                ...requestData,
                // content type: png, jpeg, pdf
                contentType,
                // callback
                (...args) => {
                    // resolve promise
                    resolve(...args);
                }
            );
        });
    };

    // makes the call to the server, via some helper funcs found in the global namespace
    const fetchRender = (requestData, containerId, onRenderCallback) => {
        // the 'url_prefix' field found throughout the BEAT code is confusing,
        // but it seems to always be an empty string now
        const urlPrefix = '';//UrlService.getApiSegment().split('/').filter(s => s.length > 0).join('/');

        // this func *cannot* be promisified! the callback is fired whenever the plot is re-rendered.
        beat.experiments.utils.displayPlot(
            // url_prefix
            urlPrefix,
            // element to append render to
            document.querySelector(`#${containerId}`),
            // spread out the request data to fill the next 3 spots
            ...requestData,
            // dont replace inner content
            false,
            // callback
            (plotter, config, merged) => {
                onRenderCallback(plotter, config, merged);
            }
        );
    };

    // helper func to process a request to plot
    const processDownload = (group, itemId, contentType) => {
        const reqData = constructPlotInfo(group, itemId);

        return fetchDownload(reqData, contentType);
    };

    // helper func to process a request to plot
    const processItem = (group, itemId, containerId, onRenderCallback) => {
        const reqData = constructPlotInfo(group, itemId);

        fetchRender(reqData, containerId, onRenderCallback);
    };

    // used if we arent ready to directly service requests
    const addItemToQueue = (group, itemId, containerId, onRenderCallback) => {
        queue.push([
            group,
            itemId,
            containerId,
            onRenderCallback
        ]);
    };

    // called by ReportService,
    // or by someone having this report-level info.
    // processes the queue and sets the plotService state such that we dont need to use
    // the queue after this
    ps.processQueue = (rsPlotters, rsDefaultPlotters, rsPlotterParameters, rsReportNumber) => {
        rsPlotters.forEach(p => ps.plotters.push(p));
        rsDefaultPlotters.forEach(dp => ps.defaultPlotters.push(dp));
        rsPlotterParameters.forEach(pp => ps.plotterParameters.push(pp));
        ps.reportNumber = rsReportNumber;

        noQueueNeeded = true;
        const promises = queue.map(q => processItem(...q));
        queue.length = 0;

        return promises;
    };

    // chooses whether to add the plot request to a queue or service directly
    // args: group, itemId, containerId, onRenderCallback
    ps.addPlot = (...args) => noQueueNeeded ? processItem(...args) : addItemToQueue(...args);

    // args: group, itemId, contentType
    ps.downloadPlot = (...args) => processDownload(...args);

    ps.getPlotter = getPlotter;
    ps.getPlotterConfig = getPlotterConfig;
    ps.getPossiblePlotters = getPossiblePlotters;
    ps.getPossibleConfigs = getPossibleConfigs;

    return ps;
}]);
