/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * ReportService
 * Desc:
 *  Consumes the "report" object from the API and digests it into helper
 *  funcs and report-wide info. Basically an adaptor & bootstrap-er.
 */
angular.module('reportApp').factory('ReportService', ['GroupsService', 'plotterFactory', 'PlotService', 'reportFactory', 'UrlService', 'ErrorService', function(GroupsService, plotterFactory, PlotService, reportFactory, UrlService, ErrorService){
    const rs = {};

    rs.isAnonymous = undefined;
    rs.isOwner = undefined;
    rs.status = undefined;
    rs.number = undefined;
    rs.author = undefined;
    rs.name = undefined;

    rs.plotters = [];
    rs.defaultPlotters = [];
    rs.plotterParameters = [];

    // processed the report data received from the server,
    // and bootstraps the state of various services
    rs.processReport = (report) => {

        // useful info about the app
        rs.isAnonymous = report.anonymous;
        rs.isOwner = report.is_owner;
        rs.status = report.status;
        rs.number = report.number;
        rs.author = report.author;
        rs.name = report.name.split('/').length > 1 ? report.name.split('/')[1] : null;

        // start up our GroupsService
        GroupsService.loadGroups(report.content.groups);
        // if the report should not change,
        // add a nice layer of immutability by freezing the Group tree
        // (see GroupsService for more info)
        const isEditable = rs.isOwner && rs.status === 'editable' && !rs.isAnonymous;
        GroupsService.setEditable(isEditable);

        // fetch all our plotter data
        // these three fetches do not depend on eachother
        const pPlotters = plotterFactory.getPlotters('')
        .then(res => {
            res.data.forEach(p => rs.plotters.push(p));
        });

        const pDefaults = plotterFactory.getDefaultPlotters('')
        .then(res => {
            res.data.forEach(p => rs.defaultPlotters.push(p));
        });

        const pParams = plotterFactory.getPlotterParameter('')
        .then(res => {
            res.data.forEach(p => rs.plotterParameters.push(p));
        });

        // process the fetched plot info
        return Promise.all([pPlotters, pDefaults, pParams])
        .then(() => PlotService.processQueue(rs.plotters, rs.defaultPlotters, rs.plotterParameters, rs.number))
        ;
    };

    // fetch the report data using either the by-name or by-number scheme,
    // according to what URLService found
    rs.fetchReport = () => {
        const nameSeg = UrlService.getNameSegment();
        const numSeg = UrlService.getNumberSegment();

        if(nameSeg){
            // the nameSeg is something like 'report/<user>/<report name>/'
            let [user, reportId] = nameSeg.split('/').filter(s => s.length > 0).slice(1);

            return reportFactory.getReportInformation(user, reportId, '')
            .then(res => rs.processReport(res.data));
        } else if(numSeg) {
            // the numSeg is something like 'report/<report number>/'
            let [number] = numSeg.split('/').filter(s => s.length > 0).slice(1);

            return reportFactory.getReportInformationFromNumber(number, '')
            .then(res => rs.processReport(res.data));
        } else {
            throw new Error('UrlService could not parse the current URL');
        }
    };

    // publish the report
    rs.publishReport = (openSourceAlgs) => {
        return reportFactory.publishReport(
            '',
            rs.author,
            rs.name,
            openSourceAlgs.length && openSourceAlgs.length > 0 ? data : undefined
        )
        .catch(error => {
            throw error;
        });
        ;
    };

    // lock the report
    rs.lockReport = () => {
        return reportFactory.lockReport(
            rs.author,
            rs.name,
            ''
        )
        .catch(error => {
            throw error;
        });
        ;
    };

    rs.fetchReport()
    .catch(e => ErrorService.logError(e, `Could not load report.`));

    return rs;
}]);
