#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.core.urlresolvers import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from django.contrib.auth.models import User
from django.conf import settings

import simplejson as json

from .models import Report

from ..dataformats.models import DataFormat
from ..databases.models import Database
from ..algorithms.models import Algorithm
from ..toolchains.models import Toolchain
from ..experiments.models import Experiment
from ..experiments.models import Block
from ..attestations.models import Attestation
from ..plotters.models import Plotter
from ..plotters.models import PlotterParameter

from ..common.models import Shareable
from ..common.testutils import tearDownModule
from ..backend.models import Environment
from ..backend.models import Queue

from itertools import chain
from datetime import datetime

import os
import shutil


#----------------------------------------------------------


class ReportTestCase(APITestCase):

    DATABASE = {
        "root_folder": "/path/to/root_folder",
        "protocols": [
            {
                "name": "triple",
                "template": "test",
                "sets": [
                    {
                        "name": "default",
                        "template": "set",
                        "view": "dummy",
                        "outputs": {
                            "output1": "johndoe/single_integer/1",
                            "output2": "johndoe/single_integer/1",
                            "output3": "johndoe/single_integer/1"
                            }
                        }
                    ]
                }
            ]
        }


    TOOLCHAIN_ONE_ANALYZER = {
        "blocks": [ {
            "name": "addition1",
            "inputs": [
                "a",
                "b"
                ],
            "outputs": [
                "sum"
                ],
            "synchronized_channel": "dataset1"
            },
            {
                "name": "addition2",
                "inputs": [
                    "a",
                    "b"
                    ],
                "outputs": [
                    "sum"
                    ],
                "synchronized_channel": "dataset1"
                }
            ],
        "datasets": [ {
            "name": "dataset1",
            "outputs": [
                "output1",
                "output2",
                "output3"
                ]
            }
            ],
        "connections": [ {
                "from": "dataset1.output1",
                "to": "addition1.a",
                "channel": "dataset1"
            },
            {
                "from": "dataset1.output2",
                "to": "addition1.b",
                "channel": "dataset1"
                },
            {
                "from": "addition1.sum",
                "to": "addition2.a",
                "channel": "dataset1"
                },
            {
                "from": "dataset1.output3",
                "to": "addition2.b",
                "channel": "dataset1"
                },
            {
                "to": "analysis.input",
                "from": "addition2.sum",
                "channel": "dataset1"
                }
            ],
        "analyzers": [
                {
                    "inputs": [
                        "input"
                        ],
                    "synchronized_channel": "dataset1",
                    "name": "analysis"
                    }
                ],
        "representation": {
                "connections": {},
                "blocks": {},
                "channel_colors": {},
                },
        }


    TOOLCHAIN_TWO_ANALYZERS = {
        "blocks": [ {
            "name": "addition1",
            "inputs": [
                "a",
                "b"
                ],
            "outputs": [
                "sum"
                ],
            "synchronized_channel": "dataset1"
            },
            {
                "name": "addition2",
                "inputs": [
                    "a",
                    "b"
                    ],
                "outputs": [
                    "sum"
                    ],
                "synchronized_channel": "dataset1"
                }
            ],
        "datasets": [ {
            "name": "dataset1",
            "outputs": [
                "output1",
                "output2",
                "output3"
                ]
            }
            ],
        "connections": [ {
            "from": "dataset1.output1",
            "to": "addition1.a",
            "channel": "dataset1"
            },
            {
                "from": "dataset1.output2",
                "to": "addition1.b",
                "channel": "dataset1"
                },
            {
                "from": "addition1.sum",
                "to": "addition2.a",
                "channel": "dataset1"
                },
            {
                "from": "dataset1.output3",
                "to": "addition2.b",
                "channel": "dataset1"
                },
            {
                "to": "analysis.input",
                "from": "addition2.sum",
                "channel": "dataset1"
                },
            {
                "to": "analysis2.input",
                "from": "addition2.sum",
                "channel": "dataset1"
                }

            ],
        "analyzers": [
                {
                    "inputs": [
                        "input"
                        ],
                    "synchronized_channel": "dataset1",
                    "name": "analysis"
                    },
             {
                    "inputs": [
                        "input"
                        ],
                    "synchronized_channel": "dataset1",
                    "name": "analysis2"
                    }

                ],
        "representation": {
                "connections": {},
                "blocks": {},
                "channel_colors": {},
                },
        }


    ALGORITHM_DECLARATION = """{
  "language": "python",
  "splittable": false,
  "groups": [
    {
      "inputs": {
        "a": { "type": "johndoe/single_integer/1" },
        "b": { "type": "johndoe/single_integer/1" }
      },
      "outputs": {
        "sum": { "type": "johndoe/single_integer/1" }
      }
    }
  ],
  "parameters": {
  }
}"""


    ALGORITHM_CODE="""class Algorithm:

    def process(self, inputs, outputs):
        data = outputs['sum'].createData()
        data.value = inputs['a'].data.value + inputs['b'].data.value
        outputs['sum'].write(data)
        return True
"""


    ANALYZER_DECLARATION="""{
  "language": "python",
  "groups": [
    {
      "inputs": {
        "in": { "type": "johndoe/single_integer/1" }
      }
    }
  ],
  "results": {
    "out_float": { "type": "float32" },
    "out_text": { "type": "string" }
  },
  "parameters": {
  }
}"""


    ANALYZER_CODE="""class Algorithm:

    def process(self, inputs, output):
        # We don't really care
        return True
"""


    CONFIGURATION_ANALYZER1 = {
        "globals": {
        },
        "blocks": {
            "addition1": {
                "algorithm": "johndoe/sum/1",
                "parameters": {
                },
                "inputs": {
                    "a": "a",
                    "b": "b"
                },
                "outputs": {
                    "sum": "sum"
                }
            },
            "addition2": {
                "algorithm": "johndoe/sum/1",
                "parameters": {
                },
                "inputs": {
                    "a": "a",
                    "b": "b"
                },
                "outputs": {
                    "sum": "sum"
                }
            }
        },
        "datasets": {
            "dataset1": {
                "database": "integers/1",
                "protocol": "triple",
                "set": "default",
            }
        },
        "analyzers": {
            "analysis": {
                "algorithm": "johndoe/analyzer1/1",
                "parameters": {
                },
                "inputs": {
                    "in": "input"
                }
            }
        },
        "globals": {
            "environment": {
                "name": "env1",
                "version": "1.0"
            },
            "queue": "queue1"
        }
    }


    CONFIGURATION_ANALYZER2 = {
        "globals": {
        },
        "blocks": {
            "addition1": {
                "algorithm": "johndoe/sum/1",
                "parameters": {
                },
                "inputs": {
                    "a": "a",
                    "b": "b"
                },
                "outputs": {
                    "sum": "sum"
                }
            },
            "addition2": {
                "algorithm": "johndoe/sum/1",
                "parameters": {
                },
                "inputs": {
                    "a": "a",
                    "b": "b"
                },
                "outputs": {
                    "sum": "sum"
                }
            }
        },
        "datasets": {
            "dataset1": {
                "database": "integers/1",
                "protocol": "triple",
                "set": "default",
            }
        },
        "analyzers": {
            "analysis": {
                "algorithm": "johndoe/analyzer2/1",
                "parameters": {
                },
                "inputs": {
                    "in": "input"
                }
            }
        },
        "globals": {
            "environment": {
                "name": "env1",
                "version": "1.0"
            },
            "queue": "queue1"
        }
    }


    CONFIGURATION_TWO_ANALYZERS = {
        "globals": {
        },
        "blocks": {
            "addition1": {
                "algorithm": "johndoe/sum/1",
                "parameters": {
                },
                "inputs": {
                    "a": "a",
                    "b": "b"
                },
                "outputs": {
                    "sum": "sum"
                }
            },
            "addition2": {
                "algorithm": "johndoe/sum/1",
                "parameters": {
                },
                "inputs": {
                    "a": "a",
                    "b": "b"
                },
                "outputs": {
                    "sum": "sum"
                }
            }
        },
        "datasets": {
            "dataset1": {
                "database": "integers/1",
                "protocol": "triple",
                "set": "default",
            }
        },
        "analyzers": {
            "analysis": {
                "algorithm": "johndoe/analyzer1/1",
                "parameters": {
                },
                "inputs": {
                    "in": "input"
                }
            },
            "analysis2": {
                "algorithm": "johndoe/analyzer2/1",
                "parameters": {
                },
                "inputs": {
                    "in": "input"
                }
            }
        },
        "globals": {
            "environment": {
                "name": "env1",
                "version": "1.0"
            },
            "queue": "queue1"
        }
    }


    def setUp(self):
        self.tearDown()


        # Create the users
        self.password = '1234'

        self.johndoe = User.objects.create_user('johndoe', 'johndoe@test.org', self.password)
        self.jackdoe = User.objects.create_user('jackdoe', 'jackdoe@test.org', self.password)
        self.plot = User.objects.create_user('plot', 'plotdoe@test.org', self.password)


        # Create an environment and queue
        environment = Environment(name='env1', version='1.0')
        environment.save()
        environment.share()

        queue = Queue(name='queue1', memory_limit=1024, time_limit=60, cores_per_slot=1, max_slots_per_user=10)
        queue.save()

        queue.environments.add(environment)

        # Create a data format for user plot
        (dataformat_plot, errors) = DataFormat.objects.create_dataformat(
                author=self.plot,
                name='single_integer',
                short_description='description',
                declaration={
                    "value": "int32"
                    },
                )
        dataformat_plot.share()

        # Create a data format
        (dataformat, errors) = DataFormat.objects.create_dataformat(
                author=self.johndoe,
                name='single_integer',
                short_description='description',
                declaration={
                    "value": "int32"
                    },
                )

        dataformat.share()


        # Create the database
        database, errors = Database.objects.create_database(
            'integers',
            declaration=ReportTestCase.DATABASE,
            )

        database.share()
        assert database, errors


        # Create the algorithm
        (self.algorithm, errors) = Algorithm.objects.create_algorithm(
            author=self.johndoe,
            name='sum',
            short_description='description',
            declaration=ReportTestCase.ALGORITHM_DECLARATION,
            code=ReportTestCase.ALGORITHM_CODE
        )

        assert not errors, 'Algorithm errors: %s' % errors

        self.algorithm.share(True)


        # Create the analyzers
        (self.analyzer1, errors) = Algorithm.objects.create_algorithm(
            author=self.johndoe,
            name='analyzer1',
            short_description='description',
            declaration=ReportTestCase.ANALYZER_DECLARATION,
            code=ReportTestCase.ANALYZER_CODE)

        assert not errors, 'Analyzer errors: %s' % errors

        self.analyzer1.share(True)

        (self.analyzer2, errors) = Algorithm.objects.create_algorithm(
            author=self.johndoe,
            name='analyzer2',
            short_description='description',
            declaration=ReportTestCase.ANALYZER_DECLARATION,
            code=ReportTestCase.ANALYZER_CODE)

        assert not errors, 'Analyzer errors: %s' % errors


        # Create the toolchains
        (self.toolchain_one_analyzer, errors) = Toolchain.objects.create_toolchain(self.johndoe,
                'toolchain_one_analyzer', 'short description 1',
                declaration=ReportTestCase.TOOLCHAIN_ONE_ANALYZER)

        assert not errors, 'Toolchain errors: %s' % errors

        self.toolchain_one_analyzer.share()


        (self.toolchain_two_analyzers, errors) = Toolchain.objects.create_toolchain(self.johndoe,
                'toolctoolchain_two_analyzershain2', 'short description 1',
                declaration=ReportTestCase.TOOLCHAIN_TWO_ANALYZERS)

        assert not errors, 'Toolchain errors: %s' % errors


        # Create some experiments
        self.experiment_analyzer1 = self.create_experiment(self.johndoe, self.toolchain_one_analyzer,
                                                           'experiment_analyzer1',
                                                           ReportTestCase.CONFIGURATION_ANALYZER1)

        self.experiment_analyzer2 = self.create_experiment(self.johndoe, self.toolchain_one_analyzer,
                                                           'experiment_analyzer2',
                                                           ReportTestCase.CONFIGURATION_ANALYZER2)

        self.experiment_two_analyzers = self.create_experiment(self.johndoe, self.toolchain_two_analyzers,
                                                           'experiment_two_analyzers',
                                                           ReportTestCase.CONFIGURATION_TWO_ANALYZERS)

        self.inaccessible_experiment = self.create_experiment(self.jackdoe, self.toolchain_one_analyzer,
                                                              'inaccessible_experiment',
                                                              ReportTestCase.CONFIGURATION_ANALYZER1)

        self.no_attestation_other_user_accessible_experiment = self.create_experiment(self.jackdoe, self.toolchain_one_analyzer,
                                                              'no_attestation_other_user_accessible_experiment',
                                                              ReportTestCase.CONFIGURATION_ANALYZER1)

        self.no_attestation_other_user_accessible_experiment.share(users=[self.johndoe])
        self.attestation_other_user_accessible_experiment = self.create_experiment(self.jackdoe, self.toolchain_one_analyzer,
                                                              'attestation_other_user_accessible_experiment',
                                                              ReportTestCase.CONFIGURATION_ANALYZER1)
        self.attestation_other_user_accessible_experiment.share(users=[self.johndoe])
        self.no_attestation_other_user_public_experiment = self.create_experiment(self.jackdoe, self.toolchain_one_analyzer,
                                                              'no_attestation_other_user_public_experiment',
                                                              ReportTestCase.CONFIGURATION_ANALYZER1)

        self.no_attestation_other_user_public_experiment.share()
        Attestation.objects.create_attestation(self.attestation_other_user_accessible_experiment)

        # Create some plotters
        (plotter_1, errors) = Plotter.objects.create_plotter(author=self.plot, name="isoroc", declaration={
            "dataformat":"plot/single_integer/1",
            "description": "descr",
            "language":"python",
            "parameters":{}
            },
        )
        plotter_1.share(True)
        # Create some plotterparameters
        plotterparameter_1 = PlotterParameter(author=self.plot, name="plotterparameter", data={}, plotter=plotter_1)
        plotterparameter_1.save()

        # Create some reports
        (self.reportdoe_editable, details) = Report.objects.create_object(author=self.johndoe, name="report_editable")

        (self.reportdoe_locked, details) = Report.objects.create_object(author=self.johndoe, name="report_locked",
                                                                        experiments=[ self.experiment_analyzer1.fullname() ])
        self.reportdoe_locked.status = Report.LOCKED
        self.reportdoe_locked.save()

        (self.reportdoe_published, details) = Report.objects.create_object(author=self.johndoe, name="report_published",
                                                                           experiments=[ self.experiment_analyzer1.fullname() ])
        self.reportdoe_published.status = Report.PUBLISHED
        self.reportdoe_published.publication_date = datetime.now()
        self.reportdoe_published.save()

        (self.reportjack_editable, details) = Report.objects.create_object(author=self.jackdoe, name="reportjack_editable")


    def tearDown(self):
        for path in [settings.TOOLCHAINS_ROOT, settings.EXPERIMENTS_ROOT,
                     settings.DATAFORMATS_ROOT, settings.ALGORITHMS_ROOT,
                     settings.CACHE_ROOT]:
            if os.path.exists(path):
                shutil.rmtree(path)


    def create_experiment(self, author, toolchain, name, configuration, status=Experiment.DONE):
        (experiment, toolchain, errors) = Experiment.objects.create_experiment(
            author=author,
            toolchain=toolchain,
            name=name,
            short_description='',
            description='',
            declaration=configuration
        )

        if status != Experiment.PENDING:
            self.client.login(username=author.username, password=self.password)

            url = reverse('api_experiments:start', args=[author.username, toolchain.author.username,
                                                         toolchain.name, toolchain.version,
                                                         experiment.name])
            response = self.client.post(url)
            self.assertEqual(response.status_code, 200)

            self.client.logout()

            for block in experiment.blocks.all():
                block.status = Block.DONE
                block.save()

            experiment.status = status
            experiment.start_date = datetime.now()
            experiment.end_date = datetime.now()
            experiment.save()

        return experiment


#----------------------------------------------------------


class ReportListTestCase(ReportTestCase):

    def setUp(self):
        super(ReportListTestCase, self).setUp()

        self.url = reverse('api_reports:all')

    def test_anonymous_user(self):
        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(json.loads(response.content)), 1)
        for content in json.loads(response.content):
            report_name = content["name"]
            report_found_author = report_name.split("/")[0]
            report_found_status = report_name.split("/")[1].split("_")[1]
            self.assertEqual(report_found_author, self.johndoe.username)
            self.assertEqual(content['author'], report_found_author)
            self.assertEqual(report_found_status, 'published')
            self.assertEqual(content['status'], report_found_status)
            self.assertEqual(content['is_owner'], False)
            self.assertEqual(content['short_description'], "")
            self.assertEqual(content['description'], "")
            self.assertEqual(content['creation_date'] is not None, True)
            self.assertEqual(content['number'] is not None, True)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(json.loads(response.content)), 3)
        for content in json.loads(response.content):
            report_name = content["name"]
            report_found_author = report_name.split("/")[0]
            report_found_status = report_name.split("/")[1].split("_")[1]
            self.assertEqual(report_found_author, self.johndoe.username)
            self.assertEqual(content['author'], report_found_author)
            self.assertEqual(report_found_status in ['editable', 'published', 'locked'], True)
            self.assertEqual(content['status'], report_found_status)
            self.assertEqual(content['is_owner'], True)
            self.assertEqual(content['short_description'], "")
            self.assertEqual(content['description'], "")
            self.assertEqual(content['creation_date'] is not None, True)
            self.assertEqual(content['number'] is not None, True)


#----------------------------------------------------------


class ReportUserListTestCase(ReportTestCase):

    def setUp(self):
        super(ReportUserListTestCase, self).setUp()

        self.url = reverse('api_reports:list_create', kwargs={'owner_name': self.johndoe.username})

    def test_anonymous_user(self):
        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(json.loads(response.content)), 1)
        for content in json.loads(response.content):
            report_name = content["name"]
            report_found_author = report_name.split("/")[0]
            report_found_status = report_name.split("/")[1].split("_")[1]
            self.assertEqual(report_found_author, self.johndoe.username)
            self.assertEqual(content['author'], report_found_author)
            self.assertEqual(report_found_status, 'published')
            self.assertEqual(content['status'], report_found_status)
            self.assertEqual(content['is_owner'], False)
            self.assertEqual(content['short_description'], "")
            self.assertEqual(content['description'], "")
            self.assertEqual(content['creation_date'] is not None, True)
            self.assertEqual(content['number'] is not None, True)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(json.loads(response.content)), 3)
        for content in json.loads(response.content):
            report_name = content["name"]
            report_found_author = report_name.split("/")[0]
            report_found_status = report_name.split("/")[1].split("_")[1]
            self.assertEqual(report_found_author, self.johndoe.username)
            self.assertEqual(content['author'], report_found_author)
            self.assertEqual(report_found_status in ['editable', 'published', 'locked'], True)
            self.assertEqual(content['status'], report_found_status)
            self.assertEqual(content['is_owner'], True)
            self.assertEqual(content['short_description'], "")
            self.assertEqual(content['description'], "")
            self.assertEqual(content['creation_date'] is not None, True)
            self.assertEqual(content['number'] is not None, True)


#----------------------------------------------------------


class ReportCreationTestCase(ReportTestCase):

    def setUp(self):
        super(ReportCreationTestCase, self).setUp()

        self.url = reverse('api_reports:list_create', kwargs={'owner_name': self.johndoe.username})

        self.data = {
            "name": "report_new",
            'short_description': '',
            'description': '',
            'experiments': [],
            "content": {}
        }

        self.data_existing_name = {
            "name": "report_editable",
            'short_description': '',
            'description': '',
            'experiments': [],
            "content": {}
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        expected_answer = {
            'name': '%s/report_new' % self.johndoe.username,
            'url': reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                           'report_name': 'report_new'
                          }),
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_logged_in_user_existing_name(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data_existing_name, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_logged_in_user__one_experiment(self):
        expected_answer = {
            'name': '%s/report_new' % self.johndoe.username,
            'url': reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                           'report_name': 'report_new'
                          }),
        }

        data = {
            "name": "report_new",
            'short_description': '',
            'description': '',
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(json.loads(response.content), expected_answer)

        self.assertEqual(Report.objects.get(name='report_new').experiments.count(), 1)

    def test_logged_in_user__two_compatible_experiments(self):
        expected_answer = {
            'name': '%s/report_new' % self.johndoe.username,
            'url': reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                           'report_name': 'report_new'
                          }),
        }

        data = {
            "name": "report_new",
            'short_description': '',
            'description': '',
            'experiments': [
                self.experiment_analyzer1.fullname(),
                self.experiment_two_analyzers.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(json.loads(response.content), expected_answer)

        self.assertEqual(Report.objects.get(name='report_new').experiments.count(), 2)

    def test_logged_in_user__two_incompatible_experiments(self):
        data = {
            "name": "report_new",
            'short_description': '',
            'description': '',
            'experiments': [
                self.experiment_analyzer1.fullname(),
                self.experiment_analyzer2.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(len(response.content) > 0)

    def test_logged_in_user__one_experiment__two_analyzers(self):
        expected_answer = {
            'name': 'johndoe/report_new',
            'url': '/api/v1/reports/johndoe/report_new/'
        }

        data = {
            "name": "report_new",
            'short_description': '',
            'description': '',
            'experiments': [
                self.experiment_two_analyzers.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(json.loads(response.content), expected_answer)


#----------------------------------------------------------


class EditableReportRetrievalTestCase(ReportTestCase):

    def setUp(self):
        super(EditableReportRetrievalTestCase, self).setUp()

        self.report = self.reportdoe_editable

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        expected_answer = {
            "name": self.report.fullname(),
            "number": self.report.number,
            "short_description": u"",
            "description": u"",
            "html_description": u"",
            "is_owner": True,
            "author": self.johndoe.username,
            "status": "editable",
            "creation_date": self.report.creation_date.isoformat(),
            "publication_date": None,
            "experiments": [],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class LockedReportRetrievalTestCase(ReportTestCase):

    def setUp(self):
        super(LockedReportRetrievalTestCase, self).setUp()

        self.report = self.reportdoe_locked

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        expected_answer = {
            "name": self.report.fullname(),
            "number": self.report.number,
            "short_description": u"",
            "description": u"",
            "html_description": u"",
            "is_owner": True,
            "author": self.johndoe.username,
            "status": "locked",
            "creation_date": self.report.creation_date.isoformat(),
            "publication_date": None,
            "experiments": [
                self.experiment_analyzer1.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class PublishedReportRetrievalTestCase(ReportTestCase):

    def setUp(self):
        super(PublishedReportRetrievalTestCase, self).setUp()

        self.report = self.reportdoe_published

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        expected_answer = {
            "name": self.report.fullname(),
            "number": self.report.number,
            "short_description": u"",
            "description": u"",
            "html_description": u"",
            "is_owner": False,
            "author": self.johndoe.username,
            "status": "published",
            "creation_date": self.report.creation_date.isoformat(),
            "publication_date": self.report.publication_date.isoformat(),
            "experiments": [
                self.experiment_analyzer1.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_logged_in_user(self):
        expected_answer = {
            "name": self.report.fullname(),
            "number": self.report.number,
            "short_description": u"",
            "description": u"",
            "html_description": u"",
            "is_owner": True,
            "author": self.johndoe.username,
            "status": "published",
            "creation_date": self.report.creation_date.isoformat(),
            "publication_date": self.report.publication_date.isoformat(),
            "experiments": [
                self.experiment_analyzer1.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_different_user(self):
        expected_answer = {
            "name": self.report.fullname(),
            "number": self.report.number,
            "short_description": u"",
            "description": u"",
            "html_description": u"",
            "is_owner": False,
            "author": self.johndoe.username,
            "status": "published",
            "creation_date": self.report.creation_date.isoformat(),
            "publication_date": self.report.publication_date.isoformat(),
            "experiments": [
                self.experiment_analyzer1.fullname(),
            ],
            "content": {},
            "analyzer": None
        }

        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.get(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)


#----------------------------------------------------------


class EditableReportUpdateTestCase(ReportTestCase):

    def setUp(self):
        super(EditableReportUpdateTestCase, self).setUp()

        self.report = self.reportdoe_editable

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

        self.data = {
            "short_description": "shorty",
            "description": "lengthy",
            "experiments": [],
            "content": {}
        }

    def test_anonymous_user(self):
        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        expected_answer = {
            "name": self.report.fullname(),
            "short_description": "shorty",
            "url": self.url
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class LockedReportUpdateTestCase(ReportTestCase):

    def setUp(self):
        super(LockedReportUpdateTestCase, self).setUp()

        self.report = self.reportdoe_locked

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

        self.data = {
            "short_description": "shorty",
            "description": "lengthy",
            "experiments": [],
            "content": {}
        }

    def test_anonymous_user(self):
        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class PublishedReportUpdateTestCase(ReportTestCase):

    def setUp(self):
        super(PublishedReportUpdateTestCase, self).setUp()

        self.report = self.reportdoe_published

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

        self.data = {
            "short_description": "shorty",
            "description": "lengthy",
            "experiments": [],
            "content": {}
        }

    def test_anonymous_user(self):
        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.put(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class EditableReportDeletionTestCase(ReportTestCase):

    def setUp(self):
        super(EditableReportDeletionTestCase, self).setUp()

        self.report = self.reportdoe_editable

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class LockedReportDeletionTestCase(ReportTestCase):

    def setUp(self):
        super(LockedReportDeletionTestCase, self).setUp()

        self.report = self.reportdoe_locked

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class PublishedReportDeletionTestCase(ReportTestCase):

    def setUp(self):
        super(PublishedReportDeletionTestCase, self).setUp()

        self.report = self.reportdoe_published

        self.url = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.delete(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class EditableReportLockingTestCase(ReportTestCase):

    def setUp(self):
        super(EditableReportLockingTestCase, self).setUp()

        self.report = self.reportdoe_editable

        self.url = reverse('api_reports:lock', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

        self.url_add_experiment = reverse('api_reports:add_experiments', kwargs={'owner_name' : self.johndoe.username,
                                                      'report_name': self.report.name
                                                     })

        self.data_experiment = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

        self.data_experiment_with_other_user_unaccessible_experiment = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
                self.no_attestation_other_user_accessible_experiment.fullname(),
            ]
        }

        self.data_experiment_with_other_user_accessible_experiment = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
                self.attestation_other_user_accessible_experiment.fullname(),
            ]
        }

        self.data_experiment_with_other_user_public_experiment = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
                self.no_attestation_other_user_public_experiment.fullname(),
            ]
        }

        self.url_update_content = reverse('api_reports:object', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

        self.data_update_content = {
            "short_description": "shorty",
            "description": "lengthy",
            "experiments": [],
            "content": {
                        "chart1":{
                            "data":{
                                "plotter": "plot/isoroc/1"
                            },
                        "selected_template":"plot/plotterparameter/1"
                        }
            }
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user_empty_report_no_experiments(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        expected_answer = "Report is empty and has no experiments"
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_logged_in_user_empty_report_with_experiments(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url_add_experiment, self.data_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)


        response = self.client.post(self.url, format='json')
        expected_answer = "Report is empty"
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_logged_in_user_non_empty_report_with_experiments(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        expected_answer = {
            "name": self.report.fullname(),
            "short_description": "shorty",
            "url": self.url_update_content
        }

        response = self.client.put(self.url_update_content, self.data_update_content, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

        response = self.client.post(self.url_add_experiment, self.data_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Report.objects.get(name=self.report.name).status, Report.LOCKED)

    def test_logged_in_user_non_empty_report_unaccessible_attestation(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        expected_answer = {
            "name": self.report.fullname(),
            "short_description": "shorty",
            "url": self.url_update_content
        }

        response = self.client.put(self.url_update_content, self.data_update_content, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

        response = self.client.post(self.url_add_experiment, self.data_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        response = self.client.post(self.url_add_experiment, self.data_experiment_with_other_user_unaccessible_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        expected_answer = {
            "unusable_experiments":["jackdoe/johndoe/toolchain_one_analyzer/1/no_attestation_other_user_accessible_experiment"]
        }

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(json.loads(response.content), expected_answer)

        self.assertEqual(Report.objects.get(name=self.report.name).status, Report.EDITABLE)

    def test_logged_in_user_non_empty_report_accessible_attestation_not_public_experiment(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        expected_answer = {
            "name": self.report.fullname(),
            "short_description": "shorty",
            "url": self.url_update_content
        }

        response = self.client.put(self.url_update_content, self.data_update_content, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

        response = self.client.post(self.url_add_experiment, self.data_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        response = self.client.post(self.url_add_experiment, self.data_experiment_with_other_user_accessible_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user_non_empty_report_accessible_attestation_public_experiment(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        expected_answer = {
            "name": self.report.fullname(),
            "short_description": "shorty",
            "url": self.url_update_content
        }

        response = self.client.put(self.url_update_content, self.data_update_content, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

        response = self.client.post(self.url_add_experiment, self.data_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        response = self.client.post(self.url_add_experiment, self.data_experiment_with_other_user_public_experiment, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(Report.objects.get(name=self.report.name).status, Report.LOCKED)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class LockedReportLockingTestCase(ReportTestCase):

    def setUp(self):
        super(LockedReportLockingTestCase, self).setUp()

        self.report = self.reportdoe_locked

        self.url = reverse('api_reports:lock', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class PublishedReportLockingTestCase(ReportTestCase):

    def setUp(self):
        super(PublishedReportLockingTestCase, self).setUp()

        self.report = self.reportdoe_published

        self.url = reverse('api_reports:lock', kwargs={'owner_name' : self.johndoe.username,
                                                  'report_name': self.report.name
                                                 })

    def test_anonymous_user(self):
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class EditableReportPublishTestCase(ReportTestCase):

    def setUp(self):
        super(EditableReportPublishTestCase, self).setUp()

        self.report = self.reportdoe_editable

        self.url = reverse('api_reports:publish', kwargs={'owner_name' : self.johndoe.username,
                                                     'report_name': self.report.name
                                                    })

    def test_anonymous_user(self):
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class LockedReportPublishTestCase(ReportTestCase):

    def setUp(self):
        super(LockedReportPublishTestCase, self).setUp()

        self.report = self.reportdoe_locked

        self.url = reverse('api_reports:publish', kwargs={'owner_name' : self.johndoe.username,
                                                     'report_name': self.report.name
                                                    })

    def test_anonymous_user(self):
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Report.objects.get(name=self.report.name).status, Report.PUBLISHED)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class PublishedReportPublishTestCase(ReportTestCase):

    def setUp(self):
        super(PublishedReportPublishTestCase, self).setUp()

        self.report = self.reportdoe_published

        self.url = reverse('api_reports:publish', kwargs={'owner_name' : self.johndoe.username,
                                                     'report_name': self.report.name
                                                    })

    def test_anonymous_user(self):
        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class EditableReportAddExperimentsTestCase(ReportTestCase):

    def setUp(self):
        super(EditableReportAddExperimentsTestCase, self).setUp()

        self.report = self.reportdoe_editable

        self.url = reverse('api_reports:add_experiments', kwargs={'owner_name' : self.johndoe.username,
                                                      'report_name': self.report.name
                                                     })

        self.data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user__one_experiment__one_analyzer(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_logged_in_user__one_inaccessible_experiment(self):
        expected_answer = {
            "inaccessible_experiments": [
                self.inaccessible_experiment.fullname(),
            ]
        }

        data = {
            'experiments': [
                self.inaccessible_experiment.fullname(),
            ]
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

    def test_logged_in_user__two_experiments__compatible_analyzers(self):
        data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
                self.experiment_two_analyzers.fullname(),
            ]
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_logged_in_user__two_experiments__incompatible_analyzers(self):
        data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
                self.experiment_analyzer2.fullname(),
            ]
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue('content' not in response)

    def test_logged_in_user__one_experiment__two_analyzers(self):
        expected_answer = {
            "common_analyzers": [
                "johndoe/analyzer1/1",
                "johndoe/analyzer2/1"
            ]
        }

        data = {
            'experiments': [
                self.experiment_two_analyzers.fullname(),
            ]
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue('content' not in response)

    def test_logged_in_user__compatible_experiment_addition(self):
        data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        data = {
            'experiments': [
                self.experiment_two_analyzers.fullname(),
            ]
        }

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue('content' not in response)


    def test_logged_in_user__incompatible_experiment_addition(self):
        expected_answer = {
            "incompatible_experiments": [
                self.experiment_analyzer2.fullname(),
            ]
        }

        data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        data = {
            'experiments': [
                self.experiment_analyzer2.fullname(),
            ]
        }

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue('content' not in response)


    def test_with_unknown_experiments(self):
        unknown_experiment_name = 'unknown_author/unknown_toolchain_author/unknown_toolchain_name/0/unknown_name'
        expected_answer = {
            "unknown_experiments": [
                unknown_experiment_name
            ]
        }

        data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
                unknown_experiment_name
            ]
        }

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(json.loads(response.content), expected_answer)

#----------------------------------------------------------


class LockedReportAddExperimentsTestCase(ReportTestCase):

    def setUp(self):
        super(LockedReportAddExperimentsTestCase, self).setUp()

        self.report = self.reportdoe_locked

        self.url = reverse('api_reports:add_experiments', kwargs={'owner_name' : self.johndoe.username,
                                                      'report_name': self.report.name
                                                     })

        self.data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class PublishedReportAddExperimentsTestCase(ReportTestCase):

    def setUp(self):
        super(PublishedReportAddExperimentsTestCase, self).setUp()

        self.report = self.reportdoe_published

        self.url = reverse('api_reports:add_experiments', kwargs={'owner_name' : self.johndoe.username,
                                                      'report_name': self.report.name
                                                     })

        self.data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class EditableReportRemoveExperimentsTestCase(ReportTestCase):

    def setUp(self):
        super(EditableReportRemoveExperimentsTestCase, self).setUp()

        self.report = self.reportdoe_editable

        self.url = reverse('api_reports:remove_experiments', kwargs={'owner_name' : self.johndoe.username,
                                                         'report_name': self.report.name
                                                        })

        self.data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_with_unknown_experiments(self):
        data = self.data
        data['experiments'].append('unknown_author/unknown_toolchain_author/unknown_toolchain_name/0/unknown_name')

        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)



#----------------------------------------------------------


class LockedReportRemoveExperimentsTestCase(ReportTestCase):

    def setUp(self):
        super(LockedReportRemoveExperimentsTestCase, self).setUp()

        self.report = self.reportdoe_locked

        self.url = reverse('api_reports:remove_experiments', kwargs={'owner_name' : self.johndoe.username,
                                                         'report_name': self.report.name
                                                        })

        self.data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


#----------------------------------------------------------


class PublishedReportRemoveExperimentsTestCase(ReportTestCase):

    def setUp(self):
        super(PublishedReportRemoveExperimentsTestCase, self).setUp()

        self.report = self.reportdoe_published

        self.url = reverse('api_reports:remove_experiments', kwargs={'owner_name' : self.johndoe.username,
                                                         'report_name': self.report.name
                                                        })

        self.data = {
            'experiments': [
                self.experiment_analyzer1.fullname(),
            ]
        }

    def test_anonymous_user(self):
        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_logged_in_user(self):
        self.client.login(username=self.johndoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_different_user(self):
        self.client.login(username=self.jackdoe.username, password=self.password)

        response = self.client.post(self.url, self.data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
