#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


"""\
Starts the worker process.

Usage:
  %(prog)s [-v ... | --verbose ...] [--settings=<file>] [--period=<seconds>]
           [--environments=<path>] [--name=<name>]
  %(prog)s (-h | --help)
  %(prog)s (-V | --version)


Options:
  -h, --help                        Show this help message
  -V, --version                     Show program's version number
  -v, --verbose                     Increases the output verbosity level
  -S <file>, --settings=<file>      The module name to the Django settings
                                    file [default: beat.web.settings.settings]
  -e <path>, --environments=<path>  The path to the installation root of
                                    available environments.
  -n <name>, --name=<name>          The unique name of this worker on the
                                    database. This is typically the assigned
                                    hostname of the node, but not necessarily
                                    [default: %(hostname)s]
  -p <seconds, --period=<seconds>   The time, in seconds, in which this worker
                                    will probe the database for jobs to run or
                                    cancel. If not set, use the value available
                                    on the Django settings file, at the
                                    variable `WORKER_INTERVAL`.


Examples:

  To start the worker do the following:

    $ %(prog)s

  You can pass the ``-v`` flag to start the worker with the logging level set
  to ``INFO`` or ``-vv`` to set it to ``DEBUG``. By default, the logging level
  is set to ``WARNING`` if no ``-v`` flag is passed.

"""

import os
import sys
import time
import socket
import signal
import docopt
import logging

stop = False

def main(user_input=None):

    arguments = docopt.docopt(
      __doc__ % dict(
        prog=os.path.basename(sys.argv[0]),
        hostname=socket.gethostname(),
        ),
      )

    os.environ.setdefault('DJANGO_SETTINGS_MODULE', arguments['--settings'])
    from django.conf import settings
    from django import setup
    setup()

    logger = logging.getLogger('beat.web')
    if arguments['--verbose'] == 1: logger.setLevel(logging.INFO)
    elif arguments['--verbose'] >= 2: logger.setLevel(logging.DEBUG)

    # installs SIGTERM handler
    def handler(signum, frame):
        #ignore further signals
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        logger.info("Signal %d caught, terminating...", signum)
        global stop
        stop = True

    signal.signal(signal.SIGTERM, handler)
    signal.signal(signal.SIGINT, handler)

    from ..backend import utils
    from ..backend.models import Worker

    try:
        worker = Worker.objects.get(name=arguments['--name'])
    except Worker.DoesNotExist:
        logger.error("Cannot find worker `%s' in database, aborting",
            arguments['--name'])
        sys.exit(1)
    else:
        logger.info("Found worker `%s' in database, proceeding...",
            arguments['--name'])

    # figure out paths to programs I need to use
    process = utils.resolve_process_path()
    logger.debug("(path) process: `%s'", process)

    from django.utils import six
    paths = arguments['--environments']
    if isinstance(paths, six.string_types):
        paths = paths.split(os.pathsep)
    environments = utils.find_environments(paths)
    logger.debug("Environments: %s", ", ".join(environments))

    # check environments
    missing, unused = worker.check_environments(environments)
    if unused:
        logger.info("The following environments where found on your " \
            "setup, but will not be used with the current queue " \
            "configuration: %s" % ", ".join(unused))
    if missing:
        raise RuntimeError("The following environments are currently " \
            "missing from your setup: %s" % ", ".join(missing))
    else:
        logger.info("All required software environments were found")

    timing = int(arguments['--period']) \
        if arguments['--period'] else settings.WORKER_INTERVAL
    logger.info("Working at `%s' every %d seconds", arguments['--name'], timing)

    global stop
    with worker:

        while not stop:

            start = time.time()
            logger.debug("Starting work cycle...")
            worker.work(environments, process)
            duration = time.time() - start
            if duration < timing:
                time.sleep(timing - duration)

    logger.info("Gracefully exiting worker `%s'" % arguments['--name'])
