/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/**
 * beat.search.controls.js
 * Implementation of search controllers
 */

if (beat === undefined) var beat = {};
if (beat.search === undefined) beat.search = {};
if (beat.search.controls === undefined) beat.search.controls = {};


/**
 * Controls user display settings on the search display page
 *
 * Pre-conditions:
 *
 *   The webpage in which this controller operates, must have as many
 *   display-able fields as defined in the selector options (on the values).
 *   Each display-able DOM element must contain two classes: "result" and the
 *   one established at the value of the concerned selector option. All results
 *   should be, by default, hidden. This controller will enable them as
 *   required by default display settings or user saved settings.
 *
 *   Displaying/Hiding happens through 2 events: "beat.show" and "beat.hide",
 *   for which objects of type $(".result." + opt.val()) must respond to,
 *   showing and hiding as appropriate.
 *
 * The job:
 *
 *   One of the jobs of this controller is to monitor the select field and make
 *   sure the objects are displayed according to the user wishes. The second
 *   job plays a role when the user is logged-in: if the user wishes to save
 *   the search, this controller scrapes displayed objects and yields a
 *   settings list (respecting the input "options" format that can be used to
 *   replay the search later.
 *
 * Parameters
 *
 *   selector (String): A jQuery selector for the 'select' DOM element managing
 *     possible display options.
 *
 *   table (String): A jQuery selector for the 'table' holding tabbed results
 *     on the current search display page.
 *
 *   options (Array): An array of saved options stating the user preferences
 *     for a possibly saved search. This is empty in case of a new search.
 */
beat.search.controls.DisplaySettings = function(selector, table, options) {

  this.select = $(selector);
  this.options = options;

  //makes selector a chosen box
  this.select.chosen({
    search_contains: true,
    display_disabled_options: false,
    placeholder_text_multiple: 'Select results to display...',
  });

  //makes the table a datatable
  this.table = $(table).DataTable({
    paging: false,
    searching: false,
    bInfo: false,
    order: [[ 3, "desc" ]],
  });

  //unselects all selected options, let our options loop below handle them
  if (Array.isArray(this.options.analyzers) && this.options.analyzers.length) {
    this.select.find('option:selected').attr('selected', false);
    this.select.trigger('chosen:updated');
  }

  //looks-up pre-selected options and applies visibility
  this.select.find('option:selected').each(function(index, element) {
    var val = $(element).val();
    $('.result.' + val).trigger('beat.show');
  });

  //process option overrides for the user (in case the search was saved)
  var local_table = this.table; //so it is accessible inside here
  options.forEach(function(element) {
    if (element.charts !== undefined) {
      element.charts.forEach(function(chel) {
        //if equivalent element is not selected, select it, updates the
        //chosen display and triggers show event
        $(selector).find("option[data-old-settings~='" + chel.name + "']").each(
            function(index, opt) {
              $(opt).attr('selected', true);
              $(selector).trigger('chosen:updated');
              $('.result.' + $(opt).val()).trigger('beat.show', {
                plotter:chel.plotter,
                parameter:chel.parameter,
              });
            });
      });
    }

    var ordering = {};
    var unsorted_pos = 1000; //for unsorted items, keep the order of appearance
    if (element.columns !== undefined) {
      element.columns.forEach(function(coel) {
        //if equivalent element is not selected, select it, updates the
        //chosen display and triggers show event
        $(selector).find("option[data-old-settings~='" + coel.name + "']").each(
            function(index, opt) {
              $(opt).attr('selected', true);
              $(selector).trigger('chosen:updated');
              $('.result.' + $(opt).val()).trigger('beat.show');
            });

        //retrieves column position
        var colidx = $("th[data-old-settings~='" + coel.name + "']").index();
        if (coel.sort_index !== null && coel.sort_index >= 0 && colidx >= 0) {
          ordering[coel.sort_index] = [colidx, coel.ascending?'asc':'desc'];
        } else if (colidx >= 0) { //still consider it, but after sorted items
          ordering[unsorted_pos] = [colidx, coel.ascending?'asc':'desc'];
          unsorted_pos += 1;
        }
      });
    }
    //sets the ordering up as required by datatables
    if (!$.isEmptyObject(ordering)) {
      var datatables_ordering = []
      Object.keys(ordering).sort().forEach(function(key) {
        datatables_ordering.push(ordering[key]);
      });
      //apply to datatables
      local_table.order(datatables_ordering).draw();
    }
  });

  //hooks select/deselect so we react accordingly
  this.select.chosen().change(function(e, par) {
    if (par.selected !== undefined) {
      $('.result.' + par.selected).trigger('beat.show');
    }
    else if (par.deselected !== undefined) {
      $('.result.' + par.deselected).trigger('beat.hide');
    }
  });
}


/**
 * Scrapes the currently displayed page features and yields a list of options
 * to save the search.
 */
beat.search.controls.DisplaySettings.prototype.scrape = function() {

  var retval = {analyzers: [], charts: [], columns: []};

  //get the current ordering
  var order = this.table.order();

  //for each selected option, scrape details
  this.select.find('option:selected').each(function(index, element) {
    var el = $(element);
    var item = {name: el.data('old-settings')};

    if (el.data('result-type') == 'charts') {
      //scrape plotter and parameter
      var data = $('.' + el.val()).data();
      item['plotter'] = data.plotter;
      item['parameter'] = data.parameter;
      retval.charts.push(item);
    }
    else { //columns
      //scrape column order and sort index
      var colidx = $("th[data-old-settings~='" + el.data('old-settings') + "']").index();
      var found = false; //user has set a preference for this column
      order.forEach(function(order_element, order_index) {
        if (order_element[0] == colidx) {
          found = true;
          item['ascending'] = order_element[1] == 'asc';
          item['sort_index'] = order_index;
        }
      });
      if (!found) {
        item['ascending'] = true;
        item['sort_index'] = null;
      }
      retval.columns.push(item);
    }

    //insert analyzer if not already in
    if (retval.analyzers.indexOf(el.data('analyzer')) < 0) {
      retval.analyzers.push(el.data('analyzer'));
    }
  });

  return [retval];

}


/**
 * Controls a filter of any of these contexts: "algorithm", "analyzer",
 * "any-field", "database-name", "experiment-author", "protocol-name",
 * "toolchain-author", "toolchain"
 *
 * Parameters:
 *
 *   container (Object): A jQuery object at which we'll attach ourselves
 *   options (Array): A preset array of options from the user (maybe empty)
 *   fun (Function): A function to hook the ENTER button press on added text
 *     inputs (normally meaning "submit" the form)
 */
beat.search.controls.SimpleFilter = function(container, options, fun) {

  this.container = container;

  //attaches self to container
  this.dom_element = $(document.createElement('div'));
  this.dom_element.addClass('form-group form-group-sm search-filter');
  this.container.append(this.dom_element);

  this.not_select = $(document.createElement('select'));
  this.not_select.addClass('form-control');
  this.not_select.append($(document.createElement('option')).val('false').text('that contain').attr('selected', true));
  this.not_select.append($(document.createElement('option')).val('true').text("that don't contain"));
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Negate'));
  this.dom_element.append(this.not_select);

  this.exactly_select = $(document.createElement('select'));
  this.exactly_select.addClass('form-control');
  this.exactly_select.append($(document.createElement('option')).val('false').text('substring').attr('selected', true));
  this.exactly_select.append($(document.createElement('option')).val('true').text("exactly"));
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Exactly'));
  this.dom_element.append(this.exactly_select);

  this.text_input = $(document.createElement('input'));
  this.text_input.attr('type', 'text');
  this.text_input.addClass('form-control');
  this.text_input.keydown(function (event) {
    var keypressed = event.keyCode || event.which;
    if (keypressed == 13) fun();
  });
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Value'));
  this.dom_element.append(this.text_input);

  if (options === undefined) return;

  //apply user options
  switch(options.operator) {
    case 'contains':
      this.text_input.val(options.value);
      break;
    case 'contains-not':
      this.text_input.val(options.value);
      this.not_select.val('true');
      break;
    case 'contains-any-of':
      this.text_input.val(options.value.join());
      break;
    case 'contains-not-any-of':
      this.not_select.val('true');
      this.text_input.val(options.value.join());
      break;
    case 'is':
      this.exactly_select.val('true');
      this.text_input.val(options.value);
      break;
    case 'is-not':
      this.exactly_select.val('true');
      this.not_select.val('true');
      this.text_input.val(options.value);
      break;
    case 'is-any-of':
      this.exactly_select.val('true');
      this.text_input.val(options.value.join());
      break;
    case 'is-not-any-of':
      this.exactly_select.val('true');
      this.not_select.val('true');
      this.text_input.val(options.value.join());
      break;
  }

}


/**
 * Dettaches and detroyes representation from the DOM
 */
beat.search.controls.SimpleFilter.prototype.remove = function() {
  this.dom_element.remove();
}

/**
 * Scrapes the final state of this filter (for saving purposes)
 *
 * Returns
 *
 *   Object: containing the fields in order to recreate this filter
 */
beat.search.controls.SimpleFilter.prototype.scrape = function() {

  //updates the help text
  var negate = this.not_select.val() == 'true';
  var exact  = this.exactly_select.val() == 'true';
  var list   = this.text_input.val().indexOf(',') != -1;

  //decides what is the equivalent operator for the search
  var op = '';
  if (!negate && !exact && !list) op = 'contains';
  if (!negate && !exact &&  list) op = 'contains-any-of';
  if (!negate &&  exact && !list) op = 'is';
  if (!negate &&  exact &&  list) op = 'is-any-of';
  if ( negate && !exact && !list) op = 'contains-not';
  if ( negate && !exact &&  list) op = 'contains-not-any-of';
  if ( negate &&  exact && !list) op = 'is-not';
  if ( negate &&  exact &&  list) op = 'is-not-any-of';

  var val = this.text_input.val().split(',').filter(function(e){return e.trim();});

  return {
    operator: op,
    name: null, //not used in these contexts
    value: list?val:(val[0]||''),
  }

}


/**
 * Controls a filter for experiments. Includes the same base options as for the
 * SimpleFilter, but add the option to handle attestations
 *
 * Parameters:
 *
 *   container (Object): A jQuery object at which we'll attach ourselves
 *   options (Array): A preset array of options from the user (maybe empty)
 *   fun (Function): A function to hook the ENTER button press on added text
 *     inputs (normally meaning "submit" the form)
 */
beat.search.controls.ExperimentFilter = function(container, options, fun) {

  this.container = container;

  //attaches self to container
  this.dom_element = $(document.createElement('div'));
  this.dom_element.addClass('form-group form-group-sm search-filter');
  this.container.append(this.dom_element);

  this.master_select = $(document.createElement('select'));
  this.master_select.addClass('form-control');
  this.master_select.append($(document.createElement('option')).val('contain').text('that contain').attr('selected', true));
  this.master_select.append($(document.createElement('option')).val('not-contain').text("that don't contain"));
  this.master_select.append($(document.createElement('option')).val('attested').text('that are attested'));
  this.master_select.append($(document.createElement('option')).val('not-attested').text('that aren\'t attested'));
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Options'));
  this.dom_element.append(this.master_select);

  this.exactly_select = $(document.createElement('select'));
  this.exactly_select.addClass('form-control');
  this.exactly_select.append($(document.createElement('option')).val('false').text('substring').attr('selected', true));
  this.exactly_select.append($(document.createElement('option')).val('true').text("exactly"));
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Exactly'));
  this.dom_element.append(this.exactly_select);

  this.text_input = $(document.createElement('input'));
  this.text_input.attr('type', 'text');
  this.text_input.addClass('form-control');
  this.text_input.keydown(function (event) {
    var keypressed = event.keyCode || event.which;
    if (keypressed == 13) fun();
  });
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Value'));
  this.dom_element.append(this.text_input);

  //rig visibility of exactly_select and text_input to the master_select
  //if 'attested' or 'not-attested' chosen, no need to show them
  var _self = this;
  function _fix_inputs() {
    var val = _self.master_select.val();
    if (val == 'attested' || val == 'not-attested') {
      _self.exactly_select.hide();
      _self.text_input.hide();
    }
    else {
      _self.exactly_select.show();
      _self.text_input.show();
    }
  }
  this.master_select.on('change', _fix_inputs);

  if (options === undefined) return;

  //apply user options
  switch(options.operator) {
    case 'contains':
      this.text_input.val(options.value);
      break;
    case 'contains-not':
      this.text_input.val(options.value);
      this.master_select.val('not-contain');
      break;
    case 'contains-any-of':
      this.text_input.val(options.value.join());
      break;
    case 'contains-not-any-of':
      this.master_select.val('not-contain');
      this.text_input.val(options.value.join());
      break;
    case 'is':
      this.exactly_select.val('true');
      this.text_input.val(options.value);
      break;
    case 'is-not':
      this.exactly_select.val('true');
      this.master_select.val('not-contain');
      this.text_input.val(options.value);
      break;
    case 'is-any-of':
      this.exactly_select.val('true');
      this.text_input.val(options.value.join());
      break;
    case 'is-not-any-of':
      this.exactly_select.val('true');
      this.master_select.val('not-contain');
      this.text_input.val(options.value.join());
      break;
    case 'has-attestation':
      this.master_select.val('attested');
      _fix_inputs();
      break;
    case 'has-no-attestation':
      this.master_select.val('not-attested');
      _fix_inputs();
      break;
  }

}


/**
 * Dettaches and detroyes representation from the DOM
 */
beat.search.controls.ExperimentFilter.prototype.remove = function() {
  this.dom_element.remove();
}

/**
 * Scrapes the final state of this filter (for saving purposes)
 *
 * Returns
 *
 *   Object: containing the fields in order to recreate this filter
 */
beat.search.controls.ExperimentFilter.prototype.scrape = function() {

  var master = this.master_select.val();

  //if one is interested on attestations, returns immediately
  if (master == 'attested')
    return {operator:'has-attestation', name:null, value:null};
  else if (master == 'not-attested')
    return {operator:'has-no-attestation', name:null, value:null};

  //else, does the same as the simple filter
  var negate = master == 'not-contain';
  var exact  = this.exactly_select.val() == 'true';
  var list   = this.text_input.val().indexOf(',') != -1;

  //decides what is the equivalent operator for the search
  var op = '';
  if (!negate && !exact && !list) op = 'contains';
  if (!negate && !exact &&  list) op = 'contains-any-of';
  if (!negate &&  exact && !list) op = 'is';
  if (!negate &&  exact &&  list) op = 'is-any-of';
  if ( negate && !exact && !list) op = 'contains-not';
  if ( negate && !exact &&  list) op = 'contains-not-any-of';
  if ( negate &&  exact && !list) op = 'is-not';
  if ( negate &&  exact &&  list) op = 'is-not-any-of';

  var val = this.text_input.val().split(',').filter(function(e){return e.trim();});

  return {
    operator: op,
    name: null, //not used in these contexts
    value: list?val:(val[0]||''),
  }

}


/**
 * Controls a filter for experiment dates.
 *
 * Parameters:
 *
 *   container (Object): A jQuery object at which we'll attach ourselves
 *   options (Array): A preset array of options from the user (maybe empty)
 *   fun (Function): A function to hook the ENTER button press on added text
 *     inputs (normally meaning "submit" the form)
 */
beat.search.controls.ExperimentDateFilter = function(container, options, fun) {

  this.container = container;

  //attaches self to container
  this.dom_element = $(document.createElement('div'));
  this.dom_element.addClass('form-group form-group-sm search-filter');
  this.container.append(this.dom_element);

  this.master_select = $(document.createElement('select'));
  this.master_select.addClass('form-control');
  this.master_select.append($(document.createElement('option')).val('>').text('later than').attr('selected', true));
  this.master_select.append($(document.createElement('option')).val('>=').text('on or later than').attr('selected', true));
  this.master_select.append($(document.createElement('option')).val('<').text("earlier than"));
  this.master_select.append($(document.createElement('option')).val('<=').text("on or earlier than"));
  this.master_select.append($(document.createElement('option')).val('is').text('exactly on'));
  this.master_select.append($(document.createElement('option')).val('is-not').text("aren't exactly on"));
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Date Comparison Operator'));
  this.dom_element.append(this.master_select);

  var input_group = $(document.createElement('div')).addClass('input-group date');
  this.dom_element.append(input_group);
  this.text_input = $(document.createElement('input'));
  this.text_input.attr('type', 'text');
  this.text_input.addClass('form-control');
  this.text_input.keydown(function (event) {
    var keypressed = event.keyCode || event.which;
    if (keypressed == 13) fun();
  });
  input_group.append($(document.createElement('label')).addClass('sr-only').text('Date'));
  input_group.append(this.text_input);
  var icon = $(document.createElement('div')).addClass('input-group-addon');
  input_group.append(icon);
  icon.append($(document.createElement('span')).addClass("fa fa-calendar"));
  this.input_group = input_group;

  //use datepicker for this field
  this.datepicker_options = {
    weekStart: 1,
    todayBtn: "linked",
    clearBtn: true,
    todayHighlight: true,
    format: 'yyyy-mm-dd',
    multidateSeparator: ','
  };

  //on master select, controls multidate possibilities
  var _self = this;
  function _update_text_field() {
    var val = _self.master_select.val();
    switch (val) {
      case 'is':
      case 'is-not':
        _self.input_group.datepicker('remove');
        _self.input_group.datepicker($.extend(true,
              {multidate: true, autoclose: false}, _self.datepicker_options));
        _self.text_input.attr('placeholder', 'Click to set one or more dates...');
        break;
      default:
        _self.input_group.datepicker('remove');
        _self.input_group.datepicker($.extend(true,
              {multidate: false, autoclose: true}, _self.datepicker_options));
        _self.text_input.attr('placeholder', 'Click to set a date...');
    }
  }
  this.master_select.on('change', _update_text_field);
  _update_text_field();

  if (options === undefined) return;

  //apply user options
  switch(options.operator) {
    case '>':
    case '>=':
    case '<':
    case '<=':
    case 'is':
    case 'is-not':
      this.input_group.datepicker('update', options.value);
      this.master_select.val(options.operator);
      break;
    case 'is-any-of':
      this.input_group.datepicker('update', options.value.join());
      this.master_select.val('is');
      break;
    case 'is-not-any-of':
      this.input_group.datepicker('update', options.value.join());
      this.master_select.val('is-not');
      break;
  }
  _update_text_field();
}


/**
 * Dettaches and detroyes representation from the DOM
 */
beat.search.controls.ExperimentDateFilter.prototype.remove = function() {
  this.dom_element.remove();
}

/**
 * Scrapes the final state of this filter (for saving purposes)
 *
 * Returns
 *
 *   Object: containing the fields in order to recreate this filter
 */
beat.search.controls.ExperimentDateFilter.prototype.scrape = function() {

  var master = this.master_select.val();
  var list   = this.text_input.val().indexOf(',') != -1;

  //decides what is the equivalent operator for the search
  var op = null;
  if      (master == 'is'     && !list) op = 'is';
  else if (master == 'is'     &&  list) op = 'is-any-of';
  else if (master == 'is-not' && !list) op = 'is-not';
  else if (master == 'is-not' &&  list) op = 'is-not-any-of';
  else op = master; //<, <=, > or >=

  var val = this.text_input.val().split(',').filter(function(e){return e.trim();});

  return {
    operator: op,
    name: null, //not used in these contexts
    value: list?val:(val[0]||''),
  }

}


/**
 * Controls a filter for results
 *
 * Parameters:
 *
 *   container (Object): A jQuery object at which we'll attach ourselves
 *   options (Array): A preset array of options from the user (maybe empty)
 *   fun (Function): A function to hook the ENTER button press on added text
 *     inputs (normally meaning "submit" the form)
 */
beat.search.controls.ResultFilter = function(container, options, fun) {

  this.container = container;

  //attaches self to container
  this.dom_element = $(document.createElement('div'));
  this.dom_element.addClass('form-group form-group-sm search-filter');
  this.container.append(this.dom_element);

  this.result_input = $(document.createElement('input'));
  this.result_input.attr('type', 'text');
  this.result_input.addClass('form-control result-filter');
  this.result_input.attr('placeholder', 'Choose analyzer output...');
  this.dom_element.append(this.result_input);

  this.master_select = $(document.createElement('select'));
  this.master_select.addClass('form-control');
  this.master_select.append($(document.createElement('option')).val('>').text('bigger than').attr('selected', true));
  this.master_select.append($(document.createElement('option')).val('>=').text('bigger or equal to').attr('selected', true));
  this.master_select.append($(document.createElement('option')).val('<').text("smaller than"));
  this.master_select.append($(document.createElement('option')).val('<=').text("smaller or equal to"));
  this.master_select.append($(document.createElement('option')).val('is').text('is exactly'));
  this.master_select.append($(document.createElement('option')).val('is-not').text("isn't"));
  this.dom_element.append($(document.createElement('label')).addClass('sr-only').text('Value Comparison Operator'));
  this.dom_element.append(this.master_select);

  this.text_input = $(document.createElement('input'));
  this.text_input.attr('type', 'text');
  this.text_input.addClass('form-control result-filter');
  this.text_input.attr('placeholder', 'Choose a value...');
  this.dom_element.append(this.text_input);
  this.text_input.keydown(function (event) {
    var keypressed = event.keyCode || event.which;
    if (keypressed == 13) fun();
  });

  if (options === undefined) return;

  //apply user options
  this.result_input.val(options.name);
  switch(options.operator) {
    case '>':
    case '>=':
    case '<':
    case '<=':
    case 'is':
    case 'is-not':
      this.text_input.val(options.value);
      this.master_select.val(options.operator);
      break;
    case 'is-any-of':
      this.text_input.val(options.value.join());
      this.master_select.val('is');
      break;
    case 'is-not-any-of':
      this.text_input.val(options.value.join());
      this.master_select.val('is-not');
      break;
  }
}


/**
 * Dettaches and detroyes representation from the DOM
 */
beat.search.controls.ResultFilter.prototype.remove = function() {
  this.dom_element.remove();
}

/**
 * Scrapes the final state of this filter (for saving purposes)
 *
 * Returns
 *
 *   Object: containing the fields in order to recreate this filter
 */
beat.search.controls.ResultFilter.prototype.scrape = function() {

  var master = this.master_select.val();
  var list   = this.text_input.val().indexOf(',') != -1;

  //decides what is the equivalent operator for the search
  var op = null;
  if      (master == 'is'     && !list) op = 'is';
  else if (master == 'is'     &&  list) op = 'is-any-of';
  else if (master == 'is-not' && !list) op = 'is-not';
  else if (master == 'is-not' &&  list) op = 'is-not-any-of';
  else op = master; //<, <=, > or >=

  var val = this.text_input.val().split(',').filter(function(e){return e.trim();});

  return {
    operator: op,
    name: this.result_input.val(),
    value: list?val:(val[0]||''),
  }

}


/**
 * Controls filters on the user search view page
 *
 * The job:
 *
 *   This controller helps the user in adding or removing filters from the
 *   search page. By clicking 'Apply', the user re-submits the search with the
 *   changed filter set.
 *
 * Parameters
 *
 *   container (Object): A jQuery object at which we'll attach ourselves
 *   options (Array): A preset array of options from the user (maybe empty)
 *   fun (Function): A function to hook the ENTER button press on added text
 *     inputs (normally meaning "submit" the form)
 */
beat.search.controls.Filters = function(container, options, fun) {

  var forms = []; //all form rows
  var master_selectors = []; //controls for the type of filter
  var controllers = []; //individual option controllers

  var selectors = {
    'any-field': 'Any fields/anywhere',
    'algorithm': 'Algorithms',
    'analyzer': 'Analyzers',
    'database-name': 'Database names',
    'experiment': 'Experiments',
    'experiment-date': 'Experiment dates',
    'experiment-author': 'Experiment authors',
    'experiment-result': 'Experiment results',
    'protocol-name': 'Protocol names',
    'toolchain-author': 'Toolchain authors',
    'toolchain': 'Toolchains',
  };

  var simple_types = [
    'any-field',
    'algorithm',
    'analyzer',
    'database-name',
    'experiment-author',
    'protocol-name',
    'toolchain-author',
    'toolchain',
  ];

  //attaches self to container
  var dom_element = $(document.createElement('div'));
  container.append(dom_element);

  //private function to add a filter row, the optional value `opt' allows the
  //function to properly configure the row with the user preferences. The
  //optional value `index' allows the user to specify the insertion position,
  //that will happen *at* that element. So, if you specify 2, the new row
  //will have index == 2.
  function _add_row(opt, index) {

    var form = $(document.createElement('div'));
    form.addClass('form-inline');

    //inserts the new form respecting the requested position
    if (index === undefined || dom_element.children().length <= index)
      dom_element.append(form);
    else dom_element.children(':eq(' + (index-1) + ')').after(form);

    var form_group = $(document.createElement('div')).addClass('form-group form-group-sm action-buttons');
    form.append(form_group);

    //buttons for adding and removing filters
    var add_button = $(document.createElement('a')).addClass('btn btn-default btn-sm btn-add').html('<i class="fa fa-plus"></i>');
    add_button.attr('title', 'Add another filter after this one');
    add_button.data('toggle', 'tooltip');
    add_button.data('placement', 'top');
    add_button.tooltip();
    form_group.append(add_button);
    var remove_button = $(document.createElement('a')).addClass('btn btn-default btn-sm btn-delete').html('<i class="fa fa-times"></i>');
    remove_button.attr('title', 'Remove this filter');
    remove_button.data('toggle', 'tooltip');
    remove_button.data('placement', 'top');
    remove_button.tooltip();
    form_group.append(remove_button);

    var master_select = $(document.createElement('select')).addClass('form-control');
    form_group.append(master_select);
    for (var key in selectors) {
      master_select.append($(document.createElement('option')).val(key).text(selectors[key]));
    }

    var controller = null;
    if (opt !== undefined) {
      if (simple_types.indexOf(opt.context) >= 0) {
        master_select.val(opt.context);
        controller = new beat.search.controls.SimpleFilter(form, opt, fun);
      }
      else if (opt.context == 'experiment') {
        master_select.val(opt.context);
        controller = new beat.search.controls.ExperimentFilter(form, opt, fun);
      }
      else if (opt.context == 'experiment-date') {
        master_select.val(opt.context);
        controller = new beat.search.controls.ExperimentDateFilter(form, opt, fun);
      }
      else if (opt.context == 'experiment-result') {
        master_select.val(opt.context);
        controller = new beat.search.controls.ResultFilter(form, opt, fun);
      }
    }
    else {
      //the default
      master_select.val('any-field');
      controller = new beat.search.controls.SimpleFilter(form, [], fun);
    }

    //rig master select boxes - on change, fix controller
    master_select.on('change', function() {
      var val = $(this).val();
      var ind = $(this).parent().parent().index(); //form-inline index
      controllers[ind].remove(); //remove DOM elements
      if (simple_types.indexOf(val) >= 0) {
        controllers[ind] = new beat.search.controls.SimpleFilter(form, opt, fun);
      }
      else if (val == 'experiment') {
        controllers[ind] = new beat.search.controls.ExperimentFilter(form, opt, fun);
      }
      else if (val == 'experiment-date') {
        controllers[ind] = new beat.search.controls.ExperimentDateFilter(form, opt, fun);
      }
      else if (val == 'experiment-result') {
        controllers[ind] = new beat.search.controls.ResultFilter(form, opt, fun);
      }
    });

    //Carefully insert the new controller, master_selector and forms so
    //it properly relates to the display order. We use that to find them
    //back upon user actions.
    if (index !== undefined) {
      controllers.splice(index, 0, controller);
      master_selectors.splice(index, 0, master_select);
      forms.splice(index, 0, form);
    }
    else {
      controllers.push(controller);
      master_selectors.push(master_select);
      forms.push(form);
    }

    //rig buttons so adding and removing works as expected
    remove_button.on('click', function() {
      var ind = $(this).parent().parent().index(); //form-inline index
      //remove controllers and then, dom element
      controllers.splice(ind, 1);
      master_selectors.splice(ind, 1);
      forms[ind].remove();
      forms.splice(ind, 1);
      //if no more lines left, insert a new one
      if (!controllers.length) _add_row();
    });

    add_button.on('click', function() {
      _add_row(undefined, $(this).parent().parent().index()+1);
    });

  }

  options.forEach(_add_row);

  if (!options || (Array.isArray(options) && !options.length)) { //we add at least one filter row with default settings
    _add_row();
  }

  //register object properties
  this.container = container;
  this.dom_element = dom_element;
  this.forms = forms;
  this.master_selectors = master_selectors;
  this.controllers = controllers;

}


/**
 * Returns the currently displayed configuration.
 */
beat.search.controls.Filters.prototype.scrape = function() {
  var retval = [];
  var len = this.master_selectors.length;
  for (i=0; i<len; ++i) {
    var obj = {context: this.master_selectors[i].val()};
    $.extend(obj, this.controllers[i].scrape());
    retval.push(obj);
  }
  return retval;
}


/**
 * Uses bootstrap3 to display a modal window to save a search. The user is
 * presented with a "typeahead" input box so they can overwrite the current
 * search or create a new term. The typeahead auto-completion works for the
 * list of searches the user current has on their portfoglio. If the search was
 * saved before, the name will be pre-selected.
 *
 * Parameters:
 *
 *   filters (Filters): An instance of the Filters class as declared above
 *   settings (DisplaySettings): An instance of the DisplaySettings as declared
 *     above
 *   leaderboard (Boolean): Indicates if we should create a leaderboard
 *     associated to this search.
 *   list_url (String): A string with the list URL to retrieve existing search
 *     names
 *   save_url (String): A string with the save URL to use for saving the search
 *   username (String): The name of the user requesting the operation
 *   name (String, Optional): If set, indicates the name of the currently saved
 *     search that relates to the currently displayed results.
 *
 **/
beat.search.controls.modal_save = function (filters, settings, leaderboard, list_url, save_url, username, name) {

  //retrieve list of existing searches - if that succeeds, construct modal form
  function saveAs() {

    return $.ajax({
      type: 'GET',
      url: list_url + '?fields=name,update_url,short_description',

    }).pipe(function(data) {

      var message = $(document.createElement('div'));
      message.append($(document.createElement('p')).text('By clicking Save, you will save the current settings under the defined search name. If the search record already exists on your account, it will be overwritten. You can cancel the operation by clicking Cancel.'));
      var form_group = $(document.createElement('div'));
      form_group.addClass('form-group has-feedback');
      message.append(form_group);
      var input_group = $(document.createElement('div'));
      input_group.addClass('input-group');
      input_group.append($(document.createElement('span')).addClass('input-group-addon').text(username + ' /'));
      var input = $(document.createElement('input')).addClass('form-control');
      input.attr('type', 'text');
      input.attr('placeholder', 'Start typing for auto-completion...');
      if (name) input.val(name.split('/')[1]);
      input_group.append(input);
      glyph = $(document.createElement('span'));
      glyph.addClass('glyphicon glyphicon-ok form-control-feedback');
      glyph.attr('aria-hidden', 'true');
      input_group.append(glyph);
      form_group.append(input_group);

      var available_searches = {};
      data.forEach(function(i){available_searches[i.name.split('/')[1]]=i;});
      var available_search_keys = Object.keys(available_searches);

      //visual feedback if new or overwriting...
      var help = $(document.createElement('span')).addClass('help');
      form_group.append(help);

      var update_help = function() {
        var currval = input.val();
        if (!currval.length) {
          help.text('');
          form_group.removeClass('has-warning');
          form_group.removeClass('has-success');
          form_group.addClass('has-error');
          glyph.removeClass('glyphicon-ok');
          glyph.removeClass('glyphicon-warning-sign');
          glyph.addClass('glyphicon-remove');
        }
        else if (available_search_keys.indexOf(currval) >= 0) {
          var destname = username + '/' + currval;
          help.text('Action: Overwritting existing search "' + destname + '"');
          form_group.removeClass('has-error');
          form_group.removeClass('has-success');
          form_group.addClass('has-warning');
          glyph.removeClass('glyphicon-ok');
          glyph.removeClass('glyphicon-remove');
          glyph.addClass('glyphicon-warning-sign');
        }
        else {
          var destname = username + '/' + currval;
          help.text('Action: Saving new search "' + destname + '"');
          form_group.removeClass('has-error');
          form_group.removeClass('has-warning');
          form_group.addClass('has-success');
          glyph.removeClass('glyphicon-remove');
          glyph.removeClass('glyphicon-warning-sign');
          glyph.addClass('glyphicon-ok');
        }
      }
      update_help(); //checks currently set value
      input.on('keyup', update_help);

      //using bootstrap3-typeahead
      input.typeahead({
        source: Object.keys(available_searches),
        items: 5, //how many to show at a time
        minLength: 0, //show suggestions even if no text available
        autoSelect: false, //user may want to leave it where it is
        afterSelect: update_help, //updates after the user selected an item
        showHintOnFocus: true, //dropdown as soon as focus on
      });

      BootstrapDialog.show({
        title: '<i class="fa fa-floppy-o fa-lg"> Select a name',
        message: message,
        type: BootstrapDialog.TYPE_PRIMARY,
        buttons: [
        {
          label: 'Cancel',
          cssClass: 'btn-default',
          action: function(the_dialog) {
            the_dialog.close();
            return false;
          },
        },
        {
          label: 'Save',
          cssClass: 'btn-primary',
          action: function(the_dialog) {
            var val = input.val();
            if (!val) {
              BootstrapDialog.alert({
                title: '<i class="fa fa-warning"></i> Error',
                message: 'You must set a name to save your search settings',
                type: BootstrapDialog.TYPE_WARNING,
              });
              return false;
            }
            the_dialog.close();

            if (val) {

              var post_info = {
                filters: filters.scrape(),
                settings: settings.scrape(),
                leaderboard: leaderboard,
              };

              var url = save_url;
              var method = 'POST';
              var idx = available_search_keys.indexOf(val);
              if (idx >= 0) {
                url = available_searches[val].update_url;
                method = 'PUT';
              }
              else { //also include the name
                post_info.name = val;
              }

              var d = $.ajax({
                type: method,
                url: url,
                data: JSON.stringify(post_info),
                contentType: "application/json; charset=utf-8",
                dataType: "json"
              });

              d.done(function(data, status) {
                if (data.url !== undefined) {
                  //loads new location so the user can set doc/sharing
                  window.location.href = data.url;
                  return true;
                }
                else {
                  //if a message pane is available, use it to display success
                  $('#messages').append(make_alert('<i class="fa fa-check"></i> <span>Successfuly stored search "' + username + '/' + val + '"</span>', true, 'success'));
                }
                return true;
              });

              d.fail(process_error);

            } //val is set
          },
        },
        ],
      });
    });
  }

  return saveAs().fail(process_error);

}
