#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import re
import datetime

from django.db.models import Q
from django.db.models.query import QuerySet

UNLIKELY_STRING = "ZaqXsw<>?:{}|!@#$%^&*()-={}"


#-------------------------------------------------------------------------------
# Helper class to create an AND'ed combination of Q filters
#
# It is destined to create filters corresponding to contribution names,
# containing optional '*' wildcards characters.
#-------------------------------------------------------------------------------
class FilterGenerator(object):

    FILTER_IEXACT      = 0
    FILTER_ICONTAINS   = 1
    FILTER_ISTARTSWITH = 2
    FILTER_IENDSWITH   = 3


    def __init__(self):
        self.filter = None


    #---------------------------------------------------------------------------
    # Process a contribution name, optionally containing '*' wildcards
    # characters, and returns the corresponding filter
    #
    # The contribution name can be in one of the following forms:
    #
    #   - author/name/version
    #   - author/name
    #   - name/version
    #   - name
    #
    # The 'exact' parameters prevents '*' wildcards to be added at the
    # beginning and/or end of the relevant contribution name parts.
    #---------------------------------------------------------------------------
    def process_contribution_name(self, contribution_name, exact=False):
        self.filter = None

        (author, name, version) = self.parse_contribution_name(contribution_name, exact)

        self.add(author, FilterGenerator.create_author_filter)
        self.add(name, FilterGenerator.create_name_filter)

        if version is not None:
            self.filter = self.filter & Q(version=version)

        return self.filter


    #---------------------------------------------------------------------------
    # Process a dataset name, optionally containing '*' wildcards characters,
    # and returns the corresponding filter
    #
    # The dataset name can be in one of the following forms:
    #
    #   - database.protocol.set
    #   - database.protocol
    #   - database
    #
    # The 'exact' parameters prevents '*' wildcards to be added at the
    # beginning and/or end of the relevant dataset name parts.
    #---------------------------------------------------------------------------
    def process_dataset_name(self, dataset_name, exact=False):
        self.filter = None

        (database, version, protocol, dataset) = self.parse_dataset_name(dataset_name, exact)

        self.add(database, FilterGenerator.create_name_filter)

        if version is not None:
            self.filter = self.filter & Q(version=version)

        self.add(protocol, FilterGenerator.create_protocol_filter)
        self.add(dataset, FilterGenerator.create_dataset_filter)

        return self.filter


    #---------------------------------------------------------------------------
    # Parse a contribution name, optionally containing '*' wildcards characters
    #
    # The contribution name can be in one of the following forms:
    #
    #   - author/name/version
    #   - author/name
    #   - name/version
    #   - name
    #
    # This method returns a tuple (author, name, version), with irrelevant
    # fields set to None.
    #
    # The 'exact' parameters prevents '*' wildcards to be added at the
    # beginning and/or end of the relevant contribution name parts.
    #---------------------------------------------------------------------------
    def parse_contribution_name(self, contribution_name, exact=False):
        parts = contribution_name.split('/')

        author  = None
        name    = None
        version = None

        if len(parts) == 3:
            author = parts[0]

            if not(exact) and (author[0] != '*'):
                author = '*' + author

            name = parts[1]

            try:
                version = int(parts[2])
            except:
                pass

        elif len(parts) == 2:
            try:
                version = int(parts[1])
                name = parts[0]

                if not(exact) and (name[0] != '*'):
                    name = '*'  + name
            except:
                author = parts[0]
                if not(exact) and (author[0] != '*'):
                    author = '*' + author

                name = parts[1]
                if not(exact) and (name[-1] != '*'):
                    name = name + '*'

        else:
            name = parts[0]

            if not(exact):
                if name[0] != '*':
                    name = '*'  + name

                if name[-1] != '*':
                    name = name + '*'

        return (author, name, version)


    #---------------------------------------------------------------------------
    # Parse a dataset name, optionally containing '*' wildcards characters
    #
    # The dataset name can be in one of the following forms:
    #
    #   - database/version.protocol.set
    #   - database.protocol.set
    #   - database/version.protocol
    #   - database.protocol
    #   - database/version
    #   - database
    #
    # This method returns a tuple (database, version, protocol, dataset), with
    # irrelevant fields set to None.
    #
    # The 'exact' parameters prevents '*' wildcards to be added at the
    # beginning and/or end of the relevant dataset name parts.
    #---------------------------------------------------------------------------
    def parse_dataset_name(self, dataset_name, exact=False):
        parts = re.split(r'[\.@]', dataset_name)

        database = None
        version  = None
        protocol = None
        dataset  = None

        if len(parts) >= 3:
            database = parts[0]
            if not(exact) and (database[0] != '*'):
                database = '*' + database

            protocol = parts[1]

            dataset = '.'.join(parts[2:])
            if not(exact) and (dataset[-1] != '*'):
                dataset = dataset + '*'

        elif len(parts) == 2:
            database = parts[0]
            if not(exact) and (database[0] != '*'):
                database = '*' + database

            protocol = parts[1]
            if not(exact) and (protocol[-1] != '*'):
                protocol = protocol + '*'

        else:
            database = parts[0]

            if not(exact):
                if database[0] != '*':
                    database = '*'  + database

                if database[-1] != '*':
                    database = database + '*'

        if database is not None:
            parts = database.split('/')
            if len(parts) == 2:
                if parts[1][-1] == '*':
                    parts[0] += '*'
                    parts[1] = parts[1][:-1]

                database = parts[0]

                try:
                    version = int(parts[1])
                except:
                    pass

        return (database, version, protocol, dataset)


    #---------------------------------------------------------------------------
    # Process a (string) part of a contribution name, and invoke the provided
    # callback with the correct creation information
    #
    # The signature of the creation callback is:
    #   creation_callback(filter_type, value)
    #
    # The creation callback must return a Q filter corresponding to the
    # supplied type.
    #
    # Note: the value can be None, in which case the filter isn't changed.
    #---------------------------------------------------------------------------
    def add(self, value, creation_callback):

        def _add(new_filter):
            if self.filter is None:
                self.filter = new_filter
            else:
                self.filter = self.filter & new_filter

        if value is None:
            return

        start = (value[0] == '*')
        end   = (value[-1] == '*')

        parts = filter(lambda x: len(x) > 0, value.split('*'))

        if len(parts) == 0:
            return
        elif len(parts) == 1:
            if start and end:
                _add(creation_callback(FilterGenerator.FILTER_ICONTAINS, parts[0]))
            elif start:
                _add(creation_callback(FilterGenerator.FILTER_IENDSWITH, parts[0]))
            elif end:
                _add(creation_callback(FilterGenerator.FILTER_ISTARTSWITH, parts[0]))
            else:
                _add(creation_callback(FilterGenerator.FILTER_IEXACT, parts[0]))
        else:
            if start:
                _add(creation_callback(FilterGenerator.FILTER_ICONTAINS, parts[0]))
            else:
                _add(creation_callback(FilterGenerator.FILTER_ISTARTSWITH, parts[0]))

            if end:
                _add(creation_callback(FilterGenerator.FILTER_ICONTAINS, parts[-1]))
            else:
                _add(creation_callback(FilterGenerator.FILTER_IENDSWITH, parts[-1]))

            for part in parts[1:-1]:
                _add(creation_callback(FilterGenerator.FILTER_ICONTAINS, part))


    #---------------------------------------------------------------------------
    # Creation callback for 'author__username__*'
    #---------------------------------------------------------------------------
    @staticmethod
    def create_author_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(author__username__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(author__username__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(author__username__iendswith=value)
        else:
            return Q(author__username__iexact=value)


    #---------------------------------------------------------------------------
    # Creation callback for 'name__*'
    #---------------------------------------------------------------------------
    @staticmethod
    def create_name_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(name__iendswith=value)
        else:
            return Q(name__iexact=value)


    #---------------------------------------------------------------------------
    # Creation callback for 'toolchain__author__username__*'
    #---------------------------------------------------------------------------
    @staticmethod
    def create_toolchain_author_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(toolchain__author__username__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(toolchain__author__username__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(toolchain__author__username__iendswith=value)
        else:
            return Q(toolchain__author__username__iexact=value)


    #---------------------------------------------------------------------------
    # Creation callback for 'toolchain__name__*'
    #---------------------------------------------------------------------------
    @staticmethod
    def create_toolchain_name_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(toolchain__name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(toolchain__name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(toolchain__name__iendswith=value)
        else:
            return Q(toolchain__name__iexact=value)


    #---------------------------------------------------------------------------
    # Creation callback for 'protocols__name__*'
    #---------------------------------------------------------------------------
    @staticmethod
    def create_protocol_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(protocols__name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(protocols__name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(protocols__name__iendswith=value)
        else:
            return Q(protocols__name__iexact=value)


    #---------------------------------------------------------------------------
    # Creation callback for 'protocols__sets__name__*'
    #---------------------------------------------------------------------------
    @staticmethod
    def create_dataset_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(protocols__sets__name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(protocols__sets__name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(protocols__sets__name__iendswith=value)
        else:
            return Q(protocols__sets__name__iexact=value)


#------------------------------------------------


def OR(filters):
    return reduce(lambda a, b: a | b, filters)


#------------------------------------------------


def generate_database_filter(database, django_operator):

    def _create_database_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(referenced_datasets__protocol__database__name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(referenced_datasets__protocol__database__name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(referenced_datasets__protocol__database__name__iendswith=value)
        else:
            return Q(referenced_datasets__protocol__database__name__iexact=value)

    def _create_protocol_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(referenced_datasets__protocol__name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(referenced_datasets__protocol__name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(referenced_datasets__protocol__name__iendswith=value)
        else:
            return Q(referenced_datasets__protocol__name__iexact=value)

    def _create_dataset_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(referenced_datasets__name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(referenced_datasets__name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(referenced_datasets__name__iendswith=value)
        else:
            return Q(referenced_datasets__name__iexact=value)


    generator = FilterGenerator()

    (database, version, protocol, dataset) = generator.parse_dataset_name(database, (django_operator == 'exact'))

    generator.add(database, _create_database_filter)

    if version is not None:
        generator.filter = generator.filter & Q(referenced_datasets__protocol__database__version=version)

    generator.add(protocol, _create_protocol_filter)
    generator.add(dataset, _create_dataset_filter)

    return generator.filter


def filter_database_name(queryset, operator, value):
    if operator == 'contains':
        return queryset.filter(generate_database_filter(value, 'contains'))

    elif operator == 'contains-not':
        return queryset.exclude(generate_database_filter(value, 'contains'))

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: generate_database_filter(x, 'contains'), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_database_filter(x, 'contains'), value)))

    elif operator == 'is':
        return queryset.filter(generate_database_filter(value, 'exact'))

    elif operator == 'is-not':
        return queryset.exclude(generate_database_filter(value, 'exact'))

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: generate_database_filter(x, 'exact'), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_database_filter(x, 'exact'), value)))

    return queryset


#------------------------------------------------


def generate_protocol_filter(protocol, django_operator):
    splitted = re.split(r'[\.@]', protocol)
    if len(splitted) <= 2:
        protocol = '*.' + protocol

    return generate_database_filter(protocol, django_operator)


def filter_protocol_name(queryset, operator, value):
    if operator == 'contains':
        return queryset.filter(generate_protocol_filter(value, 'contains'))

    elif operator == 'contains-not':
        return queryset.exclude(generate_protocol_filter(value, 'contains'))

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: generate_protocol_filter(x, 'contains'), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_protocol_filter(x, 'contains'), value)))

    elif operator == 'is':
        return queryset.filter(generate_protocol_filter(value, 'exact'))

    elif operator == 'is-not':
        return queryset.exclude(generate_protocol_filter(value, 'exact'))

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: generate_protocol_filter(x, 'exact'), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_protocol_filter(x, 'exact'), value)))

    return queryset


#------------------------------------------------


def generate_dataset_filter(dataset, django_operator):
    splitted = re.split(r'[\.@]', dataset)
    if len(splitted) == 1:
        dataset = '*.' + dataset

    return generate_protocol_filter(dataset, django_operator)


def filter_dataset_name(queryset, operator, value):
    if operator == 'contains':
        return queryset.filter(generate_dataset_filter(value, 'contains'))

    elif operator == 'contains-not':
        return queryset.exclude(generate_dataset_filter(value, 'contains'))

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: generate_dataset_filter(x, 'contains'), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_dataset_filter(x, 'contains'), value)))

    elif operator == 'is':
        return queryset.filter(generate_dataset_filter(value, 'exact'))

    elif operator == 'is-not':
        return queryset.exclude(generate_dataset_filter(value, 'exact'))

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: generate_dataset_filter(x, 'exact'), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_dataset_filter(x, 'exact'), value)))

    return queryset


#------------------------------------------------


def generate_algorithm_filter(algorithm, analyzer_flag, django_operator):

    def _create_algorithm_author_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(blocks__algorithm__author__username__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(blocks__algorithm__author__username__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(blocks__algorithm__author__username__iendswith=value)
        else:
            return Q(blocks__algorithm__author__username__iexact=value)

    def _create_algorithm_name_filter(filter_type, value):
        if filter_type == FilterGenerator.FILTER_ICONTAINS:
            return Q(blocks__algorithm__name__icontains=value)
        elif filter_type == FilterGenerator.FILTER_ISTARTSWITH:
            return Q(blocks__algorithm__name__istartswith=value)
        elif filter_type == FilterGenerator.FILTER_IENDSWITH:
            return Q(blocks__algorithm__name__iendswith=value)
        else:
            return Q(blocks__algorithm__name__iexact=value)


    generator = FilterGenerator()

    (author, name, version) = generator.parse_contribution_name(algorithm, (django_operator == 'exact'))

    generator.add(author, _create_algorithm_author_filter)
    generator.add(name, _create_algorithm_name_filter)

    if version is not None:
        generator.filter = generator.filter & Q(blocks__algorithm__version=version)

    generator.filter = generator.filter & Q(blocks__analyzer=analyzer_flag)

    return generator.filter


def filter_algorithm_fullname(queryset, operator, value, analyzer):

    if operator == 'contains':
        return queryset.filter(generate_algorithm_filter(value, analyzer, 'contains'))

    elif operator == 'contains-not':
        return queryset.exclude(generate_algorithm_filter(value, analyzer, 'contains'))

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: generate_algorithm_filter(x, analyzer, 'contains'), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_algorithm_filter(x, analyzer, 'contains'), value)))

    elif operator == 'is':
        return queryset.filter(generate_algorithm_filter(value, analyzer, 'exact'))

    elif operator == 'is-not':
        return queryset.exclude(generate_algorithm_filter(value, analyzer, 'exact'))

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: generate_algorithm_filter(x, analyzer, 'exact'), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_algorithm_filter(x, analyzer, 'exact'), value)))

    return queryset


#------------------------------------------------


def filter_toolchain_author(queryset, operator, value):
    if operator == 'contains':
        return queryset.filter(toolchain__author__username__icontains=value)

    elif operator == 'contains-not':
        return queryset.exclude(toolchain__author__username__icontains=value)

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: Q(toolchain__author__username__icontains=x), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: Q(toolchain__author__username__icontains=x), value)))

    elif operator == 'is':
        return queryset.filter(toolchain__author__username__iexact=value)

    elif operator == 'is-not':
        return queryset.exclude(toolchain__author__username__iexact=value)

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: Q(toolchain__author__username__iexact=x), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: Q(toolchain__author__username__iexact=x), value)))

    return queryset


#------------------------------------------------


def generate_toolchain_filter(toolchain, django_operator):
    generator = FilterGenerator()

    (author, name, version) = generator.parse_contribution_name(toolchain, (django_operator == 'exact'))

    generator.add(author, FilterGenerator.create_toolchain_author_filter)
    generator.add(name, FilterGenerator.create_toolchain_name_filter)

    if version is not None:
        generator.filter = generator.filter & Q(toolchain__version=version)

    return generator.filter


def filter_toolchain_fullname(queryset, operator, value):

    if operator == 'contains':
        return queryset.filter(generate_toolchain_filter(value, 'contains'))

    elif operator == 'contains-not':
        return queryset.exclude(generate_toolchain_filter(value, 'contains'))

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: generate_toolchain_filter(x, 'contains'), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_toolchain_filter(x, 'contains'), value)))

    elif operator == 'is':
        return queryset.filter(generate_toolchain_filter(value, 'exact'))

    elif operator == 'is-not':
        return queryset.exclude(generate_toolchain_filter(value, 'exact'))

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: generate_toolchain_filter(x, 'exact'), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_toolchain_filter(x, 'exact'), value)))

    return queryset


#------------------------------------------------


def filter_experiment_author(queryset, operator, value):
    if operator == 'contains':
        return queryset.filter(author__username__icontains=value)

    elif operator == 'contains-not':
        return queryset.exclude(author__username__icontains=value)

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: Q(author__username__icontains=x), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: Q(author__username__icontains=x), value)))

    elif operator == 'is':
        return queryset.filter(author__username__iexact=value)

    elif operator == 'is-not':
        return queryset.exclude(author__username__iexact=value)

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: Q(author__username__iexact=x), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: Q(author__username__iexact=x), value)))

    return queryset


#------------------------------------------------


def generate_experiment_filter(experiment, django_operator):
    experiment_elems = experiment.split('/')
    if len(experiment_elems) == 5:
        experiment_authorname, toolchain_authorname, toolchain_name, toolchain_version, experiment_name = experiment_elems
        try:
            toolchain_version = int(toolchain_version)
        except:
            return Q(name__iexact=UNLIKELY_STRING)
        if django_operator == 'contains':
            return Q(author__username__icontains=experiment_authorname) & Q(toolchain__author__username__icontains=toolchain_authorname) & \
                   Q(toolchain__name__icontains=toolchain_name) & Q(toolchain__version=toolchain_version) & \
                   Q(name__icontains=experiment_name)
        elif django_operator == 'exact':
            return Q(author__username__iexact=experiment_authorname) & Q(toolchain__author__username__iexact=toolchain_authorname) & \
                   Q(toolchain__name__iexact=toolchain_name) & Q(toolchain__version=toolchain_version) & \
                   Q(name__iexact=experiment_name)
    elif len(experiment_elems) == 4:
        toolchain_authorname, toolchain_name, toolchain_version, experiment_name = experiment_elems
        experiment_authorname = toolchain_authorname
        try:
            toolchain_version = int(toolchain_version)
        except:
            return Q(name__iexact=UNLIKELY_STRING)
        if django_operator == 'contains':
            return Q(toolchain__author__username__icontains=toolchain_authorname) & \
                   Q(toolchain__name__icontains=toolchain_name) & Q(toolchain__version=toolchain_version) & \
                   Q(name__icontains=experiment_name)
        elif django_operator == 'exact':
            return Q(toolchain__author__username__iexact=toolchain_authorname) & \
                   Q(toolchain__name__iexact=toolchain_name) & Q(toolchain__version=toolchain_version) & \
                   Q(name__iexact=experiment_name)
    elif len(experiment_elems) == 3:
        toolchain_name, toolchain_version, experiment_name = experiment_elems
        try:
            toolchain_version = int(toolchain_version)
        except:
            return Q(name__iexact=UNLIKELY_STRING)
        if django_operator == 'contains':
            return Q(toolchain__name__icontains=toolchain_name) & Q(toolchain__version=toolchain_version) & \
                   Q(name__icontains=experiment_name)
        elif django_operator == 'exact':
            return Q(toolchain__name__iexact=toolchain_name) & Q(toolchain__version=toolchain_version) & \
                   Q(name__iexact=experiment_name)
    elif len(experiment_elems) == 2:
        toolchain_version, experiment_name = experiment_elems
        try:
            toolchain_version = int(toolchain_version)
        except:
            return Q(name__iexact=UNLIKELY_STRING)
        if django_operator == 'contains':
            return Q(toolchain__version=toolchain_version) & Q(name__icontains=experiment_name)
        elif django_operator == 'exact':
            return Q(toolchain__version=toolchain_version) & Q(name__iexact=experiment_name)
    elif len(experiment_elems) == 1:
        experiment_name = experiment_elems[0]
        if django_operator == 'contains':
            return Q(name__icontains=experiment_name)
        elif django_operator == 'exact':
            return Q(name__iexact=experiment_name)
    return Q(name__iexact=UNLIKELY_STRING)


def filter_experiment(queryset, operator, value):
    if operator == 'contains':
        return queryset.filter(generate_experiment_filter(value, 'contains'))

    elif operator == 'contains-not':
        return queryset.exclude(generate_experiment_filter(value, 'contains'))

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: generate_experiment_filter(x, 'contains'), value)))

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_experiment_filter(x, 'contains'), value)))

    elif operator == 'is':
        return queryset.filter(generate_experiment_filter(value, 'exact'))

    elif operator == 'is-not':
        return queryset.exclude(generate_experiment_filter(value, 'exact'))

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: generate_experiment_filter(x, 'exact'), value)))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_experiment_filter(x, 'exact'), value)))

    elif operator == 'has-attestation':
        return queryset.filter(attestation__isnull=False)

    elif operator == 'has-no-attestation':
        return queryset.filter(attestation__isnull=True)

    return queryset


#------------------------------------------------


def _make_date(value):
    """Creates a datetime.date representation from user input value


    Parameters:

      value (str): The value to convert into datetime representation on the
        form %Y-%m-%d (e.g. 2015-13-25).


    Returns:

      datetime.date: Representing the string value set on the input.

    """

    return datetime.datetime.strptime(value, '%Y-%m-%d').date()


def _make_date_range(value):
  """Creates a datetime.datetime range that spawns the hole day


    Parameters:

      value (str): The value to convert into datetime representation on the
        form %Y-%m-%d (e.g. 2015-13-25).


    Returns:

      datetime.datetime: Representing the string value set on the input, at
        0h0m0s.
      datetime.datetime: Representing the string value set on the input, at
        23h59m59s.

  """

  d = _make_date(value)
  return (datetime.combine(d, datetime.time.min),
          datetime.combine(d, datetime.time.max))


def filter_experiment_date(queryset, operator, value):
    if operator == 'is':
        return queryset.filter(end_date__range=_make_date_range(value))

    elif operator == 'is-not':
        return queryset.exclude(end_date__range=_make_date_range(value))

    elif operator == 'is-any-of':
        return queryset.filter(OR([Q(end_date__range=_make_date_range(k)) for k in value]))

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR([Q(end_date__range=_make_date_range(k)) for k in value]))

    elif operator == '>':
        return queryset.filter(end_date__gt=datetime.datetime.combine(_make_date(value), datetime.time.max))

    elif operator == '>=':
        return queryset.filter(end_date__gte=datetime.datetime.combine(_make_date(value), datetime.time.min))

    elif operator == '<':
        return queryset.filter(end_date__lt=datetime.datetime.combine(_make_date(value), datetime.time.min))

    elif operator == '<=':
        return queryset.filter(end_date__lte=datetime.datetime.combine(_make_date(value), datetime.time.max))

    return queryset


#------------------------------------------------


def filter_experiment_result(queryset, name, operator, value):
    if operator == 'is':
        return queryset.filter(blocks__analyzer=True,
                               blocks__results__name__iexact=name,
                               blocks__results__data_value__iexact=value
                              )

    elif operator == 'is-not':
        return queryset.exclude(blocks__analyzer=True,
                                blocks__results__name__iexact=name,
                                blocks__results__data_value__iexact=value
                               )

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: Q(blocks__results__data_value__iexact=x), value)),
                               blocks__analyzer=True,
                               blocks__results__name__iexact=name
                              )

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: Q(blocks__results__data_value__iexact=x), value)),
                                blocks__analyzer=True,
                                blocks__results__name__iexact=name
                               )

    elif operator == '>':
        return queryset.filter(blocks__analyzer=True,
                               blocks__results__name__iexact=name,
                               blocks__results__data_value__gt=value
                              )

    elif operator == '>=':
        return queryset.filter(blocks__analyzer=True,
                               blocks__results__name__iexact=name,
                               blocks__results__data_value__gte=value
                              )

    elif operator == '<':
        return queryset.filter(blocks__analyzer=True,
                               blocks__results__name__iexact=name,
                               blocks__results__data_value__lt=value
                              )

    elif operator == '<=':
        return queryset.filter(blocks__analyzer=True,
                               blocks__results__name__iexact=name,
                               blocks__results__data_value__lte=value
                              )

    return queryset


#------------------------------------------------


def filter_any_field(queryset, operator, value):
    if operator == 'contains':
        return queryset.filter(generate_database_filter(value, 'contains') |
                               Q(referenced_datasets__protocol__name__icontains=value) |
                               Q(referenced_datasets__name__icontains=value) |
                               generate_algorithm_filter(value, False, 'contains') |
                               generate_algorithm_filter(value, True, 'contains') |
                               generate_toolchain_filter(value, 'contains') |
                               Q(toolchain__author__username__icontains=value) |
                               Q(author__username__icontains=value) |
                               generate_experiment_filter(value, 'contains')
                              )

    elif operator == 'contains-not':
        return queryset.exclude(generate_database_filter(value, 'contains') |
                                Q(referenced_datasets__protocol__name__icontains=value) |
                                Q(referenced_datasets__name__icontains=value) |
                                generate_algorithm_filter(value, False, 'contains') |
                                generate_algorithm_filter(value, True, 'contains') |
                                generate_toolchain_filter(value, 'contains') |
                                Q(toolchain__author__username__icontains=value) |
                                Q(author__username__icontains=value) |
                                generate_experiment_filter(value, 'contains')
                               )

    elif operator == 'contains-any-of':
        return queryset.filter(OR(map(lambda x: generate_database_filter(x, 'contains'), value)) |
                               OR(map(lambda x: Q(referenced_datasets__protocol__name__icontains=x), value)) |
                               OR(map(lambda x: Q(referenced_datasets__name__icontains=x), value)) |
                               OR(map(lambda x: generate_algorithm_filter(x, False, 'contains'), value)) |
                               OR(map(lambda x: generate_algorithm_filter(x, True, 'contains'), value)) |
                               OR(map(lambda x: generate_toolchain_filter(x, 'contains'), value)) |
                               OR(map(lambda x: Q(toolchain__author__username__icontains=x), value)) |
                               OR(map(lambda x: Q(author__username__icontains=x), value)) |
                               OR(map(lambda x: generate_experiment_filter(x, 'contains'), value))
                              )

    elif operator == 'contains-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_database_filter(x, 'contains'), value)) |
                                OR(map(lambda x: Q(referenced_datasets__protocol__name__icontains=x), value)) |
                                OR(map(lambda x: Q(referenced_datasets__name__icontains=x), value)) |
                                OR(map(lambda x: generate_algorithm_filter(x, False, 'contains'), value)) |
                                OR(map(lambda x: generate_algorithm_filter(x, True, 'contains'), value)) |
                                OR(map(lambda x: generate_toolchain_filter(x, 'contains'), value)) |
                                OR(map(lambda x: Q(toolchain__author__username__icontains=x), value)) |
                                OR(map(lambda x: Q(author__username__icontains=x), value)) |
                                OR(map(lambda x: generate_experiment_filter(x, 'contains'), value))
                               )

    elif operator == 'is':
        return queryset.filter(generate_database_filter(value, 'exact') |
                               Q(referenced_datasets__protocol__name__iexact=value) |
                               Q(referenced_datasets__name__iexact=value) |
                               generate_algorithm_filter(value, False, 'exact') |
                               generate_algorithm_filter(value, True, 'exact') |
                               generate_toolchain_filter(value, 'exact') |
                               Q(toolchain__author__username__iexact=value) |
                               Q(author__username__iexact=value) |
                               generate_experiment_filter(value, 'exact')
                              )

    elif operator == 'is-not':
        return queryset.exclude(generate_database_filter(value, 'exact') |
                                Q(referenced_datasets__protocol__name__iexact=value) |
                                Q(referenced_datasets__name__iexact=value) |
                                generate_algorithm_filter(value, False, 'exact') |
                                generate_algorithm_filter(value, True, 'exact') |
                                generate_toolchain_filter(value, 'exact') |
                                Q(toolchain__author__username__iexact=value) |
                                Q(author__username__iexact=value) |
                                generate_experiment_filter(value, 'exact')
                               )

    elif operator == 'is-any-of':
        return queryset.filter(OR(map(lambda x: generate_database_filter(x, 'exact'), value)) |
                               OR(map(lambda x: Q(referenced_datasets__protocol__name__iexact=x), value)) |
                               OR(map(lambda x: Q(referenced_datasets__name__iexact=x), value)) |
                               OR(map(lambda x: generate_algorithm_filter(x, False, 'exact'), value)) |
                               OR(map(lambda x: generate_algorithm_filter(x, True, 'exact'), value)) |
                               OR(map(lambda x: generate_toolchain_filter(x, 'exact'), value)) |
                               OR(map(lambda x: Q(toolchain__author__username__iexact=x), value)) |
                               OR(map(lambda x: Q(author__username__iexact=x), value)) |
                               OR(map(lambda x: generate_experiment_filter(x, 'exact'), value))
                              )

    elif operator == 'is-not-any-of':
        return queryset.exclude(OR(map(lambda x: generate_database_filter(x, 'exact'), value)) |
                                OR(map(lambda x: Q(referenced_datasets__protocol__name__iexact=x), value)) |
                                OR(map(lambda x: Q(referenced_datasets__name__iexact=x), value)) |
                                OR(map(lambda x: generate_algorithm_filter(x, False, 'exact'), value)) |
                                OR(map(lambda x: generate_algorithm_filter(x, True, 'exact'), value)) |
                                OR(map(lambda x: generate_toolchain_filter(x, 'exact'), value)) |
                                OR(map(lambda x: Q(toolchain__author__username__iexact=x), value)) |
                                OR(map(lambda x: Q(author__username__iexact=x), value)) |
                                OR(map(lambda x: generate_experiment_filter(x, 'exact'), value))
                               )

    return queryset


#------------------------------------------------


def apply_filter(queryset, filter_entry):
    try:
        if filter_entry['context'] == 'database-name':
            return filter_database_name(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'protocol-name':
            return filter_protocol_name(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'dataset-name':
            return filter_dataset_name(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'algorithm':
            return filter_algorithm_fullname(queryset, filter_entry['operator'], filter_entry['value'], False)

        elif filter_entry['context'] == 'analyzer':
            return filter_algorithm_fullname(queryset, filter_entry['operator'], filter_entry['value'], True)

        elif filter_entry['context'] == 'toolchain-author':
            return filter_toolchain_author(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'toolchain':
            return filter_toolchain_fullname(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'experiment-author':
            return filter_experiment_author(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'experiment':
            return filter_experiment(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'experiment-date':
            return filter_experiment_date(queryset, filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'experiment-result':
            return filter_experiment_result(queryset, filter_entry['name'],
                                            filter_entry['operator'], filter_entry['value'])

        elif filter_entry['context'] == 'any-field':
            return filter_any_field(queryset, filter_entry['operator'], filter_entry['value'])

    except:
        return QuerySet()

    return queryset
