/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.toolchains.editor.js

   Implementation of a panel that allows to modify a toolchain
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.toolchains === undefined)
    beat.toolchains = {}

if (beat.toolchains.editor === undefined)
    beat.toolchains.editor = {}


/*************************************** CONSTANTS **************************************/

// Mouse actions
beat.toolchains.editor.ACTION_NONE                   = 0;
beat.toolchains.editor.ACTION_TRANSLATE_VIEW         = 1;
beat.toolchains.editor.ACTION_TRANSLATE_SELECTION    = 2;
beat.toolchains.editor.ACTION_SELECTION              = 3;
beat.toolchains.editor.ACTION_BLOCK_CREATION         = 4;
beat.toolchains.editor.ACTION_TEXT_EDITION           = 5;
beat.toolchains.editor.ACTION_CONNECTION_CREATION    = 6;
beat.toolchains.editor.ACTION_COLOR_PICKING          = 7;
beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION = 8;
beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION   = 9;
beat.toolchains.editor.ACTION_SCROLLBAR              = 10;
beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION  = 11;

// Component types
beat.toolchains.editor.TYPE_PROTOCOL   = 'protocol';
beat.toolchains.editor.TYPE_DATASET    = 'dataset';
beat.toolchains.editor.TYPE_BLOCK      = 'block';
beat.toolchains.editor.TYPE_ANALYZER   = 'analyzer';
beat.toolchains.editor.TYPE_TOOLCHAIN  = 'toolchain';
beat.toolchains.editor.TYPE_CONNECTION = 'connection';

// Hovered components
beat.toolchains.editor.HOVERED_NONE                     = 0;
beat.toolchains.editor.HOVERED_SYNCHRONIZATION_CHANNEL  = 1;
beat.toolchains.editor.HOVERED_BLOCK                    = 2;
beat.toolchains.editor.HOVERED_SELECTED_BLOCK           = 3;
beat.toolchains.editor.HOVERED_ADD_INPUT_BUTTON         = 4;
beat.toolchains.editor.HOVERED_ADD_OUTPUT_BUTTON        = 5;
beat.toolchains.editor.HOVERED_DELETE_INPUT_BUTTON      = 6;
beat.toolchains.editor.HOVERED_DELETE_OUTPUT_BUTTON     = 7;
beat.toolchains.editor.HOVERED_INPUT_HANDLE             = 8;
beat.toolchains.editor.HOVERED_OUTPUT_HANDLE            = 9;
beat.toolchains.editor.HOVERED_CONNECTION               = 10;
beat.toolchains.editor.HOVERED_CONNECTION_WITH_CHANNELS = 11;
beat.toolchains.editor.HOVERED_INPUT_HANDLE_CONNECTION  = 12;



/******************************** CLASS: ToolchainEditor ********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// editor_id: ID of the DOM element to use
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainEditor = function(editor_id, images_url, list_selector,
                                                  smart_selector, import_settings_dialog,
                                                  height_offset)
{
    // Initialisations
    this.id                    = editor_id;
    this.element               = $('#' + editor_id)[0];
    this.toolchain             = null;
    this.database_templates    = null;
    this.importable_toolchains = null;
    this.list_selector         = list_selector;
    this.smart_selector        = smart_selector;
    this.height_offset         = (height_offset === undefined ? 0 : height_offset);

    this.view           = new beat.toolchains.editor.ToolchainView($(this.element).find('canvas.grid')[0], images_url);
    this.toolbar        = new beat.toolchains.editor.Toolbar($(this.element).find('div.toolbar')[0]);
    this.components_bar = new beat.toolchains.editor.ComponentsBar($(this.element).find('div.components_bar')[0]);
    this.infobar        = new beat.toolchains.editor.InfoBar($(this.element).find('div.infobar')[0], this.view);

    this.blocks_counter    = 0;
    this.analyzers_counter = 0;

    this.callbacks = {
        onSave:  null,
        onClose: null,
    };


    // Handle size changes
    var editor = this;

    $(window).resize(function() {
        editor._onResized();
    });

    editor._onResized();

    this.infobar.setAction(beat.toolchains.editor.ACTION_NONE);


    // Handle events from our components
    this.view.callbacks.onComponentCreated = function(type, data) {
        editor.components_bar.reset();

        if (type == beat.toolchains.editor.TYPE_DATASET)
            editor.toolchain.addDataset(data.block, data.declaration);
        else if (type == beat.toolchains.editor.TYPE_BLOCK)
            editor.toolchain.addBlock(data.block, data.declaration);
        else if (type == beat.toolchains.editor.TYPE_ANALYZER)
            editor.toolchain.addAnalyzer(data.block, data.declaration);
        else if (type == beat.toolchains.editor.TYPE_CONNECTION)
            editor.toolchain.addConnection(data);

        editor.toolbar.enableSaveButton();
    };

    this.view.callbacks.onComponentCreationCancelled = function() {
        editor.components_bar.reset();
    };

    this.view.callbacks.onComponentRenamed = function(current_name, new_name) {
        editor.toolchain.renameComponent(current_name, new_name);
        editor.toolbar.enableSaveButton();
    };

    this.view.callbacks.onComponentDeleted = function(component_name) {
        editor.toolchain.deleteComponent(component_name);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onConnectionDeleted = function(connection) {
        editor.toolchain.deleteConnection(connection);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onComponentInputCreated = function(component_name, input_name) {
        editor.toolchain.addInputToComponent(component_name, input_name);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onComponentInputRenamed = function(component_name, current_name, new_name) {
        editor.toolchain.renameComponentInput(component_name, current_name, new_name);
        editor.toolbar.enableSaveButton();
    };

    this.view.callbacks.onComponentInputDeleted = function(component_name, input_name) {
        editor.toolchain.deleteInputFromComponent(component_name, input_name);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onComponentOutputCreated = function(component_name, output_name) {
        editor.toolchain.addOutputToComponent(component_name, output_name);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onComponentOutputRenamed = function(component_name, current_name, new_name) {
        editor.toolchain.renameComponentOutput(component_name, current_name, new_name);
        editor.toolbar.enableSaveButton();
    };

    this.view.callbacks.onComponentOutputDeleted = function(component_name, output_name) {
        editor.toolchain.deleteOutputFromComponent(component_name, output_name);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onComponentSynchronizationChannelChanged = function(component_name, channel) {
        editor.toolchain.setSynchronizationChannelOfComponent(component_name, channel);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onConnectionSynchronizationChannelChanged = function(connection, channel) {
        editor.toolchain.setSynchronizationChannelOfConnection(connection, channel);
        editor.toolbar.enableSaveButton();
    }

    this.view.callbacks.onComponentMoved = function() {
        editor.toolbar.enableSaveButton();
    };

    this.view.callbacks.onActionChanged = function(action, hovered) {
        editor.infobar.setAction(action, hovered);
    };

    this.view.callbacks.onToolchainModified = function() {
        var declaration = JSON.parse(JSON.stringify(editor.toolchain.declaration));
        declaration.representation = editor.view.getJSON();

        editor.view.versioning_system.push(declaration);
    };

    this.view.callbacks.onConnectionCreationCheck = function(connection) {
        if (connection.input.connections.length > 1)
            return false;

        var from = connection.output.block.name + '.' + connection.output.name;
        var to = connection.input.block.name + '.' + connection.input.name;

        for (var i = 0; i < editor.toolchain.declaration.connections.length; ++i)
        {
            if ((editor.toolchain.declaration.connections[i].from == from) &&
                (editor.toolchain.declaration.connections[i].to == to))
            {
                return false;
            }
        }

        return true;
    };

    this.view.callbacks.onToolchainChanged = function(toolchain) {
        editor.toolchain = toolchain;
    };


    this.components_bar.callbacks.onCreationStarted = function(type) {
        var name = null;

        $('input').blur();

        if (type == beat.toolchains.editor.TYPE_PROTOCOL)
        {
            editor.list_selector.onEntrySelected = function(name, data) {
                var datasets = data.sets;
                for (var i = 0; i < datasets.length; ++i)
                {
                    var counter = 1;
                    var name = datasets[i].name;

                    while (true)
                    {
                        var dataset = editor.toolchain.dataset(name);
                        if (dataset)
                        {
                            ++counter;
                            name = datasets[i].name + '_' + counter;
                        }
                        else
                        {
                            break;
                        }
                    }

                    editor.view.startCreation(beat.toolchains.editor.TYPE_DATASET,
                                              name, data.templates[datasets[i].template]);
                }
            };

            editor.list_selector.onCancel = function() {
                editor.components_bar.reset();
                Mousetrap.unpause();
            };

            var entries = [];
            var db_template_names = Object.keys(editor.database_templates);
            db_template_names.sort();

            for (var i = 0; i < db_template_names.length; ++i)
            {
                var db_template = db_template_names[i];

                var name = '';
                var parts = db_template.split('_');
                for (var j = 0; j < parts.length; ++j)
                    name += parts[j][0].toUpperCase() + parts[j].substr(1) + ' ';

                name += '<span style="font-style: italic; color: #888888; float: right; margin-right: 10px;">(' + Object.keys(editor.database_templates[db_template].sets).length + ' sets)</span>';

                entries.push({
                    name: name,
                    data: editor.database_templates[db_template],
                });
            }

            Mousetrap.pause();

            editor.list_selector.display(entries, 100);
        }

        else if (type == beat.toolchains.editor.TYPE_DATASET)
        {
            editor.list_selector.onEntrySelected = function(name, data) {
                var counter = 1;
                name = name.split(' - ')[1];
                var name2 = name;

                while (true)
                {
                    var dataset = editor.toolchain.dataset(name2);
                    if (dataset)
                    {
                        ++counter;
                        name2 = name + '_' + counter;
                    }
                    else
                    {
                        break;
                    }
                }

                editor.view.startCreation(beat.toolchains.editor.TYPE_DATASET, name2, data);
            };

            editor.list_selector.onCancel = function() {
                editor.components_bar.reset();
                Mousetrap.unpause();
            };

            var entries = [];
            var db_template_names = Object.keys(editor.database_templates);
            db_template_names.sort();

            for (var i = 0; i < db_template_names.length; ++i)
            {
                var db_template = db_template_names[i];

                var db_template_name = '<span style="font-style: italic; color: #888888;">';
                var parts = db_template.split('_');
                for (var j = 0; j < parts.length; ++j)
                    db_template_name += parts[j][0].toUpperCase() + parts[j].substr(1) + ' ';

                db_template_name += '</span>'

                var datasets = editor.database_templates[db_template].sets;
                for (var j = 0; j < datasets.length; ++j)
                {
                    entries.push({
                        name: db_template_name + ' - ' + datasets[j].name,
                        data: editor.database_templates[db_template].templates[datasets[j].template],
                    });
                }
            }

            Mousetrap.pause();

            editor.list_selector.display(entries, 100);
        }

        else if (type == beat.toolchains.editor.TYPE_BLOCK)
        {
            while (true)
            {
                name = 'block_' + (editor.blocks_counter + 1);

                var block = editor.toolchain.block(name);
                if (block)
                    ++editor.blocks_counter;
                else
                    break;
            }

            ++editor.blocks_counter;

            editor.view.startCreation(type, name);
        }

        else if (type == beat.toolchains.editor.TYPE_ANALYZER)
        {
            while (true)
            {
                name = 'analyzer_' + (editor.analyzers_counter + 1);

                var analyzer = editor.toolchain.analyzer(name);
                if (analyzer)
                    ++editor.analyzers_counter;
                else
                    break;
            }

            ++editor.analyzers_counter;

            editor.view.startCreation(type, name);
        }

        else if (type == beat.toolchains.editor.TYPE_TOOLCHAIN)
        {
            editor.smart_selector.onEntrySelected = function(name, data) {
                import_settings_dialog.display(name);
            };

            editor.smart_selector.onCancel = function() {
                editor.components_bar.reset();
                Mousetrap.unpause();
            };

            var entries = [];

            for (var i = 0; i < editor.importable_toolchains.length; ++i)
            {
                entries.push({
                    name: editor.importable_toolchains[i].name,
                    description: editor.importable_toolchains[i].short_description,
                });
            }

            Mousetrap.pause();

            editor.smart_selector.display(entries, 100);
        }
    };


    import_settings_dialog.onToolchainSelected = function(toolchain)
    {
        var fixed_names = editor.toolchain.components().map(function(x) { return x.name; });

        var new_names = toolchain.components().map(function(x) { return x.name; });
        for (var i = 0; i < new_names.length; ++i)
        {
            if (fixed_names.indexOf(new_names[i]) == -1)
                continue;

            var number = 2;
            var prefix = new_names[i] + '_';

            var result = /(\d)+$/i.exec(new_names[i]);
            if (result !== null)
            {
                number = Number(result[0]) + 1;
                prefix = new_names[i].substr(0, new_names[i].length - result[0].length);
            }

            while ((fixed_names.indexOf(prefix + number) >= 0) || (new_names.indexOf(prefix + number) >= 0))
                number++;

            toolchain.renameComponent(new_names[i], prefix + number);
            new_names[i] = prefix + number;
        }

        editor.view.startCreation(beat.toolchains.editor.TYPE_TOOLCHAIN, null, toolchain);
    };


    import_settings_dialog.onCancel = function() {
        editor.components_bar.reset();
        Mousetrap.unpause();
    };


    this.components_bar.callbacks.onCreationCancelled = function() {
        editor.view.cancelCreation();
    };


    this.toolbar.callbacks.onSave = function() {
        var declaration = JSON.parse(JSON.stringify(editor.toolchain.declaration));
        declaration.representation = editor.view.getJSON();

        if (editor.callbacks.onSave)
            editor.callbacks.onSave(declaration);
    };


    this.toolbar.callbacks.onClose = function() {
        if (editor.callbacks.onClose)
            editor.callbacks.onClose();
    };


    this.toolbar.callbacks.onUndo = function() {
        if (editor.view.undo())
            editor.toolbar.enableSaveButton();
    };


    this.toolbar.callbacks.onRedo = function() {
        if (editor.view.redo())
            editor.toolbar.enableSaveButton();
    };


    this.toolbar.callbacks.onZoomIn = function() {
        editor.view.zoomIn();
    };


    this.toolbar.callbacks.onZoomOut = function() {
        editor.view.zoomOut();
    };


    this.toolbar.callbacks.onCenterView = function() {
        editor.view.centerView();
    };


    this.view.versioning_system.callbacks.onUpdate = function(can_undo, can_redo) {
        editor.toolbar.update(can_undo, can_redo);
    };
}


//---------------------------------------------------------


beat.toolchains.editor.ToolchainEditor.prototype._updated = function()
{
    if ((this.toolchain !== null) && (this.database_templates !== null) &&
        (this.importable_toolchains !== null))
    {
        $(this.element).removeClass('progress');
        this.toolbar.show();
        this.components_bar.show();
        this.view.setToolchain(this.toolchain);
        this.view.refresh();
    }
}


//----------------------------------------------------------------------------------------
// Set the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainEditor.prototype.setToolchain = function(toolchain)
{
    this.toolchain      = toolchain;
    this.blocks_counter = toolchain.blocks.length;

    this._updated();
}


//----------------------------------------------------------------------------------------
// Set the database templates
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainEditor.prototype.setDatabaseTemplates = function(templates)
{
    this.database_templates = templates;

    this._updated();
}


//----------------------------------------------------------------------------------------
// Set the list of importable toolchains
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainEditor.prototype.setImportableToolchains = function(toolchains)
{
    this.importable_toolchains = toolchains;

    this._updated();
}


//----------------------------------------------------------------------------------------
// Called when the editor is resized
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainEditor.prototype._onResized = function()
{
    var height = $(window).height() - $(this.element).position().top - 20;

    this.element.style.height = '' + (height + this.height_offset) + 'px';

    var toolbar_height = $(this.element).find(' .toolbar').height();
    var infobar_height = $(this.element).find(' .infobar').height();

    var new_height = this.height_offset + height - infobar_height - toolbar_height;

    this.view.setSize($(this.element).width() - 104, new_height);
    this.components_bar.setSize(100, new_height);
}



/************************************ CLASS: Toolbar ************************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// element: The DOM element to use
//----------------------------------------------------------------------------------------
beat.toolchains.editor.Toolbar = function(element)
{
    // Initialisations
    this.element            = element;
    this.save_button        = $(this.element).find('#save')[0];
    this.close_button       = $(this.element).find('#close')[0];
    this.undo_button        = $(this.element).find('#undo')[0];
    this.redo_button        = $(this.element).find('#redo')[0];
    this.zoom_in_button     = $(this.element).find('#zoom-in')[0];
    this.zoom_out_button    = $(this.element).find('#zoom-out')[0];
    this.center_view_button = $(this.element).find('#center-view')[0];

    this.callbacks = {
        onSave:       null,
        onClose:      null,
        onUndo:       null,
        onRedo:       null,
        onZoomIn:     null,
        onZoomOut:    null,
        onCenterView: null,
    };


    var toolbar = this;

    $(this.close_button).on('click', function(event) {
        if (toolbar.callbacks.onClose)
        {
            var can_close = true;

            if (!$(toolbar.save_button).hasClass('disabled'))
                can_close = confirm("You didn't saved your last modifications to the toolchain. If you close the editor, those changes will be lost.\n\nAre you sure you want to close the editor?");

            if (can_close)
                toolbar.callbacks.onClose();
        }

        return false;
    });

    $(this.undo_button).on('click', function(event) {
        if (toolbar.callbacks.onUndo)
            toolbar.callbacks.onUndo();

        return false;
    });

    $(this.redo_button).on('click', function(event) {
        if (toolbar.callbacks.onRedo)
            toolbar.callbacks.onRedo();

        return false;
    });

    $(this.zoom_in_button).on('click', function(event) {
        if (toolbar.callbacks.onZoomIn)
            toolbar.callbacks.onZoomIn();

        return false;
    });

    $(this.zoom_out_button).on('click', function(event) {
        if (toolbar.callbacks.onZoomOut)
            toolbar.callbacks.onZoomOut();

        return false;
    });

    $(this.center_view_button).on('click', function(event) {
        if (toolbar.callbacks.onCenterView)
            toolbar.callbacks.onCenterView();

        return false;
    });
}


//----------------------------------------------------------------------------------------
// Show the toolbar
//----------------------------------------------------------------------------------------
beat.toolchains.editor.Toolbar.prototype.show = function()
{
    $(this.element).show();
}


//----------------------------------------------------------------------------------------
// Enable the "Save" button
//----------------------------------------------------------------------------------------
beat.toolchains.editor.Toolbar.prototype.enableSaveButton = function()
{
    var toolbar = this;

    $(this.save_button).removeClass('disabled');

    $(this.save_button).off('click');

    $(this.save_button).on('click', function(event) {
        if (toolbar.callbacks.onSave)
        {
            toolbar.callbacks.onSave();
            $(toolbar.save_button).off('click');
            $(toolbar.save_button).addClass('disabled');
        }
        return false;
    });
}


//----------------------------------------------------------------------------------------
// Show the toolbar
//----------------------------------------------------------------------------------------
beat.toolchains.editor.Toolbar.prototype.update = function(can_undo, can_redo)
{
    if (can_undo)
        $(this.undo_button).removeClass('disabled');
    else
        $(this.undo_button).addClass('disabled');

    if (can_redo)
        $(this.redo_button).removeClass('disabled');
    else
        $(this.redo_button).addClass('disabled');
}



/********************************* CLASS: ComponentsBar *********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// element: The DOM element to use
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ComponentsBar = function(element)
{
    // Initialisations
    this.element      = element;
    this.current_type = null;

    this.callbacks = {
        onCreationStarted:   null,
        onCreationCancelled: null,
    };


    var components_bar = this;

    $(this.element).find('.creation').mousedown(function(event) {
        if (event.which != 1)
            return false;

        var previous = components_bar.current_type;

        if (components_bar.current_type !== null)
        {
            if (components_bar.callbacks.onCreationCancelled)
                components_bar.callbacks.onCreationCancelled();

            $(components_bar.element).find('.button.selected').removeClass('selected');

            components_bar.current_type = null;
        }

        if (!components_bar.current_type && (previous != this.id))
        {
            if (components_bar.callbacks.onCreationStarted)
                components_bar.callbacks.onCreationStarted(this.id);

            $(this).addClass('selected');

            components_bar.current_type = this.id;
        }

        return false;
    });
}


//----------------------------------------------------------------------------------------
// Show the components bar
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ComponentsBar.prototype.show = function()
{
    $(this.element).show();
}


//----------------------------------------------------------------------------------------
// Reset the components bar
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ComponentsBar.prototype.reset = function()
{
    $(this.element).find('.button.selected').removeClass('selected');
    this.current_type = null;
}


//----------------------------------------------------------------------------------------
// Resize the components bar
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ComponentsBar.prototype.setSize = function(width, height)
{
    if (height < 5 * 70)
    {
        $(this.element).find('.creation').removeClass('large');
        $(this.element).find('.creation .image').hide();
        $(this.element).find('.creation .text').show();
    }
    else
    {
        $(this.element).find('.creation').addClass('large');
        $(this.element).find('.creation .image').show();
        $(this.element).find('.creation .text').hide();
    }

    this.element.style.width  = '' + width + 'px';
    this.element.style.height = '' + (height - 3) + 'px';
}



/************************************ CLASS: InfoBar ************************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// element: The DOM element to use
//----------------------------------------------------------------------------------------
beat.toolchains.editor.InfoBar = function(element, view)
{
    // Initialisations
    this.element = element;
    this.view    = view;
}


//----------------------------------------------------------------------------------------
// Enable the "Save" button
//----------------------------------------------------------------------------------------
beat.toolchains.editor.InfoBar.prototype.setAction = function(action, hovered)
{
    if (hovered === undefined)
        hovered = beat.toolchains.editor.HOVERED_NONE;

    switch (action)
    {
    case beat.toolchains.editor.ACTION_NONE:
        var text = '';

        switch (hovered)
        {
        case beat.toolchains.editor.HOVERED_NONE:
            text += 'Select: <span class="key spaced">Left click</span>';
            break;

        case beat.toolchains.editor.HOVERED_SYNCHRONIZATION_CHANNEL:
            text += 'Change the color: <span class="key spaced">Left click</span>';
            break;

        case beat.toolchains.editor.HOVERED_BLOCK:
            text += 'Select: <span class="key spaced">Left click</span>' +
                    'Add to/Remove from selection: <span class="key">Shift</span> + <span class="key spaced">Left click</span>' +
                    'Move the block: <span class="key spaced">Left mouse button</span>' +
                    'Rename: <span class="key spaced">Double click</span>';
            break;

        case beat.toolchains.editor.HOVERED_SELECTED_BLOCK:
            text += 'Add to/Remove from selection: <span class="key">Shift</span> + <span class="key spaced">Left click</span>' +
                    'Move the block: <span class="key spaced">Left mouse button</span>' +
                    'Rename: <span class="key spaced">Double click</span>';
            break;

        case beat.toolchains.editor.HOVERED_ADD_INPUT_BUTTON:
            text += 'Add an input: <span class="key spaced">Left click</span>';
            break;

        case beat.toolchains.editor.HOVERED_ADD_OUTPUT_BUTTON:
            text += 'Add an output: <span class="key spaced">Left click</span>';
            break;

        case beat.toolchains.editor.HOVERED_DELETE_INPUT_BUTTON:
            text += 'Delete the input: <span class="key spaced">Left click</span>';
            break;

        case beat.toolchains.editor.HOVERED_DELETE_OUTPUT_BUTTON:
            text += 'Delete the output: <span class="key spaced">Left click</span>';
            break;

        case beat.toolchains.editor.HOVERED_INPUT_HANDLE:
            text += 'Create a new connection: <span class="key spaced">Left mouse button</span>' +
                    'Rename: <span class="key spaced">Double click</span>';
            break;

        case beat.toolchains.editor.HOVERED_OUTPUT_HANDLE:
            text += 'Create a new connection: <span class="key spaced">Left mouse button</span>' +
                    'Rename: <span class="key spaced">Double click</span>';
            break;

        case beat.toolchains.editor.HOVERED_CONNECTION:
            text += 'Translate the connection segment: <span class="key spaced">Left mouse button</span>' +
                    'Delete: <span class="key spaced">Delete</span>';
            break;

        case beat.toolchains.editor.HOVERED_CONNECTION_WITH_CHANNELS:
            text += 'Translate the connection segment: <span class="key spaced">Left mouse button</span>' +
                    'Delete: <span class="key spaced">Delete</span>' +
                    'Change the synchronization channel: <span class="key spaced">Right click</span>';
            break;

        case beat.toolchains.editor.HOVERED_INPUT_HANDLE_CONNECTION:
            text += 'Disconnect the connection: <span class="key spaced">Left mouse button</span>';
            break;
        }

        if ((this.view.selection.length > 0) && (hovered != beat.toolchains.editor.HOVERED_CONNECTION))
            text += 'Delete the selection: <span class="key spaced">Delete</span>';

        var middle_button_modifier = (navigator.platform.toLowerCase().indexOf('linux') >= 0 ? 'Ctrl' : 'Alt');

        text += 'Move the canvas: <span class="key">Middle mouse button</span> or <span class="key">' +
                middle_button_modifier + '</span> + <span class="key">Left mouse button</span> or ' +
                '<span class="key spaced">Mouse wheel</span>' +
                'Zoom: <span class="key">Ctrl</span> + <span class="key spaced">Mouse wheel</span>';

        if (navigator.platform.toLowerCase().indexOf('mac') >= 0)
        {
            text += ' Undo: <span class="key">Command</span> + <span class="key spaced">Z</span> ' +
                    'Redo: <span class="key">Shift</span> + <span class="key">Command</span> + <span class="key spaced">Z</span>';
        }
        else
        {
            text += ' Undo: <span class="key">Ctrl</span> + <span class="key spaced">Z</span> ' +
                    'Redo: <span class="key">Ctrl</span> + <span class="key spaced">Y</span>';
        }

        $(this.element).html(text);

        break;

    case beat.toolchains.editor.ACTION_TRANSLATE_VIEW:
        $(this.element).html('Move the mouse to move the canvas');
        break;

    case beat.toolchains.editor.ACTION_TRANSLATE_SELECTION:
        $(this.element).html('Move the mouse to translate the selected components');
        break;

    case beat.toolchains.editor.ACTION_SELECTION:
        $(this.element).html('Move the mouse to adjust the selection box');
        break;

    case beat.toolchains.editor.ACTION_BLOCK_CREATION:
        $(this.element).html('Place the new component: <span class="key spaced">Left click</span>' +
                             'Cancel: <span class="key">Esc</span>');
        break;

    case beat.toolchains.editor.ACTION_TEXT_EDITION:
        $(this.element).html('Save: <span class="key spaced">Enter</span>' +
                             'Cancel: <span class="key spaced">Esc</span>'
                            );
        break;

    case beat.toolchains.editor.ACTION_CONNECTION_CREATION:
        $(this.element).html('Draw the new connection using the <span class="key spaced">Left mouse button</span>' +
                             'Cancel: <span class="key">Esc</span>');
        break;

    case beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION:
        $(this.element).html('Move the mouse to translate the selected connection segment horizontally' +
                             '<span class="spaced"></span>' +
                             'Cancel: <span class="key spaced">Esc</span>' +
                             'Delete the connection: <span class="key">Delete</span>'
                            );
        break;

    case beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION:
        $(this.element).html('Move the mouse to translate the selected connection segment vertically' +
                             '<span class="spaced"></span>' +
                             'Cancel: <span class="key spaced">Esc</span>' +
                             'Delete the connection: <span class="key">Delete</span>'
                            );
        break;

    case beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION:
        $(this.element).html('Place the new components: <span class="key spaced">Left click</span>' +
                             'Cancel: <span class="key">Esc</span>');
        break;

    default:
        $(this.element).html("");
        break;
    }
}



/********************************* CLASS: ToolchainView *********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// canvas: The canvas element to use
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView = function(canvas, images_url)
{
    // Check that canvas is supported by the browser
    if (!canvas.getContext)
    {
        this.context = null;
        return;
    }


    // Initialisations
    this.canvas               = canvas;
    this.textbox              = $(this.canvas.parentElement).find('.textbox')[0];
    this.zoom                 = 1.0;
    this.context              = this.canvas.getContext("2d");
    this.current_action       = beat.toolchains.editor.ACTION_NONE;
    this.mouse_position       = null;
    this.blocks               = [];
    this.connections          = [];
    this.selection            = [];
    this.clipboard            = [];
    this.highlighted_block    = null;
    this.hovered_block        = null;
    this.hovered_connection   = null;
    this.hovered_segment      = null;
    this.color_picker_anchor  = $(this.canvas.parentElement).find('.color_picker_anchor')[0];;
    this.color_picker_dataset = null;
    this.needsRepaint         = false;
    this.action_data          = null;
    this.channel_colors       = {};
    this.bounding_box         = new beat.toolchains.editor.ToolchainBoundingBox();
    this.versioning_system    = new beat.toolchains.editor.VersioningSystem();

    this.visible_area = {
        top:    0,
        left:   0,
        width:  0,
        height: 0,
    };

    this.selection_box = {
        top:    0,
        left:   0,
        width:  0,
        height: 0,
    };

    this.horizontal_scrollbar = {
        position: 0,
        size:     0,
    };

    this.vertical_scrollbar = {
        position: 0,
        size:     0,
    };

    this.images = {
        add:                null,
        add_highlighted:    null,
        delete:             null,
        delete_highlighted: null,
    };

    this.callbacks = {
        onComponentCreated:                        null,
        onComponentCreationCancelled:              null,
        onComponentRenamed:                        null,
        onComponentDeleted:                        null,
        onConnectionDeleted:                       null,
        onComponentInputCreated:                   null,
        onComponentInputRenamed:                   null,
        onComponentInputDeleted:                   null,
        onComponentOutputCreated:                  null,
        onComponentOutputRenamed:                  null,
        onComponentOutputDeleted:                  null,
        onComponentSynchronizationChannelChanged:  null,
        onConnectionSynchronizationChannelChanged: null,
        onComponentMoved:                          null,
        onActionChanged:                           null,
        onToolchainModified:                       null,
        onToolchainChanged:                        null,
    };


    // Disable the context menu on the left click
    this.canvas.oncontextmenu = function (e) {
        e.preventDefault();
    };


    // Mouse handling
    var view = this;

    $(this.canvas).mousedown(function(event) {
        $('input').blur();
        view._onMouseDown(event);
        return false;
    });

    $(this.canvas).mousemove(function(event) {
        view._onMouseMove(event);
        return false;
    });

    $(this.canvas).mouseup(function(event) {
        view._onMouseUp(event);
        return false;
    });

    $(this.canvas).click(function(event) {
        return false;
    });

    $(this.canvas).dblclick(function(event) {
        view._onDoubleClick(event);
        return false;
    });

    $(this.canvas).mouseenter(function(event) {
        view._onMouseEnter(event);
        return false;
    });

    $(this.canvas).mouseout(function(event) {
        view._onMouseOut(event);
        return false;
    });

    $(this.canvas).bind('mousewheel DOMMouseScroll', function(event) {
        view._onMouseWheel(event);
        return false;
    });


    // Keyboard events
    Mousetrap.bind('esc', function(event) {
        return view._onEscapeKeyUp(event);
    }, 'keyup');

    Mousetrap.bind(['del', 'backspace'], function(event) {
        return view._onDeleteKeyUp(event);
    }, 'keyup');

    if (navigator.platform.toLowerCase().indexOf('mac') != -1)
    {
        Mousetrap.bind(['command+z'], function(event) {
            view.undo();
            return false;
        }, 'keydown');

        Mousetrap.bind(['shift+command+z'], function(event) {
            view.redo();
            return false;
        }, 'keydown');
    }
    else
    {
        Mousetrap.bind(['ctrl+z'], function(event) {
            view.undo();
            return false;
        }, 'keydown');

        Mousetrap.bind(['ctrl+y'], function(event) {
            view.redo();
            return false;
        }, 'keydown');
    }

    // Backspace key: Disable the standard browser behavior (go back at the previous page)
    Mousetrap.bind('backspace', function() {
        if (this.current_action != beat.toolchains.editor.ACTION_TEXT_EDITION)
            return false;
    }, 'keydown');


    // Load the needed images
    function _loadImage(name, filename)
    {
        var img = new Image();

        img.onload = function() {
            view.images[name] = this;
            view.refresh();
        };

        img.src = images_url + '/' + filename;
    }

    _loadImage('add',                'svg/hollow/add.svg');
    _loadImage('add_highlighted',    'svg/filled/add.svg');
    _loadImage('delete',             'svg/hollow/remove.svg');
    _loadImage('delete_highlighted', 'svg/filled/remove.svg');


    // Initialize the color picker
    $(this.color_picker_anchor).spectrum({
        showInput: true,
        className: "full-spectrum",
        showInitial: true,
        showPalette: true,
        showSelectionPalette: true,
        maxPaletteSize: 10,
        preferredFormat: "hex",
        localStorageKey: "spectrum.beat",

        change: function(color) {
            view.channel_colors[view.color_picker_dataset.name] = color.toHexString();
            view.refresh();
            view.callbacks.onToolchainModified();
        },

        show: function() {
            Mousetrap.unpause();
        },

        hide: function() {
            view.color_picker_dataset = null;
            view.current_action = beat.toolchains.editor.ACTION_NONE;
            Mousetrap.pause();

            if (view.callbacks.onActionChanged)
                view.callbacks.onActionChanged(view.current_action);
        },

        palette: [
            ["rgb(0, 0, 0)", "rgb(67, 67, 67)", "rgb(102, 102, 102)",
            "rgb(204, 204, 204)", "rgb(217, 217, 217)","rgb(255, 255, 255)"],
            ["rgb(152, 0, 0)", "rgb(255, 0, 0)", "rgb(255, 153, 0)", "rgb(255, 255, 0)", "rgb(0, 255, 0)",
            "rgb(0, 255, 255)", "rgb(74, 134, 232)", "rgb(0, 0, 255)", "rgb(153, 0, 255)", "rgb(255, 0, 255)"],
            ["rgb(230, 184, 175)", "rgb(244, 204, 204)", "rgb(252, 229, 205)", "rgb(255, 242, 204)", "rgb(217, 234, 211)",
            "rgb(208, 224, 227)", "rgb(201, 218, 248)", "rgb(207, 226, 243)", "rgb(217, 210, 233)", "rgb(234, 209, 220)",
            "rgb(221, 126, 107)", "rgb(234, 153, 153)", "rgb(249, 203, 156)", "rgb(255, 229, 153)", "rgb(182, 215, 168)",
            "rgb(162, 196, 201)", "rgb(164, 194, 244)", "rgb(159, 197, 232)", "rgb(180, 167, 214)", "rgb(213, 166, 189)",
            "rgb(204, 65, 37)", "rgb(224, 102, 102)", "rgb(246, 178, 107)", "rgb(255, 217, 102)", "rgb(147, 196, 125)",
            "rgb(118, 165, 175)", "rgb(109, 158, 235)", "rgb(111, 168, 220)", "rgb(142, 124, 195)", "rgb(194, 123, 160)",
            "rgb(166, 28, 0)", "rgb(204, 0, 0)", "rgb(230, 145, 56)", "rgb(241, 194, 50)", "rgb(106, 168, 79)",
            "rgb(69, 129, 142)", "rgb(60, 120, 216)", "rgb(61, 133, 198)", "rgb(103, 78, 167)", "rgb(166, 77, 121)",
            "rgb(91, 15, 0)", "rgb(102, 0, 0)", "rgb(120, 63, 4)", "rgb(127, 96, 0)", "rgb(39, 78, 19)",
            "rgb(12, 52, 61)", "rgb(28, 69, 135)", "rgb(7, 55, 99)", "rgb(32, 18, 77)", "rgb(76, 17, 48)"]
        ]
    });
}


//----------------------------------------------------------------------------------------
// Indicates if the view is correctly initialized
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.isInitialized = function()
{
    return (this.context != null);
}


//----------------------------------------------------------------------------------------
// Set the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.setToolchain = function(toolchain)
{
    this.versioning_system.reset();

    this.changeToolchain(toolchain);

    // Update the visible area to encompass most of the toolchain
    this.centerView();

    this.callbacks.onToolchainModified();
}


//----------------------------------------------------------------------------------------
// Set the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.changeToolchain = function(toolchain)
{
    // Initialisations
    this.blocks               = [];
    this.connections          = [];
    this.selection            = [];
    this.highlighted_block    = null;
    this.hovered_block        = null;
    this.hovered_connection   = null;
    this.hovered_segment      = null;
    this.action_data          = null;
    this.channel_colors       = {};
    this.bounding_box         = new beat.toolchains.editor.ToolchainBoundingBox();


    var unknown_row = 0;

    if (toolchain.declaration.representation && toolchain.declaration.representation.channel_colors)
        this.channel_colors = toolchain.declaration.representation.channel_colors;


    // Process the datasets, blocks and analyzers
    var components = toolchain.components();
    for (var i = 0; i < components.length; ++i)
    {
        var component = components[i];

        unknown_row = this._createBlockRepresentation(toolchain, component, unknown_row);

        this.blocks.push(component.representation);

        this.bounding_box.update(component.representation, this.zoom);
    }


    // Process the connections
    for (var i = 0; i < toolchain.connections.length; ++i)
    {
        var connection = toolchain.connections[i];

        this._createConnectionRepresentation(toolchain, connection);

        this.connections.push(connection.representation);

        this.bounding_box.update(connection.representation, this.zoom);
    }
}


//---------------------------------------------------------


beat.toolchains.editor.ToolchainView.prototype._createBlockRepresentation = function(toolchain, block, default_row)
{
    var representation = new beat.toolchains.common.BlockRepresentation(block);
    block.representation = representation;

    var infos = (toolchain.declaration.representation ? toolchain.declaration.representation.blocks[block.name]
                                                      : null);
    if (infos)
    {
        representation.left   = infos.col;
        representation.top    = infos.row;
        representation.width  = infos.width;
        representation.height = infos.height;
    }
    else
    {
        representation.left   = 0;
        representation.top    = default_row;
        default_row += beat.toolchains.common.BLOCK_DEFAULT_HEIGHT + 1;
    }

    if ((block.type == beat.toolchains.models.DATASET_BLOCK) && !this.channel_colors[block.name])
        this.channel_colors[block.name] = this._getUnusedChannelColor();

    return default_row;
}


//---------------------------------------------------------


beat.toolchains.editor.ToolchainView.prototype._createConnectionRepresentation = function(toolchain, connection)
{
    var representation = new beat.toolchains.common.ConnectionRepresentation(connection);

    var infos = (toolchain.declaration.representation ? toolchain.declaration.representation.connections[connection.output.block.name + '.' +
                                                                                                         connection.output.name + '/' +
                                                                                                         connection.input.block.name + '.' +
                                                                                                         connection.input.name]
                                                      : null);
    if (infos)
    {
        var last_row = infos[0].row;

        for (var j = 0; j < infos.length; ++j)
        {
            var point_infos = infos[j];

            var fixed = (j <= 1) || (j >= infos.length - 2);
            var vertical = (point_infos.row != last_row);

            var point = new beat.toolchains.common.Point(true, point_infos.col, point_infos.row, vertical, fixed);
            representation.points.push(point);

            last_row = point_infos.row;
        }

        representation._cleanup();
    }
    else
    {
        var output_block_representation = null;
        var input_block_representation = null;

        for (var j = 0; j < this.blocks.length; ++j)
        {
            if (connection.output.block == this.blocks[j].block)
            {
                output_block_representation = this.blocks[j];
                if (input_block_representation)
                    break;
            }

            else if (connection.input.block == this.blocks[j].block)
            {
                input_block_representation = this.blocks[j];
                if (output_block_representation)
                    break;
            }
        }

        representation.points.push(new beat.toolchains.common.Point(
                true, output_block_representation.left + output_block_representation.width,
                output_block_representation.top + 2 + connection.output.index,
                false, true
            ));

        representation.points.push(new beat.toolchains.common.Point(
                true, output_block_representation.left + output_block_representation.width,
                output_block_representation.top + 2 + connection.output.index,
                false, true
            ));

        representation.points.push(new beat.toolchains.common.Point(
                true, output_block_representation.left + output_block_representation.width + 1,
                output_block_representation.top + 2 + connection.output.index,
                false, false
            ));

        var middle = Math.round((output_block_representation.left + output_block_representation.width +
                                input_block_representation.left) / 2);

        representation.points.push(new beat.toolchains.common.Point(
                true, middle,
                output_block_representation.top + 2 + connection.output.index,
                false, false
            ));

        representation.points.push(new beat.toolchains.common.Point(
                true, middle,
                input_block_representation.top + 2 + connection.input.index,
                true, false
            ));

        representation.points.push(new beat.toolchains.common.Point(
                true, input_block_representation.left - 1,
                input_block_representation.top + 2 + connection.input.index,
                false, true
            ));

        representation.points.push(new beat.toolchains.common.Point(
                true, input_block_representation.left,
                input_block_representation.top + 2 + connection.input.index,
                false, true
            ));

        representation._cleanup();
    }

    connection.representation = representation;
}


//----------------------------------------------------------------------------------------
// Schedule a repaint of the content of the canvas
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.refresh = function()
{
    if (this.needsRepaint == false)
    {
        var view = this;
        this.needsRepaint = true;

        requestAnimationFrame(function(timestamp) {
            view._refresh();
        });
    }
}


//----------------------------------------------------------------------------------------
// Update the content of the canvas
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._refresh = function()
{
    this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);

    this._drawGrid();

    this.context.save();

    this.context.translate(Math.floor(-this.visible_area.left * this.zoom),
                           Math.floor(-this.visible_area.top * this.zoom));


    if (this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION)
    {
        for (var i = 0; i < this.action_data.length; ++i)
            this._drawTranslatedBlockTarget(this.action_data[i].representation);
    }
    else if (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION)
    {
        for (var i = 0; i < this.action_data[0].blocks.length; ++i)
            this._drawTranslatedBlockTarget(this.action_data[0].blocks[i]);
    }
    else if (this.current_action == beat.toolchains.editor.ACTION_TRANSLATE_SELECTION)
    {
        for (var i = 0; i < this.selection.length; ++i)
        {
            var representation = this.selection[i];
            if (representation instanceof beat.toolchains.common.BlockRepresentation)
                this._drawTranslatedBlockTarget(representation);
        }
    }


    for (var i = 0; i < this.blocks.length; ++i)
        this._drawBlock(this.blocks[i]);

    for (var i = 0; i < this.connections.length; ++i)
        this._drawConnection(this.connections[i]);


    if (this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION)
    {
        for (var i = 0; i < this.action_data.length; ++i)
            this._drawBlock(this.action_data[i].representation, true);
    }
    else if (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION)
    {
        for (var i = 0; i < this.action_data[0].blocks.length; ++i)
            this._drawBlock(this.action_data[0].blocks[i], true);

        for (var i = 0; i < this.action_data[0].connections.length; ++i)
            this._drawConnection(this.action_data[0].connections[i], true);
    }
    else if (this.current_action == beat.toolchains.editor.ACTION_CONNECTION_CREATION)
    {
        this._drawConnection(this.action_data.representation, true);
    }

    this.context.restore();


    if (this.current_action == beat.toolchains.editor.ACTION_SELECTION)
    {
        this.context.save();

        this.context.strokeStyle = '#0000CC';
        this.context.lineWidth = 2;
        this.context.strokeRect(this.selection_box.left, this.selection_box.top,
                                this.selection_box.width, this.selection_box.height);

        this.context.restore();
    }

    this._drawScrollbars();

    this.needsRepaint = false;
}


//----------------------------------------------------------------------------------------
// Resize the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.setSize = function(width, height)
{
    this.canvas.style.width  = '' + width + 'px';
    this.canvas.style.height = '' + height + 'px';

    this.canvas.width  = width;
    this.canvas.height = height

    this.visible_area.width  = this.canvas.width / this.zoom;
    this.visible_area.height = this.canvas.height / this.zoom;

    this.refresh();
}


//----------------------------------------------------------------------------------------
// Resize the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.startCreation = function(type, name, data)
{
    if (type == beat.toolchains.editor.TYPE_DATASET)
    {
        this.current_action = beat.toolchains.editor.ACTION_BLOCK_CREATION;

        if (this.action_data === null)
            this.action_data = [];

        var entry = {
            declaration: {
                name: name,
                outputs: data,
            },
            block:          null,
            representation: null,
        };

        entry.block = new beat.toolchains.models.Block(entry.declaration,
                                                       beat.toolchains.models.DATASET_BLOCK);

        this.action_data.push(entry);

        Mousetrap.unpause();
    }

    else if (type == beat.toolchains.editor.TYPE_BLOCK)
    {
        this.current_action = beat.toolchains.editor.ACTION_BLOCK_CREATION;

        if (this.action_data === null)
            this.action_data = [];

        var entry = {
            declaration: {
                name: name,
                inputs: [ 'input' ],
                outputs: [ 'output' ],
                synchronized_channel: null,
            },
            block:          null,
            representation: null,
        };

        entry.block = new beat.toolchains.models.Block(entry.declaration,
                                                       beat.toolchains.models.PROCESSING_BLOCK);

        this.action_data.push(entry);
    }

    else if (type == beat.toolchains.editor.TYPE_ANALYZER)
    {
        this.current_action = beat.toolchains.editor.ACTION_BLOCK_CREATION;

        if (this.action_data === null)
            this.action_data = [];

        var entry = {
            declaration: {
                name: name,
                inputs: [ 'input' ],
                synchronized_channel: null,
            },
            block:          null,
            representation: null,
        };

        entry.block = new beat.toolchains.models.Block(entry.declaration,
                                                       beat.toolchains.models.ANALYZER_BLOCK);

        this.action_data.push(entry);
    }
    else if (type == beat.toolchains.editor.TYPE_TOOLCHAIN)
    {
        this.current_action = beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION;

        if (this.action_data === null)
            this.action_data = [];

        var entry = {
            toolchain:      data,
            bounding_box:   new beat.toolchains.editor.ToolchainBoundingBox(),
            blocks:         [],
            connections:    [],
        };

        this.action_data.push(entry);

        // Create the visual representation of the the datasets, blocks and analyzers
        var unknown_row = 0;

        var components = entry.toolchain.components();

        for (var i = 0; i < components.length; ++i)
        {
            var component = components[i];

            unknown_row = this._createBlockRepresentation(entry.toolchain, component, unknown_row);

            this.action_data[0].bounding_box.update(component.representation, this.zoom);

            component.representation.coords_in_pixels = true;

            component.representation.left = beat.toolchains.common.cellsToPixels(component.representation.left, this.zoom);
            component.representation.top = beat.toolchains.common.cellsToPixels(component.representation.top, this.zoom);

            this.action_data[0].blocks.push(component.representation);
        }

        // Create the visual representation of the the connections
        for (var i = 0; i < entry.toolchain.connections.length; ++i)
        {
            var connection = entry.toolchain.connections[i];

            this._createConnectionRepresentation(entry.toolchain, connection);

            this.action_data[0].bounding_box.update(connection.representation, this.zoom);

            connection.representation.selected = beat.toolchains.common.SELECTION_FULL;

            connection.representation.startTranslation(this.zoom);

            this.action_data[0].connections.push(connection.representation);
        }

        $(this.canvas).addClass('action-translation');

        Mousetrap.unpause();
    }

    this._deselectAll();
    this.refresh();

    if (this.callbacks.onActionChanged)
        this.callbacks.onActionChanged(this.current_action);
}


//----------------------------------------------------------------------------------------
// Start the creation of a connection (from either an input or output)
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.startConnectionCreation = function(
                                                                    col, row, input, output)
{
    this.current_action = beat.toolchains.editor.ACTION_CONNECTION_CREATION;

    this.action_data = {
        representation: new beat.toolchains.common.ConnectionRepresentation(null),
        source_input:   input,
        source_output:  output,
        source_block:   (input ? input.block : output.block),
    }

    this.action_data.representation.addPoint(new beat.toolchains.common.Point(true, col, row, false, true), this.zoom);

    if (output)
        this.action_data.representation.addPoint(new beat.toolchains.common.Point(true, col + 1, row, false, true), this.zoom);
    else
        this.action_data.representation.addPoint(new beat.toolchains.common.Point(true, col - 1, row, false, true), this.zoom);

    var x = this.mouse_position.x + Math.floor(this.visible_area.left * this.zoom);
    var y = this.mouse_position.y + Math.floor(this.visible_area.top * this.zoom);

    this.action_data.representation.addPoint(new beat.toolchains.common.Point(false, x, y, false), this.zoom);
    this.action_data.representation.addPoint(new beat.toolchains.common.Point(false, x, y, true), this.zoom);

    this._deselectAll();
    this.refresh();

    if (this.callbacks.onActionChanged)
        this.callbacks.onActionChanged(this.current_action);
}


//----------------------------------------------------------------------------------------
// Rewire an existing connection
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.rewireConnection = function(connection_representation, disconnect_input)
{
    this.current_action = beat.toolchains.editor.ACTION_CONNECTION_CREATION;

    if (disconnect_input)
    {
        this.action_data = {
            representation: connection_representation,
            source_input:   null,
            source_output:  connection_representation.connection.output,
            source_block:   connection_representation.connection.output.block,
        }
    }
    else
    {
        this.action_data = {
            representation: connection_representation,
            source_input:   connection_representation.connection.input,
            source_output:  null,
            source_block:   connection_representation.connection.input.block,
        }

        for (var i = 0; i < connection_representation.points.length - 1; ++i)
            connection_representation.points[i].vertical = connection_representation.points[i+1].vertical;

        connection_representation.points[connection_representation.points.length - 1].vertical = false;

        connection_representation.points.reverse();
    }


    var nb_points = connection_representation.points.length;

    if (connection_representation.getSegment(nb_points - 3).isVertical())
    {
        connection_representation.points[nb_points - 2].toPixels(this.zoom);
        connection_representation.points[nb_points - 1].toPixels(this.zoom);

        connection_representation.points[nb_points - 2].fixed = false;
        connection_representation.points[nb_points - 1].fixed = false;
    }
    else if (nb_points == 4)
    {
        var new_point = new beat.toolchains.common.Point(true, Math.floor((connection_representation.points[2].col + connection_representation.points[1].col) / 2),
                                                         connection_representation.points[2].row, false
                                                        );

        connection_representation.points.splice(2, 0, new_point);

        connection_representation.points[3].col = new_point.col;
        connection_representation.points[3].vertical = true;
        connection_representation.points[3].fixed = false;

        connection_representation.points[3].toPixels(this.zoom);
        connection_representation.points[4].toPixels(this.zoom);
    }
    else
    {
        connection_representation.points[nb_points - 3].toPixels(this.zoom);
        connection_representation.points[nb_points - 1].toPixels(this.zoom);
        connection_representation.points[nb_points - 1].fixed = false;

        connection_representation.points.splice(nb_points - 2, 1);
    }

    connection_representation.active_segment_index = connection_representation.nbSegments() - 1;
    connection_representation.active_segment = connection_representation.getSegment(connection_representation.active_segment_index);
    connection_representation.active_segment.can_modify_length = true;

    this.callbacks.onConnectionDeleted(connection_representation.connection);
    this.connections.splice(this.connections.indexOf(connection_representation), 1);
    connection_representation.connection = null;

    this._deselectAll();
    this.refresh();

    if (this.callbacks.onActionChanged)
        this.callbacks.onActionChanged(this.current_action);
}


//----------------------------------------------------------------------------------------
// Called when the creation of a new component is cancelled
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.cancelCreation = function()
{
    this.current_action = beat.toolchains.editor.ACTION_NONE;
    this.action_data    = null;

    $(this.canvas).removeClass('action-translation');

    this.refresh();

    if (this.callbacks.onActionChanged)
        this.callbacks.onActionChanged(this.current_action);
}


//----------------------------------------------------------------------------------------
// Undo the last changes
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.undo = function()
{
    var declaration = this.versioning_system.previous();
    if (declaration === null)
        return false;

    var toolchain = new beat.toolchains.models.Toolchain(declaration);
    this.changeToolchain(toolchain);

    this.refresh();

    if (this.callbacks.onToolchainChanged !== null)
        this.callbacks.onToolchainChanged(toolchain);

    return true;
}


//----------------------------------------------------------------------------------------
// Redo the previous changes
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.redo = function()
{
    var declaration = this.versioning_system.next();
    if (declaration === null)
        return false;

    var toolchain = new beat.toolchains.models.Toolchain(declaration);
    this.changeToolchain(toolchain);

    this.refresh();

    if (this.callbacks.onToolchainChanged !== null)
        this.callbacks.onToolchainChanged(toolchain);

    return true;
}


//----------------------------------------------------------------------------------------
// Zoom in
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.zoomIn = function(event)
{
    if (this.zoom < 1.5)
    {
        var previous_zoom = this.zoom;
        this.zoom = Math.min(this.zoom + 0.1, 1.5);
        this.visible_area.width  = this.canvas.width / this.zoom;
        this.visible_area.height = this.canvas.height / this.zoom;
        this._updateSelectionOrActionDataPosition(previous_zoom);
    }
}


//----------------------------------------------------------------------------------------
// Zoom out
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.zoomOut = function(event)
{
    if (this.zoom > 0.5)
    {
        var previous_zoom = this.zoom;
        this.zoom = Math.max(this.zoom - 0.1, 0.5);
        this.visible_area.width  = this.canvas.width / this.zoom;
        this.visible_area.height = this.canvas.height / this.zoom;
        this._updateSelectionOrActionDataPosition(previous_zoom);
    }
}


//----------------------------------------------------------------------------------------
// Called when the mouse wheel is scrolled
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.centerView = function(event)
{
    this.zoom = 1.0;

    this._recomputeBoundingBox();

    var bbox_width = this.bounding_box.right - this.bounding_box.left;
    var bbox_height = this.bounding_box.bottom - this.bounding_box.top;

    var zoom_x = 1.0;
    var zoom_y = 1.0;

    if (this.canvas.width < bbox_width)
        zoom_x = Math.floor(10 * this.canvas.width / bbox_width) / 10;

    if (this.canvas.height < bbox_height)
        zoom_y = Math.floor(10 * this.canvas.height / bbox_height) / 10;

    this.zoom = Math.max(Math.min(zoom_x, zoom_y, 1.0), 0.5);

    this._recomputeBoundingBox();

    this.visible_area.width  = this.canvas.width / this.zoom;
    this.visible_area.height = this.canvas.height / this.zoom;

    this.visible_area.top  = this.bounding_box.top / this.zoom - beat.toolchains.common.CELL_SPACING;
    this.visible_area.left = this.bounding_box.left / this.zoom - beat.toolchains.common.CELL_SPACING;

    this.refresh();
}


//----------------------------------------------------------------------------------------
// Returns the JSON representation of the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype.getJSON = function()
{
    var result = {
        blocks: {},
        connections: {},
        channel_colors: JSON.parse(JSON.stringify(this.channel_colors)),
    };

    for (var i = 0; i < this.blocks.length; ++i)
    {
        var representation = this.blocks[i];

        result.blocks[representation.block.name] = {
            col:    representation.left,
            row:    representation.top,
            width:  representation.width,
            height: representation.height,
        };
    }

    for (var i = 0; i < this.connections.length; ++i)
    {
        var representation = this.connections[i];

        var points = [];
        for (var j = 0; j < representation.points.length; ++j)
        {
            var point = representation.points[j];

            points.push({
                col: point.col,
                row: point.row,
            });
        }

        result.connections[representation.connection.output.block.name + '.' + representation.connection.output.name + '/' +
                           representation.connection.input.block.name + '.' + representation.connection.input.name] = points;
    }

    return result;
}


//----------------------------------------------------------------------------------------
// Called when a button of the mouse is pressed in the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onMouseDown = function(event)
{
    // If some text edition is active, cancel it
    if (this.current_action == beat.toolchains.editor.ACTION_TEXT_EDITION)
    {
        if (this.action_data.edited_block)
        {
            var checked = this._checkBlockName(this.textbox.value, this.blocks,
                                               this.action_data.edited_block.name);

            if (checked.valid)
            {
                if (this.callbacks.onComponentRenamed)
                {
                    if (this.action_data.edited_block.isDataset())
                    {
                        this.channel_colors[checked.name] = this.channel_colors[this.action_data.edited_block.name];
                        delete this.channel_colors[this.action_data.edited_block.name];
                    }

                    this.callbacks.onComponentRenamed(this.action_data.edited_block.name,
                                                      checked.name);
                }
            }
        }
        else if (this.action_data.edited_input)
        {
            var block = this.action_data.edited_input.block;
            var input = this.action_data.edited_input;

            var checked = this._checkEndPointName(this.textbox.value, block, input.name);

            if (checked.valid)
            {
                if (this.callbacks.onComponentInputRenamed)
                    this.callbacks.onComponentInputRenamed(block.name, input.name, checked.name);
            }
        }
        else if (this.action_data.edited_output)
        {
            var block = this.action_data.edited_output.block;
            var output = this.action_data.edited_output;

            var checked = this._checkEndPointName(this.textbox.value, block, output.name);

            if (checked.valid)
            {
                if (this.callbacks.onComponentOutputRenamed)
                    this.callbacks.onComponentOutputRenamed(block.name, output.name, checked.name);
            }
        }

        $(this.textbox).off();
        $(this.textbox).hide();

        this.current_action = beat.toolchains.editor.ACTION_NONE;
        this.action_data = null;
    }

    // If some color picking is active, cancel it
    else if (this.current_action == beat.toolchains.editor.ACTION_COLOR_PICKING)
    {
        $(this.color_picker_anchor).spectrum("hide");
    }

    // Nothing to do if we are already doing some other action
    else if (this.current_action != beat.toolchains.editor.ACTION_NONE)
    {
        return;
    }


    var middle_button_modifier = (navigator.platform.toLowerCase().indexOf('linux') >= 0 ? event.ctrlKey : event.altKey);

    // Left mouse button (and NOT middle_button_modifier+click!)
    if ((event.which == 1) && !middle_button_modifier)
    {
        this.mouse_position = this._getMousePosition(event);

        // Determine if there is an component under the mouse
        var infos = this._getComponentAt(this.mouse_position.x, this.mouse_position.y);

        // Horizontal scrollbar
        if (this.mouse_position.y >= this.canvas.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
        {
            var view = this;

            function _incremental_horizontal_scroll()
            {
                if (view.mouse_position.y < view.canvas.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
                    return;

                if (view.mouse_position.x >= view.canvas.width - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
                    return;

                if (view.mouse_position.x < view.horizontal_scrollbar.position)
                    view.visible_area.left -= 2 * beat.toolchains.common.CELL_SPACING / view.zoom;
                else if (view.mouse_position.x > view.horizontal_scrollbar.position + view.horizontal_scrollbar.size)
                    view.visible_area.left += 2 * beat.toolchains.common.CELL_SPACING / view.zoom;

                view.refresh();
            }

            // Corner
            if (this.mouse_position.x >= this.canvas.width - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
            {
                // Nothing to do
            }

            // Left of the bar
            else if (this.mouse_position.x < this.horizontal_scrollbar.position)
            {
                this.visible_area.left -= beat.toolchains.common.CELL_SPACING / this.zoom;

                this.current_action = beat.toolchains.editor.ACTION_SCROLLBAR;
                this.action_data = {
                    horizontal:  true,
                    incremental: true,
                    timeout:     setInterval(_incremental_horizontal_scroll, 200),
                };
            }

            // Right of the bar
            else if (this.mouse_position.x > this.horizontal_scrollbar.position + this.horizontal_scrollbar.size)
            {
                this.visible_area.left += beat.toolchains.common.CELL_SPACING / this.zoom;

                this.current_action = beat.toolchains.editor.ACTION_SCROLLBAR;
                this.action_data = {
                    horizontal:  true,
                    incremental: true,
                    timeout:     setInterval(_incremental_horizontal_scroll, 200),
                };
            }

            // On the bar
            else
            {
                this.current_action = beat.toolchains.editor.ACTION_SCROLLBAR;
                this.action_data = {
                    horizontal:  true,
                    incremental: false,
                };
            }
        }

        // Vertical scrollbar
        else if (this.mouse_position.x >= this.canvas.width - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
        {
            var view = this;

            function _incremental_vertical_scroll()
            {
                if (view.mouse_position.x < view.canvas.width - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
                    return;

                if (view.mouse_position.y >= view.canvas.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
                    return;

                if (view.mouse_position.y < view.vertical_scrollbar.position)
                    view.visible_area.top -= 2 * beat.toolchains.common.CELL_SPACING / view.zoom;
                else if (view.mouse_position.y > view.vertical_scrollbar.position + view.vertical_scrollbar.size)
                    view.visible_area.top += 2 * beat.toolchains.common.CELL_SPACING / view.zoom;

                view.refresh();
            }

            // Left of the bar
            if (this.mouse_position.y < this.vertical_scrollbar.position)
            {
                this.visible_area.top -= beat.toolchains.common.CELL_SPACING / this.zoom;

                this.current_action = beat.toolchains.editor.ACTION_SCROLLBAR;
                this.action_data = {
                    horizontal:  false,
                    incremental: true,
                    timeout:     setInterval(_incremental_vertical_scroll, 200),
                };
            }

            // Right of the bar
            else if (this.mouse_position.y > this.vertical_scrollbar.position + this.vertical_scrollbar.size)
            {
                this.visible_area.top += beat.toolchains.common.CELL_SPACING / this.zoom;

                this.current_action = beat.toolchains.editor.ACTION_SCROLLBAR;
                this.action_data = {
                    horizontal:  false,
                    incremental: true,
                    timeout:     setInterval(_incremental_vertical_scroll, 200),
                };
            }

            // On the bar
            else
            {
                this.current_action = beat.toolchains.editor.ACTION_SCROLLBAR;
                this.action_data = {
                    horizontal:  false,
                    incremental: false,
                };
            }
        }

        // No component: setup a selection box
        else if (infos === null)
        {
            if (!event.shiftKey)
                this._deselectAll();

            this.current_action = beat.toolchains.editor.ACTION_SELECTION;

            this.selection_box.left   = this.mouse_position.x;
            this.selection_box.top    = this.mouse_position.y;
            this.selection_box.width  = 0;
            this.selection_box.height = 0;

            $(this.canvas).addClass('action-selection');
        }

        // Existing connection
        else if (infos.component instanceof beat.toolchains.common.ConnectionRepresentation)
        {
            var segment = infos.component.getSegment(infos.segment);

            if (segment.isVertical())
            {
                this.current_action = beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION;
                $(this.canvas).addClass('action-horizontal-translation');
            }
            else
            {
                this.current_action = beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION;
                $(this.canvas).addClass('action-vertical-translation');
            }

            this.action_data = {
                connection: infos.component,
            };

            this.hovered_segment = infos.component.startSegmentTranslation(infos.segment, this.zoom);

            this._deselectAll();
        }

        // Existing input handle
        else if (infos.input_handle !== null)
        {
            // Rewiring of an existing connection
            if (event.shiftKey && (infos.component.block.inputs[infos.input_handle].connections.length == 1))
            {
                this.rewireConnection(infos.component.block.inputs[infos.input_handle].connections[0].representation, true);
            }

            // No existing connection: start a new one
            else if (infos.component.block.inputs[infos.input_handle].connections.length == 0)
            {
                this.startConnectionCreation(infos.component.left,
                                             infos.component.top + 2 + infos.input_handle,
                                             infos.component.block.inputs[infos.input_handle],
                                             null
                                            );
            }

            $(this.canvas).addClass('action-translation');
        }

        // Existing output handle
        else if (infos.output_handle !== null)
        {
            // Rewiring of an existing connection
            if (event.shiftKey && (infos.component.block.outputs[infos.output_handle].connections.length == 1))
            {
                this.rewireConnection(infos.component.block.outputs[infos.output_handle].connections[0].representation, false);
            }

            // No existing connection: start a new one
            else
            {
                this.startConnectionCreation(infos.component.left + infos.component.width,
                                             infos.component.top + 2 + infos.output_handle,
                                             null,
                                             infos.component.block.outputs[infos.output_handle]
                                            );
            }

            $(this.canvas).addClass('action-translation');
        }

        // Existing component
        else
        {
            // Determine if the component is already selected
            var index = this.selection.indexOf(infos.component);

            // Mouse is over a 'synchronization channel' button: check that it make sense
            // for the block (a dataset, or more than one valid channel)
            if (infos.synchronization_channel && !infos.component.block.isDataset())
            {
                var valid_channels = infos.component.block.inputs.map(function(x) { return x.channel; }).unique()
                                                          .filter(function(x) { return (x !== null); });

                if (valid_channels.length <= 1)
                    infos.synchronization_channel = false;
            }

            // Mouse is still over a 'synchronization channel' button
            if (infos.synchronization_channel)
            {
                // Block not selected: select it
                if (index == -1)
                {
                    // Clean the current selection first
                    this._deselectAll();
                    this.selection = [infos.component];

                    infos.component.selected = true;

                    var selected_connections = this._selectConnectionsOfBlock(infos.component.block);

                    if (selected_connections.length > 0)
                        this.selection = this.selection.concat(selected_connections);

                    // Ensure that the newly selected component is drawn after all the
                    // others
                    var index2 = this.blocks.indexOf(infos.component);
                    this.blocks.splice(index2, 1);
                    this.blocks.push(infos.component);

                    this.hovered_block = null;
                }

                // For a dataset: show the color picker dialog
                if (infos.component.block.isDataset())
                {
                    this.color_picker_anchor.style.left = (event.clientX + 5) + 'px';
                    this.color_picker_anchor.style.top  = event.clientY + 'px';

                    this.current_action = beat.toolchains.editor.ACTION_COLOR_PICKING;

                    this.color_picker_dataset = infos.component.block;
                    $(this.color_picker_anchor).spectrum("set", this.channel_colors[this.color_picker_dataset.name]);
                    $(this.color_picker_anchor).spectrum("toggle");
                    $(this.canvas).removeClass('action-button');
                }

                // For other types of block: show the channel selector
                else
                {
                    var entries = [];
                    var valid_channels = infos.component.block.inputs.map(function(x) { return x.channel; }).unique()
                                                                     .filter(function(x) { return (x !== null); });
                    for (var i = 0; i < valid_channels.length; ++i)
                    {
                        var channel = valid_channels[i];

                        entries.push({
                            name: valid_channels[i],
                            data: infos.component,
                            color: this.channel_colors[valid_channels[i]],
                        });
                    }

                    if (entries.length > 1)
                    {
                        var view = this;
                        var list_selector = new beat.ui.ListSelector('channel_selector', true);

                        list_selector.onEntrySelected = function(name, data) {
                            if (view.callbacks.onComponentSynchronizationChannelChanged)
                                view.callbacks.onComponentSynchronizationChannelChanged(data.block.name, name);

                            view.current_action = beat.toolchains.editor.ACTION_NONE;
                            view.refresh();

                            view.callbacks.onToolchainModified();

                            Mousetrap.unpause();
                        };

                        list_selector.onCancel = function() {
                            Mousetrap.unpause();
                        };

                        Mousetrap.pause();

                        list_selector.display(entries, $(this.canvas).offset().left + this.mouse_position.x,
                                              $(this.canvas).offset().top + this.mouse_position.y, 200);

                        $(this.canvas).removeClass('action-button');
                    }
                }
            }

            // Not selected: select it and prepare the translation of the selection
            else if (index == -1)
            {
                // Select the block
                this._selectBlock(infos.component, !event.shiftKey);

                this.current_action = beat.toolchains.editor.ACTION_TRANSLATE_SELECTION;

                for (var i = 0; i < this.selection.length; ++i)
                {
                    if (this.selection[i].type() == beat.toolchains.editor.TYPE_CONNECTION)
                        this.selection[i].startTranslation(this.zoom);
                }

                this.hovered_block = null;

                $(this.canvas).addClass('action-translation');
            }

            // Selected + SHIFT key: deselect the component
            else if (event.shiftKey)
            {
                if (infos.component.type() == beat.toolchains.editor.TYPE_CONNECTION)
                    infos.component.selected = beat.toolchains.common.SELECTION_NONE;
                else
                    infos.component.selected = false;

                this.selection.splice(index, 1);

                var deselected_connections = this._deselectConnectionsOfBlock(infos.component.block);

                for (var i = 0; i < deselected_connections.length; ++i)
                    this.selection.splice(deselected_connections[i], 1);

                // Ensure that the remaining selected components are drawn after the
                // deselected one
                if (this.selection.length > 0)
                {
                    var index2 = this.blocks.indexOf(infos.component);
                    this.blocks.splice(index2, 1);

                    index2 = this.blocks.indexOf(this.selection[0])
                    this.blocks.splice(index2, 0, infos.component);
                }
            }

            // Selected and mouse over the 'new input' button: Add an input
            else if (infos.add_input_button)
            {
                if (this.callbacks.onComponentInputCreated)
                {
                    $(this.canvas).removeClass('action-button');
                    infos.component.highlighted_button = null;
                    this.highlighted_block = null;

                    var input_names = infos.component.block.inputNames();

                    var index = input_names.length + 1;
                    var input_name = 'input_' + index;

                    while (input_names.indexOf(input_name) >= 0)
                    {
                        ++index;
                        input_name = 'input_' + index;
                    }

                    this.callbacks.onComponentInputCreated(infos.component.block.name, input_name);

                    infos.component.height = Math.max(infos.component.block.nbInputs(),
                                                      infos.component.block.nbOutputs()) + 2;

                    this.refresh();

                    this.callbacks.onToolchainModified();
                }
            }

            // Selected and mouse over the 'new output' button: Add an output
            else if (infos.add_output_button)
            {
                if (this.callbacks.onComponentOutputCreated)
                {
                    $(this.canvas).removeClass('action-button');
                    infos.component.highlighted_button = null;
                    this.highlighted_block = null;

                    var output_names = infos.component.block.outputNames();

                    var index = output_names.length + 1;
                    var output_name = 'output_' + index;

                    while (output_names.indexOf(output_name) >= 0)
                    {
                        ++index;
                        output_name = 'output_' + index;
                    }

                    this.callbacks.onComponentOutputCreated(infos.component.block.name, output_name);

                    infos.component.height = Math.max(infos.component.block.nbInputs(),
                                                      infos.component.block.nbOutputs()) + 2;

                    this.refresh();

                    this.callbacks.onToolchainModified();
                }
            }

            // Selected and mouse over one 'delete input' button: Remove the input
            else if (infos.delete_input_button !== null)
            {
                if (this.callbacks.onComponentInputDeleted)
                {
                    $(this.canvas).removeClass('action-button');
                    infos.component.input_coords.splice(0, 1);
                    infos.component.highlighted_button = null;
                    this.highlighted_block = null;

                    var input = infos.component.block.inputs[infos.delete_input_button];

                    // Delete the connection representations (if applicable)
                    if (input.connections.length == 1)
                    {
                        var connection_representation = input.connections[0].representation;
                        this.connections.splice(this.connections.indexOf(connection_representation), 1);
                        this.selection.splice(this.selection.indexOf(connection_representation), 1);
                        delete input.connections[0].representation;
                    }

                    this.callbacks.onComponentInputDeleted(infos.component.block.name,
                                                           input.name);

                    infos.component.height = Math.max(infos.component.block.nbInputs(),
                                                      infos.component.block.nbOutputs()) + 2;

                    // Update the position of the connection representations of the
                    // inputs that were moved up due to the removal
                    for (var i = infos.delete_input_button; i < infos.component.block.inputs.length; ++i)
                    {
                        var input = infos.component.block.inputs[i];
                        if (input.connections.length == 0)
                            continue;

                        var connection_representation = input.connections[0].representation;

                        var previous_selection = connection_representation.selection;
                        connection_representation.selection = beat.toolchains.common.SELECTION_INPUT;

                        connection_representation.startTranslation(this.zoom);
                        connection_representation.translate(0, -beat.toolchains.common.CELL_SPACING * this.zoom, this.zoom);
                        connection_representation.endTranslation(this.zoom);

                        connection_representation.selection = previous_selection;
                    }

                    this.refresh();

                    this.callbacks.onToolchainModified();
                }
            }

            // Selected and mouse over one 'delete output' button: Remove the output
            else if (infos.delete_output_button !== null)
            {
                if (this.callbacks.onComponentOutputDeleted)
                {
                    $(this.canvas).removeClass('action-button');
                    infos.component.output_coords.splice(0, 1);
                    infos.component.highlighted_button = null;
                    this.highlighted_block = null;

                    var output = infos.component.block.outputs[infos.delete_output_button];

                    // Delete the connection representations (if applicable)
                    for (var i = 0; i < output.connections.length; ++i)
                    {
                        var connection_representation = output.connections[i].representation;
                        this.connections.splice(this.connections.indexOf(connection_representation), 1);
                        this.selection.splice(this.selection.indexOf(connection_representation), 1);
                        delete output.connections[i].representation;
                    }

                    this.callbacks.onComponentOutputDeleted(infos.component.block.name,
                                                            output.name);

                    infos.component.height = Math.max(infos.component.block.nbInputs(),
                                                      infos.component.block.nbOutputs()) + 2;

                    // Update the position of the connection representations of the
                    // outputs that were moved up due to the removal
                    for (var i = infos.delete_output_button; i < infos.component.block.outputs.length; ++i)
                    {
                        var output = infos.component.block.outputs[i];

                        for (var j = 0; j < output.connections.length; ++j)
                        {
                            var connection_representation = output.connections[j].representation;

                            var previous_selection = connection_representation.selection;
                            connection_representation.selection = beat.toolchains.common.SELECTION_OUTPUT;

                            connection_representation.startTranslation(this.zoom);
                            connection_representation.translate(0, -beat.toolchains.common.CELL_SPACING * this.zoom, this.zoom);
                            connection_representation.endTranslation(this.zoom);

                            connection_representation.selection = previous_selection;
                        }
                    }

                    this.refresh();

                    this.callbacks.onToolchainModified();
                }
            }

            // Selected: prepare the translation of the selection
            else
            {
                this.current_action = beat.toolchains.editor.ACTION_TRANSLATE_SELECTION;

                for (var i = 0; i < this.selection.length; ++i)
                {
                    if (this.selection[i].type() == beat.toolchains.editor.TYPE_CONNECTION)
                        this.selection[i].startTranslation(this.zoom);
                }

                $(this.canvas).addClass('action-translation');
            }

            this.hovered_block = null;
        }

        // The content of the view must be repaint
        this.refresh();
    }

    // Middle mouse button: prepare the translation of the whole grid
    else if ((event.which == 2) || (middle_button_modifier && (event.which == 1)))
    {
        this.current_action = beat.toolchains.editor.ACTION_TRANSLATE_VIEW;
        this.mouse_position = this._getMousePosition(event);
        $(this.canvas).addClass('action-translation');
    }

    // Right mouse button
    else if (event.which == 3)
    {
        this.mouse_position = this._getMousePosition(event);
        var infos = this._getComponentAt(this.mouse_position.x, this.mouse_position.y);

        // On a connection: Display the channel selection dialog
        if ((infos != null) && (infos.component instanceof beat.toolchains.common.ConnectionRepresentation))
        {
            var entries = [];
            var valid_channels = infos.component.connection.output.block.inputs.map(function(x) { return x.channel; }).unique()
                                                                               .filter(function(x) { return (x !== null); });
            for (var i = 0; i < valid_channels.length; ++i)
            {
                var channel = valid_channels[i];

                entries.push({
                    name: valid_channels[i],
                    data: infos.component.connection,
                    color: this.channel_colors[valid_channels[i]],
                });
            }

            if (entries.length > 1)
            {
                var view = this;
                var list_selector = new beat.ui.ListSelector('channel_selector', true);

                list_selector.onEntrySelected = function(name, data) {
                    if (view.callbacks.onConnectionSynchronizationChannelChanged)
                        view.callbacks.onConnectionSynchronizationChannelChanged(data, name);

                    view.current_action = beat.toolchains.editor.ACTION_NONE;
                    view.refresh();

                    view.callbacks.onToolchainModified();

                    Mousetrap.unpause();
                };

                list_selector.onCancel = function() {
                    Mousetrap.unpause();
                };

                Mousetrap.pause();

                list_selector.display(entries, $(this.canvas).offset().left + this.mouse_position.x,
                                      $(this.canvas).offset().top + this.mouse_position.y, 200);
            }
        }
    }

    if (this.callbacks.onActionChanged)
        this.callbacks.onActionChanged(this.current_action);
}


//----------------------------------------------------------------------------------------
// Called when the mouse is moved over the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onMouseMove = function(event)
{
    // Retrieve the position of the mouse
    var mouse_position = this._getMousePosition(event);

    // No action: update the cursor according to the component under it
    if (this.current_action == beat.toolchains.editor.ACTION_NONE)
    {
        if (this.highlighted_block !== null)
        {
            $(this.canvas).removeClass('action-button');
            this.highlighted_block.highlighted_button = null
            this.highlighted_block = null;
            this.refresh();
        }

        if (this.hovered_block !== null)
        {
            this.hovered_block = null;
            this.refresh();
        }

        if (this.hovered_connection !== null)
        {
            this.hovered_connection = null;
            this.hovered_segment    = null;
            this.refresh();
        }


        var infos = this._getComponentAt(mouse_position.x, mouse_position.y);

        // Not over any component
        if (infos === null)
        {
            $(this.canvas).removeClass('action-translation');
            $(this.canvas).removeClass('action-horizontal-translation');
            $(this.canvas).removeClass('action-vertical-translation');
            $(this.canvas).removeClass('action-button');
            $(this.canvas).removeClass('action-new-connection');

            if (this.callbacks.onActionChanged)
                this.callbacks.onActionChanged(this.current_action, beat.toolchains.editor.HOVERED_NONE);
        }

        // Over a block
        else if (infos.component instanceof beat.toolchains.common.BlockRepresentation)
        {
            $(this.canvas).removeClass('action-translation');
            $(this.canvas).removeClass('action-horizontal-translation');
            $(this.canvas).removeClass('action-vertical-translation');
            $(this.canvas).removeClass('action-button');
            $(this.canvas).removeClass('action-new-connection');

            if (infos.synchronization_channel)
            {
                var valid_channels = [];

                if (!infos.component.block.isDataset())
                {
                    valid_channels = infos.component.block.inputs.map(function(x) { return x.channel; }).unique()
                                                                 .filter(function(x) { return (x !== null); });
                }

                if ((valid_channels.length > 1) || infos.component.block.isDataset())
                {
                    $(this.canvas).addClass('action-button');
                    this.highlighted_block = infos.component;
                    infos.component.highlighted_button = -1;

                    if (this.callbacks.onActionChanged)
                    {
                        this.callbacks.onActionChanged(this.current_action,
                                                     beat.toolchains.editor.HOVERED_SYNCHRONIZATION_CHANNEL);
                    }
                }
                else
                {
                    $(this.canvas).addClass('action-translation');
                    this.hovered_block = infos.component;

                    if (this.callbacks.onActionChanged)
                    {
                        if (this.selection.length > 0)
                        {
                            this.callbacks.onActionChanged(this.current_action,
                                                         beat.toolchains.editor.HOVERED_SELECTED_BLOCK);
                        }
                        else
                        {
                            this.callbacks.onActionChanged(this.current_action,
                                                         beat.toolchains.editor.HOVERED_BLOCK);
                        }
                    }
                }

                this.refresh();
            }
            else if (infos.add_input_button)
            {
                $(this.canvas).addClass('action-button');
                this.highlighted_block = infos.component;
                infos.component.highlighted_button = -1;
                this.refresh();

                if (this.callbacks.onActionChanged)
                {
                    this.callbacks.onActionChanged(this.current_action,
                                                 beat.toolchains.editor.HOVERED_ADD_INPUT_BUTTON);
                }
            }
            else if (infos.add_output_button)
            {
                $(this.canvas).addClass('action-button');
                this.highlighted_block = infos.component;
                infos.component.highlighted_button = -2;
                this.refresh();

                if (this.callbacks.onActionChanged)
                {
                    this.callbacks.onActionChanged(this.current_action,
                                                 beat.toolchains.editor.HOVERED_ADD_OUTPUT_BUTTON);
                }
            }
            else if (infos.delete_input_button !== null)
            {
                $(this.canvas).addClass('action-button');
                this.highlighted_block = infos.component;
                infos.component.highlighted_button = infos.delete_input_button;
                this.refresh();

                if (this.callbacks.onActionChanged)
                {
                    this.callbacks.onActionChanged(this.current_action,
                                                 beat.toolchains.editor.HOVERED_DELETE_INPUT_BUTTON);
                }
            }
            else if (infos.delete_output_button !== null)
            {
                $(this.canvas).addClass('action-button');
                this.highlighted_block = infos.component;
                infos.component.highlighted_button = infos.component.input_coords.length + infos.delete_output_button;
                this.refresh();

                if (this.callbacks.onActionChanged)
                {
                    this.callbacks.onActionChanged(this.current_action,
                                                 beat.toolchains.editor.HOVERED_DELETE_OUTPUT_BUTTON);
                }
            }
            else if (infos.input_handle !== null)
            {
                if (infos.component.block.inputs[infos.input_handle].connections.length > 0)
                {
                    $(this.canvas).addClass('action-translation');

                    if (this.callbacks.onActionChanged)
                    {
                        this.callbacks.onActionChanged(this.current_action,
                                                     beat.toolchains.editor.HOVERED_INPUT_HANDLE_CONNECTION);
                    }
                }
                else
                {
                    $(this.canvas).addClass('action-new-connection');

                    if (this.callbacks.onActionChanged)
                    {
                        this.callbacks.onActionChanged(this.current_action,
                                                     beat.toolchains.editor.HOVERED_INPUT_HANDLE);
                    }
                }

                this.hovered_block = infos.component;
                this.refresh();
            }
            else if (infos.output_handle !== null)
            {
                $(this.canvas).addClass('action-new-connection');
                this.hovered_block = infos.component;
                this.refresh();

                if (this.callbacks.onActionChanged)
                {
                    this.callbacks.onActionChanged(this.current_action,
                                                 beat.toolchains.editor.HOVERED_OUTPUT_HANDLE);
                }
            }
            else
            {
                $(this.canvas).addClass('action-translation');
                this.hovered_block = infos.component;
                this.refresh();

                if (this.callbacks.onActionChanged)
                {
                    if (this.selection.length > 0)
                    {
                        this.callbacks.onActionChanged(this.current_action,
                                                     beat.toolchains.editor.HOVERED_SELECTED_BLOCK);
                    }
                    else
                    {
                        this.callbacks.onActionChanged(this.current_action,
                                                     beat.toolchains.editor.HOVERED_BLOCK);
                    }
                }
            }
        }

        // Over a connection
        else if (infos.component instanceof beat.toolchains.common.ConnectionRepresentation)
        {
            $(this.canvas).removeClass('action-translation');
            $(this.canvas).removeClass('action-horizontal-translation');
            $(this.canvas).removeClass('action-vertical-translation');
            $(this.canvas).removeClass('action-button');

            var segment = infos.component.getSegment(infos.segment);

            this.hovered_connection = infos.component;
            this.hovered_segment    = infos.segment;

            if (segment.isVertical())
                $(this.canvas).addClass('action-horizontal-translation');
            else
                $(this.canvas).addClass('action-vertical-translation');

            this.refresh();

            if (this.callbacks.onActionChanged)
            {
                var valid_channels = [];

                if (infos.component.connection.output.block.inputs !== null)
                {
                    valid_channels = infos.component.connection.output.block.inputs.map(function(x) { return x.channel; }).unique()
                                                                                   .filter(function(x) { return (x !== null); });
                }

                if (valid_channels.length > 1)
                {
                    this.callbacks.onActionChanged(this.current_action,
                                                 beat.toolchains.editor.HOVERED_CONNECTION_WITH_CHANNELS);
                }
                else
                {
                    this.callbacks.onActionChanged(this.current_action,
                                                 beat.toolchains.editor.HOVERED_CONNECTION);
                }
            }
        }

        return false;
    }

    if (!this.mouse_position)
        return false;

    // Translation of the view
    if (this.current_action == beat.toolchains.editor.ACTION_TRANSLATE_VIEW)
    {
        var diff_x = mouse_position.x - this.mouse_position.x;
        var diff_y = mouse_position.y - this.mouse_position.y;

        if (diff_x != 0)
            this.visible_area.left -= diff_x / this.zoom;

        if (diff_y != 0)
            this.visible_area.top -= diff_y / this.zoom;

        this.refresh();
    }

    // Creation of a new block
    else if (this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION)
    {
        // Just in case the "mouseenter" event didn't fired first
        if (!this.action_data[0].representation)
        {
            for (var i = 0; i < this.action_data.length; ++i)
            {
                var representation = new beat.toolchains.common.BlockRepresentation(this.action_data[i].block);
                this.action_data[i].block.representation = representation;
                representation.coords_in_pixels = true;

                this.action_data[i].representation = representation;
            }
            $(this.canvas).addClass('action-translation');
        }

        // Note: coordinates are in [pixels] at this point
        var left = this.mouse_position.x + this.visible_area.left * this.zoom -
                   0.5 * beat.toolchains.common.cellsToPixels(this.action_data[0].representation.width, this.zoom);
        var top = this.mouse_position.y + this.visible_area.top * this.zoom -
                  0.5 * beat.toolchains.common.cellsToPixels(this.action_data[0].representation.height, this.zoom);

        for (var i = 0; i < this.action_data.length; ++i)
        {
            this.action_data[i].representation.left = left;
            this.action_data[i].representation.top = top;

            top += beat.toolchains.common.cellsToPixels(this.action_data[i].representation.height + 1, this.zoom);
        }

        this.refresh();
    }

    // Creation of a connection
    else if (this.current_action == beat.toolchains.editor.ACTION_CONNECTION_CREATION)
    {
        var x = mouse_position.x + Math.floor(this.visible_area.left * this.zoom);
        var y = mouse_position.y + Math.floor(this.visible_area.top * this.zoom);

        this.action_data.representation.update(x, y, this.zoom);

        this.hovered_block = null;

        var infos = this._getComponentAt(mouse_position.x, mouse_position.y);
        if (infos !== null)
        {
            if (infos.component.block != this.action_data.source_block)
                this.hovered_block = infos.component;
        }

        this.refresh();
    }

    // Importation of a toolchain
    else if (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION)
    {
        // Note: coordinates are in [pixels] at this point
        var diff_x = mouse_position.x - this.mouse_position.x;
        var diff_y = mouse_position.y - this.mouse_position.y;

        for (var i = 0; i < this.action_data[0].blocks.length; ++i)
        {
            this.action_data[0].blocks[i].left += diff_x;
            this.action_data[0].blocks[i].top  += diff_y;
        }

        for (var i = 0; i < this.action_data[0].connections.length; ++i)
            this.action_data[0].connections[i].translate(diff_x, diff_y, this.zoom);

        this.action_data[0].bounding_box.left   += diff_x;
        this.action_data[0].bounding_box.right  += diff_x;
        this.action_data[0].bounding_box.top    += diff_y;
        this.action_data[0].bounding_box.bottom += diff_y;

        this.refresh();
    }

    // Translation of the selected components
    else if (this.current_action == beat.toolchains.editor.ACTION_TRANSLATE_SELECTION)
    {
        // Note: coordinates are in [pixels] at this point
        var diff_x = mouse_position.x - this.mouse_position.x;
        var diff_y = mouse_position.y - this.mouse_position.y;

        for (var i = 0; i < this.selection.length; ++i)
        {
            if (this.selection[i].type() == beat.toolchains.editor.TYPE_CONNECTION)
            {
                this.selection[i].translate(diff_x, diff_y, this.zoom);
            }
            else
            {
                if (!this.selection[i].coords_in_pixels)
                {
                    this.selection[i].coords_in_pixels = true;
                    this.selection[i].left             = beat.toolchains.common.cellsToPixels(this.selection[i].left, this.zoom);
                    this.selection[i].top              = beat.toolchains.common.cellsToPixels(this.selection[i].top, this.zoom);
                }

                this.selection[i].left += diff_x;
                this.selection[i].top  += diff_y;
            }
        }

        this.refresh();
    }

    // Translation of a vertical connection segment
    else if (this.current_action == beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION)
    {
        var diff_x = mouse_position.x - this.mouse_position.x;
        this.action_data.connection.translateSegment(diff_x);

        this.refresh();
    }

    // Translation of a horizontal connection segment
    else if (this.current_action == beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION)
    {
        var diff_y = mouse_position.y - this.mouse_position.y;
        this.action_data.connection.translateSegment(diff_y);

        this.refresh();
    }

    // Selection box
    else if (this.current_action == beat.toolchains.editor.ACTION_SELECTION)
    {
        this.selection_box.width  = mouse_position.x - this.selection_box.left;
        this.selection_box.height = mouse_position.y - this.selection_box.top;

        this.refresh();
    }

    // Horizontal scrollbar
    else if ((this.current_action == beat.toolchains.editor.ACTION_SCROLLBAR) && this.action_data.horizontal)
    {
        if (!this.action_data.incremental)
        {
            var diff_x = mouse_position.x - this.mouse_position.x;
            this.visible_area.left += diff_x / this.zoom;
            this.refresh();
        }
    }

    // Vertical scrollbar
    else if ((this.current_action == beat.toolchains.editor.ACTION_SCROLLBAR) && !this.action_data.horizontal)
    {
        if (!this.action_data.incremental)
        {
            var diff_y = mouse_position.y - this.mouse_position.y;
            this.visible_area.top += diff_y / this.zoom;
            this.refresh();
        }
    }

    this.mouse_position = mouse_position;
}


//----------------------------------------------------------------------------------------
// Called when a button of the mouse is released in the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onMouseUp = function(event)
{
    // Left mouse button
    if ((event.which == 1) && (this.current_action != beat.toolchains.editor.ACTION_TRANSLATE_VIEW) &&
        (this.current_action != beat.toolchains.editor.ACTION_SCROLLBAR))
    {
        // Creation of new blocks
        if (this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION)
        {
            this.current_action = beat.toolchains.editor.ACTION_NONE;
            $(this.canvas).removeClass('action-translation');

            // Compute their final position (in [cells])
            for (var i = 0; i < this.action_data.length; ++i)
            {
                this.action_data[i].representation.left = beat.toolchains.common.pixelsToCells(this.action_data[i].representation.left, this.zoom);
                this.action_data[i].representation.top = beat.toolchains.common.pixelsToCells(this.action_data[i].representation.top, this.zoom);
                this.action_data[i].representation.coords_in_pixels = false;
                this.blocks.push(this.action_data[i].representation);

                this.bounding_box.update(this.action_data[i].representation, this.zoom);

                // Invoke the corresponding callback (if any)
                if (this.callbacks.onComponentCreated)
                {
                    this.callbacks.onComponentCreated(this.action_data[i].block.isDataset()
                                                            ? beat.toolchains.editor.TYPE_DATASET
                                                            : (this.action_data[i].block.isAnalyzer()
                                                                    ? beat.toolchains.editor.TYPE_ANALYZER
                                                                    : beat.toolchains.editor.TYPE_BLOCK),
                                                      this.action_data[i]);
                }

                // New dataset -> new channel color
                if (this.action_data[i].block.isDataset())
                    this.channel_colors[this.action_data[i].block.name] = this._getUnusedChannelColor();
            }

            // Select the new blocks
            this.selection = [];
            for (var i = 0; i < this.action_data.length; ++i)
            {
                this.action_data[i].representation.selected = true;
                this.selection.push(this.action_data[i].representation);
            }

            this.action_data    = null;
            this.mouse_position = null;

            this.refresh();

            if (this.callbacks.onActionChanged)
            {
                this.callbacks.onActionChanged(this.current_action,
                                               beat.toolchains.editor.HOVERED_SELECTED_BLOCK);
            }

            this.callbacks.onToolchainModified();
        }

        // Creation of a new connection
        else if (this.current_action == beat.toolchains.editor.ACTION_CONNECTION_CREATION)
        {
            var infos = this._getComponentAt(this.mouse_position.x, this.mouse_position.y);

            // On an input handle
            if ((infos !== null) && ((infos.input_handle !== null) || (infos.input !== null)))
            {
                // Only if if appropriate: finalize the connection
                if ((this.action_data.source_input == null) &&
                    (this.action_data.source_block != infos.component.block))
                {
                    var input_index = (infos.input_handle !== null ? infos.input_handle : infos.input);

                    this.action_data.representation.finalize(
                            new beat.toolchains.common.Point(true, infos.component.left,
                                                             infos.component.top + 2 + input_index,
                                                             false, true),
                            true, this.zoom
                        );

                    var channel = (this.action_data.source_block.isDataset() ? this.action_data.source_block.name
                                                                             : this.action_data.source_block.channel);

                    var valid = (infos.component.block.inputs[input_index].connections.length == 0);
                    var connection = null;

                    if (valid)
                    {
                        connection = new beat.toolchains.models.Connection(this.action_data.source_output,
                                                                            infos.component.block.inputs[input_index],
                                                                            channel
                                                                           );

                        if (this.callbacks.onConnectionCreationCheck)
                            valid = this.callbacks.onConnectionCreationCheck(connection);
                    }

                    if (valid)
                    {
                        connection.representation = this.action_data.representation;
                        this.action_data.representation.connection = connection;
                        this.action_data.representation.selected = beat.toolchains.common.SELECTION_NONE;

                        this.connections.push(this.action_data.representation);

                        this.bounding_box.update(connection.representation, this.zoom);

                        // Invoke the corresponding callback (if any)
                        if (this.callbacks.onComponentCreated)
                        {
                            this.callbacks.onComponentCreated(beat.toolchains.editor.TYPE_CONNECTION,
                                                              connection);
                        }

                        this.action_data = null;
                        this.mouse_position = null;

                        this.current_action = beat.toolchains.editor.ACTION_NONE;

                        if (this.callbacks.onActionChanged)
                        {
                            this.callbacks.onActionChanged(this.current_action,
                                                           beat.toolchains.editor.HOVERED_INPUT_HANDLE);
                        }

                        this.callbacks.onToolchainModified();
                    }
                    else
                    {
                        if (connection != null)
                            connection._destroy();

                        this.current_action = beat.toolchains.editor.ACTION_NONE;
                        $(this.canvas).removeClass('action-translation');

                        // Invoke the corresponding callback (if any)
                        if (this.callbacks.onComponentCreationCancelled)
                            this.callbacks.onComponentCreationCancelled();

                        this.action_data    = null;
                        this.mouse_position = null;

                        this.refresh();

                        if (this.callbacks.onActionChanged)
                            this.callbacks.onActionChanged(this.current_action);
                    }
                }
            }

            // On an output handle
            else if ((infos !== null) && ((infos.output_handle !== null) || (infos.output !== null)))
            {
                // Only if if appropriate: finalize the connection
                if ((this.action_data.source_input != null) &&
                    (this.action_data.source_block != infos.component.block))
                {
                    var output_index = (infos.output_handle !== null ? infos.output_handle : infos.output);

                    this.action_data.representation.finalize(
                            new beat.toolchains.common.Point(true, infos.component.left + infos.component.width,
                                                             infos.component.top + 2 + output_index,
                                                             false, true),
                            false, this.zoom
                        );

                    var channel = (infos.component.block.isDataset() ? infos.component.block.name
                                                                     : infos.component.block.channel);

                    var connection = new beat.toolchains.models.Connection(infos.component.block.outputs[output_index],
                                                                           this.action_data.source_input,
                                                                           channel
                                                                          );

                    connection.representation = this.action_data.representation;
                    this.action_data.representation.connection = connection;
                    this.action_data.representation.selected = beat.toolchains.common.SELECTION_NONE;

                    this.connections.push(this.action_data.representation);

                    this.bounding_box.update(connection.representation, this.zoom);

                    // Invoke the corresponding callback (if any)
                    if (this.callbacks.onComponentCreated)
                    {
                        this.callbacks.onComponentCreated(beat.toolchains.editor.TYPE_CONNECTION,
                                                          connection);
                    }

                    this.action_data = null;
                    this.mouse_position = null;

                    this.current_action = beat.toolchains.editor.ACTION_NONE;

                    if (this.callbacks.onActionChanged)
                    {
                        this.callbacks.onActionChanged(this.current_action,
                                                       beat.toolchains.editor.HOVERED_OUTPUT_HANDLE);
                    }

                    this.callbacks.onToolchainModified();
                }
            }

            // Anywhere else: continue the creation of the connection
            else
            {
                this.action_data.representation.addPoint(
                        new beat.toolchains.common.Point(
                            false, this.mouse_position.x + this.visible_area.left * this.zoom,
                            this.mouse_position.y + this.visible_area.top * this.zoom
                        ),
                        this.zoom
                    );
            }

            this.refresh();
        }


        // Importation of a toolchain
        else if (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION)
        {
            this.current_action = beat.toolchains.editor.ACTION_NONE;
            $(this.canvas).removeClass('action-translation');

            // Compute the final position of the blocks (in [cells])
            for (var i = 0; i < this.action_data[0].blocks.length; ++i)
            {
                var representation = this.action_data[0].blocks[i];

                representation.left = beat.toolchains.common.pixelsToCells(representation.left, this.zoom);
                representation.top = beat.toolchains.common.pixelsToCells(representation.top, this.zoom);
                representation.coords_in_pixels = false;
                this.blocks.push(representation);

                this.bounding_box.update(representation, this.zoom);

                // Invoke the corresponding callback (if any)
                if (this.callbacks.onComponentCreated)
                {
                    this.callbacks.onComponentCreated(representation.block.isDataset()
                                                            ? beat.toolchains.editor.TYPE_DATASET
                                                            : (representation.block.isAnalyzer()
                                                                    ? beat.toolchains.editor.TYPE_ANALYZER
                                                                    : beat.toolchains.editor.TYPE_BLOCK),
                                                      {
                                                          block: representation.block,
                                                          declaration: this.action_data[0].toolchain.componentDeclaration(representation.block.name)
                                                      }
                    );
                }
            }

            // Compute the final position of the connections (in [cells])
            for (var i = 0; i < this.action_data[0].connections.length; ++i)
            {
                var representation = this.action_data[0].connections[i];

                representation.endTranslation(this.zoom);

                this.connections.push(representation);

                this.bounding_box.update(representation, this.zoom);

                // Invoke the corresponding callback (if any)
                if (this.callbacks.onComponentCreated)
                {
                    this.callbacks.onComponentCreated(beat.toolchains.editor.TYPE_CONNECTION,
                                                      representation.connection
                    );
                }
            }

            // Select the new blocks
            this.selection = [];
            for (var i = 0; i < this.action_data[0].blocks.length; ++i)
            {
                this.action_data[0].blocks[i].selected = true;
                this.selection.push(this.action_data[0].blocks[i]);
            }

            // Select the new connections
            for (var i = 0; i < this.action_data[0].connections.length; ++i)
                this.selection.push(this.action_data[0].connections[i]);

            this.action_data    = null;
            this.mouse_position = null;

            this.refresh();

            if (this.callbacks.onActionChanged)
            {
                this.callbacks.onActionChanged(this.current_action,
                                               beat.toolchains.editor.HOVERED_SELECTED_BLOCK);
            }

            this.callbacks.onToolchainModified();
        }

        // Translation of the selection
        else if (this.current_action == beat.toolchains.editor.ACTION_TRANSLATE_SELECTION)
        {
            // Compute the final position of the selected blocks (in [cells])
            for (var i = 0; i < this.selection.length; ++i)
            {
                if (this.selection[i].type() != beat.toolchains.editor.TYPE_CONNECTION)
                {
                    if (this.selection[i].coords_in_pixels)
                    {
                        this.selection[i].left = beat.toolchains.common.nearestCell(this.selection[i].left, this.zoom);
                        this.selection[i].top = beat.toolchains.common.nearestCell(this.selection[i].top, this.zoom);
                        this.selection[i].coords_in_pixels = false;
                    }
                }
            }

            // Compute the final position of the selected connections (in [cells])
            for (var i = 0; i < this.selection.length; ++i)
            {
                if (this.selection[i].type() == beat.toolchains.editor.TYPE_CONNECTION)
                    this.selection[i].endTranslation(this.zoom);
            }

            this.current_action = beat.toolchains.editor.ACTION_NONE;

            this.mouse_position = null;

            this._recomputeBoundingBox();

            this.refresh();

            if (this.callbacks.onComponentMoved)
                this.callbacks.onComponentMoved();

            if (this.callbacks.onActionChanged)
            {
                this.callbacks.onActionChanged(this.current_action,
                                               beat.toolchains.editor.HOVERED_SELECTED_BLOCK);
            }

            this.callbacks.onToolchainModified();
        }

        // Translation of a vertical connection segment
        else if (this.current_action == beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION)
        {
            this.current_action = beat.toolchains.editor.ACTION_NONE;
            $(this.canvas).removeClass('action-horizontal-translation');

            this.hovered_segment = this.action_data.connection.endSegmentTranslation(this.zoom);
            this.action_data = null;

            this._recomputeBoundingBox();

            this.refresh();

            if (this.callbacks.onComponentMoved)
                this.callbacks.onComponentMoved();

            if (this.callbacks.onActionChanged)
                this.callbacks.onActionChanged(this.current_action);

            this.callbacks.onToolchainModified();
        }

        // Translation of a horizontal connection segment
        else if (this.current_action == beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION)
        {
            this.current_action = beat.toolchains.editor.ACTION_NONE;
            $(this.canvas).removeClass('action-vertical-translation');

            this.hovered_segment = this.action_data.connection.endSegmentTranslation(this.zoom);
            this.action_data = null;

            this._recomputeBoundingBox();

            this.refresh();

            if (this.callbacks.onComponentMoved)
                this.callbacks.onComponentMoved();

            if (this.callbacks.onActionChanged)
                this.callbacks.onActionChanged(this.current_action);

            this.callbacks.onToolchainModified();
        }

        // Selection box
        else if (this.current_action == beat.toolchains.editor.ACTION_SELECTION)
        {
            this.current_action = beat.toolchains.editor.ACTION_NONE;

            this.mouse_position = null;
            $(this.canvas).removeClass('action-selection');

            // Check the position and size of the selection box
            this.selection_box.left += Math.floor(this.visible_area.left * this.zoom);
            this.selection_box.top += Math.floor(this.visible_area.top * this.zoom);

            if (this.selection_box.width < 0)
            {
                this.selection_box.left  += this.selection_box.width;
                this.selection_box.width = -this.selection_box.width;
            }
            else if (this.selection_box.width == 0)
            {
                this.selection_box.width = 1;
            }

            if (this.selection_box.height < 0)
            {
                this.selection_box.top    += this.selection_box.height;
                this.selection_box.height = -this.selection_box.height;
            }
            else if (this.selection_box.height == 0)
            {
                this.selection_box.height = 1;
            }

            // Add all the blocks inside the selection box to the current selection
            // and ensures that they are drawn after all the others
            var selected_blocks = [];
            for (var i = 0; i < this.blocks.length; ++i)
            {
                var representation = this.blocks[i];
                if (representation.selected)
                    continue;

                var block_x      = beat.toolchains.common.cellsToPixels(representation.left, this.zoom);
                var block_y      = beat.toolchains.common.cellsToPixels(representation.top, this.zoom);
                var block_width  = beat.toolchains.common.cellsToPixels(representation.width, this.zoom);
                var block_height = beat.toolchains.common.cellsToPixels(representation.height, this.zoom);

                if ((block_x + block_width >= this.selection_box.left) &&
                    (this.selection_box.left + this.selection_box.width >= block_x) &&
                    (block_y + block_height >= this.selection_box.top) &&
                    (this.selection_box.top + this.selection_box.height >= block_y))
                {
                    representation.selected = true;
                    this.selection.push(representation);

                    this.blocks.splice(i, 1);
                    --i;

                    selected_blocks.push(representation);
                }
            }

            this.blocks = this.blocks.concat(selected_blocks);

            // Add all the connections inside the selection box to the current selection
            // and ensures that they are drawn after all the others
            var selected_connections = [];
            for (var i = 0; i < this.selection.length; ++i)
            {
                var representation = this.selection[i];
                if (representation.type() == beat.toolchains.editor.TYPE_CONNECTION)
                    continue;

                var selected = this._selectConnectionsOfBlock(representation.block);

                if (selected.length > 0)
                    selected_connections = selected_connections.concat(selected);
            }

            this.selection = this.selection.concat(selected_connections);

            this.refresh();

            if (this.callbacks.onActionChanged)
                this.callbacks.onActionChanged(this.current_action);
        }
    }

    // Translation of the view
    else if (this.current_action == beat.toolchains.editor.ACTION_TRANSLATE_VIEW)
    {
        // Middle mouse button or Alt+Click (even if "Alt" isn't pressed anymore)
        if ((event.which == 2) || (event.which == 1))
        {
            this.current_action = beat.toolchains.editor.ACTION_NONE;
            $(this.canvas).removeClass('action-translation');
            this.mouse_position = null;
        }

        if (this.callbacks.onActionChanged)
            this.callbacks.onActionChanged(this.current_action);
    }

    // Scrollbar usage
    else if ((event.which == 1) && (this.current_action == beat.toolchains.editor.ACTION_SCROLLBAR))
    {
        clearInterval(this.action_data.timeout);
        this.current_action = beat.toolchains.editor.ACTION_NONE;
        this.action_data = null;
        this.refresh();
    }
}


//----------------------------------------------------------------------------------------
// Called when a double-click occurs in the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onDoubleClick = function(event)
{
    var view = this;

    function _startConnectorEdition(text, x, y, width, height)
    {
        view.textbox.style.left            = x + 'px';
        view.textbox.style.top             = y + 'px';
        view.textbox.style.width           = width + 'px';
        view.textbox.style.height          = (height + 2) + 'px';
        view.textbox.style.border          = '1px solid #FF5500';
        view.textbox.style.backgroundColor = '#F6F810';
        view.textbox.style.color           = '#000000';
        view.textbox.style.paddingLeft     = Math.floor(4 * view.zoom) + 'px';
        view.textbox.style.paddingRight    = Math.floor(4 * view.zoom) + 'px';
        view.textbox.style.fontSize        = ((beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT - 4) * view.zoom) + 'px';
        view.textbox.style.fontWeight      = 'normal';
        view.textbox.style.textAlign       = 'left';
        view.textbox.value                 = text;

        $(view.textbox).show();
        $(view.textbox).focus();

        if (view.textbox.setSelectionRange)
            view.textbox.setSelectionRange(text.length, text.length);

        view.current_action = beat.toolchains.editor.ACTION_TEXT_EDITION;

        if (view.callbacks.onActionChanged)
            view.callbacks.onActionChanged(view.current_action);
    }


    function _startBlockNameEdition(text, x, y, width, height)
    {
        view.textbox.style.left            = x + 'px';
        view.textbox.style.top             = y + 'px';
        view.textbox.style.width           = width + 'px';
        view.textbox.style.height          = (height + 4) + 'px';
        view.textbox.style.border          = 0;
        view.textbox.style.backgroundColor = '#FFCC00';
        view.textbox.style.color           = '#000000';
        view.textbox.style.paddingLeft     = Math.floor(4 * view.zoom) + 'px';
        view.textbox.style.paddingRight    = Math.floor(4 * view.zoom) + 'px';
        view.textbox.style.fontSize        = height + 'px';
        view.textbox.style.fontWeight      = 'bold';
        view.textbox.style.textAlign       = 'center';
        view.textbox.value                 = text;

        $(view.textbox).show();
        $(view.textbox).focus();

        if (view.textbox.setSelectionRange)
            view.textbox.setSelectionRange(text.length, text.length);

        view.current_action = beat.toolchains.editor.ACTION_TEXT_EDITION;

        if (view.callbacks.onActionChanged)
            view.callbacks.onActionChanged(view.current_action);
    }


    // Determine on which component the double-click occured
    var mouse_position = this._getMousePosition(event);

    var infos = this._getComponentAt(mouse_position.x, mouse_position.y);
    if (infos === null)
        return;

    $(this.canvas).removeClass('action-translation');


    var canvas_position = $(this.canvas).position();
    canvas_position.left = canvas_position.left + this.canvas.clientLeft;
    canvas_position.top = canvas_position.top + this.canvas.clientTop;


    $(this.textbox).on('keydown', function(event) {
        // Escape key
        if (event.keyCode == 27)
        {
            $(this).off();
            $(this).hide();
            view.current_action = beat.toolchains.editor.ACTION_NONE;

            if (view.callbacks.onActionChanged)
                view.callbacks.onActionChanged(view.current_action);

            return false;
        }
    });

    // On an input
    if ((infos.input !== null) || (infos.input_handle !== null))
    {
        if (infos.input === null)
            infos.input = infos.input_handle;

        // Ensure that the block is selected
        if ((this.selection.length != 1) || (this.selection[0] != infos.component.block))
            this._selectBlock(infos.component, true);

        var coords = infos.component.input_coords[infos.input];

        _startConnectorEdition(infos.component.block.inputs[infos.input].name,
                               canvas_position.left - Math.ceil(this.visible_area.left * this.zoom) +
                               beat.toolchains.common.cellsToPixels(infos.component.left, this.zoom) + coords.x,
                               canvas_position.top - Math.ceil(this.visible_area.top * this.zoom) +
                               beat.toolchains.common.cellsToPixels(infos.component.top, this.zoom) + coords.y,
                               beat.toolchains.common.cellsToPixels(infos.component.width, this.zoom) * 0.5 - 10 * this.zoom,
                               coords.height);

        this.action_data = {
            edited_block:  null,
            edited_input:  infos.component.block.inputs[infos.input],
            edited_output: null,
        };

        $(this.textbox).on('keyup', function(event) {
            var checked = view._checkEndPointName(this.value, infos.component.block,
                                                  infos.component.block.inputs[infos.input].name);

            if (checked.valid)
            {
                // Enter key
                if ((event.keyCode == 13) && view.callbacks.onComponentInputRenamed)
                {
                    view.callbacks.onComponentInputRenamed(infos.component.block.name,
                                                           infos.component.block.inputs[infos.input].name,
                                                           checked.name);
                    $(this).off();
                    $(this).hide();
                    view.refresh();
                    view.current_action = beat.toolchains.editor.ACTION_NONE;
                    view.action_data = null;

                    view.callbacks.onToolchainModified();

                    if (view.callbacks.onActionChanged)
                        view.callbacks.onActionChanged(view.current_action);

                    return;
                }

                view.textbox.style.backgroundColor = '#F6F810';
                view.textbox.style.color           = '#000000';
            }
            else
            {
                view.textbox.style.backgroundColor = '#FF0000';
                view.textbox.style.color           = '#FFFFFF';
            }
        });

        $(this.textbox).on('focusout', function() {
            var checked = view._checkEndPointName(this.value, infos.component.block,
                                                 infos.component.block.inputs[infos.input].name);

            if (checked.valid)
            {
                if (view.callbacks.onComponentInputRenamed)
                {
                    view.callbacks.onComponentInputRenamed(infos.component.block.name,
                                                           infos.component.block.inputs[infos.input].name,
                                                           checked.name);
                }
            }

            $(this).off();
            $(this).hide();
            view.refresh();
            view.current_action = beat.toolchains.editor.ACTION_NONE;
            view.action_data = null;

            view.callbacks.onToolchainModified();

            if (view.callbacks.onActionChanged)
                view.callbacks.onActionChanged(view.current_action);
        });
    }

    // On an output
    else if ((infos.output !== null) || (infos.output_handle !== null))
    {
        if (infos.component.block.isDataset())
            return;

        if (infos.output === null)
            infos.output = infos.output_handle;

        // Ensure that the block is selected
        if ((this.selection.length != 1) || (this.selection[0] != infos.component.block))
            this._selectBlock(infos.component, true);

        var coords = infos.component.output_coords[infos.output];

        _startConnectorEdition(infos.component.block.outputs[infos.output].name,
                               canvas_position.left - Math.ceil(this.visible_area.left * this.zoom) +
                               beat.toolchains.common.cellsToPixels(infos.component.left, this.zoom) +
                               beat.toolchains.common.cellsToPixels(infos.component.width, this.zoom) * 0.5 + 10 * this.zoom,
                               canvas_position.top - Math.ceil(this.visible_area.top * this.zoom) +
                               beat.toolchains.common.cellsToPixels(infos.component.top, this.zoom) + coords.y,
                               beat.toolchains.common.cellsToPixels(infos.component.width, this.zoom) * 0.5 - 10 * this.zoom,
                               coords.height);

        this.action_data = {
            edited_block:  null,
            edited_input:  null,
            edited_output: infos.component.block.outputs[infos.output],
        };

        $(this.textbox).on('keyup', function(event) {
            var checked = view._checkEndPointName(this.value, infos.component.block,
                                                  infos.component.block.outputs[infos.output].name);

            if (checked.valid)
            {
                // Enter key
                if ((event.keyCode == 13) && view.callbacks.onComponentOutputRenamed)
                {
                    view.callbacks.onComponentOutputRenamed(infos.component.block.name,
                                                            infos.component.block.outputs[infos.output].name,
                                                            checked.name);
                    $(this).off();
                    $(this).hide();
                    view.refresh();
                    view.current_action = beat.toolchains.editor.ACTION_NONE;
                    view.action_data = null;

                    view.callbacks.onToolchainModified();

                    if (view.callbacks.onActionChanged)
                        view.callbacks.onActionChanged(view.current_action);

                    return;
                }

                view.textbox.style.backgroundColor = '#F6F810';
                view.textbox.style.color           = '#000000';
            }
            else
            {
                view.textbox.style.backgroundColor = '#FF0000';
                view.textbox.style.color           = '#FFFFFF';
            }
        });

        $(this.textbox).on('focusout', function() {
            var checked = view._checkEndPointName(this.value, infos.component.block,
                                                  infos.component.block.outputs[infos.output].name);

            if (checked.valid)
            {
                if (view.callbacks.onComponentOutputRenamed)
                {
                    view.callbacks.onComponentOutputRenamed(infos.component.block.name,
                                                            infos.component.block.outputs[infos.output].name,
                                                            checked.name);
                }
            }

            $(this).off();
            $(this).hide();
            view.refresh();
            view.current_action = beat.toolchains.editor.ACTION_NONE;
            view.action_data = null;

            view.callbacks.onToolchainModified();

            if (view.callbacks.onActionChanged)
                view.callbacks.onActionChanged(view.current_action);
        });
    }

    // On a block
    else
    {
        var radius = (infos.component.block.isDataset() ? 18 : 14) * this.zoom;

        _startBlockNameEdition(infos.component.block.name,
                               canvas_position.left - Math.ceil(this.visible_area.left * this.zoom) +
                               beat.toolchains.common.cellsToPixels(infos.component.left, this.zoom) + 1 + radius + 4 * this.zoom,
                               canvas_position.top - Math.ceil(this.visible_area.top * this.zoom) +
                               beat.toolchains.common.cellsToPixels(infos.component.top, this.zoom) + Math.floor(4 * this.zoom),
                               beat.toolchains.common.cellsToPixels(infos.component.width, this.zoom) - 2 - 2 * Math.floor(4 * this.zoom) - 2 * radius - 8 * this.zoom,
                               beat.toolchains.common.CELL_SPACING / 2 * this.zoom);

        this.action_data = {
            edited_block:  infos.component.block,
            edited_input:  null,
            edited_output: null,
        };

        $(this.textbox).on('keyup', function(event) {
            var checked = view._checkBlockName(this.value, view.blocks,
                                               infos.component.block.name);

            if (checked.valid)
            {
                // Enter key
                if ((event.keyCode == 13) && view.callbacks.onComponentRenamed)
                {
                    if (infos.component.block.isDataset())
                    {
                        view.channel_colors[checked.name] = view.channel_colors[infos.component.block.name];
                        delete view.channel_colors[infos.component.block.name];
                    }

                    view.callbacks.onComponentRenamed(infos.component.block.name,
                                                      checked.name);

                    $(this).off();
                    $(this).hide();
                    view.refresh();
                    view.current_action = beat.toolchains.editor.ACTION_NONE;
                    view.action_data = null;

                    view.callbacks.onToolchainModified();

                    if (view.callbacks.onActionChanged)
                        view.callbacks.onActionChanged(view.current_action);

                    return;
                }

                view.textbox.style.backgroundColor = '#FFCC00';
                view.textbox.style.color           = '#000000';
            }
            else
            {
                view.textbox.style.backgroundColor = '#FF0000';
                view.textbox.style.color           = '#FFFFFF';
            }
        });

        $(this.textbox).on('focusout', function() {
            var checked = view._checkBlockName(this.value, view.blocks,
                                               infos.component.block.name);

            if (checked.valid)
            {
                if (view.callbacks.onComponentRenamed)
                {
                    if (infos.component.block.isDataset())
                    {
                        view.channel_colors[checked.name] = view.channel_colors[infos.component.block.name];
                        delete view.channel_colors[infos.component.block.name];
                    }

                    view.callbacks.onComponentRenamed(infos.component.block.name,
                                                      checked.name);
                }
            }

            $(this).off();
            $(this).hide();
            view.refresh();
            view.current_action = beat.toolchains.editor.ACTION_NONE;
            view.action_data = null;

            if (view.callbacks.onActionChanged)
                view.callbacks.onActionChanged(view.current_action);

            view.callbacks.onToolchainModified();
        });
    }
}


//----------------------------------------------------------------------------------------
// Called when the 'Escape' key is released on the keyboard
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onEscapeKeyUp = function(event)
{
    if ((this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION) ||
        (this.current_action == beat.toolchains.editor.ACTION_CONNECTION_CREATION) ||
        (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION))
    {
        this.current_action = beat.toolchains.editor.ACTION_NONE;
        $(this.canvas).removeClass('action-translation');

        // Invoke the corresponding callback (if any)
        if (this.callbacks.onComponentCreationCancelled)
            this.callbacks.onComponentCreationCancelled();

        this.action_data    = null;
        this.mouse_position = null;

        this.refresh();

        if (this.callbacks.onActionChanged)
            this.callbacks.onActionChanged(this.current_action);

        return false;
    }

    else if (this.current_action == beat.toolchains.editor.ACTION_COLOR_PICKING)
    {
        $(this.color_picker_anchor).spectrum("hide");
    }

    else if ((this.current_action == beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION) ||
             (this.current_action == beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION))
    {
        this.action_data.connection.cancelSegmentTranslation(this.zoom);
        this.action_data = null;

        this.hovered_connection = null;
        this.hovered_segment = null;

        this.current_action = beat.toolchains.editor.ACTION_NONE;

        $(this.canvas).removeClass('action-vertical-translation');
        $(this.canvas).removeClass('action-horizontal-translation');

        this.refresh();

        if (this.callbacks.onActionChanged)
            this.callbacks.onActionChanged(this.current_action);

        return false;
    }
}


//----------------------------------------------------------------------------------------
// Called when the 'Delete' key is released on the keyboard
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onDeleteKeyUp = function(event)
{
    if (this.current_action == beat.toolchains.editor.ACTION_NONE)
    {
        if ((this.selection.length == 0) && (this.hovered_connection == null))
            return false;

        while (this.selection.length > 0)
        {
            var representation = this.selection[0];

            // Block
            if (representation instanceof beat.toolchains.common.BlockRepresentation)
            {
                // Delete the connection representations (if applicable)
                for (var i = 0; i < representation.block.nbInputs(); ++i)
                {
                    var input = representation.block.inputs[i];

                    if (input.connections.length == 1)
                    {
                        var connection_representation = input.connections[0].representation;
                        this.connections.splice(this.connections.indexOf(connection_representation), 1);
                        this.selection.splice(this.selection.indexOf(connection_representation), 1);
                        delete input.connections[0].representation;

                        if (this.hovered_connection == connection_representation)
                        {
                            this.hovered_connection = null;
                            this.hovered_segment = null;
                        }
                    }
                }

                for (var i = 0; i < representation.block.nbOutputs(); ++i)
                {
                    var output = representation.block.outputs[i];

                    for (var j = 0; j < output.connections.length; ++j)
                    {
                        var connection_representation = output.connections[j].representation;
                        this.connections.splice(this.connections.indexOf(connection_representation), 1);
                        this.selection.splice(this.selection.indexOf(connection_representation), 1);
                        delete output.connections[0].representation;

                        if (this.hovered_connection == connection_representation)
                        {
                            this.hovered_connection = null;
                            this.hovered_segment = null;
                        }
                    }
                }

                delete representation.block.representation;

                this.callbacks.onComponentDeleted(representation.block.name);

                this.blocks.splice(this.blocks.indexOf(representation), 1);

                if (representation.block.isDataset())
                    delete this.channel_colors[representation.block.name];
            }

            // Connection
            else if (representation instanceof beat.toolchains.common.ConnectionRepresentation)
            {
                this.callbacks.onConnectionDeleted(representation.connection);
                this.connections.splice(this.connections.indexOf(representation), 1);

                if (this.hovered_connection == representation)
                {
                    this.hovered_connection = null;
                    this.hovered_segment = null;
                }
            }

            this.selection.splice(0, 1);
        }

        if (this.hovered_connection)
        {
            this.callbacks.onConnectionDeleted(this.hovered_connection.connection);
            this.connections.splice(this.connections.indexOf(this.hovered_connection), 1);

            this.hovered_connection = null;
            this.hovered_segment = null;

            $(this.canvas).removeClass('action-vertical-translation');
            $(this.canvas).removeClass('action-horizontal-translation');
        }

        this.refresh();

        this.callbacks.onToolchainModified();
    }

    else if ((this.current_action == beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION) ||
             (this.current_action == beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION))
    {
        this.callbacks.onConnectionDeleted(this.action_data.connection.connection);
        this.connections.splice(this.connections.indexOf(this.action_data.connection), 1);

        this.action_data = null;

        this.hovered_connection = null;
        this.hovered_segment = null;

        this.current_action = beat.toolchains.editor.ACTION_NONE;

        $(this.canvas).removeClass('action-vertical-translation');
        $(this.canvas).removeClass('action-horizontal-translation');

        this.refresh();

        this.callbacks.onToolchainModified();

        if (this.callbacks.onActionChanged)
            this.callbacks.onActionChanged(this.current_action);
    }

    return false;
}


//----------------------------------------------------------------------------------------
// Called when the mouse enter the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onMouseEnter = function(event)
{
    this.mouse_position = this._getMousePosition(event);

    // Creation of new blocks: update their position
    if (this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION)
    {
        // The very first time: create their visual representation
        if (!this.action_data[0].representation)
        {
            for (var i = 0; i < this.action_data.length; ++i)
            {
                var representation = new beat.toolchains.common.BlockRepresentation(this.action_data[i].block);
                this.action_data[i].block.representation = representation;
                representation.coords_in_pixels = true;
                this.action_data[i].representation = representation;
            }
            $(this.canvas).addClass('action-translation');
        }

        var left = this.mouse_position.x + this.visible_area.left * this.zoom -
                   0.5 * beat.toolchains.common.cellsToPixels(this.action_data[0].representation.width, this.zoom);
        var top = this.mouse_position.y + this.visible_area.top * this.zoom -
                  0.5 * beat.toolchains.common.cellsToPixels(this.action_data[0].representation.height, this.zoom);

        for (var i = 0; i < this.action_data.length; ++i)
        {
            this.action_data[i].representation.left = left;
            this.action_data[i].representation.top = top;

            top += beat.toolchains.common.cellsToPixels(this.action_data[i].representation.height + 1, this.zoom);
        }

        this.refresh();
    }


    // Importation of a toolchain: update its position
    else if (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION)
    {
        var diff_x = this.visible_area.left * this.zoom + this.mouse_position.x - this.action_data[0].bounding_box.width() / 2 - this.action_data[0].bounding_box.left;
        var diff_y = this.visible_area.top * this.zoom + this.mouse_position.y - this.action_data[0].bounding_box.height() / 2 - this.action_data[0].bounding_box.top;

        for (var i = 0; i < this.action_data[0].blocks.length; ++i)
        {
            this.action_data[0].blocks[i].left += diff_x;
            this.action_data[0].blocks[i].top  += diff_y;
        }

        for (var i = 0; i < this.action_data[0].connections.length; ++i)
            this.action_data[0].connections[i].translate(diff_x, diff_y, this.zoom);

        this.action_data[0].bounding_box.left   += diff_x;
        this.action_data[0].bounding_box.right  += diff_x;
        this.action_data[0].bounding_box.top    += diff_y;
        this.action_data[0].bounding_box.bottom += diff_y;

        this.refresh();
    }
}


//----------------------------------------------------------------------------------------
// Called when the mouse leaves the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onMouseOut = function(event)
{
    // Translation of the view: stop it
    if (this.current_action == beat.toolchains.editor.ACTION_TRANSLATE_VIEW)
    {
        this.current_action = beat.toolchains.editor.ACTION_NONE;
        $(this.canvas).removeClass('action-translation');
        this.mouse_position = null;

        if (this.callbacks.onActionChanged)
            this.callbacks.onActionChanged(this.current_action);
    }

    // Scrollbar usage
    else if (this.current_action == beat.toolchains.editor.ACTION_SCROLLBAR)
    {
        clearInterval(this.action_data.timeout);
        this.current_action = beat.toolchains.editor.ACTION_NONE;
        this.action_data = null;
    }
}


//----------------------------------------------------------------------------------------
// Called when the mouse wheel is scrolled
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._onMouseWheel = function(event)
{
    var previous_zoom = this.zoom;
    var must_redraw = false;

    // With Ctrl key: control the zoom
    if (event.ctrlKey)
    {
        // Scroll up
        if (event.originalEvent.wheelDelta > 0 || event.originalEvent.detail < 0)
            this.zoom = Math.min(this.zoom + 0.1, 1.5);

        // Scroll down
        else
            this.zoom = Math.max(this.zoom - 0.1, 0.5);

        must_redraw = (previous_zoom != this.zoom);
    }

    // Without modifier key: translate the view
    else
    {
        var mouse_position = this._getMousePosition(event);

        // Mouse over the horizontal scrollbar: translate horizontally
        if (mouse_position.y >= this.canvas.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
        {
            // Scroll up
            if (event.originalEvent.wheelDelta > 0 || event.originalEvent.detail < 0)
                this.visible_area.left -= beat.toolchains.common.CELL_SPACING / this.zoom;

            // Scroll down
            else
                this.visible_area.left += beat.toolchains.common.CELL_SPACING / this.zoom;
        }

        // Mouse over the vertical scrollbar: translate vertically
        else
        {
            // Scroll up
            if (event.originalEvent.wheelDelta > 0 || event.originalEvent.detail < 0)
                this.visible_area.top -= beat.toolchains.common.CELL_SPACING / this.zoom;

            // Scroll down
            else
                this.visible_area.top += beat.toolchains.common.CELL_SPACING / this.zoom;
        }

        must_redraw = true;
    }

    // Avoid unnecessary refreshes
    if (must_redraw)
        this._updateSelectionOrActionDataPosition(previous_zoom);
}


//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._updateSelectionOrActionDataPosition = function(previous_zoom)
{
    this.visible_area.width  = this.canvas.width / this.zoom;
    this.visible_area.height = this.canvas.height / this.zoom;

    // If we are currently creating new blocks, correctly update their position
    if (this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION)
    {
        if (this.action_data[0].representation)
        {
            var top = this.action_data[0].representation.top + beat.toolchains.common.cellsToPixels(this.action_data[0].representation.height + 1, this.zoom);

            for (var i = 1; i < this.action_data.length; ++i)
            {
                this.action_data[i].representation.top = top;
                top += beat.toolchains.common.cellsToPixels(this.action_data[i].representation.height + 1, this.zoom);
            }
        }
    }

    // If we are currently importing a toolchain, correctly update the position
    // of all the blocks and connections
    else if (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION)
    {
        for (var i = 0; i < this.action_data[0].blocks.length; ++i)
        {
            var representation = this.action_data[0].blocks[i];
            representation.top  = representation.top / previous_zoom * this.zoom;
            representation.left = representation.left / previous_zoom * this.zoom;
        }

        for (var i = 0; i < this.action_data[0].connections.length; ++i)
        {
            var representation = this.action_data[0].connections[i];
            representation.scale(this.zoom / previous_zoom, this.zoom);
        }
    }

    // If we are currently translating selected components, correctly update their
    // positions
    else if (this.current_action == beat.toolchains.editor.ACTION_TRANSLATE_SELECTION)
    {
        for (var i = 0; i < this.selection.length; ++i)
        {
            var representation = this.selection[i];

            if (representation.type() == beat.toolchains.editor.TYPE_CONNECTION)
            {
                representation.scale(this.zoom / previous_zoom, this.zoom);
            }
            else
            {
                representation.top  = representation.top / previous_zoom * this.zoom;
                representation.left = representation.left / previous_zoom * this.zoom;
            }
        }
    }

    // If we are currently creating a new connection, correctly update the position
    // of each point
    else if (this.current_action == beat.toolchains.editor.ACTION_CONNECTION_CREATION)
    {
        this.action_data.representation.scale(this.zoom / previous_zoom, this.zoom);
        this.action_data.representation.update(this.mouse_position.x, this.mouse_position.y, this.zoom);
    }

    // If we are currently importing a toolchain, correctly update the position
    // of all the blocks and connections
    else if ((this.current_action == beat.toolchains.editor.ACTION_HORIZONTAL_TRANSLATION) ||
             (this.current_action == beat.toolchains.editor.ACTION_VERTICAL_TRANSLATION))
    {
        this.action_data.connection.scale(this.zoom / previous_zoom, this.zoom);
    }

    this.refresh();
}


//----------------------------------------------------------------------------------------
// Resize the view
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._getMousePosition = function(event)
{
    var canvas_position = $(this.canvas).offset();

    return {
        x: event.clientX - Math.round(canvas_position.left) - this.canvas.clientLeft + window.scrollX,
        y: event.clientY - Math.round(canvas_position.top) - this.canvas.clientTop + window.scrollY,
    };
}


//----------------------------------------------------------------------------------------
// Returns details about the visual representation at the given position
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._getComponentAt = function(x, y)
{
    x += this.visible_area.left * this.zoom;
    y += this.visible_area.top * this.zoom;

    var col = beat.toolchains.common.pixelsToCells(x, this.zoom);
    var row = beat.toolchains.common.pixelsToCells(y, this.zoom);

    for (var i = this.connections.length - 1; i >= 0; --i)
    {
        var representation = this.connections[i];

        if ((representation.bounding_box.top > row) || (representation.bounding_box.bottom < row) ||
            (representation.bounding_box.left > col) || (representation.bounding_box.right < col))
            continue;

        for (var j = 1; j < representation.nbSegments() - 1; ++j)
        {
            var segment = representation.getSegment(j);

            if (segment.isVertical())
            {
                var segment_x  = beat.toolchains.common.cellsToPixels(segment.point1.col, this.zoom);
                var segment_y1 = beat.toolchains.common.cellsToPixels(segment.point1.row, this.zoom);
                var segment_y2 = beat.toolchains.common.cellsToPixels(segment.point2.row, this.zoom);

                if (segment_y1 > segment_y2)
                {
                    var tmp = segment_y2;
                    segment_y2 = segment_y1;
                    segment_y1 = tmp;
                }

                if ((x >= segment_x - 2) && (x <= segment_x + 2) && (y >= segment_y1) && (y <= segment_y2))
                {
                    return {
                        component:               representation,
                        synchronization_channel: false,
                        input:                   null,
                        output:                  null,
                        input_handle:            null,
                        output_handle:           null,
                        add_input_button:        false,
                        add_output_button:       false,
                        delete_input_button:     null,
                        delete_output_button:    null,
                        segment:                 j,
                    };
                }
            }
            else
            {
                var segment_x1 = beat.toolchains.common.cellsToPixels(segment.point1.col, this.zoom);
                var segment_x2 = beat.toolchains.common.cellsToPixels(segment.point2.col, this.zoom);
                var segment_y  = beat.toolchains.common.cellsToPixels(segment.point1.row, this.zoom);

                if (segment_x1 > segment_x2)
                {
                    var tmp = segment_x2;
                    segment_x2 = segment_x1;
                    segment_x1 = tmp;
                }

                if ((y >= segment_y - 2) && (y <= segment_y + 2) && (x >= segment_x1) && (x <= segment_x2))
                {
                    return {
                        component:               representation,
                        synchronization_channel: false,
                        input:                   null,
                        output:                  null,
                        input_handle:            null,
                        output_handle:           null,
                        add_input_button:        false,
                        add_output_button:       false,
                        delete_input_button:     null,
                        delete_output_button:    null,
                        segment:                 j,
                    };
                }
            }
        }
    }

    for (var i = this.blocks.length - 1; i >= 0; --i)
    {
        var representation = this.blocks[i];

        if ((representation.top > row) || (representation.top + representation.height <= row))
            continue;

        var x2 = x - beat.toolchains.common.cellsToPixels(representation.left, this.zoom);
        var y2 = y - beat.toolchains.common.cellsToPixels(representation.top, this.zoom);

        // Input connector handle?
        if (((col == representation.left - 1) || (col == representation.left)) &&
             (x2 >= -beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH * this.zoom) &&
             (x2 <= beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH * this.zoom))
        {
            for (var j = 0; j < representation.input_coords.length; ++j)
            {
                var coords = representation.input_coords[j];

                if ((coords.y <= y2) && (coords.y + coords.height > y2))
                {
                    return {
                        component:               representation,
                        synchronization_channel: false,
                        input:                   null,
                        output:                  null,
                        input_handle:            j,
                        output_handle:           null,
                        add_input_button:        false,
                        add_output_button:       false,
                        delete_input_button:     null,
                        delete_output_button:    null,
                        segment:                 null,
                    };
                }
            }
        }

        // Output connector handle?
        else if (((col == representation.left + representation.width - 1) || (col == representation.left + representation.width)) &&
                 (x2 <= beat.toolchains.common.cellsToPixels(representation.width, this.zoom) + beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH * this.zoom) &&
                 (x2 >= beat.toolchains.common.cellsToPixels(representation.width, this.zoom) - beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH * this.zoom))
        {
            for (var j = 0; j < representation.output_coords.length; ++j)
            {
                var coords = representation.output_coords[j];

                if ((coords.y <= y2) && (coords.y + coords.height > y2))
                {
                    return {
                        component:               representation,
                        synchronization_channel: false,
                        input:                   null,
                        output:                  null,
                        input_handle:            null,
                        output_handle:           j,
                        add_input_button:        false,
                        add_output_button:       false,
                        delete_input_button:     null,
                        delete_output_button:    null,
                        segment:                 null,
                    };
                }
            }
        }

        // The block itself?
        if ((representation.left <= col) && (representation.left + representation.width > col))
        {
            var result = {
                component:               representation,
                synchronization_channel: false,
                input:                   null,
                output:                  null,
                input_handle:            null,
                output_handle:           null,
                add_input_button:        false,
                add_output_button:       false,
                delete_input_button:     null,
                delete_output_button:    null,
                segment:                 null,
            };

            if (representation.block.channel || representation.block.isDataset())
            {
                var coords = {
                    x: beat.toolchains.common.cellsToPixels(representation.width, this.zoom) - (beat.toolchains.common.BLOCK_CHANNEL_RADIUS + 6) * this.zoom,
                    y: (beat.toolchains.common.BLOCK_CHANNEL_RADIUS + 4) * this.zoom,
                    radius: beat.toolchains.common.BLOCK_CHANNEL_RADIUS * this.zoom,
                };

                if ((coords.x - coords.radius < x2) && (coords.x + coords.radius > x2) &&
                    (coords.y - coords.radius < y2) && (coords.y + coords.radius > y2))
                {
                    result.synchronization_channel = true;
                }
            }

            if (!result.synchronization_channel)
            {
                for (var j = 0; j < representation.input_coords.length; ++j)
                {
                    var coords = representation.input_coords[j];

                    if ((coords.x <= x2) && (coords.x + coords.width > x2) &&
                        (coords.y <= y2) && (coords.y + coords.height > y2))
                    {
                        result.input = j;
                        break;
                    }
                    else if (representation.selected &&
                             (coords.delete_button_dim !== null) &&
                             (coords.x + coords.width + (beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT / 2 - 1) * this.zoom <= x2) &&
                             (coords.x + coords.width + (beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT / 2 - 1) * this.zoom + coords.delete_button_dim > x2) &&
                             (coords.y + 0.5 * this.zoom <= y2) &&
                             (coords.y + 0.5 * this.zoom + coords.delete_button_dim > y2))
                    {
                        result.delete_input_button = j;
                        break;
                    }
                }
            }

            if (!result.synchronization_channel && (result.input === null) && (result.delete_input_button === null))
            {
                for (var j = 0; j < representation.output_coords.length; ++j)
                {
                    var coords = representation.output_coords[j];
                    if ((coords.x <= x2) && (coords.x + coords.width > x2) &&
                        (coords.y <= y2) && (coords.y + coords.height > y2))
                    {
                        result.output = j;
                        break;
                    }
                    else if (representation.selected &&
                             (coords.delete_button_dim !== null) &&
                             (coords.x - coords.delete_button_dim <= x2) &&
                             (coords.x > x2) &&
                             (coords.y + 0.5 * this.zoom <= y2) &&
                             (coords.y + 0.5 * this.zoom + coords.delete_button_dim > y2))
                    {
                        result.delete_output_button = j;
                        break;
                    }
                }
            }

            if (representation.selected && !result.synchronization_channel && (result.input === null) &&
                (result.delete_input_button === null) && (result.output === null) &&
                (result.delete_output_button === null))
            {
                if ((representation.add_input_button_coords !== null) &&
                    (representation.add_input_button_coords.x <= x2) &&
                    (representation.add_input_button_coords.x + representation.add_input_button_coords.dim > x2) &&
                    (representation.add_input_button_coords.y <= y2) &&
                    (representation.add_input_button_coords.y + representation.add_input_button_coords.dim > y2))
                {
                    result.add_input_button = true;
                }
                else if ((representation.add_output_button_coords !== null) &&
                         (representation.add_output_button_coords.x <= x2) &&
                         (representation.add_output_button_coords.x + representation.add_output_button_coords.dim > x2) &&
                         (representation.add_output_button_coords.y <= y2) &&
                         (representation.add_output_button_coords.y + representation.add_output_button_coords.dim > y2))
                {
                    result.add_output_button = true;
                }
            }

            return result;
        }
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Add a block to the selection
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._selectBlock = function(block_representation, replace)
{
    if (replace)
    {
        this._deselectAll();
        this.selection = [block_representation];
    }
    else
    {
        this.selection.push(block_representation);
    }

    block_representation.selected = true;

    var selected_connections = this._selectConnectionsOfBlock(block_representation.block);

    if (selected_connections.length > 0)
        this.selection = this.selection.concat(selected_connections);

    // Ensure that the newly selected component is drawn after all the others
    var index2 = this.blocks.indexOf(block_representation);
    this.blocks.splice(index2, 1);
    this.blocks.push(block_representation);
}


//----------------------------------------------------------------------------------------
// Deselect everything
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._deselectAll = function()
{
    for (var i = 0; i < this.selection.length; ++i)
    {
        if (this.selection[i].type() == beat.toolchains.editor.TYPE_CONNECTION)
            this.selection[i].selected = beat.toolchains.common.SELECTION_NONE;
        else
            this.selection[i].selected = false;
    }

    this.selection = [];
}


//----------------------------------------------------------------------------------------
// Smart selection of the connections of a given block
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._selectConnectionsOfBlock = function(block)
{
    var selected_connections = [];

    for (var i = 0; i < block.nbInputs(); ++i)
    {
        var input = block.inputs[i];

        if (input.connections.length == 1)
        {
            var connection_representation = input.connections[0].representation;

            if (connection_representation.selected == beat.toolchains.common.SELECTION_NONE)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_INPUT;
                selected_connections.push(connection_representation);
            }
            else if (connection_representation.selected == beat.toolchains.common.SELECTION_OUTPUT)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_FULL;
            }
        }
    }

    for (var i = 0; i < block.nbOutputs(); ++i)
    {
        var output = block.outputs[i];

        for (var j = 0; j < output.connections.length; ++j)
        {
            var connection_representation = output.connections[j].representation;

            if (connection_representation.selected == beat.toolchains.common.SELECTION_NONE)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_OUTPUT;
                selected_connections.push(connection_representation);
            }
            else if (connection_representation.selected == beat.toolchains.common.SELECTION_INPUT)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_FULL;
            }
        }
    }

    return selected_connections;
}


//----------------------------------------------------------------------------------------
// Smart deselection of the connections of a given block
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._deselectConnectionsOfBlock = function(block)
{
    var deselected_connections = [];

    for (var i = 0; i < block.nbInputs(); ++i)
    {
        var input = block.inputs[i];

        if (input.connections.length == 1)
        {
            var connection_representation = input.connections[0].representation;

            if (connection_representation.selected == beat.toolchains.common.SELECTION_INPUT)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_NONE;
                deselected_connections.push(connection_representation);
            }
            else if (connection_representation.selected == beat.toolchains.common.SELECTION_FULL)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_OUTPUT;
            }
        }
    }

    for (var i = 0; i < block.nbOutputs(); ++i)
    {
        var output = block.outputs[i];

        for (var j = 0; j < output.connections.length; ++j)
        {
            var connection_representation = output.connections[j].representation;

            if (connection_representation.selected == beat.toolchains.common.SELECTION_OUTPUT)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_NONE;
                deselected_connections.push(connection_representation);
            }
            else if (connection_representation.selected == beat.toolchains.common.SELECTION_FULL)
            {
                connection_representation.selected = beat.toolchains.common.SELECTION_INPUT;
            }
        }
    }

    return deselected_connections;
}


//----------------------------------------------------------------------------------------
// Draw the grid
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._drawGrid = function()
{
    this.context.save();

    this.context.fillStyle = '#000000';

    var increment = this.zoom * beat.toolchains.common.CELL_SPACING;

    var x0 = null;
    var y = null;

    if (this.visible_area.left >= 0)
    {
        x0 = increment - (this.visible_area.left % beat.toolchains.common.CELL_SPACING) * this.zoom;
    }
    else
    {
        x0 = - (this.visible_area.left % beat.toolchains.common.CELL_SPACING) * this.zoom;
        if (x0 == 0)
            x0 = increment;
    }

    if (this.visible_area.top >= 0)
    {
        y = increment - (this.visible_area.top % beat.toolchains.common.CELL_SPACING) * this.zoom;
    }
    else
    {
        y = - (this.visible_area.top % beat.toolchains.common.CELL_SPACING) * this.zoom;
        if (y == 0)
            y = increment;
    }

    while (y <= this.canvas.height)
    {
        var x = x0;

        while (x <= this.canvas.width)
        {
            this.context.fillRect(Math.floor(x), Math.floor(y), 1, 1);
            x += increment;
        }

        y += increment;
    }

    this.context.restore();
}


//----------------------------------------------------------------------------------------
// Draw a block
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._drawTranslatedBlockTarget = function(block_representation)
{
    if (!block_representation || !block_representation.coords_in_pixels)
        return;

    this.context.save();

    // Transformations
    var col = beat.toolchains.common.nearestCell(block_representation.left, this.zoom);
    var row = beat.toolchains.common.nearestCell(block_representation.top, this.zoom);

    this.context.translate(beat.toolchains.common.cellsToPixels(col, this.zoom),
                           beat.toolchains.common.cellsToPixels(row, this.zoom));

    // Block box
    var width  = beat.toolchains.common.cellsToPixels(block_representation.width, this.zoom);
    var height = beat.toolchains.common.cellsToPixels(block_representation.height, this.zoom);

    if (this.current_action == beat.toolchains.editor.ACTION_BLOCK_CREATION)
        this.context.fillStyle = '#66FF66';
    else
        this.context.fillStyle = '#226622';

    this.context.globalAlpha = 0.2;

    this.context.fillRect(0.5, 0.5, width, height);

    this.context.restore();
}


//----------------------------------------------------------------------------------------
// Draw a block
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._drawBlock = function(block_representation, creation)
{
    if (!block_representation)
        return;

    this.context.save();

    // Transformations
    if (!block_representation.coords_in_pixels)
    {
        this.context.translate(beat.toolchains.common.cellsToPixels(block_representation.left, this.zoom),
                               beat.toolchains.common.cellsToPixels(block_representation.top, this.zoom));
    }
    else
    {
        this.context.translate(Math.floor(block_representation.left),
                               Math.floor(block_representation.top));
    }


    // Block box
    this.context.strokeStyle = (block_representation.selected ? '#FF5500' : '#000000');
    this.context.fillStyle   = (block_representation.selected
                                    ? '#FFCC00'
                                    : (block_representation.block.isAnalyzer()
                                            ? '#CCCCFA'
                                            : (block_representation.block.isDataset()
                                                    ? '#FAF8CC'
                                                    : '#CCCCCC')));

    this.context.lineWidth = 1;
    this.context.lineCap   = 'square';

    if (creation)
        this.context.globalAlpha = 0.2;

    var width  = beat.toolchains.common.cellsToPixels(block_representation.width, this.zoom);
    var height = beat.toolchains.common.cellsToPixels(block_representation.height, this.zoom);
    var radius = (block_representation.block.isDataset() ? 18 : 14) * this.zoom;

    this.context.beginPath();

    if (!block_representation.block.isDataset())
    {
        this.context.moveTo(0.5, 0.5);
    }
    else
    {
        this.context.moveTo(0.5, radius + 0.5);
        this.context.arc(radius + 0.5, radius + 0.5, radius, Math.PI, -Math.PI / 2, false);
    }

    this.context.lineTo(width - radius, 0.5);
    this.context.arc(width - radius + 0.5, radius + 0.5, radius, -Math.PI / 2, 0, false);
    this.context.lineTo(width + 0.5, height - radius + 0.5);
    this.context.arc(width - radius + 0.5, height - radius + 0.5, radius, 0, Math.PI / 2, false);

    if (!block_representation.block.isDataset())
    {
        this.context.lineTo(0.5, height + 0.5);
        this.context.lineTo(0.5, 0.5);
    }
    else
    {
        this.context.lineTo(radius + 0.5, height + 0.5);
        this.context.arc(radius + 0.5, height - radius + 0.5, radius, Math.PI / 2, Math.PI, false);
        this.context.lineTo(0.5, radius + 0.5);
    }

    this.context.fill();
    this.context.stroke();


    // Block title
    this.context.font         = 'Bold ' + (beat.toolchains.common.CELL_SPACING / 2 * this.zoom) + 'px Arial';
    this.context.fillStyle    = (block_representation.selected ? '#FF5500' : '#000000');
    this.context.textBaseline = 'top';
    this.context.textAlign    = 'center';

    var name = block_representation.block.name;
    var metrics = this.context.measureText(name);
    var title_width = width - 2 * radius - 4;

    var nb = 1;
    while (metrics.width > title_width - 8 * this.zoom)
    {
        ++nb;

        var offset = block_representation.block.name.length >> 1;

        name = block_representation.block.name.substr(0, offset - nb) + '...' +
               block_representation.block.name.substr(offset + nb);

        metrics = this.context.measureText(name);
    }

    this.context.fillText(name, Math.floor(width / 2), Math.floor(4 * this.zoom));


    // Inputs
    var max_connector_width = beat.toolchains.common.cellsToPixels(block_representation.width, this.zoom) * 0.5 - 20 * this.zoom;
    var connector_height = Math.floor(beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT * this.zoom);
    var connector_radius = connector_height / 2;

    this.context.strokeStyle = (block_representation.selected ? '#FF5500' : '#000000');
    this.context.font = '' + ((beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT - 4) * this.zoom) + 'px Arial';
    this.context.textBaseline = 'middle';
    this.context.textAlign = 'start';

    if (block_representation.block.isDataset())
        max_connector_width += 40 * this.zoom;

    var view = this;

    function _computeConnectorsWidth(connectors) {
        if (!connectors)
            return {
                width: 0,
                names: [],
            }

        var connectors_width = 0;
        var names = [];

        for (var i = 0; i < connectors.length; ++i)
        {
            var name = connectors[i].name;
            var metrics = view.context.measureText(name);

            var nb = 2;
            while (metrics.width > max_connector_width - 4 * view.zoom)
            {
                ++nb;
                name = connectors[i].name.substr(0, connectors[i].name.length - nb) + '...';
                metrics = view.context.measureText(name);
            }

            connectors_width = Math.min(max_connector_width,
                                        Math.max(connectors_width,
                                                 metrics.width + 4 * view.zoom));

            names.push(name);
        }

        return {
            width: connectors_width,
            names: names,
        };
    }


    function _drawConnector(connector_name, x, y, width, input, highlighted, hovered) {
        if (!input)
            x -= width;

        if (hovered && !input)
        {
            view.context.beginPath();
            view.context.arc(x + width, y + 0.5 + connector_radius,
                             connector_radius + 4, -Math.PI / 2, Math.PI / 2, false);
            view.context.fillStyle = '#555555';
            view.context.fill();
        }

        // Connector box
        view.context.beginPath();
        view.context.moveTo(x + 0.5, y + 0.5);
        view.context.lineTo(x + width, y + 0.5);
        view.context.arc(x + width, y + 0.5 + connector_radius,
                         connector_radius, -Math.PI / 2, Math.PI / 2, false);
        view.context.lineTo(x + 0.5, y + connector_height + 0.5);
        view.context.lineTo(x + 0.5, y + 0.5);

        view.context.fillStyle = '#f6f810';
        view.context.fill();
        view.context.stroke();

        // Connector name
        view.context.fillStyle = (block_representation.selected ? '#FF5500' : '#000000');
        view.context.fillText(connector_name, x + Math.floor(4 * view.zoom), y + connector_radius);

        // If necessary: delete button
        var delete_button_dim = null;
        if (highlighted !== null)
        {
            delete_button_dim = 12 * view.zoom;
            view.context.drawImage(highlighted ? view.images.delete_highlighted
                                               : view.images.delete,
                                   input ? x + width + connector_radius + (3 * view.zoom) : x - delete_button_dim - (2 * view.zoom),
                                   y + 0.5 * (6*view.zoom), delete_button_dim, delete_button_dim);
        }

        if (hovered && input)
        {
            view.context.beginPath();
            view.context.arc(x, y + 0.5 + connector_radius,
                             beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH * view.zoom,
                             Math.PI / 2, -Math.PI / 2, false);
            view.context.fillStyle = '#555555';
            view.context.fill();
        }

        return {
            x: x,
            y: y,
            width: width,
            height: connector_height,
            delete_button_dim: delete_button_dim,
        };
    }


    var hovered         = (this.hovered_block == block_representation);
    var hovered_inputs  = hovered;
    var hovered_outputs = hovered;

    if (hovered)
    {
        if (this.current_action == beat.toolchains.editor.ACTION_CONNECTION_CREATION)
        {
            hovered_inputs  = (this.action_data.source_input == null);
            hovered_outputs = !hovered_inputs;
        }
    }


    var connectors_infos = _computeConnectorsWidth(block_representation.block.inputs);
    for (var i = 0; i < connectors_infos.names.length; ++i)
    {
        var y = Math.floor(((i + 2) * beat.toolchains.common.CELL_SPACING - (beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT - 1) / 2) * this.zoom);

        var highlighted = (block_representation.selected && (connectors_infos.names.length > 1)
                                                         ? (block_representation.highlighted_button === i)
                                                         : null);

        block_representation.input_coords[i] = _drawConnector(connectors_infos.names[i],
                                                              0, y, connectors_infos.width,
                                                              true, highlighted,
                                                              hovered_inputs);
    }


    var connectors_infos = _computeConnectorsWidth(block_representation.block.outputs);
    for (var i = 0; i < connectors_infos.names.length; ++i)
    {
        var y = Math.floor(((i + 2) * beat.toolchains.common.CELL_SPACING - (beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT - 1) / 2) * this.zoom);

        var highlighted = (block_representation.selected && !block_representation.block.isDataset() && (connectors_infos.names.length > 1)
                                                         ? (block_representation.highlighted_button === i + block_representation.block.inputs.length)
                                                         : null);

        block_representation.output_coords[i] = _drawConnector(connectors_infos.names[i],
                                                               width, y, connectors_infos.width,
                                                               false, highlighted,
                                                               hovered_outputs);
    }


    // Synchronization channel
    if (block_representation.block.channel || block_representation.block.isDataset())
    {
        this.context.beginPath();
        this.context.arc(width - (beat.toolchains.common.BLOCK_CHANNEL_RADIUS + 6) * this.zoom + 0.5,
                         (beat.toolchains.common.BLOCK_CHANNEL_RADIUS + 4) * this.zoom + 0.5,
                         beat.toolchains.common.BLOCK_CHANNEL_RADIUS * this.zoom, 0, 2 * Math.PI, false);

        this.context.fillStyle = (block_representation.block.channel ? this.channel_colors[block_representation.block.channel]
                                                                     : this.channel_colors[block_representation.block.name]);
        this.context.fill();
    }


    // Only if the block is selected: add customization controls
    if (block_representation.selected && !block_representation.block.isDataset())
    {
        if ((this.images.add !== null) && (this.images.add_highlighted !== null))
        {
            // New input button
            if (!block_representation.block.isDataset())
            {
                block_representation.add_input_button_coords = {
                    x:   2 * this.zoom,
                    y:   beat.toolchains.common.cellsToPixels(block_representation.block.inputs.length + 2, this.zoom) - 16 * this.zoom,
                    dim: 12 * this.zoom,
                };

                this.context.drawImage(block_representation.highlighted_button == -1 ? this.images.add_highlighted
                                                                                     : this.images.add,
                                       block_representation.add_input_button_coords.x,
                                       block_representation.add_input_button_coords.y,
                                       block_representation.add_input_button_coords.dim,
                                       block_representation.add_input_button_coords.dim);
            }

            // New output button
            if (!block_representation.block.isAnalyzer())
            {
                block_representation.add_output_button_coords = {
                    x:   width - connectors_infos.width,
                    y:   beat.toolchains.common.cellsToPixels(block_representation.block.outputs.length + 2, this.zoom) - 16 * this.zoom,
                    dim: 12 * this.zoom,
                };

                this.context.drawImage(block_representation.highlighted_button == -2 ? this.images.add_highlighted
                                                                                     : this.images.add,
                                       block_representation.add_output_button_coords.x,
                                       block_representation.add_output_button_coords.y,
                                       block_representation.add_output_button_coords.dim,
                                       block_representation.add_output_button_coords.dim);
            }
        }
    }

    this.context.restore();
}


//----------------------------------------------------------------------------------------
// Draw a connection
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._drawConnection = function(representation, creation)
{
    function _coords(point, zoom)
    {
        if (point.in_cells)
        {
            return {
                x: beat.toolchains.common.cellsToPixels(point.col, zoom),
                y: beat.toolchains.common.cellsToPixels(point.row, zoom),
            };
        }
        else
        {
            return {
                x: point.x,
                y: point.y,
            };
        }
    }


    this.context.save();

    this.context.lineWidth = 2;
    this.context.lineCap   = 'square';


    var start = 0;
    var end = representation.nbSegments() - 1;

    if (creation && (this.current_action != beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION))
    {
        this.context.globalAlpha = 0.4;

        this.context.strokeStyle = '#FF5500';

        var segment = representation.getSegment(0);
        var point1 = _coords(segment.point1, this.zoom);

        this.context.beginPath();

        if (this.action_data.source_input)
        {
            if (this.hovered_block && (this.hovered_block.block === this.action_data.source_block))
                this.context.moveTo(point1.x - beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH * this.zoom + 0.5, point1.y + 0.5);
            else
                this.context.moveTo(point1.x + 0.5, point1.y + 0.5);
        }
        else
        {
            this.context.moveTo(point1.x + Math.floor(beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT * this.zoom) / 2 + 0.5, point1.y + 0.5);
        }
    }
    else
    {
        if (creation)
            this.context.globalAlpha = 0.4;

        this.context.strokeStyle = '#000000';

        this.context.beginPath();

        var segment = representation.getSegment(0);
        var point1 = _coords(segment.point1, this.zoom);
        var point2 = _coords(segment.point2, this.zoom);

        this.context.moveTo(point1.x + Math.floor(beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT * this.zoom) / 2 + 0.5,
                            point1.y + 0.5);

        this.context.lineTo(point2.x + 0.5, point2.y + 0.5);

        segment = representation.getSegment(representation.nbSegments() - 1);
        point1 = _coords(segment.point1, this.zoom);
        point2 = _coords(segment.point2, this.zoom);

        if (this.hovered_block && (this.hovered_block.block === representation.connection.input.block))
            point2.x -= beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH * this.zoom + 1;
        else
            point2.x -= 1;

        this.context.moveTo(point1.x + 0.5, point1.y + 0.5);
        this.context.lineTo(point2.x + 0.5, point2.y + 0.5);

        this.context.stroke();

        start = 1;
        end = representation.nbSegments() - 2;

        if (this.current_action == beat.toolchains.editor.ACTION_TOOLCHAIN_IMPORTATION)
            this.context.strokeStyle = this.channel_colors[representation.connection.input.channel];
        else if (representation.selected == beat.toolchains.common.SELECTION_FULL)
            this.context.strokeStyle = '#FF7700';
        else if (this.hovered_connection == representation)
            this.context.strokeStyle = '#FFAA00';
        else if (representation.connection.input.channel)
            this.context.strokeStyle = this.channel_colors[representation.connection.input.channel];
        else
            this.context.strokeStyle = '#000000';

        segment = representation.getSegment(1);
        point1 = _coords(segment.point1, this.zoom);

        this.context.beginPath();
        this.context.moveTo(point1.x + 0.5, point1.y + 0.5);
    }


    for (var i = start; i <= end; ++i)
    {
        var segment = representation.getSegment(i);
        var point2 = _coords(segment.point2, this.zoom);

        if ((this.hovered_connection == representation) && (i == this.hovered_segment))
            this.context.moveTo(point2.x + 0.5, point2.y + 0.5);
        else
            this.context.lineTo(point2.x + 0.5, point2.y + 0.5);
    }

    this.context.stroke();


    if (this.hovered_connection == representation)
    {
        var segment = representation.getSegment(this.hovered_segment);
        var point1 = _coords(segment.point1, this.zoom);
        var point2 = _coords(segment.point2, this.zoom);

        this.context.strokeStyle = '#FF2200';

        this.context.beginPath();
        this.context.moveTo(point1.x + 0.5, point1.y + 0.5);
        this.context.lineTo(point2.x + 0.5, point2.y + 0.5);
        this.context.stroke();
    }


    this.context.restore();
}


//----------------------------------------------------------------------------------------
// Draw the scrollbars
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._drawScrollbars = function()
{
    function _roundRect(ctx, x, y, width, height, radius, fill, stroke)
    {
        if (typeof stroke == "undefined")
            stroke = true;

        if (typeof radius === "undefined")
            radius = 5;

        ctx.beginPath();
        ctx.moveTo(x + radius, y);
        ctx.lineTo(x + width - radius, y);
        ctx.quadraticCurveTo(x + width, y, x + width, y + radius);
        ctx.lineTo(x + width, y + height - radius);
        ctx.quadraticCurveTo(x + width, y + height, x + width - radius, y + height);
        ctx.lineTo(x + radius, y + height);
        ctx.quadraticCurveTo(x, y + height, x, y + height - radius);
        ctx.lineTo(x, y + radius);
        ctx.quadraticCurveTo(x, y, x + radius, y);
        ctx.closePath();

        if (stroke)
            ctx.stroke();

        if (fill)
            ctx.fill();
    }

    this.context.save();


    // Compute all the needed positions / sizes
    var horizontal_top = this.canvas.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH;
    var vertical_left = this.canvas.width - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH;

    if (this.bounding_box.left !== null)
    {
        var bounding_box = {
            left:   this.bounding_box.left,
            right:  this.bounding_box.right,
            top:    this.bounding_box.top,
            bottom: this.bounding_box.bottom,
            width:  null,
            height: null,
        };

        bounding_box.width  = bounding_box.right - bounding_box.left;
        bounding_box.height = bounding_box.bottom - bounding_box.top;

        if (bounding_box.width < this.visible_area.width)
        {
            var diff = (this.visible_area.width - bounding_box.width) / 2;
            bounding_box.left -= diff;
            bounding_box.right += diff;
            bounding_box.width = this.visible_area.width;
        }

        if (bounding_box.height < this.visible_area.height)
        {
            var diff = (this.visible_area.height - bounding_box.height) / 2;
            bounding_box.top -= diff;
            bounding_box.bottom += diff;
            bounding_box.height = this.visible_area.height;
        }
    }
    else
    {
        var bounding_box = {
            left:   this.visible_area.left,
            right:  this.visible_area.left + this.visible_area.width,
            top:    this.visible_area.top,
            bottom: this.visible_area.top + this.visible_area.height,
            width:  this.visible_area.width,
            height: this.visible_area.height,
        };
    }

    bounding_box.width  = bounding_box.right - bounding_box.left;
    bounding_box.height = bounding_box.bottom - bounding_box.top;

    var visible_area_center_x = this.visible_area.left + this.visible_area.width / 2;
    var visible_area_center_y = this.visible_area.top + this.visible_area.height / 2;

    var bounding_box_center_x = bounding_box.left + bounding_box.width / 2;
    var bounding_box_center_y = bounding_box.top + bounding_box.height / 2;

    var extended_width  = Math.max(bounding_box.right, this.visible_area.left + this.visible_area.width) - Math.min(bounding_box.left, this.visible_area.left);
    var extended_height = Math.max(bounding_box.bottom, this.visible_area.top + this.visible_area.height) - Math.min(bounding_box.top, this.visible_area.top);

    var covered_width  = extended_width / this.zoom + beat.toolchains.common.SCROLLBAR_MARGIN * 2;
    var covered_height = extended_height / this.zoom + beat.toolchains.common.SCROLLBAR_MARGIN * 2;

    var horizontal_x = 0.5 + (visible_area_center_x - bounding_box_center_x) / extended_width * 0.5;
    var vertical_y   = 0.5 + (visible_area_center_y - bounding_box_center_y) / extended_height * 0.5;

    this.horizontal_scrollbar.size = ((this.visible_area.width - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH) / covered_width) * this.canvas.width / this.zoom;
    this.horizontal_scrollbar.position = (this.canvas.width * horizontal_x) - this.horizontal_scrollbar.size / 2;

    this.vertical_scrollbar.size = ((this.visible_area.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH) / covered_height) * this.canvas.height / this.zoom;
    this.vertical_scrollbar.position = (this.canvas.height * vertical_y) - this.vertical_scrollbar.size / 2;


    // Draw the holders
    this.context.fillStyle = '#F9F9F9';
    this.context.fillRect(0.5, horizontal_top + 0.5, this.canvas.width, beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH);
    this.context.fillRect(vertical_left + 0.5, 0.5, beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH, this.canvas.height);

    this.context.strokeStyle = '#E8E8E8';
    this.context.strokeRect(0.5, horizontal_top + 0.5, this.canvas.width - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH, 1);
    this.context.strokeRect(vertical_left + 0.5, 0.5, 1, this.canvas.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH);


    // Draw the horizontal scrollbar
    if ((this.current_action == beat.toolchains.editor.ACTION_SCROLLBAR) && this.action_data.horizontal)
        this.context.fillStyle = '#A2A2A2';
    else
        this.context.fillStyle = '#C2C2C2';

    _roundRect(this.context, this.horizontal_scrollbar.position + 0.5, horizontal_top + beat.toolchains.common.SCROLLBAR_OFFSET + 0.5,
               this.horizontal_scrollbar.size, beat.toolchains.common.SCROLLBAR_WIDTH, beat.toolchains.common.SCROLLBAR_WIDTH / 2, true, false);

    if ((this.current_action == beat.toolchains.editor.ACTION_SCROLLBAR) && !this.action_data.horizontal)
        this.context.fillStyle = '#A2A2A2';
    else
        this.context.fillStyle = '#C2C2C2';

    _roundRect(this.context, vertical_left + beat.toolchains.common.SCROLLBAR_OFFSET + 0.5, this.vertical_scrollbar.position + 0.5,
               beat.toolchains.common.SCROLLBAR_WIDTH, this.vertical_scrollbar.size, beat.toolchains.common.SCROLLBAR_WIDTH / 2, true, false);


    this.context.restore();
}


//----------------------------------------------------------------------------------------
// Returns an unused channel color
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._getUnusedChannelColor = function()
{
    var view = this;
    var used_colors = Object.keys(this.channel_colors).map(function(k){return view.channel_colors[k].toUpperCase();});

    var palette = [
        '#0000FF',
        '#6AA84F',
        '#CC0000',
        '#274E13',
        '#7F6000',
        '#4A86E8',
        '#45818E',
        '#9900FF',
        '#8E7CC3',
    ];

    for (var j = 0; j < palette.length; ++j)
    {
        var color = palette[j];
        if (used_colors.indexOf(color) == -1)
            return color;
    }

    var color = null;

    do {
        color = '#' + Math.floor(Math.random() * 256).toString(16).toUpperCase() +
                      Math.floor(Math.random() * 256).toString(16).toUpperCase() +
                      Math.floor(Math.random() * 256).toString(16).toUpperCase();
    } while (used_colors.indexOf(color) >= 0);

    return color;
}


//----------------------------------------------------------------------------------------
// Check that the provided block name is valid
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._checkBlockName = function(
                                                    name, block_representations, current_name)
{
    var sanitized = name.trim().replace(/\W/g, '_');

    var valid = (sanitized.length > 0);

    if (valid)
    {
        for (var i = 0; i < block_representations.length; ++i)
        {
            if (block_representations[i].block.name == current_name)
                continue;

            if (sanitized == block_representations[i].block.name)
            {
                valid = false;
                break;
            }
        }
    }

    return {
        name: sanitized,
        valid: valid,
    };
}


//----------------------------------------------------------------------------------------
// Check that the provided end point name is valid
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._checkEndPointName = function(name, block, current_name)
{
    var sanitized = name.trim().replace(/\W/g, '_');

    var valid = (sanitized.length > 0);

    if (valid)
    {
        for (var i = 0; i < block.nbInputs(); ++i)
        {
            if (block.inputs[i].name == current_name)
                continue;

            if (sanitized == block.inputs[i].name)
            {
                valid = false;
                break;
            }
        }
    }

    if (valid)
    {
        for (var i = 0; i < block.nbOutputs(); ++i)
        {
            if (block.outputs[i].name == current_name)
                continue;

            if (sanitized == block.outputs[i].name)
            {
                valid = false;
                break;
            }
        }
    }

    return {
        name: sanitized,
        valid: valid,
    };
}


//----------------------------------------------------------------------------------------
// Recompute the bounding box
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainView.prototype._recomputeBoundingBox = function()
{
    this.bounding_box = new beat.toolchains.editor.ToolchainBoundingBox();

    for (var i = 0; i < this.blocks.length; ++i)
        this.bounding_box.update(this.blocks[i], this.zoom);

    for (var i = 0; i < this.connections.length; ++i)
        this.bounding_box.update(this.connections[i], this.zoom);
}


/***************************** CLASS: ToolchainBoundingBox ******************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainBoundingBox = function()
{
    // Initialisations
    this.top    = null;
    this.left   = null;
    this.bottom = null;
    this.right  = null;
}


//----------------------------------------------------------------------------------------
// Update the bounding box using the visual representation of a new component or
// connection
//----------------------------------------------------------------------------------------
beat.toolchains.editor.ToolchainBoundingBox.prototype.update = function(representation, zoom)
{
    if (representation instanceof beat.toolchains.common.ConnectionRepresentation)
    {
        if ((this.top === null) || (representation.bounding_box.top < this.top))
            this.top = representation.bounding_box.top;

        if ((this.left === null) || (representation.bounding_box.left < this.left))
            this.left = representation.bounding_box.left;

        if ((this.bottom === null) || (representation.bounding_box.bottom > this.bottom))
            this.bottom = representation.bounding_box.bottom;

        if ((this.right === null) || (representation.bounding_box.right > this.right))
            this.right = representation.bounding_box.right;
    }
    else
    {
        var bbox = representation.getBoundingBoxInPixels(zoom);

        if ((this.top === null) || (bbox.top < this.top))
            this.top = bbox.top;

        if ((this.left === null) || (bbox.left < this.left))
            this.left = bbox.left;

        if ((this.bottom === null) || (bbox.bottom > this.bottom))
            this.bottom = bbox.bottom;

        if ((this.right === null) || (bbox.right > this.right))
            this.right = bbox.right;
    }
}


//---------------------------------------------------------


beat.toolchains.editor.ToolchainBoundingBox.prototype.width = function()
{
    return this.right - this.left;
}


//---------------------------------------------------------


beat.toolchains.editor.ToolchainBoundingBox.prototype.height = function()
{
    return this.bottom - this.top;
}



/******************************* CLASS: VersioningSystem ********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.editor.VersioningSystem = function()
{
    this._versions = [];
    this._current = -1;

    this.callbacks = {
        onUpdate: null,
    };
}


//----------------------------------------------------------------------------------------
// Add a new version of the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.editor.VersioningSystem.prototype.push = function(declaration)
{
    if ((this._versions.length > 0) && _.isEqual(declaration, this._versions[this._current]))
        return;

    if (this._current < this._versions.length - 1)
    {
        this._versions.splice(this._current + 1);
        this._current = this._versions.length - 1;
    }

    this._versions.push(declaration);
    ++this._current;

    if (this.callbacks.onUpdate !== null)
    {
        this.callbacks.onUpdate((this._current > 0) && (this._versions.length > 0),
                                this._current < this._versions.length - 1);
    }
}


//----------------------------------------------------------------------------------------
// Returns the previous version
//----------------------------------------------------------------------------------------
beat.toolchains.editor.VersioningSystem.prototype.previous = function()
{
    if (this._current <= 0)
        return null;

    --this._current;

    if (this.callbacks.onUpdate !== null)
    {
        this.callbacks.onUpdate((this._current > 0) && (this._versions.length > 0),
                                this._current < this._versions.length - 1);
    }

    return this._versions[this._current];
}


//----------------------------------------------------------------------------------------
// Returns the next version
//----------------------------------------------------------------------------------------
beat.toolchains.editor.VersioningSystem.prototype.next = function()
{
    if (this._current >= this._versions.length - 1)
        return null;

    ++this._current;

    if (this.callbacks.onUpdate !== null)
    {
        this.callbacks.onUpdate((this._current > 0) && (this._versions.length > 0),
                                this._current < this._versions.length - 1);
    }

    return this._versions[this._current];
}


//----------------------------------------------------------------------------------------
// Returns the next version
//----------------------------------------------------------------------------------------
beat.toolchains.editor.VersioningSystem.prototype.reset = function()
{
    this._versions = [];
    this._current = -1;

    if (this.callbacks.onUpdate !== null)
        this.callbacks.onUpdate(false, false);
}
