/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.toolchains.models.js

   Declaration of the toolchain-related classes
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.toolchains === undefined)
    beat.toolchains = {}

if (beat.toolchains.models === undefined)
    beat.toolchains.models = {}


/************************************ CLASS: EndPoint ***********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.models.EndPoint = function(block, name, index, is_output)
{
    this.block       = block;
    this.name        = name;
    this.index       = index;
    this.is_output   = is_output;
    this.dataformat  = null;
    this.channel     = null;
    this.connections = new Array();
}


//----------------------------------------------------------------------------------------
// Returns the minimal data format that is needed on this end point to fullfill the
// requirements of the connections
//----------------------------------------------------------------------------------------
beat.toolchains.models.EndPoint.prototype.neededDataformat = function(dataformats)
{
    if (this.connections.length == 0)
        return null;

    // Output
    if (this.is_output)
    {
        if (this.connections.length == 1)
            return this.connections[0].input.dataformat;

        var needed_dataformat = null;

        for (var i = 0; i < this.connections.length; ++i)
        {
            if (!needed_dataformat)
            {
                needed_dataformat = this.connections[i].input.dataformat;
            }
            else
            {
                var dataformat = this.connections[i].input.dataformat;

                if (dataformat != needed_dataformat)
                {
                    if (dataformats.areCompatible(needed_dataformat, dataformat))
                        needed_dataformat = dataformat;
                }
            }
        }

        return needed_dataformat;
    }

    // Input
    else
    {
        var output = this.connections[0].output;

        if ((output.connections.length == 1) || output.dataformat)
            return output.dataformat;

        var needed_dataformat = null;

        for (var i = 0; i < output.connections.length; ++i)
        {
            if (output.connections[i].input == this)
                continue;

            if (!needed_dataformat)
            {
                needed_dataformat = output.connections[i].input.dataformat;
            }
            else
            {
                var dataformat = output.connections[i].input.dataformat;

                if (dataformat != needed_dataformat)
                {
                    if (dataformats.areCompatible(dataformat, needed_dataformat))
                        needed_dataformat = dataformat;
                }
            }
        }

        return needed_dataformat;
    }
}



/********************************** CLASS: Connection ***********************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.models.Connection = function(output, input, channel)
{
    this.output = output;
    this.input  = input;

    this.output.channel = channel;

    this.output.connections.push(this);
    this.output.block.connections.push(this);

    this.input.channel = channel;

    this.input.connections.push(this);

    if (this.input.block !== null)
        this.input.block.connections.push(this);
}


//----------------------------------------------------------------------------------------
// Called to destroy the connection
//----------------------------------------------------------------------------------------
beat.toolchains.models.Connection.prototype._destroy = function()
{
    for (var i = 0; i < this.input.connections.length; ++i)
    {
        if (this.input.connections[i] == this)
        {
            this.input.connections.splice(i, 1);
            break;
        }
    }

    for (var i = 0; i < this.output.connections.length; ++i)
    {
        if (this.output.connections[i] == this)
        {
            this.output.connections.splice(i, 1);
            break;
        }
    }

    this.output = undefined;
    this.input  = undefined;
}



/************************************* CLASS: Block *************************************/


beat.toolchains.models.PROCESSING_BLOCK = 1;
beat.toolchains.models.DATASET_BLOCK    = 2;
beat.toolchains.models.ANALYZER_BLOCK   = 3;


//----------------------------------------------------------------------------------------
// Constructor
//
// declaration: Declaration of the block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block = function(declaration, block_type)
{
    this.name        = declaration.name;
    this.connections = new Array();
    this.type        = (block_type !== undefined ? block_type : beat.toolchains.models.PROCESSING_BLOCK);
    this.inputs      = (this.type == beat.toolchains.models.DATASET_BLOCK ? null : new Array());
    this.outputs     = (this.type == beat.toolchains.models.ANALYZER_BLOCK ? null : new Array());
    this.channel     = (declaration.synchronized_channel ? declaration.synchronized_channel : null);

    if (!this.isDataset())
    {
        for (var j = 0; j < declaration.inputs.length; ++j)
            this.inputs.push(new beat.toolchains.models.EndPoint(this, declaration.inputs[j], j, false));
    }

    if (!this.isAnalyzer())
    {
        for (var j = 0; j < declaration.outputs.length; ++j)
            this.outputs.push(new beat.toolchains.models.EndPoint(this, declaration.outputs[j], j, true));
    }
}


//----------------------------------------------------------------------------------------
// Indicates if the block is a processing one
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.isProcessingBlock = function()
{
    return (this.type == beat.toolchains.models.PROCESSING_BLOCK);
}


//----------------------------------------------------------------------------------------
// Indicates if the block is a dataset one
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.isDataset = function()
{
    return (this.type == beat.toolchains.models.DATASET_BLOCK);
}


//----------------------------------------------------------------------------------------
// Indicates if the block is an analyzer one
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.isAnalyzer = function()
{
    return (this.type == beat.toolchains.models.ANALYZER_BLOCK);
}


//----------------------------------------------------------------------------------------
// Returns the number of inputs of the block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.nbInputs = function()
{
    return (this.inputs !== null ? this.inputs.length : 0);
}


//----------------------------------------------------------------------------------------
// Returns the number of outputs of the block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.nbOutputs = function()
{
    return (this.outputs !== null ? this.outputs.length : 0);
}


//----------------------------------------------------------------------------------------
// Retrieves an input by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.input = function(name)
{
    if (this.inputs == null)
        return null;

    for (var i = 0; i < this.inputs.length; i++)
    {
        var input = this.inputs[i];

        if (input.name == name)
            return input;
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieves an output by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.output = function(name)
{
    if (this.outputs == null)
        return null;

    for (var i = 0; i < this.outputs.length; i++)
    {
        var output = this.outputs[i];

        if (output.name == name)
            return output;
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Returns a list of the blocks that directly precedes this one in the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.predecessors = function()
{
    var predecessors = new Array();

    for (var i = 0; i < this.connections.length; i++)
    {
        if (this.connections[i].input.block == this)
        {
            if (predecessors.indexOf(this) == -1)
                predecessors.push(this.connections[i].output.block);
        }
    }

    return predecessors;
}


//----------------------------------------------------------------------------------------
// Returns a list of all the blocks that precedes this one in the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.allPredecessors = function()
{
    var predecessors = new Array();

    var direct_predecessors = this.predecessors();

    var predecessors = [].concat(direct_predecessors);

    for (var i = 0; i < direct_predecessors.length; i++)
    {
        var upstream_predecessors = direct_predecessors[i].allPredecessors();

        upstream_predecessors = upstream_predecessors.filter(function(block) {
            return predecessors.indexOf(block) == -1;
        });

        predecessors = predecessors.concat(upstream_predecessors);
    }

    return predecessors;
}


//----------------------------------------------------------------------------------------
// Retrieves the name of all the inputs
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.inputNames = function()
{
    var names = [];

    if (this.inputs != null)
    {
        for (var i = 0; i < this.inputs.length; i++)
            names.push(this.inputs[i].name);

        names.sort();
    }

    return names;
}


//----------------------------------------------------------------------------------------
// Retrieves the name of all the outputs
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.outputNames = function()
{
    var names = [];

    if (this.outputs != null)
    {
        for (var i = 0; i < this.outputs.length; i++)
            names.push(this.outputs[i].name);

        names.sort();
    }

    return names;
}


//----------------------------------------------------------------------------------------
// Retrieves a list of all the channels of the inputs
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.inputChannels = function()
{
    var channels = [];

    if (this.inputs != null)
    {
        for (var i = 0; i < this.inputs.length; i++)
        {
            if (channels.indexOf(this.inputs[i].channel) == -1)
                channels.push(this.inputs[i].channel);
        }
    }

    return channels;
}


//----------------------------------------------------------------------------------------
// Retrieves the inputs belonging to the specified group
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.inputsOfChannel = function(channel)
{
    if (this.inputs == null)
        return [];

    return this.inputs.filter(function(input) {
        return input.channel == channel;
    });
}


//----------------------------------------------------------------------------------------
// Retrieves a list of all the channels of the outputs
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.outputChannels = function()
{
    var channels = [];

    if (this.outputs != null)
    {
        for (var i = 0; i < this.outputs.length; i++)
        {
            if (channels.indexOf(this.outputs[i].channel) == -1)
                channels.push(this.outputs[i].channel);
        }
    }

    return channels;
}


//----------------------------------------------------------------------------------------
// Retrieves the outputs belonging to the specified group
//----------------------------------------------------------------------------------------
beat.toolchains.models.Block.prototype.outputsOfChannel = function(channel)
{
    if (this.outputs == null)
        return [];

    return this.outputs.filter(function(output) {
        return output.channel == channel;
    });
}



/*********************************** CLASS: Toolchain ***********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Callback signatures:
//   - onComponentNameChanged(previous_name, new_name)
//   - onComponentInputNameChanged(block_name, current_name, new_name)
//   - onComponentOutputNameChanged(block_name, current_name, new_name)
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain = function(declaration)
{
    // Check the parameters
    if (declaration === undefined)
        declaration = {
            analyzers: [],
            blocks: [],
            datasets: [],
            connections: [],
        }

    // Declarations
    this.declaration = JSON.parse(JSON.stringify(declaration));
    this.analyzers   = new Array();
    this.blocks      = new Array();
    this.datasets    = new Array();
    this.connections = new Array();

    this.onComponentNameChanged       = null;
    this.onComponentInputNameChanged  = null;
    this.onComponentOutputNameChanged = null;

    // Process the blocks
    for (var i = 0; i < this.declaration.blocks.length; ++i)
    {
        this.blocks.push(new beat.toolchains.models.Block(this.declaration.blocks[i],
                                                          beat.toolchains.models.PROCESSING_BLOCK));
    }

    // Process the datasets
    for (var i = 0; i < this.declaration.datasets.length; ++i)
    {
        this.datasets.push(new beat.toolchains.models.Block(this.declaration.datasets[i],
                                                            beat.toolchains.models.DATASET_BLOCK));
    }

    // Process the analyzers
    for (var i = 0; i < this.declaration.analyzers.length; ++i)
    {
        this.analyzers.push(new beat.toolchains.models.Block(this.declaration.analyzers[i],
                                                             beat.toolchains.models.ANALYZER_BLOCK));
    }

    // Process the connections
    for (var i = 0; i < this.declaration.connections.length; ++i)
    {
        var connection_declaration = this.declaration.connections[i];

        var from         = connection_declaration.from.split('.');
        var output_block = this.component(from[0]);
        var output       = output_block.output(from[1]);

        var to           = connection_declaration.to.split('.');
        var input_block  = this.component(to[0]);
        var input        = input_block.input(to[1]);

        var connection = new beat.toolchains.models.Connection(output, input, connection_declaration.channel);
        this.connections.push(connection);
    }
}


//----------------------------------------------------------------------------------------
// Returns the number of datasets
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.nbDatasets = function()
{
    return this.declaration.datasets.length;
}


//----------------------------------------------------------------------------------------
// Returns the number of blocks
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.nbBlocks = function()
{
    return this.declaration.blocks.length;
}


//----------------------------------------------------------------------------------------
// Returns the number of analyzers
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.nbAnalyzers = function()
{
    return this.declaration.analyzers.length;
}


//----------------------------------------------------------------------------------------
// Returns the number of components
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.nbComponents = function()
{
    return this.declaration.datasets.length + this.declaration.blocks.length +
           this.declaration.analyzers.length;
}


//----------------------------------------------------------------------------------------
// Retrieve a list of all the components of the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.components = function()
{
    return [].concat(this.datasets, this.blocks, this.analyzers);
}


//----------------------------------------------------------------------------------------
// Retrieve a dataset by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.dataset = function(name)
{
    var result = this.datasets.filter(function(x) { return (x.name == name); });

    if (result.length == 1)
        return result[0];

    return null;
}


//----------------------------------------------------------------------------------------
// Add a new dataset
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.addDataset = function(dataset, declaration)
{
    if (this.dataset(dataset.name))
        return false;

    this.datasets.push(dataset);
    this.declaration.datasets.push(declaration);

    return true;
}


//----------------------------------------------------------------------------------------
// Retrieve a block by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.block = function(name)
{
    var result = this.blocks.filter(function(x) { return (x.name == name); });

    if (result.length == 1)
        return result[0];

    return null;
}


//----------------------------------------------------------------------------------------
// Add a new block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.addBlock = function(block, declaration)
{
    if (this.block(block.name))
        return false;

    this.blocks.push(block);
    this.declaration.blocks.push(declaration);

    return true;
}


//----------------------------------------------------------------------------------------
// Retrieve an analyzer by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.analyzer = function(name)
{
    var result = this.analyzers.filter(function(x) { return (x.name == name); });

    if (result.length == 1)
        return result[0];

    return null;
}


//----------------------------------------------------------------------------------------
// Add a new analyzer
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.addAnalyzer = function(analyzer, declaration)
{
    if (this.analyzer(analyzer.name))
        return false;

    this.analyzers.push(analyzer);
    this.declaration.analyzers.push(declaration);

    return true;
}


//----------------------------------------------------------------------------------------
// Retrieve a component by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.component = function(name)
{
    var component = this.dataset(name);

    if (component === null)
        component = this.block(name);

    if (component === null)
        component = this.analyzer(name);

    return component;
}


//----------------------------------------------------------------------------------------
// Delete a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.deleteComponent = function(component_name)
{
    var component = this.component(component_name);
    if (!component)
        return false;

    for (var i = 0; i < component.nbInputs(); i++)
    {
        var input = component.inputs[i];

        while (input.connections.length > 0)
            this.deleteConnection(input.connections[0]);
    }

    for (var i = 0; i < component.nbOutputs(); i++)
    {
        var output = component.outputs[i];

        while (output.connections.length > 0)
            this.deleteConnection(output.connections[0]);
    }

    var component_declaration = this.componentDeclaration(component_name);

    if (component.isProcessingBlock())
    {
        this.blocks.splice(this.blocks.indexOf(component), 1);
        this.declaration.blocks.splice(this.declaration.blocks.indexOf(component_declaration), 1);
    }
    else if (component.isDataset())
    {
        this.datasets.splice(this.datasets.indexOf(component), 1);
        this.declaration.datasets.splice(this.declaration.datasets.indexOf(component_declaration), 1);
    }
    else
    {
        this.analyzers.splice(this.analyzers.indexOf(component), 1);
        this.declaration.analyzers.splice(this.declaration.analyzers.indexOf(component_declaration), 1);
    }

    return true;
}


//----------------------------------------------------------------------------------------
// Add a new input to a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.addInputToComponent = function(
                                                                component_name, input_name)
{
    var component = this.component(component_name);
    if (!component)
        return false;

    if (component.input(input_name))
        return false;

    component.inputs.push(new beat.toolchains.models.EndPoint(component, input_name,
                                                              component.inputs.length, false));

    this.componentDeclaration(component_name).inputs.push(input_name);

    return true;
}


//----------------------------------------------------------------------------------------
// Remove an input from a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.deleteInputFromComponent = function(
                                                                component_name, input_name)
{
    var component = this.component(component_name);
    if (!component)
        return false;

    if (component.nbInputs() == 0)
        return false;

    var declaration = this.componentDeclaration(component_name);

    var diff = 0;

    for (var i = 0; i < component.inputs.length; i++)
    {
        var input = component.inputs[i];

        if (input.name == input_name)
        {
            while (input.connections.length > 0)
                this.deleteConnection(input.connections[0]);

            component.inputs.splice(i, 1);
            declaration.inputs.splice(i, 1);

            ++diff;
            --i;
        }
        else
        {
            input.index -= diff;
        }
    }

    return false;
}


//----------------------------------------------------------------------------------------
// Add a new output to a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.addOutputToComponent = function(
                                                                component_name, output_name)
{
    var component = this.component(component_name);
    if (!component)
        return false;

    if (component.output(output_name))
        return false;

    component.outputs.push(new beat.toolchains.models.EndPoint(component, output_name,
                                                               component.outputs.length, true));

    this.componentDeclaration(component_name).outputs.push(output_name);

    return true;
}


//----------------------------------------------------------------------------------------
// Remove an output from a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.deleteOutputFromComponent = function(
                                                                component_name, output_name)
{
    var component = this.component(component_name);
    if (!component)
        return false;

    if (component.nbOutputs() == 0)
        return false;

    var declaration = this.componentDeclaration(component_name);

    var diff = 0;

    for (var i = 0; i < component.outputs.length; i++)
    {
        var output = component.outputs[i];

        if (output.name == output_name)
        {
            while (output.connections.length > 0)
                this.deleteConnection(output.connections[0]);

            component.outputs.splice(i, 1);
            declaration.outputs.splice(i, 1);

            ++diff;
            --i;
        }
        else
        {
            output.index -= diff;
        }
    }

    return false;
}


//----------------------------------------------------------------------------------------
// Change the synchronization channel of a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.setSynchronizationChannelOfComponent = function(
                                                                component_name, channel)
{
    var component = this.component(component_name);
    if (!component)
        return false;

    component.channel = channel;
    this.componentDeclaration(component_name).synchronized_channel = channel;

    if (component.outputs && component.outputs.length > 0)
    {
        for (var i = 0; i < component.outputs.length; ++i)
        {
            var output = component.outputs[i];

            if (!output.channel && (output.connections.length > 0))
            {
                output.channel = channel;

                for (var j = 0; j < output.connections.length; ++j)
                    this.setSynchronizationChannelOfConnection(output.connections[j], channel);
            }
        }
    }

    return true;
}


//----------------------------------------------------------------------------------------
// Change the synchronization channel of a connection
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.setSynchronizationChannelOfConnection = function(
                                                                connection, channel)
{
    var from = connection.output.block.name + '.' + connection.output.name;
    var to   = connection.input.block.name + '.' + connection.input.name;

    var declaration = null;
    for (var i = 0; i < this.declaration.connections.length; ++i)
    {
        if ((this.declaration.connections[i].from == from) &&
            (this.declaration.connections[i].to == to))
        {
            declaration = this.declaration.connections[i];
            break;
        }
    }

    if (!declaration)
        return false;


    var previous_channel = connection.output.channel;
    var blocks_to_update = [];

    connection.output.channel = channel;
    declaration.channel       = channel;

    for (var i = 0; i < connection.output.connections.length; ++i)
    {
        connection.output.connections[i].input.channel = channel;

        var affected_block = connection.output.connections[i].input.block;
        if (affected_block.channel == null)
        {
            blocks_to_update.push(affected_block);
        }
        else if (affected_block.channel == previous_channel)
        {
            var must_reaffect = (affected_block.inputs.filter(function(x) { return (x.channel == previous_channel); }).length == 0);
            if (must_reaffect)
                blocks_to_update.push(affected_block);
        }
    }


    blocks_to_update = blocks_to_update.unique();

    for (var i = 0; i < blocks_to_update.length; ++i)
    {
        if (blocks_to_update[i].channel)
            this._checkComponent(blocks_to_update[i], channel);
        else
            this.setSynchronizationChannelOfComponent(blocks_to_update[i].name, channel);
    }

    return true;
}


//----------------------------------------------------------------------------------------
// Change the synchronization channel of a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype._checkComponent = function(component, channel)
{
    var valid_channels = component.inputs.map(function(x) { return x.channel; }).unique()
                                         .filter(function(x) { return (x !== null); });

    if (valid_channels.indexOf(component.channel) == -1)
    {
        if (!channel && (valid_channels.length > 0))
            channel = valid_channels[0];

        component.channel = channel;
        this.componentDeclaration(component.name).synchronized_channel = component.channel;
    }

    var blocks_to_update = [];

    if (component.outputs && component.outputs.length > 0)
    {
        for (var i = 0; i < component.outputs.length; ++i)
        {
            var output = component.outputs[i];

            if ((!output.channel || (valid_channels.indexOf(output.channel) == -1)) && (output.connections.length > 0))
            {
                output.channel = component.channel;

                for (var j = 0; j < output.connections.length; ++j)
                {
                    output.connections[j].input.channel = component.channel;
                    blocks_to_update.push(output.connections[j].input.block);
                }
            }
        }
    }

    blocks_to_update = blocks_to_update.unique();

    for (var i = 0; i < blocks_to_update.length; ++i)
    {
        if (blocks_to_update[i].channel)
            this._checkComponent(blocks_to_update[i], component.channel);
        else
            this.setSynchronizationChannelOfComponent(blocks_to_update[i].name, component.channel);
    }

    return true;
}


//----------------------------------------------------------------------------------------
// Add a new block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.addConnection = function(connection)
{
    var connection_declaration = {
        from:    connection.output.block.name + '.' + connection.output.name,
        to:      connection.input.block.name + '.' + connection.input.name,
        channel: connection.output.channel,
    };

    this.declaration.connections.push(connection_declaration);
    this.connections.push(connection);

    if (!connection.input.block.channel && connection.output.channel)
        this.setSynchronizationChannelOfComponent(connection.input.block.name, connection.output.channel);
}


//----------------------------------------------------------------------------------------
// Remove a connection
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.deleteConnection = function(connection)
{
    var from = connection.output.block.name + '.' + connection.output.name;
    var to   = connection.input.block.name + '.' + connection.input.name;

    for (var i = 0; i < this.declaration.connections.length; ++i)
    {
        if ((this.declaration.connections[i].from == from) &&
            (this.declaration.connections[i].to == to))
        {
            this.declaration.connections.splice(i, 1);
            break;
        }
    }

    for (var i = 0; i < this.connections.length; ++i)
    {
        if (this.connections[i] == connection)
        {
            this.connections.splice(i, 1);
            break;
        }
    }

    var block_to_update = connection.input.block;
    connection._destroy();

    this._checkComponent(block_to_update, null);
}


//----------------------------------------------------------------------------------------
// Retrieve the declaration of a dataset
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.datasetDeclaration = function(name)
{
    var result = this.declaration.datasets.filter(function(x) { return (x.name == name); });

    if (result.length == 1)
        return result[0];

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieve the declaration of a block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.blockDeclaration = function(name)
{
    var result = this.declaration.blocks.filter(function(x) { return (x.name == name); });

    if (result.length == 1)
        return result[0];

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieve the declaration of an analyzer
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.analyzerDeclaration = function(name)
{
    var result = this.declaration.analyzers.filter(function(x) { return (x.name == name); });

    if (result.length == 1)
        return result[0];

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieve the declaration of a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.componentDeclaration = function(name)
{
    var declaration = this.datasetDeclaration(name);

    if (declaration === null)
        declaration = this.blockDeclaration(name);

    if (declaration === null)
        declaration = this.analyzerDeclaration(name);

    return declaration;
}


//----------------------------------------------------------------------------------------
// Retrieve the declaration of a connection
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.connectionDeclaration = function(to)
{
    for (var i = 0; i < this.declaration.connections.length; ++i)
    {
        var declaration = this.declaration.connections[i];

        if (declaration.to == to)
            return declaration;
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Rename a component of the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.renameComponent = function(current_name, new_name)
{
    if (current_name == new_name)
        return;

    // Modify the in-memory representation
    var component = this.component(current_name);
    if (component == null)
        return;

    component.name = new_name;

    // Modify the declaration
    var declaration = null;

    if (component.isDataset())
    {
        declaration = this.datasetDeclaration(current_name);

        for (var i = 0; i < this.blocks.length; ++i)
        {
            var block = this.blocks[i];
            if (block.channel == current_name)
                block.channel = new_name;
        }

        for (var i = 0; i < this.declaration.blocks.length; ++i)
        {
            if (this.declaration.blocks[i].synchronized_channel == current_name)
                this.declaration.blocks[i].synchronized_channel = new_name;
        }

        for (var i = 0; i < this.analyzers.length; ++i)
        {
            var analyzer = this.analyzers[i];
            if (analyzer.channel == current_name)
                analyzer.channel = new_name;
        }

        for (var i = 0; i < this.declaration.analyzers.length; ++i)
        {
            if (this.declaration.analyzers[i].synchronized_channel == current_name)
                this.declaration.analyzers[i].synchronized_channel = new_name;
        }

        for (var i = 0; i < this.connections.length; ++i)
        {
            var connection = this.connections[i];
            if (connection.input.channel == current_name)
            {
                connection.input.channel = new_name;
                connection.output.channel = new_name;
            }
        }

        for (var i = 0; i < this.declaration.connections.length; ++i)
        {
            if (this.declaration.connections[i].channel == current_name)
                this.declaration.connections[i].channel = new_name;
        }

        if (this.declaration.representation && this.declaration.representation.channel_colors &&
            (this.declaration.representation.channel_colors[current_name] !== undefined))
        {
            this.declaration.representation.channel_colors[new_name] = this.declaration.representation.channel_colors[current_name];
            delete this.declaration.representation.channel_colors[current_name];
        }
    }
    else if (component.isProcessingBlock())
    {
        declaration = this.blockDeclaration(current_name);
    }
    else if (component.isAnalyzer())
    {
        declaration = this.analyzerDeclaration(current_name);
    }

    if (declaration !== null)
        declaration.name = new_name;

    if (this.declaration.representation && this.declaration.representation.blocks &&
        (this.declaration.representation.blocks[current_name] !== undefined))
    {
        this.declaration.representation.blocks[new_name] = this.declaration.representation.blocks[current_name];
        delete this.declaration.representation.blocks[current_name];
    }

    for (var i = 0; i < this.declaration.connections.length; ++i)
    {
        var from = this.declaration.connections[i].from.split('.');
        var to = this.declaration.connections[i].to.split('.');

        var representation_name = this.declaration.connections[i].from + '/' + this.declaration.connections[i].to;

        if (from[0] == current_name)
        {
            this.declaration.connections[i].from = new_name + '.' + from[1];

            if (this.declaration.representation && this.declaration.representation.connections &&
                (this.declaration.representation.connections[representation_name] !== undefined))
            {
                this.declaration.representation.connections[new_name + '.' + from[1] + '/' + this.declaration.connections[i].to] = this.declaration.representation.connections[representation_name];
                delete this.declaration.representation.connections[representation_name];
            }
        }

        if (to[0] == current_name)
        {
            this.declaration.connections[i].to = new_name + '.' + to[1];

            if (this.declaration.representation && this.declaration.representation.connections &&
                (this.declaration.representation.connections[representation_name] !== undefined))
            {
                this.declaration.representation.connections[this.declaration.connections[i].from + '/' + new_name + '.' + to[1]] = this.declaration.representation.connections[representation_name];
                delete this.declaration.representation.connections[representation_name];
            }
        }
    }

    // Invoke the callback
    if (this.onComponentNameChanged)
        this.onComponentNameChanged(current_name, new_name)
}


//----------------------------------------------------------------------------------------
// Rename an input of a component of the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.renameComponentInput = function(
                                                    component_name, current_name, new_name)
{
    if (current_name == new_name)
        return;

    // Modify the in-memory representation
    var component = this.component(component_name);
    if (component == null)
        return;

    var input = component.input(current_name);
    if (input == null)
        return;

    input.name = new_name;

    // Modify the declaration
    var declaration = null;

    if (component.isProcessingBlock())
        declaration = this.blockDeclaration(component_name);
    else if (component.isAnalyzer())
        declaration = this.analyzerDeclaration(component_name);

    if (declaration !== null)
    {
        var index = declaration.inputs.indexOf(current_name);
        declaration.inputs[index] = new_name;
    }

    for (var i = 0; i < this.declaration.connections.length; ++i)
    {
        if (this.declaration.connections[i].to == component_name + '.' + current_name)
        {
            var representation_name = this.declaration.connections[i].from + '/' + this.declaration.connections[i].to;

            this.declaration.connections[i].to = component_name + '.' + new_name;

            if (this.declaration.representation && this.declaration.representation.connections &&
                (this.declaration.representation.connections[representation_name] !== undefined))
            {
                this.declaration.representation.connections[this.declaration.connections[i].from + '/' + component_name + '.' + new_name] = this.declaration.representation.connections[representation_name];
                delete this.declaration.representation.connections[representation_name];
            }
        }
    }

    // Invoke the callback
    if (this.onComponentInputNameChanged)
        this.onComponentInputNameChanged(component_name, current_name, new_name)
}


//----------------------------------------------------------------------------------------
// Rename an output of a component of the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.models.Toolchain.prototype.renameComponentOutput = function(
                                                    component_name, current_name, new_name)
{
    if (current_name == new_name)
        return;

    // Modify the in-memory representation
    var component = this.block(component_name);
    if (component == null)
        return;

    var output = component.output(current_name);
    if (output == null)
        return;

    output.name = new_name;

    // Modify the declaration
    var declaration = null;

    if (component.isDataset())
        declaration = this.datasetDeclaration(component_name);
    else if (component.isProcessingBlock())
        declaration = this.blockDeclaration(component_name);

    if (declaration !== null)
    {
        var index = declaration.outputs.indexOf(current_name);
        declaration.outputs[index] = new_name;
    }

    for (var i = 0; i < this.declaration.connections.length; ++i)
    {
        if (this.declaration.connections[i].from == component_name + '.' + current_name)
        {
            var representation_name = this.declaration.connections[i].from + '/' + this.declaration.connections[i].to;

            this.declaration.connections[i].from = component_name + '.' + new_name;

            if (this.declaration.representation && this.declaration.representation.connections &&
                (this.declaration.representation.connections[representation_name] !== undefined))
            {
                this.declaration.representation.connections[component_name + '.' + new_name + '/' + this.declaration.connections[i].to] = this.declaration.representation.connections[representation_name];
                delete this.declaration.representation.connections[representation_name];
            }
        }
    }

    // Invoke the callback
    if (this.onComponentOutputNameChanged)
        this.onComponentOutputNameChanged(component_name, current_name, new_name)
}



/********************************* CLASS: Configuration *********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Callback signatures:
//   - onAlgorithmChanged(block_name, algorithm_name)
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration = function(declaration)
{
    // Check the parameters
    if (declaration === undefined)
        declaration = {
            datasets:     {},
            blocks:       {},
            analyzers:    {},
            globals:      {},
        }

    // Declarations
    this.declaration = JSON.parse(JSON.stringify(declaration));

    if (this.declaration.globals === undefined)
        this.declaration.globals = {};

    this.onAlgorithmChanged          = null;
    this.onAlgorithmGlobalChanged    = {};
    this.onDefaultEnvironmentChanged = [];
    this.onDefaultQueueChanged       = [];
}


//----------------------------------------------------------------------------------------
// Retrieves the number of blocks configured
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.nbBlocksConfigured = function()
{
    return Object.keys(this.declaration.blocks).length;
}


//----------------------------------------------------------------------------------------
// Retrieves the number of datasets configured
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.nbDatasetsConfigured = function()
{
    return Object.keys(this.declaration.datasets).length;
}


//----------------------------------------------------------------------------------------
// Retrieves the number of datasets configured
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.nbAnalyzersConfigured = function()
{
    return Object.keys(this.declaration.analyzers).length;
}


//----------------------------------------------------------------------------------------
// Retrieves the name of all the algorithms used in the configuration
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.algorithms = function()
{
    function _getAlgorithmNames(components)
    {
        var algorithms = [];

        var component_names = Object.keys(components);
        for (var i = 0; i < component_names.length; ++i)
        {
            var component_configuration = components[component_names[i]];

            if (component_configuration.algorithm)
                algorithms.push(component_configuration.algorithm);
        }

        return algorithms.unique();
    }

    return [].concat(_getAlgorithmNames(this.declaration.blocks),
                     _getAlgorithmNames(this.declaration.analyzers));
}


//----------------------------------------------------------------------------------------
// Retrieves the global parameters for a specific algorithm
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.algorithmGlobals = function(algorithm_name)
{
    var globals = this.declaration.globals[algorithm_name];
    if (globals === undefined)
        globals = {};

    return globals;
}


//----------------------------------------------------------------------------------------
// Set the value of a global parameter
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setAlgorithmGlobal = function(algorithm_name, parameter, value)
{
    if (this.declaration.globals[algorithm_name] === undefined)
        this.declaration.globals[algorithm_name] = {};

    var previous_value = this.declaration.globals[algorithm_name][parameter];

    this.declaration.globals[algorithm_name][parameter] = value;

    var name = algorithm_name + '.' + parameter;
    var keys = [];
    if (this.onAlgorithmGlobalChanged[name] !== undefined)
        keys = Object.keys(this.onAlgorithmGlobalChanged[name]);

    for (var i = 0; i < keys.length; ++i)
        this.onAlgorithmGlobalChanged[name][keys[i]](previous_value, value);
}


//----------------------------------------------------------------------------------------
// Retrieves the name of all the datasets
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.datasetNames = function()
{
    return Object.keys(this.declaration.datasets);
}


//----------------------------------------------------------------------------------------
// Retrieves the name of all the blocks
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.blockNames = function()
{
    return Object.keys(this.declaration.blocks);
}


//----------------------------------------------------------------------------------------
// Retrieves the declaration of the configuration of a block by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.block = function(block_name)
{
    var block_configuration = this.declaration.blocks[block_name];
    if (block_configuration === undefined)
        return null;

    return block_configuration;
}


//----------------------------------------------------------------------------------------
// Retrieves the name of all the analyzers
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.analyzerNames = function()
{
    return Object.keys(this.declaration.analyzers);
}


//----------------------------------------------------------------------------------------
// Retrieves the declaration of the configuration of an analyzer by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.analyzer = function(analyzer_name)
{
    var analyzer_configuration = this.declaration.analyzers[analyzer_name];
    if (analyzer_configuration === undefined)
        return null;

    return analyzer_configuration;
}


//----------------------------------------------------------------------------------------
// Retrieves the declaration of the configuration of a component by name
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.component = function(component_name)
{
    var component_configuration = this.block(component_name);

    if (component_configuration === null)
        component_configuration = this.analyzer(component_name);

    if (component_configuration === null)
        component_configuration = this.dataset(component_name);

    return component_configuration;
}


//----------------------------------------------------------------------------------------
// Retrieves the name of the algorithm for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.componentAlgorithm = function(component_name)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return null;

    return configuration.algorithm;
}


//----------------------------------------------------------------------------------------
// Retrieves the resolved parameters for a specific component (global + local parameters)
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.componentParameters = function(component_name)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return null;

    var global_parameters = this.algorithmGlobals(configuration.algorithm);

    var local_parameters = configuration.parameters;
    if (local_parameters === undefined)
        local_parameters = {}

    var parameters = {};

    var names = Object.keys(global_parameters);
    for (var i = 0; i < names.length; ++i)
        parameters[names[i]] = global_parameters[names[i]];

    names = Object.keys(local_parameters);
    for (var i = 0; i < names.length; ++i)
        parameters[names[i]] = local_parameters[names[i]];

    return parameters;
}


//----------------------------------------------------------------------------------------
// Indicates if some parameters were set locally for a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.isParametrized = function(component_name)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return false;

    return (configuration.parameters !== undefined) &&
           (Object.keys(configuration.parameters).length > 0);
}


//----------------------------------------------------------------------------------------
// Set a local parameter for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setComponentParameter = function(
                                                        component_name, parameter_name, value)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return false;

    var global_parameters = this.algorithmGlobals(configuration.algorithm);

    if (configuration.parameters === undefined)
        configuration.parameters = {}

    if ((global_parameters[parameter_name] !== undefined) && (global_parameters[parameter_name] == value))
    {
        delete configuration.parameters[parameter_name];
        return false;
    }
    else
    {
        configuration.parameters[parameter_name] = value;
        return true;
    }
}


//----------------------------------------------------------------------------------------
// Set a local parameter for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.resetComponentParameters = function(component_name)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return null;

    configuration.parameters = undefined;
}


//----------------------------------------------------------------------------------------
// Retrieves the name of the algorithm input mapped to a specific component input
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.algorithmInputName = function(
                                                            component_name, input_name)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return null;

    var names = Object.keys(configuration.inputs);
    for (var i = 0; i < names.length; ++i)
    {
        if (configuration.inputs[names[i]] == input_name)
            return names[i];
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieves the name of the algorithm output mapped to a specific component output
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.algorithmOutputName = function(
                                                            component_name, output_name)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return null;

    var names = Object.keys(configuration.outputs);
    for (var i = 0; i < names.length; ++i)
    {
        if (configuration.outputs[names[i]] == output_name)
            return names[i];
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieves the details about a dataset
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.dataset = function(block_name)
{
    var details = this.declaration.datasets[block_name];
    if (details === undefined)
        return null;

    return details;
}


//----------------------------------------------------------------------------------------
// Rename a block of the configuration
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.renameBlock = function(current_name, new_name)
{
    var block_configuration = this.declaration.blocks[current_name];
    if (block_configuration === undefined)
        return false;

    this.declaration.blocks[new_name] = block_configuration;
    delete this.declaration.blocks[current_name];

    return true;
}


//----------------------------------------------------------------------------------------
// Change the algorithm of a block of the configuration
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setComponentAlgorithm = function(
                                                                component, algorithm_name)
{
    var configuration = this.component(component.name);
    if (configuration === null)
    {
        if (component.isAnalyzer())
        {
            this.declaration.analyzers[component.name] = {
                algorithm: algorithm_name,
                inputs: {},
                parameters: {},
            };
        }
        else
        {
            this.declaration.blocks[component.name] = {
                algorithm: algorithm_name,
                inputs: {},
                outputs: {},
                parameters: {},
            };
        }
    }
    else
    {
        var previous_algorithm_name = configuration.algorithm;

        configuration.algorithm = algorithm_name;
        configuration.inputs = {};

        if (component.isProcessingBlock())
            configuration.outputs = {};

        this._removeUnusedGlobals(previous_algorithm_name, component.isAnalyzer());
    }

    // Invoke the callback
    if (this.onAlgorithmChanged)
        this.onAlgorithmChanged(component.name, algorithm_name)
}


//----------------------------------------------------------------------------------------
// Change the block input corresponding to an algorithm input of a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setComponentInputMapping = function(
                                                    component_name, algorithm_input, component_input)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return;

    configuration.inputs[algorithm_input] = component_input;
}


//----------------------------------------------------------------------------------------
// Change the block output corresponding to an algorithm output of a component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setComponentOutputMapping = function(
                                                    component_name, algorithm_output, component_output)
{
    var configuration = this.component(component_name);
    if (configuration === null)
        return;

    configuration.outputs[algorithm_output] = component_output;
}


//----------------------------------------------------------------------------------------
// Automatically complete the mapping between the endpoints of the component and those of
// the algorithm
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.completeComponentMapping = function(
                                                    component, algorithm, dataformats, smart_mapping)
{
    var configuration = this.component(component.name);
    if (configuration === null)
        return;

    var ambiguous = false;

    smart_mapping.startIteration();
    while (!smart_mapping.isIterationDone())
    {
        var infos = smart_mapping.next();

        var component_inputs = component.inputsOfChannel(infos.channel);
        var component_outputs = component.outputsOfChannel(infos.channel);

        var algorithm_inputs = algorithm.inputsOfChannel(infos.group);
        var algorithm_outputs = algorithm.outputsOfChannel(infos.group);

        ambiguous = this._autoAssignAlgorithmEndpoints(component_inputs, algorithm_inputs,
                                                       configuration.inputs, false,
                                                       dataformats) || ambiguous;

        for (var i = 0; i < algorithm_inputs.length; ++i)
        {
            var component_input_name = configuration.inputs[algorithm_inputs[i].name];

            for (var j = 0; j < component_inputs.length; ++j)
            {
                var input = component_inputs[j];

                if (input.name == component_input_name)
                {
                    input.dataformat = algorithm_inputs[i].dataformat;
                    break;
                }
            }
        }

        if (component.nbOutputs() > 0)
        {
            ambiguous = this._autoAssignAlgorithmEndpoints(component_outputs, algorithm_outputs,
                                                           configuration.outputs, true,
                                                           dataformats) || ambiguous;

            for (var i = 0; i < algorithm_outputs.length; ++i)
            {
                var component_output_name = configuration.outputs[algorithm_outputs[i].name];

                for (var j = 0; j < component_outputs.length; ++j)
                {
                    var output = component_outputs[j];

                    if (output.name == component_output_name)
                    {
                        output.dataformat = algorithm_outputs[i].dataformat;
                        break;
                    }
                }
            }
        }
    }

    return ambiguous;
}


//----------------------------------------------------------------------------------------
// Reset the configuration of a block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.resetBlock = function(block_name)
{
    var previous_algorithm_name = this.declaration.blocks[block_name].algorithm;

    delete this.declaration.blocks[block_name];

    if (previous_algorithm_name)
        this._removeUnusedGlobals(previous_algorithm_name, false);
}


//----------------------------------------------------------------------------------------
// Reset the configuration of a block
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.resetAnalyzer = function(block_name)
{
    var previous_algorithm_name = this.declaration.analyzers[block_name].algorithm;

    delete this.declaration.analyzers[block_name];

    if (previous_algorithm_name)
        this._removeUnusedGlobals(previous_algorithm_name, true);
}


//----------------------------------------------------------------------------------------
// Reset the configuration of a dataset
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.resetDataset = function(dataset_name)
{
    delete this.declaration.datasets[dataset_name];
}


//----------------------------------------------------------------------------------------
// Change the algorithm of a block of the configuration
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setDatasetDetails = function(
                                                                block_name, database_name,
                                                                protocol_name, dataset_name
                                                            )
{
    var block_details = this.declaration.datasets[block_name];
    if (block_details === undefined)
    {
        this.declaration.datasets[block_name] = {
            database: database_name,
            protocol: protocol_name,
            set:      dataset_name,
        };
    }
    else
    {
        block_details.database = database_name;
        block_details.protocol = protocol_name;
        block_details.set      = dataset_name;
    }
}


//----------------------------------------------------------------------------------------
// Retrieves the default environment
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.defaultEnvironment = function()
{
    if (this.declaration.globals.environment === undefined)
        return null;

    return this.declaration.globals.environment;
}


//----------------------------------------------------------------------------------------
// Retrieves the default queue
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.defaultQueue = function()
{
    if (this.declaration.globals.queue === undefined)
        return null;

    return this.declaration.globals.queue;
}


//----------------------------------------------------------------------------------------
// Set the default environment
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setDefaultEnvironment = function(environment)
{
    var previous_value = this.declaration.globals.environment;

    this.declaration.globals.environment = environment;

    var keys = Object.keys(this.onDefaultEnvironmentChanged);
    for (var i = 0; i < keys.length; ++i)
        this.onDefaultEnvironmentChanged[keys[i]](previous_value, environment);
}


//----------------------------------------------------------------------------------------
// Set the default queue
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setDefaultQueue = function(queue)
{
    var previous_value = this.declaration.globals.queue;

    this.declaration.globals.queue = queue;

    var keys = Object.keys(this.onDefaultQueueChanged);
    for (var i = 0; i < keys.length; ++i)
        this.onDefaultQueueChanged[keys[i]](previous_value, queue);
}


//----------------------------------------------------------------------------------------
// Retrieves the environment for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.blockEnvironment = function(block_name)
{
    var configuration = this.component(block_name);

    if ((configuration === undefined) || (configuration.environment == null))
        return this.defaultEnvironment();

    return configuration.environment;
}


//----------------------------------------------------------------------------------------
// Retrieves the queue for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.blockQueue = function(block_name)
{
    var configuration = this.component(block_name);

    if ((configuration === undefined) || (configuration.queue == null))
        return this.defaultQueue();

    return configuration.queue;
}


//----------------------------------------------------------------------------------------
// Retrieves the number of slots for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.blockNbSlots = function(block_name)
{
    var configuration = this.component(block_name);

    if ((configuration === undefined) || (configuration.nb_slots == null))
        return 1;

    return configuration.nb_slots;
}


//----------------------------------------------------------------------------------------
// Set the environment for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setBlockEnvironment = function(block_name, environment)
{
    var configuration = this.component(block_name);
    if (configuration === undefined || environment === null)
        return;

    configuration.environment = {
      name: environment['name'],
      version: environment['version'],
    };
}


//----------------------------------------------------------------------------------------
// Set the queue for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setBlockQueue = function(block_name, queue)
{
    var configuration = this.component(block_name);
    if (configuration === undefined)
        return;

    configuration.queue = queue;
}


//----------------------------------------------------------------------------------------
// Set the number of slots for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.setBlockNbSlots = function(block_name, nb_slots)
{
    var configuration = this.component(block_name);
    if (configuration === undefined)
        return;

    configuration.nb_slots = nb_slots;
}


//----------------------------------------------------------------------------------------
// Reset the environment, queue and number of slots for a specific component
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.resetBlockEnvironmentAndQueue = function(block_name)
{
    var configuration = this.component(block_name);
    if (configuration === undefined)
        return;

    if (configuration.environment !== undefined)
        delete configuration.environment;

    if (configuration.queue !== undefined)
        delete configuration.queue;

    if (configuration.nb_slots !== undefined)
        delete configuration.nb_slots;
}


//----------------------------------------------------------------------------------------
// Add a listener function for changes in the value of a global parameter
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.addGlobalParameterListener = function(
                                                        algorithm_name, parameter, unique_id, listener)
{
    var name = algorithm_name + '.' + parameter;

    if (this.onAlgorithmGlobalChanged[name] === undefined)
        this.onAlgorithmGlobalChanged[name] = {};

    this.onAlgorithmGlobalChanged[name][unique_id] = listener;
}


//----------------------------------------------------------------------------------------
// Remove a listener function for changes in the value of a global parameter
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.removeGlobalParameterListener = function(
                                                        algorithm_name, parameter, unique_id)
{
    var name = algorithm_name + '.' + parameter;

    if (this.onAlgorithmGlobalChanged[name] === undefined)
        return;

    delete this.onAlgorithmGlobalChanged[name][unique_id];

    if (this.onAlgorithmGlobalChanged[name].length == 0)
        delete this.onAlgorithmGlobalChanged[name];
}


//----------------------------------------------------------------------------------------
// Add a listener function for changes in the value of the default environment
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.addDefaultEnvironmentListener = function(
                                                                        unique_id, listener)
{
    this.onDefaultEnvironmentChanged[unique_id] = listener;
}


//----------------------------------------------------------------------------------------
// Remove a listener function for changes in the value of the default environment
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.removeDefaultEnvironmentListener = function(
                                                                                    unique_id)
{
    delete this.onDefaultEnvironmentChanged[unique_id];
}


//----------------------------------------------------------------------------------------
// Add a listener function for changes in the value of the default queue
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.addDefaultQueueListener = function(
                                                                        unique_id, listener)
{
    this.onDefaultQueueChanged[unique_id] = listener;
}


//----------------------------------------------------------------------------------------
// Remove a listener function for changes in the value of the default queue
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.removeDefaultQueueListener = function(unique_id)
{
    delete this.onDefaultQueueChanged[unique_id];
}


//----------------------------------------------------------------------------------------
// Indicates if a component is an analyzer
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype.is_analyzer = function(name)
{
    return this.declaration.analyzers[name] !== undefined;
}


//-----------------------------------------------


beat.toolchains.models.Configuration.prototype._autoAssignAlgorithmEndpoints = function(
                                                            block_endpoints, algorithm_endpoints,
                                                            mapping, is_output, dataformats)
{
    var ambiguous = false;

    // Determine which endpoints must be auto-assigned
    var assigned_block_endpoints = [];
    var assignable_block_endpoints = [];
    var remaining_block_endpoints = [];

    var assigned_algorithm_endpoints = Object.keys(mapping);
    for (var i = 0; i < assigned_algorithm_endpoints.length; ++i)
        assigned_block_endpoints.push(mapping[assigned_algorithm_endpoints[i]]);

    for (var i = 0; i < block_endpoints.length; ++i)
    {
        var block_endpoint = block_endpoints[i];

        // Already assigned
        if (assigned_block_endpoints.indexOf(block_endpoint.name) == 0)
            continue;

        if ((!is_output && (block_endpoint.connections[0].output.dataformat !== null)) ||
            (is_output && (block_endpoint.connections[0].input.dataformat !== null)))
        {
            assignable_block_endpoints.push(block_endpoint);
        }
        else
        {
            remaining_block_endpoints.push(block_endpoint);
        }
    }

    var free_algorithm_endpoints = algorithm_endpoints.filter(function(value) {
            return (assigned_algorithm_endpoints.indexOf(value.name) == -1);
        })
        .map(function(value) {
            return {
                name: value.name,
                dataformat: value.dataformat
            };
        });


    // Iterate through them
    for (var j = 0; j < assignable_block_endpoints.length; ++j)
    {
        var block_endpoint = assignable_block_endpoints[j];

        var distances = free_algorithm_endpoints.map(function(value) {
                var offset = (block_endpoint.name == value.name ? 0 : 1);

                if (is_output)
                {
                    return offset + 2 * dataformats.distance(value.dataformat,
                                                             block_endpoint.connections[0].input.dataformat);
                }
                else
                {
                    return offset + 2 * dataformats.distance(block_endpoint.connections[0].output.dataformat,
                                                             value.dataformat);
                }
            }, this);

        var d = 0;
        while (true)
        {
            var index = distances.indexOf(d);
            if (index != -1)
            {
                mapping[free_algorithm_endpoints[index].name] = block_endpoint.name;
                free_algorithm_endpoints.splice(index, 1);

                if (!ambiguous)
                {
                    distances.splice(index, 1);
                    if (distances.indexOf(d) != -1)
                        ambiguous = true;
                }

                break;
            }

            d += 1;
        }
    }

    // Assign the endpoints where the data format is still free, using fuzzy matching of
    // the names
    while (remaining_block_endpoints.length > 0)
    {
        var best_block_endpoint_index     = 0;
        var best_algorithm_endpoint_index = 0;
        var best_score                    = 0.0;

        for (var i = 0; i < remaining_block_endpoints.length; ++i)
        {
            var block_endpoint = remaining_block_endpoints[i];

            for (var j = 0; j < free_algorithm_endpoints.length; ++j)
            {
                var score = block_endpoint.name.score(free_algorithm_endpoints[j].name, 0.5);

                if (score > best_score)
                {
                    best_block_endpoint_index     = i;
                    best_algorithm_endpoint_index = j;
                    best_score                    = score;
                }
            }
        }

        mapping[free_algorithm_endpoints[best_algorithm_endpoint_index].name] = remaining_block_endpoints[best_block_endpoint_index].name;
        free_algorithm_endpoints.splice(best_algorithm_endpoint_index, 1);
        remaining_block_endpoints.splice(best_block_endpoint_index, 1);
    }

    return ambiguous;
}


//-----------------------------------------------


//----------------------------------------------------------------------------------------
// Remove the globals parameters of a previously used algorithm (if necessary)
//----------------------------------------------------------------------------------------
beat.toolchains.models.Configuration.prototype._removeUnusedGlobals = function(removed_algorithm_name, analyzer)
{
    if (this.declaration.globals[removed_algorithm_name] !== undefined)
    {
        var must_remove = true;
        var components = (analyzer ? this.declaration.analyzers : this.declaration.blocks);

        var component_names = Object.keys(components);
        for (var i = 0; i < component_names.length; ++i)
        {
            if (components[component_names[i]].algorithm == removed_algorithm_name)
            {
                must_remove = false;
                break;
            }
        }

        if (must_remove)
            delete this.declaration.globals[removed_algorithm_name];
    }
}
