/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.toolchains.viewer.js

   Implementation of a panel that allows to view, create and modify a toolchain
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.toolchains === undefined)
    beat.toolchains = {}

if (beat.toolchains.viewer === undefined)
    beat.toolchains.viewer = {}



/******************************** CLASS: ToolchainViewer ********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// editor_id: ID of the DOM element to use
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer = function(viewer_id)
{
    // Initialisations
    this.id             = viewer_id;
    this.element        = $('#' + viewer_id)[0];
    this.toolchain      = null;
    this.configuration  = null;
    this.canvas         = $(this.element).find('canvas.grid')[0];
    this.context        = null;
    this.zoom           = 1.0;
    this.translation    = false;
    this.mouse_position = null;
    this.blocks         = [];
    this.connections    = [];
    this.needsRepaint   = false;
    this.channel_colors = {};
    this.running        = false;
    this.blocks_status  = null;
    this.onClick        = null;
    this.onTransform    = null;

    this.visible_area = {
        top:    0,
        left:   0,
        width:  0,
        height: 0,
    };

    // Check that canvas is supported by the browser
    if (!this.canvas.getContext)
    {
        this.context = null;
        return;
    }

    this.context = this.canvas.getContext("2d");


    var view = this;

    $(window).resize(function() {
        view._onResized();
    });

    this._onResized();


    // Disable the context menu on the left click
    this.canvas.oncontextmenu = function (e) {
        e.preventDefault();
    };


    // Mouse handling
    $(this.canvas).mousedown(function(event) {
        view._onMouseDown(event);
        return false;
    });

    $(this.canvas).mousemove(function(event) {
        view._onMouseMove(event);
        return false;
    });

    $(this.canvas).mouseup(function(event) {
        view._onMouseUp(event);
        return false;
    });

    $(this.canvas).click(function(event) {
        if (view.onClick && !view.translation)
            view.onClick();

        view.translation = false;

        return false;
    });

    $(this.canvas).dblclick(function(event) {
        return false;
    });

    $(this.canvas).mouseenter(function(event) {
        return false;
    });

    $(this.canvas).mouseout(function(event) {
        view._onMouseOut(event);
        return false;
    });

    $(this.canvas).bind('mousewheel DOMMouseScroll', function(event) {
        view._onMouseWheel(event);
        return false;
    });
}


//----------------------------------------------------------------------------------------
// Indicates if the viewer is correctly initialized
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.isInitialized = function()
{
    return (this.context != null);
}


//----------------------------------------------------------------------------------------
// Set the toolchain
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.setToolchain = function(toolchain)
{
    function _updateBoundingBox(bounding_box, representation)
    {
        if (representation instanceof beat.toolchains.common.ConnectionRepresentation)
        {
            if ((bounding_box.top === null) || (representation.bounding_box.top < bounding_box.top))
                bounding_box.top = representation.bounding_box.top;

            if ((bounding_box.left === null) || (representation.bounding_box.left < bounding_box.left))
                bounding_box.left = representation.bounding_box.left;

            if ((bounding_box.bottom === null) || (representation.bounding_box.bottom > bounding_box.bottom))
                bounding_box.bottom = representation.bounding_box.bottom;

            if ((bounding_box.right === null) || (representation.bounding_box.right > bounding_box.right))
                bounding_box.right = representation.bounding_box.right;
        }
        else
        {
            if ((bounding_box.top === null) || (representation.top < bounding_box.top))
                bounding_box.top = representation.top;

            if ((bounding_box.left === null) || (representation.left < bounding_box.left))
                bounding_box.left = representation.left;

            if ((bounding_box.bottom === null) || (representation.top + representation.height > bounding_box.bottom))
                bounding_box.bottom = representation.top + representation.height;

            if ((bounding_box.right === null) || (representation.left + representation.width > bounding_box.right))
                bounding_box.right = representation.left + representation.width;
        }
    }


    this.toolchain = toolchain;

    $(this.element).removeClass('progress');

    var unknown_row = 0;

    var bounding_box = {
        top:    null,
        left:   null,
        bottom: null,
        right:  null,
    };

    if (toolchain.declaration.representation && toolchain.declaration.representation.channel_colors)
        this.channel_colors = toolchain.declaration.representation.channel_colors;


    // Process the datasets
    for (var i = 0; i < toolchain.datasets.length; ++i)
    {
        var dataset = toolchain.datasets[i];

        var representation = new beat.toolchains.common.BlockRepresentation(dataset);
        dataset.representation = representation;

        var infos = (toolchain.declaration.representation ? toolchain.declaration.representation.blocks[dataset.name]
                                                          : null);
        if (infos)
        {
            representation.left   = infos.col;
            representation.top    = infos.row;
            representation.width  = infos.width;
            representation.height = infos.height;
        }
        else
        {
            representation.left   = 0;
            representation.top    = unknown_row;
            unknown_row += beat.toolchains.common.BLOCK_DEFAULT_HEIGHT + 1;
        }

        if (!this.channel_colors[dataset.name])
            this.channel_colors[dataset.name] = this._getUnusedChannelColor();

        this.blocks.push(representation);

        _updateBoundingBox(bounding_box, representation);
    }


    // Process the blocks
    for (var i = 0; i < toolchain.blocks.length; ++i)
    {
        var block = toolchain.blocks[i];

        var representation = new beat.toolchains.common.BlockRepresentation(block);
        block.representation = representation;

        var infos = (toolchain.declaration.representation ? toolchain.declaration.representation.blocks[block.name]
                                                          : null);
        if (infos)
        {
            representation.left   = infos.col;
            representation.top    = infos.row;
            representation.width  = infos.width;
            representation.height = infos.height;
        }
        else
        {
            representation.left   = 0;
            representation.top    = unknown_row;
            unknown_row += beat.toolchains.common.BLOCK_DEFAULT_HEIGHT + 1;
        }

        this.blocks.push(representation);

        _updateBoundingBox(bounding_box, representation);
    }


    // Process the analyzers
    for (var i = 0; i < toolchain.analyzers.length; ++i)
    {
        var analyzer = toolchain.analyzers[i];

        var representation = new beat.toolchains.common.BlockRepresentation(analyzer);
        analyzer.representation = representation;

        var infos = (toolchain.declaration.representation ? toolchain.declaration.representation.blocks[analyzer.name]
                                                          : null);
        if (infos)
        {
            representation.left   = infos.col;
            representation.top    = infos.row;
            representation.width  = infos.width;
            representation.height = infos.height;
        }
        else
        {
            representation.left   = 0;
            representation.top    = unknown_row;
            unknown_row += beat.toolchains.common.BLOCK_DEFAULT_HEIGHT + 1;
        }

        this.blocks.push(representation);

        _updateBoundingBox(bounding_box, representation);
    }


    // Process the connections
    for (var i = 0; i < toolchain.connections.length; ++i)
    {
        var connection = toolchain.connections[i];

        var representation = new beat.toolchains.common.ConnectionRepresentation(connection);

        var infos = (toolchain.declaration.representation ? toolchain.declaration.representation.connections[connection.output.block.name + '.' +
                                                                                                             connection.output.name + '/' +
                                                                                                             connection.input.block.name + '.' +
                                                                                                             connection.input.name]
                                                          : null);
        if (infos)
        {
            var last_row = infos[0].row;

            for (var j = 0; j < infos.length; ++j)
            {
                var point_infos = infos[j];

                var fixed = (j <= 1) || (j >= infos.length - 2);
                var vertical = (point_infos.row != last_row);

                var point = new beat.toolchains.common.Point(true, point_infos.col, point_infos.row, vertical, fixed);
                representation.points.push(point);

                last_row = point_infos.row;
            }

            representation._cleanup();
        }
        else
        {
            var output_block_representation = null;
            var input_block_representation = null;

            for (var j = 0; j < this.blocks.length; ++j)
            {
                if (connection.output.block == this.blocks[j].block)
                {
                    output_block_representation = this.blocks[j];
                    if (input_block_representation)
                        break;
                }

                else if (connection.input.block == this.blocks[j].block)
                {
                    input_block_representation = this.blocks[j];
                    if (output_block_representation)
                        break;
                }
            }

            representation.points.push(new beat.toolchains.common.Point(
                    true, output_block_representation.left + output_block_representation.width,
                    output_block_representation.top + 2 + connection.output.index,
                    false, true
                ));

            representation.points.push(new beat.toolchains.common.Point(
                    true, output_block_representation.left + output_block_representation.width,
                    output_block_representation.top + 2 + connection.output.index,
                    false, true
                ));

            representation.points.push(new beat.toolchains.common.Point(
                    true, output_block_representation.left + output_block_representation.width + 1,
                    output_block_representation.top + 2 + connection.output.index,
                    false, false
                ));

            var middle = (output_block_representation.left + output_block_representation.width +
                          input_block_representation.left) / 2;

            representation.points.push(new beat.toolchains.common.Point(
                    true, middle,
                    output_block_representation.top + 2 + connection.output.index,
                    false, false
                ));

            representation.points.push(new beat.toolchains.common.Point(
                    true, middle,
                    input_block_representation.top + 2 + connection.input.index,
                    true, false
                ));

            representation.points.push(new beat.toolchains.common.Point(
                    true, input_block_representation.left - 1,
                    input_block_representation.top + 2 + connection.input.index,
                    false, true
                ));

            representation.points.push(new beat.toolchains.common.Point(
                    true, input_block_representation.left,
                    input_block_representation.top + 2 + connection.input.index,
                    false, true
                ));

            representation._cleanup();
        }

        connection.representation = representation;
        this.connections.push(representation);

        _updateBoundingBox(bounding_box, representation)
    }


    // Update the visible area to encompass most of the toolchain
    var zoom_x = Math.floor(10 * this.visible_area.width / ((bounding_box.right - bounding_box.left + 2) * beat.toolchains.common.CELL_SPACING)) / 10;
    var zoom_y = Math.floor(10 * this.visible_area.height / ((bounding_box.bottom - bounding_box.top + 2) * beat.toolchains.common.CELL_SPACING)) / 10;

    this.visible_area.top  = beat.toolchains.common.cellsToPixels(bounding_box.top - 1, this.zoom);
    this.visible_area.left = beat.toolchains.common.cellsToPixels(bounding_box.left - 1, this.zoom);
    this.zoom              = Math.max(Math.min(zoom_x, zoom_y, 1.0), 0.5);

    this.refresh();
}


//----------------------------------------------------------------------------------------
// Set the configuration
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.setConfiguration = function(configuration)
{
    this.configuration = configuration;

    if (this.toolchain)
        this.refresh();
}


//----------------------------------------------------------------------------------------
// Modify the transformations of the view
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.setTransforms = function(left, top, zoom)
{
    this.visible_area.left = left;
    this.visible_area.top  = top;
    this.zoom              = zoom;

    this.refresh();
}


//----------------------------------------------------------------------------------------
// Called when the experiment was started
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.onExperimentStarted = function()
{
    if (!this.configuration || this.running)
        return;

    this.running       = true;
    this.blocks_status = null;

    this._refresh();
}


//----------------------------------------------------------------------------------------
// Called when the experiment is done
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.onExperimentDone = function()
{
    this.running       = false;
    this.blocks_status = null;

    this._refresh();
}


//----------------------------------------------------------------------------------------
// Called when some block outputs were generated
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.updateBlocksStatus = function(blocks_status)
{
    this.blocks_status = blocks_status;
    this._refresh();
}


//----------------------------------------------------------------------------------------
// Schedule a repaint of the content of the canvas
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.refresh = function()
{
    if (this.needsRepaint == false)
    {
        var view = this;
        this.needsRepaint = true;

        requestAnimationFrame(function(timestamp) {
            view._refresh();
        });
    }
}


//----------------------------------------------------------------------------------------
// Update the content of the canvas
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._refresh = function()
{
    this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);

    this.context.save();

    this.context.translate(Math.floor(-this.visible_area.left * this.zoom),
                           Math.floor(-this.visible_area.top * this.zoom));


    for (var i = 0; i < this.blocks.length; ++i)
        this._drawBlock(this.blocks[i]);

    for (var i = 0; i < this.connections.length; ++i)
        this._drawConnection(this.connections[i]);

    this.context.restore();

    this.needsRepaint = false;
}


//----------------------------------------------------------------------------------------
// Resize the view
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype.setSize = function(width, height)
{
    this.canvas.style.width  = '' + width + 'px';
    this.canvas.style.height = '' + height + 'px';

    this.canvas.width  = width;
    this.canvas.height = height

    this.visible_area.width  = this.canvas.width / this.zoom;
    this.visible_area.height = this.canvas.height / this.zoom;

    this.refresh();
}


//----------------------------------------------------------------------------------------
// Called when the viewer is resized
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._onResized = function()
{
    var height = $(this.element).height();
    var width = $(this.element).width();

    this.canvas.style.width  = '' + width + 'px';
    this.canvas.style.height = '' + height + 'px';

    this.canvas.width  = width;
    this.canvas.height = height

    this.visible_area.width  = this.canvas.width / this.zoom;
    this.visible_area.height = this.canvas.height / this.zoom;

    this.refresh();
}


//----------------------------------------------------------------------------------------
// Called when a button of the mouse is pressed in the view
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._onMouseDown = function(event)
{
    // Nothing to do if we are already doing some other action
    if (this.translation)
        return;


    // Middle mouse button: prepare the translation of the whole grid
    if ((event.which == 2) || (event.altKey && (event.which == 1)))
    {
        this.translation = true;
        this.mouse_position = this._getMousePosition(event);
        $(this.canvas).addClass('action-translation');
    }
}


//----------------------------------------------------------------------------------------
// Called when the mouse is moved over the view
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._onMouseMove = function(event)
{
    // Retrieve the position of the mouse
    var mouse_position = this._getMousePosition(event);

    if (!this.mouse_position)
        return false;

    // Translation of the view
    if (this.translation)
    {
        var diff_x = mouse_position.x - this.mouse_position.x;
        var diff_y = mouse_position.y - this.mouse_position.y;

        if (diff_x != 0)
            this.visible_area.left -= diff_x / this.zoom;

        if (diff_y != 0)
            this.visible_area.top -= diff_y / this.zoom;

        this.refresh();

        if (this.onTransform)
            this.onTransform(this.visible_area.left, this.visible_area.top, this.zoom);
    }

    this.mouse_position = mouse_position;
}


//----------------------------------------------------------------------------------------
// Called when a button of the mouse is released in the view
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._onMouseUp = function(event)
{
    // Middle mouse button
    if ((event.which == 2) || (event.which == 1))
    {
        // Translation of the view
        if (this.translation)
        {
            this.translation = (event.which == 1);
            $(this.canvas).removeClass('action-translation');
            this.mouse_position = null;
        }
    }
}


//----------------------------------------------------------------------------------------
// Called when the mouse leaves the view
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._onMouseOut = function(event)
{
    // Translation of the view: stop it
    if (this.translation)
    {
        this.translation = false;
        $(this.canvas).removeClass('action-translation');
        this.mouse_position = null;
    }
}


//----------------------------------------------------------------------------------------
// Called when the mouse wheel is scrolled
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._onMouseWheel = function(event)
{
    // Prevent zoom during translation
    if (this.translation)
        return;

    var previous_zoom = this.zoom;
    var must_redraw = false;

    // With Ctrl key: control the zoom
    if (event.ctrlKey)
    {
        // Scroll up
        if (event.originalEvent.wheelDelta > 0 || event.originalEvent.detail < 0)
        {
            if (this.zoom < 1.5)
                this.zoom += 0.1;
        }

        // Scroll down
        else
        {
            if (this.zoom > 0.5)
                this.zoom -= 0.1;
        }

        must_redraw = (previous_zoom != this.zoom);
    }

    // Without modifier key: translate the view
    else
    {
        var mouse_position = this._getMousePosition(event);

        // Mouse over the horizontal scrollbar: translate horizontally
        if (mouse_position.y >= this.canvas.height - beat.toolchains.common.SCROLLBAR_HOLDER_WIDTH)
        {
            // Scroll up
            if (event.originalEvent.wheelDelta > 0 || event.originalEvent.detail < 0)
                this.visible_area.left -= beat.toolchains.common.CELL_SPACING / this.zoom;

            // Scroll down
            else
                this.visible_area.left += beat.toolchains.common.CELL_SPACING / this.zoom;
        }

        // Mouse over the vertical scrollbar: translate vertically
        else
        {
            // Scroll up
            if (event.originalEvent.wheelDelta > 0 || event.originalEvent.detail < 0)
                this.visible_area.top -= beat.toolchains.common.CELL_SPACING / this.zoom;

            // Scroll down
            else
                this.visible_area.top += beat.toolchains.common.CELL_SPACING / this.zoom;
        }

        must_redraw = true;
    }

    // Avoid unnecessary refreshes
    if (must_redraw)
    {
        this.visible_area.width  = this.canvas.width / this.zoom;
        this.visible_area.height = this.canvas.height / this.zoom;
        this.refresh();

        if (this.onTransform)
            this.onTransform(this.visible_area.left, this.visible_area.top, this.zoom);
    }
}


//----------------------------------------------------------------------------------------
// Resize the view
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._getMousePosition = function(event)
{
    var canvas_position = $(this.canvas).offset();

    return {
        x: event.clientX - Math.round(canvas_position.left) - this.canvas.clientLeft,
        y: event.clientY - Math.round(canvas_position.top) - this.canvas.clientTop,
    };
}


//----------------------------------------------------------------------------------------
// Draw a block
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._drawBlock = function(block_representation, creation)
{
    function _drawText(text, context, x, y, max_width, zoom)
    {
        context.textBaseline = 'top';
        context.textAlign    = 'center';

        var name = text;
        var metrics = context.measureText(name);

        var nb = 1;
        while (metrics.width > max_width - 8 * zoom)
        {
            ++nb;

            var offset = text.length >> 1;
            name = text.substr(0, offset - nb) + '...' + text.substr(offset + nb);

            metrics = context.measureText(name);
        }

        context.fillText(name, x, y);
    }


    if (!block_representation)
        return;

    this.context.save();

    // Transformations
    this.context.translate(beat.toolchains.common.cellsToPixels(block_representation.left, this.zoom),
                           beat.toolchains.common.cellsToPixels(block_representation.top, this.zoom));


    // Block box
    var block_status = null;
    if (this.blocks_status)
        block_status = this.blocks_status[block_representation.block.name];

    if (block_representation.block.isDataset())
    {
        this.context.strokeStyle = '#000000';
        this.context.fillStyle   = '#FAF8CC';
    }
    else if (this.running && (block_status === 'generated'))
    {
        this.context.strokeStyle = '#00CC00';
        this.context.fillStyle   = '#44EE44';
    }
    else if (this.running && (block_status === 'processing'))
    {
        this.context.strokeStyle = '#0000CC';
        this.context.fillStyle   = '#9999EE';
    }
    else if (block_status === 'failed')
    {
        this.context.strokeStyle = '#CC0000';
        this.context.fillStyle   = '#FF9999';
    }
    else if (block_representation.block.isAnalyzer())
    {
        this.context.strokeStyle = '#000000';
        this.context.fillStyle   = '#CCCCFA';
    }
    else
    {
        this.context.strokeStyle = '#000000';
        this.context.fillStyle   = '#CCCCCC';
    }

    this.context.lineWidth = 1;
    this.context.lineCap   = 'square';

    var width  = beat.toolchains.common.cellsToPixels(block_representation.width, this.zoom);
    var height = beat.toolchains.common.cellsToPixels(block_representation.height, this.zoom);
    var radius = (block_representation.block.isDataset() ? 18 : 14) * this.zoom;

    this.context.beginPath();

    if (!block_representation.block.isDataset())
    {
        this.context.moveTo(0.5, 0.5);
    }
    else
    {
        this.context.moveTo(0.5, radius + 0.5);
        this.context.arc(radius + 0.5, radius + 0.5, radius, Math.PI, -Math.PI / 2, false);
    }

    this.context.lineTo(width - radius, 0.5);
    this.context.arc(width - radius + 0.5, radius + 0.5, radius, -Math.PI / 2, 0, false);
    this.context.lineTo(width + 0.5, height - radius + 0.5);
    this.context.arc(width - radius + 0.5, height - radius + 0.5, radius, 0, Math.PI / 2, false);

    if (!block_representation.block.isDataset())
    {
        this.context.lineTo(0.5, height + 0.5);
        this.context.lineTo(0.5, 0.5);
    }
    else
    {
        this.context.lineTo(radius + 0.5, height + 0.5);
        this.context.arc(radius + 0.5, height - radius + 0.5, radius, Math.PI / 2, Math.PI, false);
        this.context.lineTo(0.5, radius + 0.5);
    }

    this.context.fill();
    this.context.stroke();


    // Block title
    this.context.font         = 'Bold ' + (beat.toolchains.common.CELL_SPACING / 2 * this.zoom) + 'px Arial';
    this.context.fillStyle    = '#000000';

    _drawText(block_representation.block.name, this.context, Math.floor(width / 2),
              Math.floor(4 * this.zoom), width - 2 * radius - 4, this.zoom);


    // Algorithm name (if applicable)
    if (this.configuration && !block_representation.block.isDataset())
    {
        this.context.font         = 'Italic ' + ((beat.toolchains.common.CELL_SPACING / 2 - 2) * this.zoom) + 'px Arial';
        this.context.fillStyle    = '#000000';

        _drawText(this.configuration.componentAlgorithm(block_representation.block.name),
                  this.context, Math.floor(width / 2),
                  Math.floor((beat.toolchains.common.CELL_SPACING / 2 + 10) * this.zoom),
                  width - 4, this.zoom);
    }

    // Dataset name (if applicable)
    else if (this.configuration && block_representation.block.isDataset())
    {
        this.context.font         = 'Italic ' + ((beat.toolchains.common.CELL_SPACING / 2 - 2) * this.zoom) + 'px Arial';
        this.context.fillStyle    = '#000000';
        this.context.textBaseline = 'top';
        this.context.textAlign    = 'center';

        var dataset_details = this.configuration.declaration.datasets[block_representation.block.name];

        _drawText(dataset_details.database + '.' + dataset_details.protocol,
                  this.context, Math.floor(width / 2),
                  Math.floor((beat.toolchains.common.CELL_SPACING / 2 + 6) * this.zoom),
                  width - 4, this.zoom);

        this.context.font = ((beat.toolchains.common.CELL_SPACING / 2 - 2) * this.zoom) + 'px Arial';

        _drawText(dataset_details.set,
                  this.context, Math.floor(width / 2),
                  Math.floor((beat.toolchains.common.CELL_SPACING + 6) * this.zoom),
                  width - 4, this.zoom);
    }


    // Inputs
    var max_connector_width = beat.toolchains.common.cellsToPixels(block_representation.width, this.zoom) * 0.5 - 20 * this.zoom;
    var connector_height = Math.floor(beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT * this.zoom);
    var connector_radius = connector_height / 2;

    this.context.strokeStyle = '#000000';
    this.context.font = '' + ((beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT - 4) * this.zoom) + 'px Arial';
    this.context.textBaseline = 'middle';
    this.context.textAlign = 'start';

    if (block_representation.block.isDataset())
        max_connector_width += 40 * this.zoom;

    var view = this;

    function _computeConnectorsWidth(connectors) {
        if (!connectors)
            return {
                width: 0,
                names: [],
            }

        var connectors_width = 0;
        var names = [];

        for (var i = 0; i < connectors.length; ++i)
        {
            var name = connectors[i].name;
            var metrics = view.context.measureText(name);

            var nb = 2;
            while (metrics.width > max_connector_width - 4 * view.zoom)
            {
                ++nb;
                name = connectors[i].name.substr(0, connectors[i].name.length - nb) + '...';
                metrics = view.context.measureText(name);
            }

            connectors_width = Math.min(max_connector_width,
                                        Math.max(connectors_width,
                                                 metrics.width + 4 * view.zoom));

            names.push(name);
        }

        return {
            width: connectors_width,
            names: names,
        };
    }


    function _drawConnector(connector_name, x, y, width, input, generated) {
        if (!input)
            x -= width;

        // Connector box
        view.context.beginPath();
        view.context.moveTo(x + 0.5, y + 0.5);
        view.context.lineTo(x + width, y + 0.5);
        view.context.arc(x + width, y + 0.5 + connector_radius,
                         connector_radius, -Math.PI / 2, Math.PI / 2, false);
        view.context.lineTo(x + 0.5, y + connector_height + 0.5);
        view.context.lineTo(x + 0.5, y + 0.5);

        if (generated)
        {
            view.context.strokeStyle = '#00AA00';
            view.context.fillStyle = '#00CC00';
        }
        else
        {
            view.context.strokeStyle = '#000000';
            view.context.fillStyle = '#f6f810';
        }

        view.context.fill();
        view.context.stroke();

        // Connector name
        view.context.fillStyle = '#000000';
        view.context.fillText(connector_name, x + Math.floor(4 * view.zoom), y + connector_radius);

        return {
            x: x,
            y: y,
            width: width,
            height: connector_height,
        };
    }


    var connectors_infos = _computeConnectorsWidth(block_representation.block.inputs);
    for (var i = 0; i < connectors_infos.names.length; ++i)
    {
        var y = Math.floor(((i + 2) * beat.toolchains.common.CELL_SPACING - (beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT - 1) / 2) * this.zoom);

        if (block_representation.block.inputs[i].connections.length > 0)
        {
            var output = block_representation.block.inputs[i].connections[0].output;
            var generated = (this.running &&
                             (output.block.isDataset() ||
                              ((this.blocks_status !== null) &&
                               (this.blocks_status[output.block.name] == 'generated'))));
        }
        else
        {
            var generated = false;
        }

        block_representation.input_coords[i] = _drawConnector(connectors_infos.names[i],
                                                              0, y, connectors_infos.width,
                                                              true, generated);
    }


    var connectors_infos = _computeConnectorsWidth(block_representation.block.outputs);
    for (var i = 0; i < connectors_infos.names.length; ++i)
    {
        var y = Math.floor(((i + 2) * beat.toolchains.common.CELL_SPACING - (beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT - 1) / 2) * this.zoom);

        var generated = (this.running &&
                         (block_representation.block.isDataset() ||
                          ((this.blocks_status !== null) &&
                           (this.blocks_status[block_representation.block.name] == 'generated'))));

        block_representation.output_coords[i] = _drawConnector(connectors_infos.names[i],
                                                               width, y, connectors_infos.width,
                                                               false, generated);
    }


    // Synchronization channel
    if (!this.running && (block_representation.block.channel || block_representation.block.isDataset()))
    {
        this.context.beginPath();
        this.context.arc(width - (beat.toolchains.common.BLOCK_CHANNEL_RADIUS + 6) * this.zoom + 0.5,
                         (beat.toolchains.common.BLOCK_CHANNEL_RADIUS + 4) * this.zoom + 0.5,
                         beat.toolchains.common.BLOCK_CHANNEL_RADIUS * this.zoom, 0, 2 * Math.PI, false);

        this.context.fillStyle = (block_representation.block.channel ? this.channel_colors[block_representation.block.channel]
                                                                     : this.channel_colors[block_representation.block.name]);
        this.context.fill();
    }


    this.context.restore();
}


//----------------------------------------------------------------------------------------
// Draw a connection
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._drawConnection = function(representation, creation)
{
    function _coords(point, zoom)
    {
        return {
            x: beat.toolchains.common.cellsToPixels(point.col, zoom),
            y: beat.toolchains.common.cellsToPixels(point.row, zoom),
        };
    }


    this.context.save();

    this.context.lineWidth = 2;
    this.context.lineCap   = 'square';


    var start = 0;
    var end = representation.nbSegments() - 1;

    this.context.strokeStyle = '#000000';

    this.context.beginPath();

    var segment = representation.getSegment(0);
    var point1 = _coords(segment.point1, this.zoom);
    var point2 = _coords(segment.point2, this.zoom);

    this.context.moveTo(point1.x + Math.floor(beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT * this.zoom) / 2 + 0.5,
                        point1.y + 0.5);

    this.context.lineTo(point2.x + 0.5, point2.y + 0.5);

    segment = representation.getSegment(representation.nbSegments() - 1);
    point1 = _coords(segment.point1, this.zoom);
    point2 = _coords(segment.point2, this.zoom);

    point2.x -= 1;

    this.context.moveTo(point1.x + 0.5, point1.y + 0.5);
    this.context.lineTo(point2.x + 0.5, point2.y + 0.5);

    this.context.stroke();

    start = 1;
    end = representation.nbSegments() - 2;

    if (this.running)
    {
        if (representation.connection.output.block.isDataset() ||
            (this.blocks_status &&
             (this.blocks_status[representation.connection.output.block.name] == 'generated')))
        {
            this.context.strokeStyle = '#00CC00';
        }
        else
        {
            this.context.strokeStyle = '#000000';
        }
    }
    else if (representation.connection.input.channel)
    {
        this.context.strokeStyle = this.channel_colors[representation.connection.input.channel];
    }
    else
    {
        this.context.strokeStyle = '#000000';
    }

    segment = representation.getSegment(1);
    point1 = _coords(segment.point1, this.zoom);

    this.context.beginPath();
    this.context.moveTo(point1.x + 0.5, point1.y + 0.5);


    for (var i = start; i <= end; ++i)
    {
        var segment = representation.getSegment(i);
        var point2 = _coords(segment.point2, this.zoom);

        this.context.lineTo(point2.x + 0.5, point2.y + 0.5);
    }

    this.context.stroke();

    this.context.restore();
}


//----------------------------------------------------------------------------------------
// Returns an unused channel color
//----------------------------------------------------------------------------------------
beat.toolchains.viewer.ToolchainViewer.prototype._getUnusedChannelColor = function()
{
    var view = this;
    var used_colors = Object.keys(this.channel_colors).map(function(k){return view.channel_colors[k].toUpperCase();});

    var palette = [
        '#0000FF',
        '#6AA84F',
        '#CC0000',
        '#274E13',
        '#7F6000',
        '#4A86E8',
        '#45818E',
        '#9900FF',
        '#8E7CC3',
    ];

    for (var j = 0; j < palette.length; ++j)
    {
        var color = palette[j];
        if (used_colors.indexOf(color) == -1)
            return color;
    }

    var color = null;

    do {
        color = '#' + Math.floor(Math.random() * 256).toString(16).toUpperCase() +
                      Math.floor(Math.random() * 256).toString(16).toUpperCase() +
                      Math.floor(Math.random() * 256).toString(16).toUpperCase();
    } while (used_colors.indexOf(color) >= 0);

    return color;
}
