#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.http import Http404
from django.shortcuts import get_object_or_404
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.contrib.auth.models import User
from django.db.models.functions import Coalesce

from .models import Toolchain
from ..team.models import Team
from ..reports.models import Report
from ..common.texts import Messages
from ..ui.templatetags.markup import restructuredtext

from beat.core import prototypes

import simplejson as json


@login_required
def create(request, name=None):
    """Creates a new toolchain or a new version of an existing toolchain

    The user must be authenticated before it can add a new toolchain
    """

    parameters = {'toolchain_author':  request.user.username,
                  'toolchain_name':    name,
                  'toolchain_version': 1,
                  'short_description': '',
                  'description':       '',
                  'errors':            '',
                  'edition':           False,
                  'messages':          Messages,
                 }

    # Retrieves the existing toolchain (if necessary)
    if name is not None:
        previous_versions = Toolchain.objects.filter(
                                  author=request.user,
                                  name__iexact=name,
                              ).order_by('-version')
        if len(previous_versions) == 0:
            raise Http404()

        previous_version = previous_versions[0]

        description = previous_version.description

        parameters['toolchain_version'] = previous_version.version + 1
        parameters['declaration']       = previous_version.declaration_string.replace('\n', '')
        parameters['short_description'] = previous_version.short_description
        parameters['description']       = description.replace('\n', '\\n')
        parameters['html_description']  = restructuredtext(description).replace('\n', '')
        parameters['errors']            = previous_version.errors.replace('\n', '\\n') if previous_version.errors is not None else ''
    else:
        declaration, errors = prototypes.load('toolchain')
        parameters['declaration'] = json.dumps(declaration)

    return render_to_response('toolchains/edition.html',
                              parameters,
                              context_instance=RequestContext(request))


@login_required
def fork(request, author, name, version):
    """Creates a new toolchain by forking an existing toolchain

    The user must be authenticated before it can add a new toolchain
    """

    # Retrieves the forked toolchain
    fork_of = get_object_or_404(Toolchain.objects.for_user(request.user, True),
                                author__username__iexact=author,
                                name__iexact=name,
                                version=int(version)
                               )

    description = fork_of.description

    parameters = {'toolchain_author':  request.user.username,
                  'toolchain_name':    name,
                  'toolchain_version': 1,
                  'fork_of':           fork_of,
                  'declaration':       fork_of.declaration_string.replace('\n', ''),
                  'short_description': fork_of.short_description,
                  'description':       description.replace('\n', '\\n'),
                  'errors':            fork_of.errors.replace('\n', '\\n') if fork_of.errors is not None else '',
                  'edition':           False,
                  'messages':          Messages,
                 }

    return render_to_response('toolchains/edition.html',
                              parameters,
                              context_instance=RequestContext(request))


@login_required
def edit(request, author, name, version):
    """Allows the modification of the toolchain. The Web API is used to retrieve the
    details about the toolchain and check the accessibility.

    """

    if author != request.user.username:
        raise Http404()

    # Retrieves the toolchain
    toolchain = get_object_or_404(Toolchain,
                                  author__username__iexact=author,
                                  name__iexact=name,
                                  version=int(version)
                                 )

    description = toolchain.description

    # Render the page
    return render_to_response('toolchains/edition.html',
                              {'toolchain_author':  request.user.username,
                               'toolchain_name':    name,
                               'toolchain_version': toolchain.version,
                               'declaration':       toolchain.declaration_string.replace('\n', ''),
                               'short_description': toolchain.short_description,
                               'description':       description.replace('\n', '\\n'),
                               'html_description':  restructuredtext(description).replace('\n', ''),
                               'errors':            toolchain.errors.replace('\n', '\\n') if toolchain.errors is not None else '',
                               'edition':           True,
                               'messages':          Messages,
                              },
                              context_instance=RequestContext(request))


def view(request, author, name, version=None):
    """Shows the toolchain. The Web API is used to retrieve the details about
    the toolchain and check the accessibility.
    """

    # Retrieves the forked toolchain
    if version:
        toolchain = get_object_or_404(
                Toolchain,
                author__username__iexact=author,
                name__iexact=name,
                version=int(version),
                )
    else:
        toolchain = Toolchain.objects.filter(author__username__iexact=author,
                name__iexact=name).order_by('-version')
        if not toolchain:
            raise Http404()
        else:
            toolchain = toolchain[0]

    (has_access, accessibility) = toolchain.accessibility_for(request.user)

    if not has_access:
        raise Http404()

    owner = (request.user == toolchain.author)

    reports = None
    if not request.user.is_anonymous(): #fetch user reports, if any
        reports = Report.objects.filter(author=request.user)

    # Users the object can be shared with
    users = User.objects.exclude(username__in=settings.ACCOUNTS_TO_EXCLUDE_FROM_TEAMS).order_by('username')

    # Render the page
    return render_to_response('toolchains/view.html',
                              {
                                  'toolchain': toolchain,
                                  'owner': owner,
                                  'reports': reports,
                                  'users': users,
                                  'teams': Team.objects.for_user(request.user, True)
                              },
                              context_instance=RequestContext(request))


def diff(request, author1, name1, version1, author2, name2, version2):
    """Shows the difference between two toolchains. The Web API is used to
    retrieve the details about the toolchains and check the accessibility.
    """

    toolchain1 = get_object_or_404(
            Toolchain,
            author__username__iexact=author1,
            name__iexact=name1,
            version=int(version1),
            )
    has_access, _ = toolchain1.accessibility_for(request.user)
    if not has_access: raise Http404()

    toolchain2 = get_object_or_404(
            Toolchain,
            author__username__iexact=author2,
            name__iexact=name2,
            version=int(version2),
            )
    has_access, _ = toolchain2.accessibility_for(request.user)
    if not has_access: raise Http404()

    return render_to_response('toolchains/diff.html',
                              {
                                  'toolchain1':  toolchain1,
                                  'toolchain2':  toolchain2,
                              },
                              context_instance=RequestContext(request))


def ls(request, author_name):
    '''List all accessible toolchains to the request user'''

    if not author_name: return public_ls(request)

    # check that the user exists on the system
    author = get_object_or_404(User, username=author_name)

    # orders toolchains so that the latest information is displayed first
    objects = Toolchain.objects.from_author_and_public(request.user,
            author_name).order_by('-creation_date')
    objects = Toolchain.filter_latest_versions(objects)

    return render_to_response('toolchains/list.html',
            dict(
                objects=objects,
                author=author,
                owner=(request.user==author),
                ),
            context_instance=RequestContext(request),
            )


def public_ls(request):
    '''List all publicly accessible objects'''

    # orders so that more recent are first
    objects = Toolchain.objects.public().order_by('-creation_date')
    objects = Toolchain.filter_latest_versions(objects)

    return render_to_response('toolchains/list.html',
            dict(
                objects=objects,
                author=request.user, #anonymous
                owner=False,
                ),
            context_instance=RequestContext(request),
            )
