#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import logging
logger = logging.getLogger(__name__)

import os
import shutil
import tarfile
import tempfile

from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
from django.conf import settings
from django.apps import apps, registry

from .backup import APPS

# Overrides deserialization to affect OneToOneFields for Users correctly
import django.core.serializers.base
from django.contrib.auth.models import User
from ....navigation.models import Agreement
from ....accounts.models import AccountSettings
from ....experiments.models import Experiment, Block

_original_save = django.core.serializers.base.DeserializedObject.save

def save(self, *args, **kwargs):

    if isinstance(self.object, Agreement):
        # if the user in question has an agreement, delete and write new
        user = User.objects.get(pk=self.object.user_id)
        if hasattr(user, 'agreement'): user.agreement.delete()

    elif isinstance(self.object, AccountSettings):
        # if the user in question has account settings, just update it
        user = User.objects.get(pk=self.object.owner_id)
        if hasattr(user, 'accountsettings'): user.accountsettings.delete()

    # use the built-in function for storing the new object
    _original_save(self, *args, **kwargs)

django.core.serializers.base.DeserializedObject.save = save


class Command(BaseCommand):

    help = 'Resets the current database with a (stored) backup'


    def add_arguments(self, parser):

        parser.add_argument('-e', '--exclude', dest='exclude', action='append',
            default=[], help='An app_label to exclude (use multiple ' \
                '--exclude to exclude multiple apps).')

        parser.add_argument('backup', type=str,
                help='The backup you wish to restore from')


    def handle(self, *ignored, **arguments):

        # Setup this command's logging level
        global logger
        arguments['verbosity'] = int(arguments['verbosity'])
        if arguments['verbosity'] >= 1:
            if arguments['verbosity'] == 1: logger.setLevel(logging.INFO)
            elif arguments['verbosity'] >= 2: logger.setLevel(logging.DEBUG)

        try:
            tmpdir = tempfile.mkdtemp('.backup', 'beat.web-')

            # extracts the tarball on the temporary directory
            logger.info("Reading archive `%s'" % arguments['backup'])
            with tarfile.open(arguments['backup']) as tar:
                tar.extractall(tmpdir)

            exclude = arguments.get('exclude')

            # remove uninstalled apps
            global APPS
            installed_apps = registry.apps.all_models.keys()
            exclude += [app for app in APPS if app not in installed_apps]

            arguments = dict(
                    verbosity=arguments.get('verbosity'),
                    interactive=False,
                    )

            # reset database contents
            logger.info("Loading initial data...")
            call_command('migrate', **arguments)

            # loads the initial data
            srcfile = os.path.join(tmpdir, 'initial.json')
            logger.info("Loading unspecified initial data <- `%s'" % \
                    srcfile)
            call_command('loaddata', srcfile, **arguments)

            # reset all user tokens (so they can be re-inserted)
            apps.get_model('authtoken.Token').objects.all().delete()

            # and loads the apps respecting the imposed order
            for app in APPS:
                if app in exclude: continue
                # copy prefix data
                path = os.path.join(settings.PREFIX, app)
                srcdir = os.path.join(tmpdir, 'prefix', app)
                if os.path.exists(srcdir):
                    logger.info("Restoring core objects for `%s' <- " \
                            "`%s'" % (app, srcdir))
                    if os.path.exists(path):
                        logger.info("Removing `%s'..." % path)
                        shutil.rmtree(path)
                    shutil.copytree(srcdir, path)
                else:
                    logger.info("No core objects found for `%s'" % app)

                srcfile = os.path.join(tmpdir, '%s.json' % app)
                logger.info("Loading data for `%s' <- `%s'" % \
                        (app, srcfile))
                call_command('loaddata', srcfile, **arguments)

            # creates the cache directory
            cache = settings.CACHE_ROOT
            if not os.path.exists(cache):
                os.makedirs(cache)
                logger.info("Created directory `%s'" % cache)



        finally:

            # removes the temporary directory
            shutil.rmtree(tmpdir)

            # reset all experiments in a transient state (notice Job and
            # JobSplit are not backed-up because of circular dependence issues
            # on the "experiments" app).
            transient = (Experiment.RUNNING, Experiment.CANCELING)
            for e in Experiment.objects.filter(status__in=transient): e.reset()
